#include <stdio.h>
#include <stdlib.h>
#include "Params.h"
#include "types.h"

Kernel_Params params;
void writeFromC(TYPE *, TYPE *, int, int, char*);

//copies array R of number random numbers to the gpu
void copyspringnumbers_(float * R, int * number) {
	int N = *number;
	params.random_numbers_count = N;
	printf("N= %i\n", N);
	copyArrayToGPU(&params,R);
}

void sameergpuinit_(int * Nc_f, int * L_f, int * nBlock, double * alpha_f, int* ntr, int* sgn) {
	params.Nc   = *( Nc_f );
	params.L    = *( L_f ) ;
	params.alpha = *(alpha_f);
	params.ISIZE = ( params.Nc) * ( params.L );
	params.GSIZE = params.ISIZE * params.ISIZE;
	params.nDelays = *(nBlock);
	params.ntr = *ntr;
        params.sgn = *sgn;
	printf("from gpuinit params.alpha = %f \n" , params.alpha);
	AllocateCudaMemory(&params);
	launch_seed_random(&params);
	SetCacheSize();
}

void gpufinalize_() {
}

double clock_now__() {
    struct timeval now;
    gettimeofday(&now, NULL);
    return now.tv_sec*1.0e6 + 1.0*now.tv_usec;
};

void callcuda_(TYPE * gu, TYPE * gd, int *s, int * Nc_f, int * L_f, int * nBlock, double * alpha_f, int* umode_i) {
  	printf("From callcuda_\n");  
	int N = * Nc_f;
	int L = * L_f;
	TYPE alpha = *alpha_f;
	int umode = *umode_i;
	double elapsed =0.0;
	double start,end;
#if __DEBUG__
	printf("From callcuda_ we have: alpha=%f, N=%d, L=%d \n", alpha, N,L);
	printf("First element of gu from c (gu[0]): %f \n", gu[0]);
	printf("Next element of gu from c (gu[1]): %f \n", gu[1]);
	printf("Next element of gd from c (gd[1]): %f \n", gd[1]);
#endif 
	start = clock_now__();
	CopyDataToGPU( gu, gd , s , &params);
	end = clock_now__();
	elapsed += (end - start);
	if(umode == 4) {
          launch_reg_delayed(&params);
        }
	else if (umode ==5) { 
          printf("ERROR: mode 5 not available in this version, now exiting \n");
          exit(-1);
        }
	else { 
          printf("ERROR -- CUDA umode not implemented.  \n");
        }

	start = clock_now__();
	CopyDataToCPU(gu, gd , s , &params);
	end = clock_now__();
	elapsed += (end - start);
}

//TODO: Pass the mode to the to the CUDA code from the Fortran code !
void callupdatecuda_(TYPE * gu, TYPE * gd, int *s, int * Nc_f, int * L_f, int * nBlock, int * ntimes, double * alpha_f, int* sgn, int* umode_i) {
  	printf("From callupdatecuda_\n");
	int N = * Nc_f;
	int L = * L_f;
	TYPE alpha   = *alpha_f;
	params.nTimes = *ntimes;
	int umode = *umode_i;
	double elapsed =0.0;
	double start,end;
#if __DEBUG__ 
	printf("From callcuda_ we have: alpha=%f, N=%d, L=%d \n", alpha, N,L);
	printf("First element of gu from c (gu[0]): %f \n", gu[0]);
	printf("Next element of gu from c (gu[1]): %f \n" , gu[1]);
	printf("Next element of gd from c (gd[1]): %f \n" , gd[1]);
#endif
	start = clock_now__();
	CopyDataToGPU(gu, gd , s , &params);
	end = clock_now__();
	elapsed += (end - start);
	if(umode == 4) {
	  printf("Running for CUDA delayed updates\n");
	  launch_update_delayed(&params);
         }
	else if(umode ==5) {
          printf("ERROR: mode 5 not available in this version, now exiting \n");
          exit(-1);
	}
	else 
          printf("ERROR -- CUDA umode not implemented.  cSideCallingCuda.c \n");
	
	
	//dev_mc_delayed_host_cuda_5(&params);	
	//printf("Back calling cuda_5\n");
	start = clock_now__();
	CopyDataToCPU(gu, gd, s, &params);
	elapsed += (end - start);
	//printf("copy from MC took: %f\n",elapsed/1000000);
	//printf("gu[0]=%f after\n", gu[0]);
	// Copy back the sign to the fortran code.
	*sgn = (params.sgn);
}

void callmeascuda_(TYPE * gu, TYPE * gd, int *s, double * beta_fort){

	params.beta = *( beta_fort );
	CopyDataToGPU(gu, gd , s , &params);
	//callMeasurement_host(&params);
        printf("ERROR: Measurement on CUDA not available in this version, now exiting \n");
        exit(-1);
}

void writeFromC(TYPE * gu, TYPE * gd, int N, int L, char* tag) {
  int i,j,l1,l2, ind, imd, gind;
  FILE * guOut;
  FILE * gdOut;
  char* bufferName;
	
  bufferName = (char*)malloc(100*sizeof(char));
  sprintf(bufferName,"../debugIO/from_C_gu_%s.dat",tag);
  guOut = fopen(bufferName,"w");
  if(NULL == guOut) {
    printf( "Error opening files for guOut\n" );
    return ;
  }

  sprintf(bufferName,"../debugIO/from_C_gd_%s.dat",tag);
  gdOut = fopen(bufferName,"w");
  if (NULL == gdOut) {
    printf( "Error opening files for gdOut\n" );
    return ;
   }
	for(l1 =0; l1< L ;l1++) {
	for(l2 =0; l2<L; l2++) {
	        for(j=0; j< N; j++) {
	                for(i=0; i<N; i++) {
			        ind = N*l2 + i;
			        imd = N*l1 + j;
			        gind = imd + ind*N*L;
			        fprintf(guOut, "%d %d %d %d %f \n",l1,l2,j,i, gu[gind]);
			        fprintf(gdOut, "%d %d %d %d %f \n",l1,l2,j,i, gd[gind]);
	               }
	     }
	}
       }
	fclose(guOut);
	fclose(gdOut);
	free(bufferName); 
}
