        subroutine update_single(ntimes)
c*****************************************************************************
c       Update updates the Hubbard-Statonovich fields by sweeping sequentially 
c       through them and testing if changes are required.  The Ising field is
c       updated by the metropolis or heat bath algorithm.  If changes are made 
c       it updates the greens functions appropriately.  This block should 
c       probably be compiled with an inline option.
c*****************************************************************************
        use Global
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h" 
c************************************************************
        integer :: i,j,ntimes,st,l
        real(kind) :: R,au,ad,ru,rd,time1,time2
c************************************************************
 
        time1=real(MPI_Wtime(),kind)
        do i=1,ntimes
          nslr=nslr+1
          do j=1,nt
c           Now update the j'th ising field element s(j)
            ucount=ucount+oner
            st=-s(j)
            ru=oner-(gu(j,j)-oner)*expm1(+st)
            if(isu.eq.1) then
              rd=oner-(gd(j,j)-oner)*expm1(-st)
              R=ru*rd
            else !isu=-1 and U<0 so 
              R=ru**2*exp(-twor*alpha*st)
            end if

c           sprng is David Ceperley's parallel random number generator
c           see http://www.ncsa.uiuc.edu/Apps/CMP/RNG/index_old.html
c           if(abs(R)/(oner+abs(R)).gt.sprng())then   !heatbath
c           if(abs(R)/(oner+abs(R)).gt.rand())then    !heatbath
            if(abs(R).gt.sprng())then                 !metropolis
c             accept the change
              sgn=sign(oner,R)*sgn
              acrat=acrat+oner
              s(j)=st
              au=expm1(+st)/ru
              call upg_single(j,au,gu)   
              if(isu.eq.1) then  ! then Gd must also be updated
                ad=expm1(-st)/rd
                call upg_single(j,ad,gd) 
              end if
            end if
          end do
c
c         Now we must check to see if it is time to reconfigure the
c         Gd matrix.  Reconfiguration is necessary to deal with roundoff error.
 
          if(nslr.ge.ntr)then
            call reg_single
            nslr=0
          end if 
        end do
 
        time2=real(MPI_Wtime(),kind)
        time_date=time_date+time2-time1
        if(myrank.eq.0.and.iprint.ge.9) write(lud,*) 'done update' !0
        return
        end

        subroutine reg_single
c       This subroutine recalculates the green functions given only the  
c       s-field configuration and the initial green's functions Gcstsc  
c       (which corresponds to all s=0).
 
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
 
        real(kind) :: uxterm,dxterm,au,ad,ru,rd
        integer i,j,k,l,m,n,ind,imd,ic,jc
 
c       Gcstsc is the greens function for all s=0, hence we will
c       sweep through the lattice and reconfigure it to coorespond
c       to the lattice configuration.
 
c       move Gcstsc to gu and gd
        do m=1,nl
        do n=1,nl
          k=n-m
        do ic=1,Nc
        do jc=1,Nc
          ind=ic + (n-1)*Nc
          imd=jc + (m-1)*Nc
          gu(ind,imd)=Gcst_not(k,ic,jc)
          gd(ind,imd)=Gcst_not(k,ic,jc)
        end do
        end do
        end do
        end do
        sgn=oner 
c       sweep over all the time slices and sites
        do i=1,nt
          uxterm=exp(+alpha*s(i))-oner
          ru=(oner-(gu(i,i)-oner)*uxterm)
          au=uxterm/ru
          call upg_single(i,au,gu) 
          if(isu.eq.1) then
c           U>0, so the down spin G must be updated
            dxterm=exp(-alpha*s(i))-oner
            rd=(oner-(gd(i,i)-oner)*dxterm)
            ad=dxterm/rd
            call upg_single(i,ad,gd)
            sgn=sign(oner,ru*rd)*sgn
          else ! isu=-1 ( U<0 )
            sgn=oner
          end if
        end do
 
        return
        end
        subroutine upg_single(lo,acg,gcg)
 
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
 
        integer lo,i,j
        real(kind) :: acg,d(nt),b(nt),gcg(nt,nt),time1,time2
 
c       subroutine upg updates gd after a change in one of the field
c       variables has been accepted.
 
c       FORM THE VECTORS
c       do i=1,nt
c          b(i)=gcg(i,lo)*acg
c          d(i)=gcg(lo,i)
c        end do
        b(:)=gcg(:,lo)*acg
        b(lo)=b(lo)-acg
        d(:)=gcg(lo,:)
 
c       FORM THE OUTER PRODUCT, AND THE NEW G
c       5 WAYS TO DO IT (see which is fastest on your machine)
cbw     do i=1,nt
cbw     do j=1,nt
cbw       gcg(j,i)=gcg(j,i)+b(j)*d(i)
cbw     end do
cbw     end do
c
c       This is equivalent to the level 2  blas call (usually best)
c
        call sger(nt,nt,oner,b,1,d,1,gcg,nt)
c
c       and or the L3 blas call
c
c       call sgemm('N','T',nt,nt,1,oner,b,nt,d,nt,oner,gcg,nt)
c
c       or to the loop using level 1 blas call
c        
c       do i=1,nt
c         call saxpy(nt,d(i),b,1,gcg(1,i),1)
c       end do
c
c       and to the equivalent fortran90 call
c       do i=1,nt
c         gcg(:,i)=gcg(:,i) +b(:)*d(i)
c       end do

        return
        end
