!
! Brute-force search for cluster geometries
!
! Examine all possible lattice vectors up to some maximum size;
! sort by Betts perfection and cubicity. Code mostly from DCA code tables.f
!
! Outputs:
!
! standard out = all clusters
!
! gen_clusters.outf = FORTRAN source (!) tables.f style of best bipartide and
! non-bipartide cluster of each size.
!
! Notes:
!
! Took 25 minutes on a 3GHz P4 with parameters maxshell=3,maxstore=1000,maxc=40
!
! Took several hours on a 3GHz P4 with parameters maxshell=4,maxstore=1000,maxc=40
! Execution time scales as ~maxshell^9. Quicker if maxstore or maxc are small
!
! P. R. C. Kent Feb 2005

      module m_matrix_utils

      interface Inverse
      module procedure Inverse_3x3
      end interface

      interface determinant
      module procedure Ideterminant
      module procedure Rdeterminant
      end interface

      contains

      function Inverse_3x3(a)   
      integer :: a(3,3)
      double precision :: Inverse_3x3(3,3),r1
c
c       [ a(1,1) a(1,2) a(1,3) ]
c     A=[ a(2,1) a(2,2) a(2,3) ]
c       [ a(3,1) a(3,2) a(3,3) ]
c
c      -1        i+j 
c     A_ji = (-1)   C_ij  /|A|    C_ij is the ij cofactor of A
c     
      
      r1=   a(1,1)*a(2,2)*a(3,3)  ! the determinent of a
     &    + a(1,2)*a(2,3)*a(3,1)
     &    + a(1,3)*a(2,1)*a(3,2)
     &    - a(3,1)*a(2,2)*a(1,3)
     &    - a(3,2)*a(2,3)*a(1,1)
     &    - a(3,3)*a(2,1)*a(1,2) 
      
      Inverse_3x3(1,1) = +( a(2,2)*a(3,3)-a(3,2)*a(2,3) )/r1
      Inverse_3x3(1,2) = -( a(1,2)*a(3,3)-a(3,2)*a(1,3) )/r1
      Inverse_3x3(1,3) = +( a(1,2)*a(2,3)-a(2,2)*a(1,3) )/r1
      
      Inverse_3x3(2,1) = -( a(2,1)*a(3,3)-a(3,1)*a(2,3) )/r1
      Inverse_3x3(2,2) = +( a(1,1)*a(3,3)-a(3,1)*a(1,3) )/r1
      Inverse_3x3(2,3) = -( a(1,1)*a(2,3)-a(2,1)*a(1,3) )/r1
      
      Inverse_3x3(3,1) = +( a(2,1)*a(3,2)-a(3,1)*a(2,2) )/r1
      Inverse_3x3(3,2) = -( a(1,1)*a(3,2)-a(3,1)*a(1,2) )/r1
      Inverse_3x3(3,3) = +( a(1,1)*a(2,2)-a(2,1)*a(1,2) )/r1

      return
      end function Inverse_3x3

      function Ideterminant(a)
      integer::  a(3,3),Ideterminant

      Ideterminant=
     &         a(1,1)*a(2,2)*a(3,3)
     &       + a(1,2)*a(2,3)*a(3,1)
     &       + a(1,3)*a(2,1)*a(3,2)
     &       - a(3,1)*a(2,2)*a(1,3)
     &       - a(3,2)*a(2,3)*a(1,1)
     &       - a(3,3)*a(2,1)*a(1,2)
      return
      end function Ideterminant
      
      function Rdeterminant(a)
      real(8)::  a(3,3),Rdeterminant

      Rdeterminant=
     &         a(1,1)*a(2,2)*a(3,3)
     &       + a(1,2)*a(2,3)*a(3,1)
     &       + a(1,3)*a(2,1)*a(3,2)
     &       - a(3,1)*a(2,2)*a(1,3)
     &       - a(3,2)*a(2,3)*a(1,1)
     &       - a(3,3)*a(2,1)*a(1,2)
      return
      end function Rdeterminant

      end module m_matrix_utils

      module m_cluster

      integer ixy(48),ixz(48),iyz(48),isx(48),isy(48),isz(48)

c      interface transform3d
c      module procedure gtransform3D
c      module procedure Rtransform3D
c      module procedure Ktransform3D
c      end interface transform3d

      contains


      function gtransform3D(a,igroup)
c
c  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
c  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
c  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
c
      implicit none
      integer :: i,a(3,3),g(3,3),gtransform3D(3,3),igroup
      
      do i=1,3  
         gtransform3D(i,1)=a(i,1)
         gtransform3D(i,2)=(1-iyz(igroup))*a(i,2) + iyz(igroup)*a(i,3)
         gtransform3D(i,3)=(1-iyz(igroup))*a(i,3) + iyz(igroup)*a(i,2)
      end do
               
      do i=1,3  
         g(i,1)= (1-ixz(igroup))*gtransform3D(i,1)
     &         + ixz(igroup)*gtransform3D(i,3)
         g(i,2)=                gtransform3D(i,2)
         g(i,3)= (1-ixz(igroup))*gtransform3D(i,3) 
     &         + ixz(igroup)*gtransform3D(i,1)
      end do

      do i=1,3  
         gtransform3D(i,1)=isx(igroup)*((1-ixy(igroup))*g(i,1) 
     &                  + ixy(igroup)*g(i,2))
         gtransform3D(i,2)=isy(igroup)*((1-ixy(igroup))*g(i,2) 
     &                  + ixy(igroup)*g(i,1))
         gtransform3D(i,3)=isz(igroup)*                 g(i,3)         
      end do
      
      return
      end function gtransform3d
        
      function Rtransform3D(Rc,igroup)
c
c  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
c  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
c  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
c
      implicit none
      integer :: i,Rc(3),g(3),Rtransform3D(3),igroup
      
      Rtransform3D(1)=            Rc(1)
      Rtransform3D(2)=(1-iyz(igroup))*Rc(2) + iyz(igroup)*Rc(3)
      Rtransform3D(3)=(1-iyz(igroup))*Rc(3) + iyz(igroup)*Rc(2)
      
      g(1)=(1-ixz(igroup))*Rtransform3D(1) + ixz(igroup)*Rtransform3D(3)
      g(2)=                Rtransform3D(2)
      g(3)=(1-ixz(igroup))*Rtransform3D(3) + ixz(igroup)*Rtransform3D(1)
      
      Rtransform3D(1)=isx(igroup)*((1-ixy(igroup))*g(1) 
     &             + ixy(igroup)*g(2))
      Rtransform3D(2)=isy(igroup)*((1-ixy(igroup))*g(2) 
     &             + ixy(igroup)*g(1))
      Rtransform3D(3)=isz(igroup)*                 g(3) 
      
      return
      end function rtransform3d
      
      function Ktransform3D(Kc,igroup)
c     
c  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
c  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
c  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
c
      implicit none
      integer :: i,igroup
      real(8) :: Kc(3),g(3),Ktransform3D(3)
      
      Ktransform3D(1)=                Kc(1)
      Ktransform3D(2)=(1-iyz(igroup))*Kc(2) + iyz(igroup)*Kc(3)
      Ktransform3D(3)=(1-iyz(igroup))*Kc(3) + iyz(igroup)*Kc(2)
      

      g(1)=(1-ixz(igroup))*Ktransform3D(1) + ixz(igroup)*Ktransform3D(3)
      g(2)=                Ktransform3D(2)
      g(3)=(1-ixz(igroup))*Ktransform3D(3) + ixz(igroup)*Ktransform3D(1)
      
      Ktransform3D(1)=isx(igroup)*((1-ixy(igroup))*g(1) 
     &             + ixy(igroup)*g(2))
      Ktransform3D(2)=isy(igroup)*((1-ixy(igroup))*g(2) 
     &             + ixy(igroup)*g(1))
      Ktransform3D(3)=isz(igroup)*                 g(3) 
      
      return
      end function ktransform3d

      function betts_perfection(a)
      use m_matrix_utils
      implicit none

      integer betts_perfection(2)
      integer, intent(in) :: a(3,3)

      integer :: a1(3,3)
      double precision :: c1(3,3)

      integer :: Nc,I_F, I_FShell
      integer :: i1,i,j,k,iv1(3),iv2(3),i2,i3
      integer :: n,m,p

      integer, parameter :: ndim=3
      integer, parameter :: nshellmax=50
      double precision, parameter :: eps=1d-6

      integer :: nneigh(nshellmax),nneighl(nshellmax),
     &     nneighp(nshellmax),ishellmax

      double precision :: rv1(3),rv2(3)

      do i=1,3
        a1(:,i)=a(i,:)
      end do
      c1=Inverse(a1)

      Nc=abs(determinant(a))

c      if(iprint.ge.5.and.myrank.eq.0) then
c        write(*,*) '  ' !0
c        write(*,*) 'neighbors of ic' !0
c        write(*,*) 
c     &    'shell  nneigh(shell)  nneighl(shell) nneighp(shell)' !0
c      end if
      i1 = 2*(Nc+1)**(1d0/ndim); rv1=0d0; nneigh=0; ishellmax=0
      nneighl(:)=0 
      nneighp(:)=0
      
      do i=-i1,i1
      do j=-i1,i1
      do k=-i1,i1
        iv1=(/ i, j, k/)
        i2=sum(abs(iv1)) ! the lattice neighbor index (according to Betts)
        if(i2.gt.0) then
          if(i2.le.nshellmax) then
            nneighl(i2)=nneighl(i2)+1
          end if
          do n=1,ndim
            rv1(n)=iv1(n) + 0.5d0*sum(a(:,n))
          end do
          rv2=matmul(c1,rv1)
          if(rv2(1).lt.1d0-eps.and.rv2(1).gt.0d0-eps.and.
     &       rv2(2).lt.1d0-eps.and.rv2(2).gt.0d0-eps.and.
     &       rv2(3).lt.1d0-eps.and.rv2(3).gt.0d0-eps ) then
c            This point is within the cluster. Now we need to find
c            the shortest distance between this point and icR0, and
c            the corresponding cluster neighbor index i3.
             i3=i2
             do n=-1,1
             do m=-1,1
             do p=-1,1
               iv2=iv1 + n*a(1,:) + m*a(2,:) + p*a(3,:)
               if(sum(abs(iv2)).lt.i2) i3=sum(abs(iv2))
             end do !p
             end do !m
             end do !n
             nneigh(i3)=nneigh(i3)+1
             if(i3.gt.ishellmax) ishellmax=i3
          end if
        end if
      end do !k
      end do !j
      end do !i
      
      i2=0
      do while (sum(nneighp).lt.Nc-1)
        i2=i2+1
        nneighp(i2)=nneighl(i2)
      end do
      nneighp(i2)=Nc-1-sum(nneighp(1:i2-1))
      
      if(sum(nneigh)+1.ne.Nc.or.sum(nneighp)+1.ne.Nc) then
!        write(*,*) 'error in nneigh',nc
!        write (*,*) a(1,:)
!        write (*,*) a(2,:)
!        write (*,*) a(3,:)
!        write(*,*) 'sum(nneigh)=',sum(nneigh)
!        write(*,*) 'sum(nneighp)=',sum(nneighp)
!        stop
! HACK: if problem assume "bad" cluster
         betts_perfection(1)=666
         return
      end if
        
c      if(iprint.ge.5.and.myrank.eq.0) then
c        do i1=1,ishellmax
c          write(*,"(2x,i3,8x,i3,10x,i3,10x,i3)") 
c     &      i1,nneigh(i1),nneighl(i1),nneighp(i1)
c        end do
c        write(*,*) 'ishellmax=',ishellmax
c      end if
c
c     Now determine the perfection.  It seems that Betts defines the
c     ferromagnetic imperfection I_F as the numer of additional spins in
c     the shells beyond j plus the number missing the shells before j,
c     and not counting those in the jth shell.  The integer j seems to
c     be the number which yeilds the smallest I_F

      do j=1,ishellmax
        i3=0
        do i1=1,j-1
          i3=i3+abs(nneigh(i1)-nneighl(i1))
        end do
        do i1=j+1,ishellmax
          i3=i3+nneigh(i1)
        end do
        if(j.gt.1) then
          if(i3.lt.I_F) then
             I_F=i3 
             I_FShell=j
           end if
        else
          i_F=i3
          i_FShell=j
        end if
      end do
      betts_perfection(1)=i_F
      betts_perfection(2)=i_FShell
cc      write(*,*) betts_perfection

c      
c      if(myrank.eq.0) 
c     &  write(*,*) 'ferromagnetic imperfection I_F,I_FShell',I_F,I_FShell
     
c     Betts proposes a simpler algorithm to calculate the perfection,
c     involving the difference of two quantities.
c
c     F_f = SUM k nneigh(k)
c            k
c
c     F_p = SUM k nneighp(k)
c            k

c      i2=0; i3=0
c      do i1=1,ishellmax
c        i2=i2+i1*nneigh(i1)
c        i3=i3+i1*nneighp(i1)
c      end do
cc      betts_perfection(1)=i2-i3
cc      betts_perfection(2)=I_FShell

      end function

      function cubicity(a)

      double precision :: cubicity,r1,r2,
     &     corner(3,8),Bettsd(4),Bettsf(6),Bettsl(3),
     &     meand,meanf,meanl
      integer,intent(in) :: a(3,3)

c     Calculate the cubicity of the cluster      
c
c       c8-------c7    body diagonals   face diagonals
c      /|       /|     --------------   --------------
c     / |      / |       d1 (c1,c7)       f1 (c1,c3)
c    /  |     /  |       d2 (c2,c8)       f2 (c2,c4)
c   c5-------c6  |       d3 (c3,c5)       f3 (c1,c6)
c   |   c4---|---c3      d4 (c4,c6)       f4 (c2,c5)
c a3|  /     |  /                         f5 (c1,c9)
c   | / a2   | /                          f6 (c4,c5)
c   |/       |/
c   c1-------c2
c        a1
c
      corner(:,1) = 0.
      corner(:,2) = a(1,:)
      corner(:,3) = a(1,:)+a(2,:)
      corner(:,4) = a(2,:)
      corner(:,5) = a(3,:)
      corner(:,6) = a(3,:)+a(1,:)
      corner(:,7) = a(3,:)+a(1,:)+a(2,:)
      corner(:,8) = a(3,:)+a(2,:)
      
      Bettsd(1)=sqrt(sum((corner(:,1)-corner(:,7))**2))
      Bettsd(2)=sqrt(sum((corner(:,2)-corner(:,8))**2))
      Bettsd(3)=sqrt(sum((corner(:,3)-corner(:,5))**2))
      Bettsd(4)=sqrt(sum((corner(:,4)-corner(:,6))**2))
      
      Bettsf(1)=sqrt(sum((corner(:,1)-corner(:,3))**2))
      Bettsf(2)=sqrt(sum((corner(:,2)-corner(:,4))**2))
      Bettsf(3)=sqrt(sum((corner(:,1)-corner(:,6))**2))
      Bettsf(4)=sqrt(sum((corner(:,2)-corner(:,5))**2))
      Bettsf(5)=sqrt(sum((corner(:,1)-corner(:,8))**2))
      Bettsf(6)=sqrt(sum((corner(:,4)-corner(:,5))**2))
      
      Bettsl(1)=sqrt(dot_product(corner(:,2),corner(:,2)))
      Bettsl(2)=sqrt(dot_product(corner(:,4),corner(:,4)))
      Bettsl(3)=sqrt(dot_product(corner(:,5),corner(:,5)))
      
      meand=product(Bettsd)**(1d0/4d0)
      meanf=product(Bettsf)**(1d0/6d0)
      meanl=product(Bettsl)**(1d0/3d0)
      r1=sqrt(3d0)*meanl/meand
      r2=sqrt(2d0)*meanl/meanf
      cubicity=max(r1,1d0/r1)*max(r2,1d0/r2)
      end function cubicity

      subroutine initsymm
      implicit none
! Weird and wonderful symmetry operations setup per DCA code
!
!  Only 16A has all 48 operations and has the point group Oh.  The remaining clusters 
!  have lower symmetry than the cubic lattice, with symmetries designated by the
!  Schoenflies symbols (From Ibach and Luth):
!  Cj (j=2,3,4, 6) j-fold rotation axis 
!  Sj j-fold rotation-inversion axis 
!  Dj j 2-fold rotation axes perpendicular to a j-fold principle rotation axis 
!  T  4 three-and 3 two-fold rotation axes, as in a tetrahedron 
!  O  4 three-and 3 four-fold rotation axes, as in a octahedron 
!  Ci a center of inversion 
!  Cs a mirror plane
!  In addition, their are sufixes for mirror planes 
!     h: horizontal=perpendicular to the c  rotation axis, 
!     v: vertical=parallel to the main rotation axis in the plane, 
!     d: diagonal=parallel to the main rotation axis in the plane  
!        bisecting the two-fold rotation axes.
! 
!  The Clusters identified by Betts, have several different symmetries
!
!  C2h   A two fold axis, plus a single mirror plane that is perpendicular to the axis
!  Ci    Contains only the identity and inversion
!  C3i   Triagonal with a 3-fold axis and inversion (GUESS)??
!          * rotations by 2pi/3 about the axix x=y=z
!          * inversion wrt the origin.
!  Oh    The full symmetry of the cubic lattice 
!  D2h   2 2-fold axes perp. to the 2-fold rotation axis, plus a horizonal mirror plane
!  D3h   3 2-fold axes perp. to the 3-fold rotation axis, plus a horizonal mirror plane
!
!  We need to automate the process of finding which of the 48 point group operations are
!  retained by the different clusters.  To do this, we must perform all 48 cubic group
!  operations on the {a} and see if they form an equivalent parallelpiped.  
!
!  Group Operations for a cubic lattice:
!
!  x --> +/- x                  There are 2^D D! operations in D dimensions
!    \/                         or 48 operations in 3D.
!    /\
!  y --> +/- y   
!    \/
!    /\
!  z --> +/- z   
!        .
!  a'=transform3D(a,igroup,ixy,ixz,iyz,isx,isy,isz)
!  ixy=0(1) dont (do) exchange x and y
!  ixz=0(1) dont (do) exchange x and z
!  iyz=0(1) dont (do) exchange y and z
!  isx      sign of x 
!  isy      sign of y
!  isz      sign of z
!
!  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
!  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
!  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
!
      ixy( 1)=0;ixz( 1)=0; iyz( 1)=0; isx( 1)=+1; isy( 1)=+1; isz( 1)=+1 !identity  xyz e
      ixy( 2)=1;ixz( 2)=0; iyz( 2)=0; isx( 2)=+1; isy( 2)=+1; isz( 2)=+1 !ref. x=y  yxz o
      ixy( 3)=0;ixz( 3)=1; iyz( 3)=0; isx( 3)=+1; isy( 3)=+1; isz( 3)=+1 !ref. x=z  zyx o
      ixy( 4)=0;ixz( 4)=0; iyz( 4)=1; isx( 4)=+1; isy( 4)=+1; isz( 4)=+1 !ref. y=z  xzy o
      ixy( 5)=1;ixz( 5)=1; iyz( 5)=0; isx( 5)=+1; isy( 5)=+1; isz( 5)=+1 !          yzx e
      ixy( 6)=1;ixz( 6)=0; iyz( 6)=1; isx( 6)=+1; isy( 6)=+1; isz( 6)=+1 !          zxy e
      
      ixy( 7)=0;ixz( 7)=0; iyz( 7)=0; isx( 7)=-1; isy( 7)=+1; isz( 7)=+1 !
      ixy( 8)=1;ixz( 8)=0; iyz( 8)=0; isx( 8)=-1; isy( 8)=+1; isz( 8)=+1 !
      ixy( 9)=0;ixz( 9)=1; iyz( 9)=0; isx( 9)=-1; isy( 9)=+1; isz( 9)=+1 !
      ixy(10)=0;ixz(10)=0; iyz(10)=1; isx(10)=-1; isy(10)=+1; isz(10)=+1 !
      ixy(11)=1;ixz(11)=1; iyz(11)=0; isx(11)=-1; isy(11)=+1; isz(11)=+1 !
      ixy(12)=1;ixz(12)=0; iyz(12)=1; isx(12)=-1; isy(12)=+1; isz(12)=+1 !
     
      ixy(13)=0;ixz(13)=0; iyz(13)=0; isx(13)=+1; isy(13)=-1; isz(13)=+1 !
      ixy(14)=1;ixz(14)=0; iyz(14)=0; isx(14)=+1; isy(14)=-1; isz(14)=+1 !
      ixy(15)=0;ixz(15)=1; iyz(15)=0; isx(15)=+1; isy(15)=-1; isz(15)=+1 !
      ixy(16)=0;ixz(16)=0; iyz(16)=1; isx(16)=+1; isy(16)=-1; isz(16)=+1 !
      ixy(17)=1;ixz(17)=1; iyz(17)=0; isx(17)=+1; isy(17)=-1; isz(17)=+1 !
      ixy(18)=1;ixz(18)=0; iyz(18)=1; isx(18)=+1; isy(18)=-1; isz(18)=+1 !
      
      ixy(19)=0;ixz(19)=0; iyz(19)=0; isx(19)=-1; isy(19)=-1; isz(19)=+1 !
      ixy(20)=1;ixz(20)=0; iyz(20)=0; isx(20)=-1; isy(20)=-1; isz(20)=+1 !
      ixy(21)=0;ixz(21)=1; iyz(21)=0; isx(21)=-1; isy(21)=-1; isz(21)=+1 !
      ixy(22)=0;ixz(22)=0; iyz(22)=1; isx(22)=-1; isy(22)=-1; isz(22)=+1 !
      ixy(23)=1;ixz(23)=1; iyz(23)=0; isx(23)=-1; isy(23)=-1; isz(23)=+1 !
      ixy(24)=1;ixz(24)=0; iyz(24)=1; isx(24)=-1; isy(24)=-1; isz(24)=+1 !
      
      ixy(25)=0;ixz(25)=0; iyz(25)=0; isx(25)=+1; isy(25)=+1; isz(25)=-1 !
      ixy(26)=1;ixz(26)=0; iyz(26)=0; isx(26)=+1; isy(26)=+1; isz(26)=-1 !
      ixy(27)=0;ixz(27)=1; iyz(27)=0; isx(27)=+1; isy(27)=+1; isz(27)=-1 !
      ixy(28)=0;ixz(28)=0; iyz(28)=1; isx(28)=+1; isy(28)=+1; isz(28)=-1 !
      ixy(29)=1;ixz(29)=1; iyz(29)=0; isx(29)=+1; isy(29)=+1; isz(29)=-1 !
      ixy(30)=1;ixz(30)=0; iyz(30)=1; isx(30)=+1; isy(30)=+1; isz(30)=-1 !
      
      ixy(31)=0;ixz(31)=0; iyz(31)=0; isx(31)=-1; isy(31)=+1; isz(31)=-1 !
      ixy(32)=1;ixz(32)=0; iyz(32)=0; isx(32)=-1; isy(32)=+1; isz(32)=-1 !
      ixy(33)=0;ixz(33)=1; iyz(33)=0; isx(33)=-1; isy(33)=+1; isz(33)=-1 !
      ixy(34)=0;ixz(34)=0; iyz(34)=1; isx(34)=-1; isy(34)=+1; isz(34)=-1 !
      ixy(35)=1;ixz(35)=1; iyz(35)=0; isx(35)=-1; isy(35)=+1; isz(35)=-1 !
      ixy(36)=1;ixz(36)=0; iyz(36)=1; isx(36)=-1; isy(36)=+1; isz(36)=-1 !
      
      ixy(37)=0;ixz(37)=0; iyz(37)=0; isx(37)=+1; isy(37)=-1; isz(37)=-1 !
      ixy(38)=1;ixz(38)=0; iyz(38)=0; isx(38)=+1; isy(38)=-1; isz(38)=-1 !
      ixy(39)=0;ixz(39)=1; iyz(39)=0; isx(39)=+1; isy(39)=-1; isz(39)=-1 !
      ixy(40)=0;ixz(40)=0; iyz(40)=1; isx(40)=+1; isy(40)=-1; isz(40)=-1 !
      ixy(41)=1;ixz(41)=1; iyz(41)=0; isx(41)=+1; isy(41)=-1; isz(41)=-1 !
      ixy(42)=1;ixz(42)=0; iyz(42)=1; isx(42)=+1; isy(42)=-1; isz(42)=-1 !
      
      ixy(43)=0;ixz(43)=0; iyz(43)=0; isx(43)=-1; isy(43)=-1; isz(43)=-1 !
      ixy(44)=1;ixz(44)=0; iyz(44)=0; isx(44)=-1; isy(44)=-1; isz(44)=-1 !
      ixy(45)=0;ixz(45)=1; iyz(45)=0; isx(45)=-1; isy(45)=-1; isz(45)=-1 !
      ixy(46)=0;ixz(46)=0; iyz(46)=1; isx(46)=-1; isy(46)=-1; isz(46)=-1 !
      ixy(47)=1;ixz(47)=1; iyz(47)=0; isx(47)=-1; isy(47)=-1; isz(47)=-1 !
      ixy(48)=1;ixz(48)=0; iyz(48)=1; isx(48)=-1; isy(48)=-1; isz(48)=-1 !

      end subroutine initsymm

      subroutine findsymmops(a,group,ngroup)
      use m_matrix_utils
      implicit none
      integer,intent(in) :: a(3,3)
      integer,intent(out) :: group(48),ngroup
      
      integer :: lc,igroup,a1(3,3),nc
      double precision :: ainv(3,3),a1ainv(3,3)
      logical    :: denied,denied_lin_alg
      double precision, parameter :: eps=1d-6

      nc=abs(determinant(a))
      ngroup=0; Lc=2*(Nc**.3333+1)
!      if(myrank.eq.0.and.iprint.ge.5) then
!        write(lud,*) '--------------------------------------------',
!     &               '------------------' 
!        write(lud,*) '             trans(a)        igroup L  ixy ix',
!     &               'z iyz isx isy isz'
!        write(lud,*) '--------------------------------------------',
!     &               '------------------'
!      end if
      do igroup=1,48
         a1=gtransform3D(a,igroup)

c PK If lattice points generated by "a1" vectors can be found in 
c lattice points generated by "a" vectors then (a1)x(a^-1) can only 
c contain integers.
            ainv=Inverse_3x3(a)
            a1ainv=matmul(a1,ainv)
            if (all(abs(a1ainv-nint(a1ainv)).le.eps)) then 
               denied_lin_alg=.false. ! matrix only contains integers
            else
               denied_lin_alg=.true.
            end if

         denied=denied_lin_alg

         if(denied.eqv..false.) then ! a corresponding point was found
            ngroup=ngroup+1
            group(ngroup)=igroup
         end if    

!         if(myrank.eq.0.and.iprint.ge.5) then
!           allowed=.true.;if (denied) allowed=.false.
!           write(lud,"(3('(',i2,1x,i2,1x,i2,')'),2x,i2,2x,l,6(2x,i2))")
!     &     a1(1,:),a1(2,:),a1(3,:),igroup,allowed,ixy(igroup),
!     &     ixz(igroup),iyz(igroup),isx(igroup),isy(igroup),isz(igroup)
!         end if
      end do      

!      if(myrank.eq.0.and.iprint.ge.5) 
!     &  write(lud,"('cluster=',a3,' ngroup=',i2,' groups=',48(i2,1x))") 
!     &  cluster,ngroup,group(1:ngroup)


      end subroutine findsymmops


      subroutine check_kspace(a,group,ngroup,ick0,ickpi)
      use m_matrix_utils
      implicit none
      integer,intent(in) :: a(3,3),group(48),ngroup
      integer,intent(out) :: ick0,ickpi

c       Now calculate the principle translations in K-space
c   
c  
c                              _    _
c    _     _    _        2 pi aj x ak          
c    a --> g    g_i =  ------------------      plus cyclic permutations
c                      | ai . (aj x ak)|       
c                                           
c       Principle translation vectors for lattice tiling       
c                                        
c
      double precision, parameter ::pi=3.1415926535897932384626433832795
      double precision, parameter ::eps=1d-6
      double precision :: r1,rv1(3),rv2(3),g(3,3)
      integer :: i,j,k,l,ic,itest,igroup,jc
      integer :: nc,ncw,iv1(3)
      integer,parameter :: maxwedgekp=101
      double precision :: kc(3,maxwedgekp)

      r1=determinant(a)
      nc=abs(r1)

      g(1,1)=2*pi*( a(2,2)*a(3,3)-a(3,2)*a(2,3) )/r1  !a2y*a3z-a3y*a2z
      g(1,2)=2*pi*( a(2,3)*a(3,1)-a(3,3)*a(2,1) )/r1  !a2z*a3x-a3z*a2x
      g(1,3)=2*pi*( a(2,1)*a(3,2)-a(3,1)*a(2,2) )/r1  !a2x*a3y-a3x*a2y
      
      g(2,1)=2*pi*( a(3,2)*a(1,3)-a(1,2)*a(3,3) )/r1  !a3y*a1z-a1y*a3z
      g(2,2)=2*pi*( a(3,3)*a(1,1)-a(1,3)*a(3,1) )/r1  !a3z*a1x-a1z*a3x
      g(2,3)=2*pi*( a(3,1)*a(1,2)-a(1,1)*a(3,2) )/r1  !a3x*a1y-a1x*a3y
        
      g(3,1)=2*pi*( a(1,2)*a(2,3)-a(2,2)*a(1,3) )/r1  !a1y*a2z-a2y*a1z
      g(3,2)=2*pi*( a(1,3)*a(2,1)-a(2,3)*a(1,1) )/r1  !a1z*a2x-a2z*a1x
      g(3,3)=2*pi*( a(1,1)*a(2,2)-a(2,1)*a(1,2) )/r1  !a1x*a2y-a2x*a1y
      
      r1=abs(determinant(g))
      if(abs((2*pi)**3/r1-Nc).gt.eps) then
!        if(myrank.eq.0) write(lud,*) 'This cell volume aint Nc',r1
        stop
      end if
      do i=1,3
      do j=1,3
        r1=sum(a(i,:)*g(j,:))/(2*pi)
        if(i.eq.j.and.abs(r1-1d0).gt.eps) then
!         if(myrank.eq.0) write(lud,*) 'bug in g',i,j,r1
          stop
        end if
        if(i.ne.j.and.abs(r1).gt.eps) then
!         if(myrank.eq.0) write(lud,*) 'bug in g',i,j,r1
          stop
        end if
      end do
      end do
      
!      if(iprint.ge.5.and.myrank.eq.0) 
!     &  write(lud,"(' g=',3(' (',f5.2,1x,f5.2,1x,f5.2,')'))") 
!     &  (g(l,:),l=1,3)
     
     
c***********************************************************************        
c     Create the cluster momentum tables Kc(i,ic), i=x,y,z
c***********************************************************************        

c     First find the points in the irreducible wedge.  They will be
c     indexed from 1 to Ncw
!      if(iprint.ge.5.and.myrank.eq.0) then
!        write(lud,*) '  Wedge K points  '    !0
!        write(lud,*) 'ic  i  j  k  Kcx(ic)  Kcy(ic)  Kcz(ic)  ' !0
!      end if
      ic=0; Ncw=0; icK0=0; icKpi=0
      do i=-Nc,Nc    !times a1
      do j=-Nc,Nc    !times a2
      do k=-Nc,Nc    !times a3
        iv1=(/i,j,k/)
        do l=1,3         !x,y,z components 
          rv1(l)=sum(iv1(:)*g(:,l))
        end do
        if(rv1(1).gt.-pi+eps.and.rv1(1).lt.pi+eps.and.
     &     rv1(2).gt.-pi+eps.and.rv1(2).lt.pi+eps.and.
     &     rv1(3).gt.-pi+eps.and.rv1(3).lt.pi+eps) then
c          The point is in the first SC Brillouin zone.  Now see if
c          the point falls within the IRW.  First see if it is 
c          equivalent to any known point in the IRW (if so it cannot be in IRW)
          do igroup=2,ngroup ! exclude the identity
            rv2=Ktransform3D(rv1,group(igroup))
            itest=0
            do jc=1,Ncw
              if(sum(abs(Kc(:,jc)-rv2(:))).lt.eps) then ! rv2 = Kc(jc)
                 itest=1
                 exit  ! leave the jc loop
              end if
            end do
            if(itest.eq.1) exit !leave the igroup loop
          end do
          if(itest.eq.0) then ! rv1 is in 1BZ and not equivalent to any point  
             Ncw=Ncw+1        ! in the known IRW.  It must be in the IRW
             ic=ic+1
             Kc(:,ic)=rv1(:)
!             ig(ic)=i
!             jg(ic)=j
!             kg(ic)=k
             if(sum(abs(rv1-pi)).lt.eps) icKpi=ic
             if(sum(abs(rv1)).lt.eps) icK0=ic
!             if(iprint.ge.5.and.myrank.eq.0) 
!     &            write(lud,5) ic,i,j,k,Kc(1,ic),Kc(2,ic),Kc(3,ic) !0
          end if
        end if 
      end do 
      end do 
      end do 
        
!      if(icK0.eq.0) then  ! we failed to find (0,0)
!        if(myrank.eq.0) 
!     &    write(lud,*) 'failed to find (0,0) in the IRW'    !0
!        stop
!      else if(icKpi.eq.0) then  ! we failed to find (pi,pi)
!       icKpi=icK0
!       if(myrank.eq.0) then
!         write(lud,*) '*****************************************'    !0
!         write(lud,*) '* failed to find (pi,pi,pi) in the IRW  *'    !0
!         write(lud,*) '* apparently the lattice isnt bipartide *'    !0
!         write(lud,*) '* icKpi=icK0 so that the code will run  *'    !0
!         write(lud,*) '*****************************************'    !0
!       end if
!      end if


      end subroutine check_kspace

      end module m_cluster

      program gen_clusters
      use m_cluster
      use m_matrix_utils
      implicit none

! Maxmium vector size: up to (maxshell,maxshell,maxshell)
!      integer,parameter :: maxshell=4
      integer,parameter :: maxshell=3
! Maximum no of clusters of each size to store
      integer,parameter :: maxstore=1000
! Half maximum cluster size (odd clusters are excluded)
      integer,parameter :: maxc=40

! Sort on perfection first, then cubicity if true
! Reverse priority if false
      logical, parameter :: sort_on_perfection = .true.

      double precision, parameter :: tol =1d-6

      integer :: store_nc(maxc) ! Nc of stored clusters
      integer :: store_nfound(maxc) ! No clusters of each size (up to maxstore)
      integer :: store_a(3,3,maxstore,maxc) ! A of stored clusters
      double precision :: store_cubicity(maxstore,maxc)
      integer :: store_perfection(2,maxstore,maxc)

      integer :: i,j,nc
      integer :: a(3,3),aref(3,3),aref1(3,3)
      double precision :: arefI(3,3),aprod(3,3)


      integer :: a1i,a1j,a1k
      integer :: a2i,a2j,a2k
      integer :: a3i,a3j,a3k
      double precision :: cub
      integer :: perf(2),jpos,endpos

      integer :: ngroup,group(48),ick0,ickpi,jpi

! Initialise the store_nc index array
      store_nfound=0
! We examine only clusters where any(store_nc.eq.nc)
!      store_nc(1)=1
!      do i=2,maxc
!         store_nc(i)=2*(i-1)
      do i=1,maxc
         store_nc(i)=2*i
      end do

! Debugging/testing dinosaur tracks:
!    8A     ( 2, 0, 0)    ( 0, 2, 0)   ( 0, 0, 2)     Oh   48   4     yes
!    8B     ( 2, 1, 0)    ( 0, 1, 2)   ( 1,-1, 1)     C2h  4    0     no  (=C2 x Ci)
!      a(1,:)=(/2, 1, 0/); a(2,:)=(/0, 1, 2/); a(3,:)=(/1,-1, 1/)
!      a(1,:)=(/2, 0, 0/); a(2,:)=(/0, 2,0/); a(3,:)=(/0,0,2/)
!      aref(1,:)=(/2, 2, 0/); aref(2,:)=(/0, 2,2/); aref(3,:)=(/2,0,2/)
!      write (*,*) abs(determinant(aref)),cubicity(aref)
!      do i=1,3
!        aref1(:,i)=aref(i,:)
!      end do
!      arefI=Inverse(aref1)
!      write (6,*) arefI

      do a1i=1,maxshell
      do a1j=0,maxshell
      do a1k=0,maxshell
      a(1,:)=(/a1i,a1j,a1k/)

      do a2i=maxshell,-maxshell,-1
      do a2j=maxshell,-maxshell,-1
      do a2k=maxshell,-maxshell,-1
      a(2,:)=(/a2i,a2j,a2k/)


      do a3i=maxshell,-maxshell,-1
      do a3j=maxshell,-maxshell,-1
      do a3k=maxshell,-maxshell,-1
      a(3,:)=(/a3i,a3j,a3k/)
      
      nc=abs(determinant(a))
      if (nc.gt.0.and.(mod(nc,2).eq.0).or.nc.eq.1) then

         do i=1,maxc
            if (store_nc(i).eq.nc) exit
         end do
         if (i.gt.maxc) cycle


         if (sort_on_perfection) then
            cub=-1
            perf=betts_perfection(a)
            jpos=0
            do j=1,store_nfound(i)
               if (perf(1).lt.store_perfection(1,j,i)) then
                  jpos=j
                  exit
               else
                  if (perf(1).eq.store_perfection(1,j,i)) then
                     if (cub.eq.-1) cub=cubicity(a)
                     
                     if( cub.gt.store_cubicity(j,i)-tol.and.
     &                    cub.lt.store_cubicity(j,i)+tol) then
! "duplicate" cluster with identical perfection + ~identical cubicity
                        jpos=-1
                        exit
                     else if (cub.le.store_cubicity(j,i)-tol) then
                        jpos=j
                        exit
                     end if
                  end if
               end if
            end do
         else
! Only calculate the perfection once (slow):
! Use perf=-1 to indicate that the perfection has not been calculated
            cub=cubicity(a)
            perf(1)=-1
            jpos=0
            do j=1,store_nfound(i)
               if (cub.lt.store_cubicity(j,i)-tol) then
                  jpos=j
                  exit
               else
                  if (cub.gt.store_cubicity(j,i)-tol.and.
     &                 cub.lt.store_cubicity(j,i)+tol) then
                     
                     if (perf(1).eq.-1) perf=betts_perfection(a)
                     if (perf(1).eq.store_perfection(1,j,i)) then
!     "duplicate" cluster with identical perfection + ~identical cubicity
                        jpos=-1
                        exit
                     else if (perf(1).lt.store_perfection(1,j,i)) then
                        jpos=j
                        exit
                     end if
                  end if
               end if
            end do
         end if
         if (jpos.eq.0.and.j.le.maxstore) jpos=j
         
         if (jpos.gt.0) then
! jpos>0 => do insertion
            if (cub.eq.-1) cub=cubicity(a)
            if (perf(1).eq.-1) perf=betts_perfection(a)

            endpos=min(store_nfound(i),maxstore-1)
            store_cubicity(jpos+1:endpos+1,i)=
     &           store_cubicity(jpos:endpos,i)
            store_perfection(:,jpos+1:endpos+1,i)=
     &           store_perfection(:,jpos:endpos,i)
            store_a(:,:,jpos+1:endpos+1,i)=
     &           store_a(:,:,jpos:endpos,i)
            if(store_nfound(i).lt.maxstore) store_nfound(i)=
     &           store_nfound(i)+1

            store_cubicity(jpos,i)=cub
            store_perfection(:,jpos,i)=perf(:)
            store_a(:,:,jpos,i)=a
         end if

      end if

      end do
      end do
      end do

      end do
      end do
      end do

      end do
      end do
      end do

      call initsymm()

! Output the smallest cubicity, most perfect cluster as "NCA"
! the first bipartide cluster as "NCB"
      open(10,file='gen_clusters.outf')

      do i=1,maxc
         nc=store_nc(i)

         jpi=0
         do j=1,store_nfound(i)
            a=store_a(:,:,j,i)
            cub=store_cubicity(j,i)
            perf(:)=store_perfection(:,j,i)

            call findsymmops(a,group,ngroup)
            call check_kspace(a,group,ngroup,ick0,ickpi)
            write (*,9000) nc,a(1,:),a(2,:),a(3,:),cub,perf,ick0,ickpi
 9000       format(I3,1x,3('(',I2,1x,I2,1x,I2,')',1x),f12.8,4(1x,I3))


! Ugly conditonal section to write fortran code on unit 10
! Use two format statements so cluster size string has no spaces
! [Could tidy with some string manips]
            if (j.eq.1) then
            if (ickpi.ne.0) then
               jpi=ickpi
            end if
            if (nc.ge.10) then
            write (10,9010) nc,'A',cub,perf,jpi.ne.0,
     &              j,a(1,:),a(2,:),a(3,:)
            else
            write (10,9015) nc,'A',cub,perf,jpi.ne.0,
     &              j,a(1,:),a(2,:),a(3,:)
            end if
            else
               if (jpi.eq.0.and.ickpi.ne.0) then
                  jpi=ickpi
            if (nc.ge.10) then
            write (10,9010) nc,'B',cub,perf,jpi.ne.0,j,
     &              a(1,:),a(2,:),a(3,:)
            else
            write (10,9015) nc,'B',cub,perf,jpi.ne.0,j,
     &              a(1,:),a(2,:),a(3,:)
            end if
               end if
            end if
 9010       format(
     &  '      else if(cluster.eq."',I2,A1,'") then',                     
     & /'! Cubicity=',f12.8,' Perfection=',I3,' Betts Shell=',I3,
     &   ' Bipartide=',L1,' ID#=',I3,
     & /'         a(1,:)=(/',I2,',',I2,',',I2,'/); a(2,:)=(/',
     &  I2,',',I2,',',I2,'/); a(3,:)=(/',I2,',',I2,',',I2,'/)')
 9015       format(
     &  '      else if(cluster.eq."',I1,A1,'") then',                     
     & /'! Cubicity=',f12.8,' Perfection=',I3,' Betts Shell=',I3,
     &   ' Bipartide=',L1,' ID#=',I3,
     & /'         a(1,:)=(/',I2,',',I2,',',I2,'/); a(2,:)=(/',
     &  I2,',',I2,',',I2,'/); a(3,:)=(/',I2,',',I2,',',I2,'/)')
         end do
      end do
      close(10)


      end program gen_clusters
