        subroutine geod                     
c***********************************************************************
c       Starting from Sigma(iwn,K), this block makes the initial green 
c       functions.                         | K = momentum
c                                          | f = frequency
c       GcKf, GcKfsc, Gcstsc and Gcst_not  |sc = script (cluster excluded)
c                                          | s = space ; t = time
c       The last three correspond to the field configuration where all 
c       s(i)=0.0.  Gcstsc is calculated by summing its Fourier transform. 
c       Since the sum over matsubara frequencies does not converge, 
c       convergence is forced by adding and subtracting a known exact 
c       result. The coarse-graining is done in parallel.  
c***********************************************************************
        use Global
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h"
c**************************************************************
        real(kind) :: uxterm,ru,au,time1,time2
        integer :: i,j,k,l,m,n,p,ic,ind,imd,myconfig
c**************************************************************

        time1=real(MPI_Wtime(),kind)    ! A platform indep. timing call
        
c       First, coarse-grain the Green function.
        call CoarseGrain(GcKf,sigma)

c       Now make GcKfsc, the cluster excluded analog of GcKf. 
        GcKfsc=GcKf/(oner+GcKf*sigma)

c       Impose the point group and (optional) PH symmetries.              
        call Symm_PG(GcKfsc)
        if(abs(ed).lt.eps.and.abs(tprime).lt.eps) call Symm_PH(GcKfsc)

c       Now Fourier transform GcKfsc(iwn,K) --> Gcstsc(tau,R)
        call FT_Kf_to_Kt(GcKfsc,data,0) !Freq. part. data(0,ic) --> tau=0!
        do k=1,Nc                       ! see if the result is sensible
          ru=data(0,k)+data(nl,k)       ! G(0+,K)+G(beta-,K)
          if(abs(ru-oner).gt.eps) then
c           there is a problem with the Green function
            if(myrank.eq.0) then
              write(lud,*) 'problem with GcKtsc, ick=',k
              write(lud,*) 'GcKtsc(0,ick)+GcKtsc(nl,ick)=',ru
              write(lud,*) 'GcKtsc(tau,ick)',(data(l,k),l=0,nl)
            end if
            abortf=1
          end if
        end do
        write(abortstring,*) 'geod,GcKtsc'
        call abortquerry
        call FT_Kt_to_Rt(data,Gcstsc)   !Spatial part.  Gcstsc(0,ic) --> tau=0!


c       The initial green function must have two spatial indices to
c       allow for the possibility of disorder.  
        do n=-nl,nl
        do i=1,Nc
        do j=1,Nc
          Gcst_not(n,i,j)=Gcstsc(n,icrdiff(i,j))
        end do
        end do
        end do
        
        if(isv.eq.1) then
c         There is disorder, so we need to set the value of the 
c         disorder potential and modify the initial green function.
c
          if(iter.eq.1) then                     
            if(V0.gt.zeror) then                  ! Binary Disorder
              if(nproc.ge.2**Nc .and.             !
     &               mod(nproc,2**Nc).eq.0) then  ! Enumerate the disorder
                myconfig=mod(myrank,2**Nc)        ! configuration in binary
                do i=1,Nc                         ! 
                  V(i)=V0*twor*(iand(1,ishft(     ! shift myconfig i-1 
     &                    myconfig,(1-i)))-halfr) ! to the right, and and
                end do                            !
              else                                ! Sample the disorder
                do i=1,Nc                         ! P(V)= [del(V-V0) +
                  V(i)=V0*sign(oner,sprng()-halfr)!        del(V+V0)]/2
                end do                            !
              end if
            else                                  ! Continuous Disorder
              do i=1,Nc                           !
                  V(i)=twor*V0*(sprng()-halfr)      !     [ 1/2V0, -V0<V<V)
              end do                              ! P(V)= {
            end if                                !       [  0  ,  else
          end if                                  !
c         Load the initial Green function, Gcst_not, into gu.
          do m=1,nl
          do n=1,nl
          do i=1,Nc
          do j=1,Nc
            k=n-m
            ind=i + (n-1)*Nc
            imd=j + (m-1)*Nc
            gu(ind,imd)=Gcst_not(k,i,j)
          end do
          end do
          end do
          end do
c         Sweep over all the time slices and sites and make the "bare"
c         Green function in the matrix gu.
          do ic=1,Nc
            uxterm=exp(dt*V(ic))-oner
c           uxterm=dt*V(ic)
          do l=1,nl
            i=ic + (l-1)*Nc
c           uxterm=V(ic)
            ru=(oner-(gu(i,i)-oner)*uxterm)
            au=uxterm/ru
cBe sure to call single spin upg here: no guarantee blocks are setup 
            call upg_single(i,au,gu)
          end do
          end do   
          gd=gu       
c         Now load the gu we have just formed into Gcst_not
          do m=1,nl
          do n=1,nl
          do i=1,Nc
          do j=1,Nc
            k=n-m
            ind=i + (n-1)*Nc
            imd=j + (m-1)*Nc
            Gcst_not(k,i,j)=gu(ind,imd)
          end do
          end do
          end do
          end do
        end if
c       Now Gcst_not (and gu) corresponds to V_ic=V(ic) and all s(i)=0.

        time2=real(MPI_Wtime(),kind)
        if(myrank.eq.0) write(lud,*) 'time in geod= ',time2-time1        !0
        time_geod=time_geod+time2-time1
        if(myrank.eq.0.and.iprint.ge.1) then
              write(lud,*) 'done geod';flush(lud)
	endif
        
        return
        end subroutine geod


        subroutine geodosft                          
c***********************************************************************
c       This block makes the Green functions used to condition the
c       Fourier transform of gft to Gcsf, which are   
c
c       Gcsfosft and Gcstosft
c
c       The transform is done by oversampling (osft=oversampling FT)
c***********************************************************************
        use Global
        implicit none
        include 'mpif.h'   ! Is this needed??
#define SIMPLE_SPRNG       ! Is this needed??
#define FLOAT_GEN          ! Is this needed??
#define USE_MPI            ! Is this needed??
#include "sprng_f.h"       ! Is this needed??
c**************************************************************
        real(kind) :: time1,time2
        integer :: i
c**************************************************************

        if(iosft.eq.0) then ! Use sigmapt to calculate Gcsfosft and Gcstosft.
c         Calculate the coarse-grained perturbation theory greens functions
          call CoarseGrain(GcKfpt,sigmapt)
c         Fourier transform GcKfpt --> Gcsfosft and Gcstosft
          call FT_Kf_to_Kt(GcKfpt,data,1)   ! data(0,ic) --> tau=0!
          call FT_Kt_to_Rt(data,Gcstosft)   ! Gcstosft(0,ic) --> tau=0! 
          call FT_Kf_to_Rf(GcKfpt,Gcsfosft)
        else if (iosft.eq.1) then ! Use GcKf from the previous iteration.
c         Fourier transform GcKf --> Gcsfosft and Gcstosft
          call FT_Kf_to_Kt(GcKf,data,1)     ! data(0,ic) --> tau=0!
          call FT_Kt_to_Rt(data,Gcstosft)   ! Gcstosft(0,ic) --> tau=0! 
          call FT_Kf_to_Rf(GcKf,Gcsfosft)         
        else
          write(6,*) 'bad value for iosft'
        end if

        return
        end subroutine geodosft
        

        subroutine perturbation                         !parallel version
c***********************************************************************
c       This block calculates the perturbation theory self energy
c       using second order perturbation theory.  
c***********************************************************************
        use Global
        implicit none
        include 'mpif.h'
#define SIMPLE_SPRNG
#define FLOAT_GEN
#define USE_MPI
#include "sprng_f.h"
c************************************************************
        integer i,n,m,k,p,ik,ip,iq,ic,ick,maxbound,minbound,
     &          its,nleft,iproc
        complex(kind) :: csum
c************************************************************

        its=Nc/nproc
        nleft=Nc-nproc*its
        if(myrank+1.le.nleft) its = its + 1        

c       Now calculate the perturbation theory self-energy sigmapt(iwn)
c
c
c        ___                 wm,q
c       /   \               ---<---
c      |     |            /         \
c       \___/            |\         /|
c         |              |  --->---  |
c         |              | wm+wp,q+p |         =   sigma(k,wn)
c         |              |           |
c       --<-- wn,k       ------<------  wn,k
c                         wn-wp, k-p
c
c                         2  2 ___                __
c   U(1/2 -gf(tau=0))   -T U   \   Gf(wn-wp,k-p) \   Gf(wm,q) Gf(wm+wp,q+p)
c                      -----   /__               /__
c                      Nc*Nc   wp,p             wm,q
c
c       First we must calculate the bubble in the second order part.
c       Conditioning assists the convergence, using
c
c           SUM 1/(iwn(m) iwn(m+p)) = - (beta/2)^2 delta(p,0)
c
c
        chi(:,:)=zeroc

        do k=1,its
          ic = (myrank + 1) + nproc*(k-1)
          do p=-nwn/2,nwn/2
            if(p.ge.0) then
              maxbound=nwn-p-1
              minbound=-nwn+1
            else
              maxbound=nwn-1
              minbound=-nwn-p+1
            end if
            do iq=1,Nc
              do m=minbound,maxbound
                chi(p,ic)=chi(p,ic) 
     &                   - GcKfsc(m,iq)*GcKfsc(m+p,ickplus(iq,ic))
     &                   - oner/(wn(m)*wn(m+p))
              end do ! m
              if(p.eq.0) chi(0,ic)=chi(0,ic) + (halfr*beta)**2
            end do   ! iq
          end do     ! p
        end do       ! k (i.e. ic)
c       Now on every processor, pass the information around
        do ik=1,Nc
          iproc = ik - 1 - nproc*((ik-1)/nproc)
          call MPI_BCAST(chi(-nwn,ik),2*nwn+1,
     &                   MPI_COMPLEX16,
     &                   iproc,MPI_COMM_WORLD,ierr)
        end do
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)

c       Now form the self energy
        sigmapt(:,:)=zeroc

        do k=1,its
          ik = (myrank + 1) + nproc*(k-1)
        do n=0,nwn-1
          maxbound=nwn-1-abs(n)
          minbound=-nwn+1+abs(n)
          do ip=1,Nc
          do p=minbound,maxbound
            sigmapt(n,ik)=sigmapt(n,ik)+
     &             halfr*chi(p,ip)*
     &      (GcKfsc(n+p,ickplus(ik,ip))+GcKfsc(n-p,ickdiff(ik,ip)))
          end do
          end do
          sigmapt(n,ik)=u*(nfocc-halfr)+
     &                  sigmapt(n,ik)*(u*temp)**2/real(Nc**2,kind)
          sigmapt(-1-n,ik)=conjg(sigmapt(n,ik))
        end do
        end do
c       Now on every processor, pass the information around
        do ik=1,Nc
          iproc = ik - 1 - nproc*((ik-1)/nproc)
          call MPI_BCAST(sigmapt(-nwn,ik),2*nwn+1,
     &                   MPI_COMPLEX16,
     &                   iproc,MPI_COMM_WORLD,ierr)
        end do
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
        
c       Symmetrize the PT self energy.
        call Symm_PG(sigmapt)
        if(abs(ed).lt.eps.and.abs(tprime).lt.eps) 
     &    call Symm_PH(sigmapt)
        
        if(isv.eq.1) then
c         There is also a disorder contribution to sigmapt.  We
c         will approximate this with the CPA.  Since <V^n>=0 for
c         n odd, only the even in V terms survive.
c
c             X                __ X__
c            / \            __/  / \ \__
c           /   \   +   ___/    /   \   \___    + ...
c          /     \     /       /     \      \
c         o-------o   o-------o-------o------o
c
c
c         For binary diagonal disorder, <V^n> = V^n, n even
c
c                                 2
c                             G V0          where G is the local
c         Sigma_bin(iwn) = ----------       site-excluded green
c                                   2       function.
c                           1- (GV0)
c
c         For continuous diagonal disorder V is evenly distributed 
c         from -V0 < V < V0, P(V) =1/2V0.
c
c                          /           2
c                          |        G V              1    [    -1            ]
c         Sigma_con(iwn) = | P(V) ---------  dV  = ------ [tanh  (GV0) - GV0 ]
c                          |              2        (V0G)G [                  ]
c                          /      1 - (GV)
c
c         Alternatively, we could approximate the disorder contribution 
c         with just the second order result:
c
c                               2
c         Sigma_bin(iwn) = G V0
c
c                               2
c         Sigma_con(iwn) = G V0  /3
c
c         Let's use the latter, at least for now
c

          do n=0,nwn-1
            csum=zeroc
            do ick=1,Nc
              csum=csum + GcKfsc(n,ick)
            end do
            csum=csum/real(Nc)
            if(V0.gt.0) then
              sigmapt(n,:)=sigmapt(n,:) + csum*V0**2
            else
              sigmapt(n,:)=sigmapt(n,:) + csum*V0**2/3.0_kind
            end if
          end do
          
        end if   ! isv=1        
        
        if(myrank.eq.0.and.iprint.ge.1)  then
             write(lud,*) 'done perturbation' ; flush(lud)
	endif
        return
        end subroutine perturbation


        subroutine Symm_PG(Gcg)
c**********************************************************************
c       Average the green funtion Gcg over all allowed point-group
c       operations.
c                         x <---> +/-x       
c                            \ /             
c                             X                            
c                            / \                           
c                         y <---> +/-y 
c**********************************************************************
        use Global
        implicit none
c**************************************************************
        integer :: n,ic,iequiv
        complex(kind) :: Gcg(-nwn:nwn,Nc)
c**************************************************************
        chi(:,:)=zeroc              
        do n=0,nwn-1                                                 
        do ic=1,Nc      
          do iequiv=1,ngroup
            chi(n,ic)=chi(n,ic) + Gcg(n,ickequ(ic,iequiv))
          end do                    
          chi(n,ic)=chi(n,ic)/real(ngroup,kind)
          chi(-n-1,ic)=conjg(chi(n,ic))             
        end do                                                      
        end do   
                                                           
        Gcg=chi                              

        return
        end subroutine Symm_PG
        
        
        subroutine Symm_PH(Gcg)
c**********************************************************************
c       Impose particle-hole symmetry                                          
c                                                                   
c       G (iwn) = -G   (-iwn)                                       
c        k          Q-k                                             
c                                                                   
c       H is particle-hole symmetric whenever ed=0 and t'=0.  For a 
c       two-dimensional system, this means that 
c
c               +             +    i
c       C  --> C exp(iQ.i) = C (-1)
c        i      i             i
c                                                +              +
c       C  = SUM exp(ik.i) C  --> SUM exp(ik.i) C exp(-iQ.i) = C
c        k    i             i                    i              Q-k
c
c       or if ed=0, G (tau) = -G   (-tau) = G   (beta-tau)
c                    k          Q-k          Q-k
c       and
c                   G (iwn) = - G   (-iwn)  real parts opposite sign
c                    k           Q-k        imaginary parts equal sign
c
c**********************************************************************
        use Global
        implicit none
c**************************************************************
        integer :: n,ic
        complex(kind) :: Gcg(-nwn:nwn,Nc)
c**************************************************************
c       Impose particle-hole symmetry                                          
c                                                                   
c       G (iwn) = -G   (-iwn)                                       
c        k          Q-k                                             
c                                                                   
        do n=0,nwn-1                                                 
        do ic=1,Nc                                                
          chi(n,ic)=halfr*(Gcg(n,ic)-                 
     &                   Gcg(-n-1,ickdiff(icKpi,ic)))      
          chi(-n-1,ic)=conjg(chi(n,ic))                   
        end do                                                      
        end do                                   
                                                                    
        Gcg=chi                      
        
        return
        end subroutine Symm_PH


        subroutine FT_Kt_to_Rt(GcgKt,GcgRt)
c**********************************************************************
c       Fourier Transform GcgKt(l,ic) to GcgRt(l,ic).       
c**********************************************************************
        use Global
        implicit none
c***************************************************************
        integer :: i,ic,ik,idim
        real(kind) :: GcgRt(-nl:nl,Nc),GcgKt(-nl:nl,Nc),r1
        complex(kind) :: csum
c***************************************************************

        do ic=1,Nc
        do i=-nl,nl
          csum=zeroc
          do ik=1,Nc
            csum=csum + GcgKt(i,ik)*FTCoefs_K_to_R(ik,ic)
          end do
          GcgRt(i,ic)=real(csum)
        end do
        end do

        return
        end subroutine FT_Kt_to_Rt

        subroutine FT_Kf_to_Rf(GcgKf,GcgRf)     
c**********************************************************************
c       Fourier Transform GcgKf(n,ic) to GcgRf(n,ic).       
c**********************************************************************
        use Global
        implicit none
c******************************************************************
        integer :: n,ic,ik,idim
        real(kind) :: r1
        complex(kind) :: csum,GcgKf(-nwn:nwn,Nc),GcgRf(-nwn:nwn,Nc)
c******************************************************************

        do ic=1,Nc
        do n=0,nwn-1
          csum=zeroc
          do ik=1,Nc
            csum=csum + GcgKf(n,ik)*FTCoefs_K_to_R(ik,ic)
          end do
          GcgRf(n,ic)=csum
          GcgRf(-1-n,ic)=conjg(GcgRf(n,ic))
        end do
        end do
        
        end subroutine FT_Kf_to_Rf

        subroutine FT_Rf_to_Kf(GcgRf,GcgKf)     
c**********************************************************************
c       Fourier Transform GcgRf(n,ic) to GcgKf(n,ic).   Basically the
c       same as FT_Kf_to_Rf, except for the normalization and the loop
c       order. 
c**********************************************************************
        use Global
        implicit none
c******************************************************************
        integer :: n,ic,ik,idim
        real(kind) :: r1
        complex(kind) :: csum,GcgKf(-nwn:nwn,Nc),GcgRf(-nwn:nwn,Nc)
c******************************************************************

        do ik=1,Nc
        do n=0,nwn-1
          csum=zeroc
          do ic=1,Nc
            csum=csum + GcgRf(n,ic)*FTCoefs_R_to_K(ik,ic)
          end do
          GcgKf(n,ik)=csum                 ! no factor of 1/Nc
          GcgKf(-1-n,ik)=conjg(GcgKf(n,ik))  
        end do
        end do
        
        end subroutine FT_Rf_to_Kf
