        program fmain
c**************************************************************************
c       This is the main part of the DCA code.  Most blocks are 
c       generic,including this one, and can be reused in most
c       Hubbard-like codes.  The blocks specific to one model
c       are readin.f and modle_Global.f
c**************************************************************************
c       DEVELOPERS
c       Original serial code by Mark Jarrell and Niki Tahvildarzadeh, 1998.
c
c       This is a perfectly parallel version was written by Carey Huscroft 
c       1998, Mark Jarrell 1999-2003, and Alexandru Macridin 2002-2003.
c
c       We reserve all rights to this code.  Do not modify, distribute,
c       or run this code without our permission.
c
c       A very! large effort has been extended here, please do not add your 
c       name to this list unless you have dedicated a similar effort.
c
c       Parallelization details:
c       used mpi 1.2
c
c
c**************************************************************************
        use Global
	use debug
	use measureGPU
        implicit none
        include 'mpif.h'
c**************************************************************************
        integer i,j
        real(kind) ::  time1,time2, s_warmup
	real(kind) :: e_warmup, s_meas, e_meas
        character(len=1024) :: filename
c**************************************************************************
         
c
c       MPI INITIALIZATION
c

        call MPI_INIT(ierr)                             ! setup MPI
        call MPI_COMM_RANK(MPI_COMM_WORLD,myrank,ierr)  ! get myrank
        call MPI_COMM_SIZE(MPI_COMM_WORLD,nproc,ierr)   ! get the # processors

        time1=real(MPI_Wtime(),kind)
        call readin      ! Readin, open files and allocate some arrays
        call tables      ! make tables and alloc. 2-P arrays
        call sigmainit   ! Initialize sigma (read it or generate from PT).
        time_meas=zeror
        time_date=zeror
        time_put=zeror
        time_geod=zeror
        time_sumup=zeror
        write(*,*)'out of sigmainit'
        if(myrank.eq.0) call Header_bins  ! Write header for the binned data
         
  	call SAMEERGPUINIT(Nc,nl,Nblock,alpha,ntr,sgn)

        write(*,*)"Now running HFQMC, niter", niter

        do iter=1,niter
          if(isi.ge.1.and.iter.gt.1) meas=int(meas*ifac)    
          call allocate_auto !associated array sizes change in meas  
          if(myrank.eq.0)                                                   !0
     &      write(lud,*) 'iteration',iter,' runs=',nproc*run,               !0
     &        ' meas=',meas                                                 !0
          call geod         ! Make Gcstsc, G for all HHS fields s(l)=0
          call perturbation ! Find sigmapt (required GcKfsc from geod)
          call geodosft     ! Make the G's for conditioning the OSFT in put.

          write (filename,"((A,I2.2))") "beforeREG_iter_",iter
	  call printGscript(filename)

          call ginit        ! Initialize the greens functions and the fields.

c	  stop
c         Make the warmup sweeps.
          if(iter.eq.1.and.iso.eq.0) then
          s_warmup = real(MPI_Wtime(),kind)
            call update(max(20,warm))  !use a minimum of 20 warms if iter=1
          e_warmup = real(MPI_Wtime(),kind)
		print *, "Warmup took:", e_warmup - s_warmup
          else
            call update(max(skip,warm))
          end if
c         Stop after doinf the warmup
c          stop 
 
         if(myrank.eq.0.and.iprint.ge.1) then
		write(lud,*) 'done warmups';flush(lud)
	 endif

		if (umode .eq. 1 ) then 
			write (filename,"((A,I2.2))") "afterWarmup_fortran_iter_",iter
		else if (umode .eq. 4) then 
         		write (filename,"((A,I2.2))") "afterWarmup_cuda_iter_",iter
		end if
			
	  call printGuAndGd(filename)	!FOR DEBUGGING
	
          acrat=zeror    ! We only care about acrat for the measurment runs
          acount=zeror   ! We only care about acount for the measurment runs
          ucount=zeror   ! We only care about ucount for the measurment runs
c         The following is the main part of the main program.  It directs
c         measurement, and updating the lattice and the greens functions
c         between the measurements.  Note that it is PERFECTLY PARALLEL
c         unless OpenMp is used.
          nsgn=0
          do nrun=1,run
            do nmeas=1,meas
              call update(skip)
 !         	s_meas = real(MPI_Wtime(),kind)
               call measure
!            SAM: uncomment the below line to test your measurement code
	!      call cudaMeas
!		e_meas = real(MPI_Wtime(),kind)
!		print *, "Mease took:", e_meas - s_meas 
            end do           
            call sumup   ! load the data into bins and accumulators.
          end do
          call output    ! write the results into sigma.dat and data

		if (umode .eq. 1 ) then 
			write (filename,"((A,I2.2))") "afterMeas_fortran_iter_",iter
		else if (umode .eq. 4) then 
         		write (filename,"((A,I2.2))") "afterMeas_cuda_iter_",iter
		end if
			
	  call printGuAndGd(filename)	!FOR DEBUGGING

          if(myrank.eq.0) write(lud,*) 'done iteration',iter            !0
          time2=real(MPI_Wtime(),kind)
          if(myrank.eq.0) write(lud,*) 'at time',time2-time1            !0
          
          call deallocate_auto
       end do
     
        write(*,*)"HFQMC over"


        time2=real(MPI_Wtime(),kind)
        if(myrank.eq.0) then       
c         Write out the wallclock times used in different units         !0
          write(lub,*) 'cpu  ,time=  ',time2-time1                      !0
          write(lud,*) 'cpu  ,time=  ',time2-time1                      !0
          write(lud,*) 'geod ,time=  ',time_geod                        !0
          write(lud,*) 'date ,time=  ',time_date                        !0
          write(lud,*) 'meas ,time=  ',time_meas                        !0
          write(lud,*) 'put  ,time=  ',time_put                         !0
          write(lud,*) 'sumup,time=  ',time_sumup                       !0
        end if

        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
        call MPI_FINALIZE(ierr)   ! clean up
        call deallocate_1p
        if(isg.eq.1) call deallocate_2p

        if(myrank.eq.0.and.iprint.ge.1) write(lud,*) 'done all'
c       stop
        end

        subroutine abortquerry
        
c**********************************************************************
c       Check the abort flag abortf on each processor to see if
c       the calculation should be halted.  Since this is almost always
c       called only after tables (where 2P arrays are conditionally 
c       allocated), both the 1P and 2P arrays are deallocated on abort.
c**********************************************************************
        use Global
        implicit none
        include 'mpif.h'
c**********************************************************************
        integer       :: istat(MPI_STATUS_SIZE),abortfsum
c**********************************************************************

c       Make sure every proc. is here.
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
c       sum abortf to proc. 0
        call MPI_REDUCE(abortf,abortfsum,1,MPI_INTEGER,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
c       broadcast the sum to all processors
        call MPI_BCAST(abortfsum,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr) 
           
c       do we need to abort?
        if(abortfsum.ne.0) then
          if(myrank.eq.0) then
            write(lud,*) '***going down in abortquerry***',abortfsum
            write(lud,*) abortstring
            write(6,*) '***going down in abortquerry***',abortfsum
            write(6,*) abortstring
          end if
          call MPI_FINALIZE(ierr)         ! clean up
!          call deallocate_1p              ! dealocate 1P arrays
!          if(isg.eq.1) call deallocate_2p ! dealocate 2P arrays
          stop
        endif

        return
        end
