        subroutine measure
c***********************************************************************
c       This subroutine adds the present values of g, sus ... to the 
c       accumulators for a single bin.  Some measurements, such as the 
c       sign, the single-particle green function, the moment and the
c       charge cluster charge susceptiblity need to be done on each
c       iteration.  The remaining cluster susceptibilities only need
c       to be measured on the last iteration.
c***********************************************************************
        use Global
        use module_kron
        implicit none
        include 'mpif.h'
c************************************************************
        integer i,j,k,ind,imd,n,m,ic,jc,il,jl,icw,jcw,iw,jw,
     &          iQ,jQ,jcQ,icQ,neig1,neig2,indp,imdp,jk,jkp,
     &          jqk,jqkp,indk,indqk,indkp,indqkp,ik,l,lp,
     &          jmk,jmkp,indmk,indmkp,i1,i2,i3,i4,idim,i5,i6,
     &          Zbondx(Nc,nl),Zbondy(Nc,nl),Zbondxy(Nc,nl),
     &          Zbondmx(Nc,nl),Zbondmy(Nc,nl)
        real(kind) :: uno,sssm,time1,time2,r1,r2,r3,r4,r5,r6,
     &                rs,r1d,r2d,r1xx,r2xx
        complex(kindgp) :: alp,bet
!
        
        complex(kind=kindgp) :: alpha_c, beta_c
        logical, parameter :: use_original = .false.
        logical, parameter :: use_original2 = .false.
        complex(kind) :: chi_k(-nl:nl)
        integer :: ld1,ld2,ld3,ld4, nn,nna,nnb, kk
        complex(kindgp), dimension(:,:,:,:), allocatable :: T2,T3
        real(kind) :: r1r2, r1jk(Nc)
        !--------------------------------------------------------------
        ! temp arrays for susceptibilities for dynamic modulation
        ! sus_kkt,sus_kdt,sus_ddt and sus_eet
        real(kind) :: tmp_kkt(nl),tmp_kdt(nl),bow_tmp
        real(kind) :: tmp_ddt(nl),tmp_eet(nl),dbop_tmp
        ! arrays for dimer column phase and 'staggered phase',and et.al.
        complex(kind):: theta(4,Nc,0:2)
        complex(kind):: Zbond(nl,0:2)
        !--------------------------------------------------------------
        ! arrays for bond order wave susceptibility
        integer :: ib,jb,it,jt,NADJ
        integer,allocatable :: bond_link(:,:)
        complex(kind):: chiRVB_tmp(0:2),chiBOW_tmp(0:2)
        ! arrays for spin-spin Z-Z and S-S susceptibilities
        integer :: iss_flag !flag for singlet or triplet bond
        complex(kind):: chizz_tmp(0:2),chiss_tmp(0:2)
c***************************************************************************
c***************************************************************************
c       MAKE MEASUREMENTS NEEDED ON EACH ITERATION.
c***************************************************************************
c***************************************************************************

c       Measure Gd.  Due to the property G(-i)=-G(L-i) we know that the
c       time indices in the gd(i,j) matrix will have the form
c
c                        g(0)  -g(L-1)  -g(L-2)...       -g(1)
c                        g(1)   g(0)    -g(L-1)...       -g(2)
c                        g(2)   g(1)     g(0)...             .
c                         .      .         .
c                         .      .         .
c                         .      .         .
c
c
c                         .      .         .                 .
c                         .      .         .                 .
c                        g(L-1) g(L-2)...                ..g(0)
c
c       where each appearance of a particular g is a separate measurement
c       of the size Nc spatial cluster at the appropriate time indices.
c
        if(myrank.eq.0.and.iprint.ge.9) then
            write(lud,*) 'start measure';flush(lud)
        endif
        time1=real(MPI_Wtime(),kind)

        
        if(isu.eq.-1) gd=gu ! for U<0
        
        data(:,:)=zeror
        do ic=1,Nc
cdir$ concurrent
        do jc=1,Nc
        do n=1,nl
cdir$   concurrent
        do m=1,nl
#ifdef USE_F90_MERGE
           k = merge(n-m+nl+1, n-m+1, m > n)
           uno = merge(-halfr, halfr, m > n)
#else
           if (m.gt.n) then
              k = (n-m+1)+nl
              uno = -halfr
           else
              k = (n-m+1)
              uno = halfr
           endif

!            k=n-m 
!            uno=halfr
!            if(k.lt.0) then
!              uno=-halfr
!              k=k+nl
!            end if
!            k=k+1
#endif
            i=icrdiff(ic,jc)   
            ind=ic + (n-1)*Nc
            imd=jc + (m-1)*Nc
            data(k,i)=data(k,i)+(gu(ind,imd)+gd(ind,imd))*uno
        end do
        end do
        end do
        end do

        gmeas(1:nl,1:Nc)=gmeas(1:nl,1:Nc)+sgn*data(1:nl,1:Nc)

        if(iskip.eq.1.and.iauto.eq.4) then
c           Now integrate to calculate  G(k=0,wn(0)).  If we just
c           sum data over all the space time indices, then we should
c           obtain a result which is too large by Nc nl/dtau.  To obtain
c           a properly normalized result, we divide by this factor.
c           Needed for the calculation of the autocorrelation length
           r1=zeror
            do ic=1,Nc                                                 
              r2=twor/threer
              do l=1,nl            ! the integral in tau                  
                r1=r1+r2*data(l,ic)*exp(ii*wn(0)*dt*l)                  
                r2=twor-r2                                              
              end do                                                   
            end do      
            autodat(nmeas+(nrun-1)*meas)=r1*sgn*dt/real(2*nt,kind)  
            autosgn(nmeas+(nrun-1)*meas)=sgn                           
        endif

        if(iskip.eq.1.and.iauto.eq.5) then
c           Needed for the calculation of the autocorrelation length
c           of (G(t=0,r=0)
             autodat(nmeas+(nrun-1)*meas)=sgn*data(1,1)/real(nt)
             autosgn(nmeas+(nrun-1)*meas)=sgn                          
        endif


c***************************************************************************
c       Measure the sign
        sgnm=sgnm+sgn
c***************************************************************************
!       measure bond order susceptibilities
!===========================================================================
        if(ibow.ge.1.and.Nc.gt.1) then 
!
!===========================================================================
! measure bond-order-wave susceptibility for cluster
! subroutine BOW_MEAS
        if(isb.eq.8) then
           tmp_kkt = zeror
           tmp_kdt = zeror
           tmp_ddt = zeror
           tmp_eet = zeror
           bow_tmp = zeror
           dbop_tmp= zeror
           call BOW_SUSMEAS(tmp_kkt,tmp_kdt,tmp_ddt,tmp_eet,
     &bow_tmp,dbop_tmp)
           sus_kkt(:) = sus_kkt(:) + tmp_kkt(:)*sgn
           sus_kdt(:) = sus_kdt(:) + tmp_kdt(:)*sgn
           sus_ddt(:) = sus_ddt(:) + tmp_ddt(:)*sgn
           sus_eet(:) = sus_eet(:) + tmp_eet(:)*sgn
           bow_meas   = bow_meas   + bow_tmp*sgn
           dbop_meas   = dbop_meas  + dbop_tmp*sgn
        endif
!---------------------------------------------------------------------------
! create form factor or phase parameter (theta) for 2D square lattice
! for bond order calculation
        call create_theta(theta)
!---------------------------------------------------------------------------
! calculate susceptibilities for bond order:
! chiRVB <S^z.S^z> from configurations of H-S field
!---------------------------------------------------------------------------
        Zbond = zeroc
        do l = 1, nl
           do i1= 1, Nc
              i2=nneighbor(1,i1)     ! nn in the +x direction (2)  !     i3  i4
              i3=nneighbor(2,i1)     ! nn in the +y direction (3)  !     
              i4=nneighbor(2,i2)     ! nn in the +x+y direction (4)! i5  i1  i2
              i5=nneighbor(1+ndim,i1)! nn in the -x direction (2)  ! 
              i6=nneighbor(2+ndim,i1)! nn in the -y direction (2)  !     i6
              Zbond(l,:)= Zbond(l,:) 
     & +s(i1+(l-1)*Nc)*s(i2+(l-1)*Nc)*theta(1,i1,:)  
     & +s(i1+(l-1)*Nc)*s(i3+(l-1)*Nc)*theta(2,i1,:)  
     & +s(i1+(l-1)*Nc)*s(i5+(l-1)*Nc)*theta(3,i1,:)  
     & +s(i1+(l-1)*Nc)*s(i6+(l-1)*Nc)*theta(4,i1,:)
           enddo 
           chiRVB0 = chiRVB0 + Zbond(l,:)*sgn
        enddo

        do l = 1, nl
           chiRVB_tmp = zeror
           do lp = 1, nl
              k  = 1 + mod(lp+l-2,nl)
              chiRVB_tmp(:) = chiRVB_tmp(:) 
     & + conjg(Zbond(k,:))*Zbond(lp,:) 
           enddo
           chiRVB_tau(l,:) = chiRVB_tau(l,:) + real(chiRVB_tmp(:))*sgn
        enddo

!------------------------------------------------------------------------------
! 1. measure the cluster bond order wave operator (BOW)
!    B_{ij} = c^+_{i,s}c_{j,s} + c^+_{j,s}c_{i,s}, 
!    where i,j are space index and s is spin index 
!    and its susceptibility
!    <T B_{ij}(t}B_{kl}(t')>
! 2. meausure the spin bond susceptibility from
!    Wick contraction, i.e. from Green's function
!------------------------------------------------------------------------------
! NADJ: number of bond
! 1) n.n. neighbor of 2D squre, NADJ=4
! 2) n.n. neighbor of 3D cubin, NADJ=6
!
! bond_link(NADJ, Nc): link table for bond
!
        if(ndim.eq.2) NADJ=4 
        if(ndim.eq.3) NADJ=6

        allocate(bond_link(NADJ,Nc))
        do i1 = 1, Nc
           bond_link(1,i1)=nneighbor(1,i1)     ! nn in the +x direction (1)!     2
           bond_link(2,i1)=nneighbor(2,i1)     ! nn in the +y direction (2)! 3   i1  1 
           bond_link(3,i1)=nneighbor(1+ndim,i1)! nn in the -x direction (3)!     4
           bond_link(4,i1)=nneighbor(2+ndim,i1)! nn in the -y direction (4)!
           if(ndim.eq.3) then! for 3D case
              bond_link(5,i1) = nneighbor(3,i1)     !nn in the +z direction
              bond_link(6,i1) = nneighbor(3+ndim,i1)!nn in the -z direction
           endif
        enddo
       
        do l  = 1, nl
           chiBOW_tmp = zeroc
           chizz_tmp  = zeroc
           chiss_tmp  = zeroc
           do i = 1, Nc
           do ib= 1, NADJ
              il= bond_link(ib,i)

              call meas_bow(i,il,l,r1)
              chiBOW0(:) = chiBOW0(:) + r1*theta(ib,i,:)*sgn
              call meas_spinzz(i,il,l,r1)
              chizz0(:)  = chizz0(:)  + r1*theta(ib,i,:)*sgn
              iss_flag = 1 ! singlet bond
              call meas_spinSS(i,il,l,iss_flag,r1)
              chiss0(:)  = chiss0(:)  + r1*theta(ib,i,:)*sgn

              do j = 1, Nc
              do jb= 1, NADJ
                 jl=bond_link(jb,j)
                 do it = 1, nl
                    jt = 1 + mod(it+l-2,nl)

                    call bow_suscept(i,il,it,j,jl,jt,r2)
                    chiBOW_tmp(:) = chiBOW_tmp(:) 
     & + r2*conjg(theta(ib,i,:))*theta(jb,j,:)
                    call spinzz_suscept(i,il,it,j,jl,jt,r2)
                    chizz_tmp(:) = chizz_tmp(:) 
     & + r2*conjg(theta(ib,i,:))*theta(jb,j,:)
                    iss_flag = 1 ! singlet bond
                    call spinSS_suscept(i,il,it,j,jl,jt,iss_flag,r2)
                    chiss_tmp(:) = chiss_tmp(:) 
     & + r2*conjg(theta(ib,i,:))*theta(jb,j,:)

                 enddo
              enddo
              enddo
           enddo
           enddo        
           chiBOW_tau(l,:) = chiBOW_tau(l,:) + real(chiBOW_tmp(:))*sgn
           chizz_tau(l,:)  = chizz_tau(l,:)  + real(chizz_tmp(:)) *sgn
           chiss_tau(l,:)  = chiss_tau(l,:)  + real(chiss_tmp(:)) *sgn
        enddo

        deallocate(bond_link)
        end if !end if ibow
c***************************************************************************
        if(irvb.ne.0.and.Nc.gt.1.and.ndim.eq.2) then
c       Measure the z-bond moment or susceptibility.  We define the z-bond
c       moment as the sum of the near-neighbor bond singlets on the
c       cluster defined by the HHS fields which are proportional to Sz.  
c
c       M_RVB(tau) = SUM_i S(i,tau)S(i+x,tau) + S(i,tau)S(i+y,tau)
c
c                                is a linear combination of the 
c       3-----4      3      4    singlets illustrates (summed over 
c                +   |      |    sites 1 accross the cluster).  The
c                    |      |    RVB state is present if the correlations
c       1-----2      1      2    between the singlets persists for 
c                                all time (or tau [0,beta]) as measured by
c
c       M(T) = INT dtau < M_RVB(tau) M_RVB(0) > - <M_RVB>^2
c
c       Calculate <M_RVB>
c        do l=1,nl
c        do i1=1,Nc                                        ! 3    4
c         i2=nneighbor(1,i1) ! nn in the +x direction (2)  !
c         i3=nneighbor(2,i1) ! nn in the +y direction (3)  !
c         i4=nneighbor(2,i2) ! nn in the +x+y direction (4)! 1    2
c           RVB0=RVB0 + (s(i1+(l-1)*Nc)*s(i2+(l-1)*Nc)  
c     &                 +s(i1+(l-1)*Nc)*s(i3+(l-1)*Nc))
c     &               * sgn
c        end do
c        end do
c       The result is too big by Nc*nl.  Next calculate 
c       < M_RVB(tau+tau') M_RVB(tau) >'
c
c       First define the different bonds around each site in space-time

        do l=1,nl
        do i1=1,Nc
            i2=nneighbor(1,i1)     ! nn in the +x direction (2)  !     i3  i4
            i3=nneighbor(2,i1)     ! nn in the +y direction (3)  !     
            i4=nneighbor(2,i2)     ! nn in the +x+y direction (4)! i5  i1  i2
            i5=nneighbor(1+ndim,i1)! nn in the -x direction (2)  ! 
            i6=nneighbor(2+ndim,i1)! nn in the -y direction (2)  !     i6
            Zbondx(i1,l)  = s(i1+(l-1)*Nc)*s(i2+(l-1)*Nc)
            Zbondy(i1,l)  = s(i1+(l-1)*Nc)*s(i3+(l-1)*Nc)
            Zbondxy(i1,l) = s(i1+(l-1)*Nc)*s(i4+(l-1)*Nc)
            Zbondmx(i1,l) = s(i1+(l-1)*Nc)*s(i5+(l-1)*Nc)
            Zbondmy(i1,l) = s(i1+(l-1)*Nc)*s(i6+(l-1)*Nc)
        end do 
        end do

        do l=1,nl            ! tau
          r1=zeror
          r1d=zeror
          r1xx=zeror
          r3=zeror
          do i1=1,Nc                                             ! 3    4
            r1 = r1 + (Zbondx(i1,l) +Zbondy(i1,l)) ! RVB with s-wave symmetry
            r1d=r1d + (Zbondx(i1,l) -Zbondy(i1,l)) ! RVB with d-wave symmetry
            r1xx=r1xx + Zbondx(i1,l)               ! RVBxx

            rs=zeror
            do idim=1,ndim
              rs=rs+Kc(idim,icKpi)*Rc(idim,i1)
            enddo
            rs=real(exp(ii*rs))
            r3=r3+s(i1+(l-1)*Nc)*rs
          end do !i1
          r3=abs(r3)           ! AFM operator ,.i.e. ABS(S^z(Q)), Q=(pi,pi)
          do lp=1,nl           ! tau'
            k=1+mod(lp+l-2,nl) ! tau+tau'
            r2=zeror
            r2d=zeror
            r2xx=zeror
            r4=zeror    
            do i1=1,Nc                                         ! 3    4
              r2  =  r2 + (Zbondx(i1,k)+Zbondy(i1,k))   ! s-wave RVB 
              r2d = r2d + (Zbondx(i1,k)-Zbondy(i1,k))   ! d-wave RVB 
              r2xx=r2xx + Zbondx(i1,k)                  ! RVBxx
              rs=zeror
              do idim=1,ndim
                rs=rs+Kc(idim,icKpi)*Rc(idim,i1)
              enddo
              rs=real(exp(ii*rs))
              r4=r4+s(i1+(k-1)*Nc)*rs
            end do !i1
            r4=abs(r4)  ! AFM operator ,.i.e. ABS(S^z(Q)), Q=(pi,pi)

            RVB1(lp) = RVB1(lp) + r1*r2*sgn
            RVB1d(lp) = RVB1d(lp) + r1d*r2d*sgn
            RVB1xx(lp) = RVB1xx(lp) + r1xx*r2xx*sgn
            AFM1(lp)=AFM1(lp) + r3*r4*sgn
          end do !lp
          RVB0 = RVB0 + r1*sgn
          RVB0d = RVB0d + r1d*sgn
          RVB0xx = RVB0xx + r1xx*sgn
          AFM0=AFM0 + r3*sgn
        end do ! l for tau

c       Measure the Zbond xx susceptiblity
        data(:,:)=zeror
        data1(:,:)=zeror
        do ic=1,Nc ! 
        do jc=1,Nc !
            i=icrdiff(ic,jc)    
        do n=1,nl  ! 
        do m=1,nl  ! 
            k=ltaudiff(n,m)
            data(k,i)=data(k,i) + Zbondx(ic,n)*Zbondx(jc,m)
           data1(k,i)=data1(k,i) + Zbondy(ic,n)*Zbondy(jc,m)
c     &            + Zbondy(icrequ(ic, igroupxy),n)*Zbondy(icrequ(jc, igroupxy),m)
c     &                      *xyswitch
        end do
        end do
        end do
        end do
        RVBchixxm(1:nl,1:Nc)=RVBchixxm(1:nl,1:Nc)+sgn*data(1:nl,1:Nc)+
     & sgn*xyswitch*data1(1:nl,icrequ(1:Nc, igroupxy))
          
c       Measure the Zbond xy susceptibility
        data(:,:)=zeror
        data1(:,:)=zeror
        do ic=1,Nc ! 
        do jc=1,Nc !
            i=icrdiff(ic,jc)    
        do n=1,nl  ! 
        do m=1,nl  ! 
            k=ltaudiff(n,m)
            data(k,i)=data(k,i) + Zbondx(ic,n)*Zbondy(jc,m)
            data1(k,i)=data1(k,i) +Zbondy(ic,n)*Zbondx(jc,m)
c     &               + Zbondy(icrequ(ic, igroupxy),n)*Zbondx(icrequ(jc, igroupxy),m)
c     &                      *xyswitch
        end do
        end do
        end do
        end do
        RVBchixym(1:nl,1:Nc)=RVBchixym(1:nl,1:Nc)+sgn*data(1:nl,1:Nc)+
     &  sgn*xyswitch*data1(1:nl,icrequ(1:Nc, igroupxy))
        
c       Measure the HHS (Ising-like) susceptibility
        data(:,:)=zeror
        do ic=1,Nc ! 
        do jc=1,Nc !
            i=icrdiff(ic,jc)    
        do n=1,nl  ! 
        do m=1,nl  ! 
            k=ltaudiff(n,m)
            data(k,i)=data(k,i) + s(ic+(n-1)*Nc)*s(jc+(m-1)*Nc)
        end do
        end do
        end do
        end do
        ChiHHSm(1:nl,1:Nc)=ChiHHSm(1:nl,1:Nc)+sgn*data(1:nl,1:Nc)
  
        if (irvb.ge.2) then
          do i1=1,Nc      
          do l=1,nl                ! tau
            i2=nneighbor(1,i1)     ! nn in the +x direction (2)  !
            i3=nneighbor(2,i1)     ! nn in the +y direction (3)  !
            i4=nneighbor(2,i2)     ! nn in the +x+y direction (4)! 1    2
            i5=nneighbor(1+ndim,i1)! nn in the -x direction (2)  !
            i6=nneighbor(2+ndim,i1)! nn in the -y direction (2)  !       
            do lp=1,nl             ! tau'
              k=1+mod(lp+l-2,nl) ! tau+tau' 
              BB1(lp) = BB1(lp)+ Zbondx(i1,l)* Zbondx(i1,k)*sgn  ! local
              BBp1(lp) = BBp1(lp) + ( Zbondx(i1,l)*Zbondy(i1,k)  !perpendicular
     & +Zbondx(i1,l)*Zbondmy(i1,k))*sgn*halfr
              BBp2(lp) = BBp2(lp) +( Zbondx(i1,l)*s(i3+(k-1)*Nc)
     & *s(i4+(k-1)*Nc)                                           !perpendicular
     & +Zbondy(i1,l)*s(i2+(k-1)*Nc)*s(i4+(k-1)*Nc) )*sgn*halfr
                
              if (Nc.gt.4) 
     &        BBp3(lp)= BBp3(lp) +( Zbondx(i1,l)*s(i1+(k-1)*Nc)
     & *s(i5+(k-1)*Nc)                                !neighbors, same direction
     & +Zbondy(i1,l)*Zbondmy(i1,k) )*sgn*halfr
        
            end do !lp tau'
            BB0 = BB0 + ( Zbondx(i1,l)+Zbondy(i1,l) )*sgn*halfr
          enddo !l  tau
          end do !i1
        endif !(irvb.ge.2))
        end if ! (irvb.ne.0.and.Nc.gt.1.and.ndim.eq.2)

c***************************************************************************
c       Measure the cluster charge susceptibility.  It is needed to 
c       adjust ed to a fixed filing.  Calculated each iteration.
        data(:,:)=zeror

        do ic=1,Nc ! The connected graphs
cdir$ concurrent
        do jc=1,Nc !
            i=icrdiff(ic,jc)    
        do n=1,nl  ! ind /-----\ imd
!dir$   ivdep
        do m=1,nl  !     \-----/
#ifdef USE_F90_MERGE
           k = merge(n-m+nl+1, n-m+1, m > n)
#else
c            if (m .gt. n) then
c               k = (n-m+1)+nl
c            else
c               k = (n-m+1)
c            endif
           k=ltaudiff(n,m)
#endif
            ind=ic + (n-1)*Nc
            imd=jc + (m-1)*Nc
            data(k,i)=data(k,i) - 
     1               gd(ind,imd)*gd(imd,ind) -
     2               gu(ind,imd)*gu(imd,ind)
        end do
        end do
        end do
        end do

c       a fix to the r=0, tau=0 part of data(l,ic)
        do ic=1,Nc
        do n=1,nl
           ind=ic + (n-1)*Nc
           data(1,icR0)=data(1,icR0)+gu(ind,ind)+gd(ind,ind)
        end do
        end do

        do ic=1,Nc ! The disconnected graphs
cdir$ concurrent
        do jc=1,Nc !
          i=icrdiff(ic,jc)   
        do n=1,nl  ! ind /-\  /-\ imd
cdir$ ivdep
        do m=1,nl  !     \-/  \-/
#ifdef USE_F90_MERGE
          k = merge(n-m+nl+1, n-m+1, m > n)
#else
c           if (m .gt. n) then
c              k = (n-m+1) + nl
c           else
c              k = (n-m+1)
c           endif
          k=ltaudiff(n,m)
#endif
          ind=ic + (n-1)*Nc
          imd=jc + (m-1)*Nc
          data(k,i)=data(k,i)
     1              +(oner-gu(ind,ind))*(oner-gu(imd,imd))
     2              +(oner-gd(ind,ind))*(oner-gd(imd,imd))
     3              +(oner-gu(ind,ind))*(oner-gd(imd,imd))
     4              +(oner-gd(ind,ind))*(oner-gu(imd,imd))
        end do
        end do
        end do
        end do

        cham(1:nl,1:Nc)=cham(1:nl,1:Nc)+sgn*data(1:nl,1:Nc)


        if (abs(iauto).eq.3) then
         
c         Now integrate to calculate  the charge susceptibility.  If we just
c         sum data over all the space time indices, then we should
c         obtain a result which is too large by Nc nl/dtau.  To obtain
c         a properly normalized result, we divide by this factor.
c         Needed for the calculation of the autocorrelation length
          r1=zeror                                                  
          do ic=1,Nc                                                    
            r2=twor/threer
            r3=(sign(1,iauto))**sum(Rc(:,ic))
            do l=1,nl            ! the integral in tau                  
              r1=r1+r2*r3*data(l,ic)                               
              r2=twor-r2                                                
            end do                                                      
          end do                                                        
          autodat(nmeas+(nrun-1)*meas)=r1*sgn*dt/real(2*nt,kind)  
          autosgn(nmeas+(nrun-1)*meas)=sgn                           
        end if ! iauto

c***************************************************************************




c***************************************************************************
c       Measure the square average local moment
        sssm=zeror
        do ic=1,Nc
        do n=1,nl
          ind=ic + (n-1)*Nc
          sssm=sssm+halfr*(gu(ind,ind)+gd(ind,ind))-
     &                     gu(ind,ind)*gd(ind,ind)
        end do
        end do
        moom=moom+sssm*sgn
c***************************************************************************
        

c***************************************************************************
        if ((iskip.eq.1.and.abs(iauto).eq.1).or.iter.eq.niter) then
          
c         Measure the cluster longitudinal (zz) spin susceptibility. If
c         iskip=1 and iauto=1, is needed each iteration since it is used 
c         to calculate the autocorrelation length.

          data(:,:)=zeror

          do ic=1,Nc ! The connected graphs
cdir$ concurrent
          do jc=1,Nc !
             i=icrdiff(ic,jc)    
          do n=1,nl  ! ind /-----\ imd
cdir$ ivdep
          do m=1,nl  !     \-----/
#ifdef USE_F90_MERGE
             k = merge(n-m+nl+1, n-m+1, m > n)
#else
c             if (m.gt.n) then
c                k = (n-m+1) +nl
c             else
c                k = (n-m+1)
c             endif
             k=ltaudiff(n,m)
#endif
             ind=ic + (n-1)*Nc
             imd=jc + (m-1)*Nc
             data(k,i)=data(k,i) - 
     &                 gd(ind,imd)*gd(imd,ind) -
     &                 gu(ind,imd)*gu(imd,ind)
          end do
          end do
          end do
          end do

c         a fix to the r=0, tau=0 part of data(l,ic)
          do ic=1,Nc
          do n=1,nl
             ind=ic + (n-1)*Nc
             data(1,icR0)=data(1,icR0)+gu(ind,ind)+gd(ind,ind)
          end do
          end do

          do ic=1,Nc ! The disconnected graphs
cdir$ concurrent
          do jc=1,Nc !
            i=icrdiff(ic,jc)   
          do n=1,nl  ! ind /-\  /-\ imd
cdir$ ivdep
          do m=1,nl  !     \-/  \-/
#ifdef USE_F90_MERGE
            k = merge(n-m+nl+1, n-m+1, m > n)
#else
c             if (m .gt. n) then
c               k = (n-m+1) + nl
c             else
c               k = (n-m+1)
c             endif
            k=ltaudiff(n,m)
#endif
            ind=ic + (n-1)*Nc
            imd=jc + (m-1)*Nc
            data(k,i)=data(k,i)
     1                +(oner-gu(ind,ind))*(oner-gu(imd,imd))
     2                +(oner-gd(ind,ind))*(oner-gd(imd,imd))
     3                -(oner-gu(ind,ind))*(oner-gd(imd,imd))
     4                -(oner-gd(ind,ind))*(oner-gu(imd,imd))
          end do
          end do
          end do
          end do

          chmeas(1:nl,1:Nc)=chmeas(1:nl,1:Nc)+sgn*data(1:nl,1:Nc)
       
          if (abs(iauto).eq.1) then
           
c           Now integrate to calculate the susceptibility.  If we just
c           sum data over all the space time indices, then we should
c           obtain a result which is too large by Nc nl/dtau.  To obtain
c           a properly normalized result, we divide by this factor.

            r1=zeror
            do ic=1,Nc
              r2=twor/threer
              r3=(sign(1,iauto))**sum(Rc(:,ic))
              do l=1,nl      ! the integral in tau                
                r1=r1+r2*r3*data(l,ic)
                r2=twor-r2
              end do
            end do
            autodat(nmeas+(nrun-1)*meas)=r1*sgn*dt/real(2*nt,kind)  
            autosgn(nmeas+(nrun-1)*meas)=sgn
          end if ! iauto
        end if 

c***************************************************************************
c
c       Some measurements are more easily done with G(K,K',t,t')
c       rather than with G(R,t;R',t').  These include:
c
c       scmeas  cluster pair susceptibility, needed if iauto=2 or isb>=5
c       ccmeas  cluster CC and B2g Raman susceptibility, needed if isb>=4
c       b1gmeas cluster B1g Raman susceptibility, needed if isb>=6
c       b2gmeas cluster B2g Raman susceptibility, needed if isb>=7

        if((iskip.eq.1.and.iauto.eq.2).or.
     &     (iter.eq.niter.and.isg.ge.1).or.
     &     (iter.eq.niter.and.isb.ge.4).or.
     &     (iter.eq.niter.and.iauto.eq.2)) then

          call FT_RRp_toKKp(gu,temat1,temat2)  ! Gup(R,t,R',t')-->Gup(K,t,K',t')
          call FT_RRp_toKKp(gd,temat1,temat3)  ! Gdn(R,t,R',t')-->Gdn(K,t,K',t')    
          
c         Measure the cluster d-wave or s-wave pair field susceptibility (D-independent).
c
c             up
c      Q+K  --<---  Q+K'
c  g(K')                g(K) = cos(Kx)-cos(Ky)
c  tau                  tau'
c       -K  --<---  -K'
c            down
c
c         g(K) Gu(Q+K,tau;Q+K',tau') Gd(-K,tau;-K',tau') g(K)
c
          chi(:,:) = zeroc
          if (use_original2) then

          do ic=1,Nc    !Q
          do jk =1,Nc   !K
            jmk=ickdiff(ick0,jk) !-K
            r1=gdwave(jk)*halfr*(1+isu)+halfr*(1-isu) ! g=gd (gs) for U>(<)0
          do jkp=1,Nc   !K'
            jmkp=ickdiff(ick0,jkp) !-K'
            r2=gdwave(jkp)*halfr*(1+isu)+halfr*(1-isu)
            jqk  = ickplus(ic,jk)  ! K+Q
            jqkp = ickplus(ic,jkp) ! K'+Q



          do n=1,nl     !tau
            indk   = jk   + (n-1)*Nc ! K, tau
            indmk  = jmk  + (n-1)*Nc ! -K, tau
            indqk  = jqk  + (n-1)*Nc ! K+Q, tau
          do m=1,n     !tau'
            k=n-m+1    ! tau-tau'
            indkp  = jkp  + (m-1)*Nc ! K', tau' 
            indmkp = jmkp + (m-1)*Nc ! -K', tau' 
            indqkp = jqkp + (m-1)*Nc ! K'+Q, tau'
            
            chi(k,ic)=chi(k,ic) + r1*r2*( ! Cluster d-wave or s-wave susc.
c                    up K+Q,tau; K'+Q,tau'   dn   -K,tau;-K',tau'
     &                temat2(indqk,indqkp)*temat3(indmk,indmkp) +
c                    dn K+Q,tau; K'+Q,tau'   up   -K,tau;-K',tau'              
     &                temat3(indqk,indqkp)*temat2(indmk,indmkp) )
          end do
          do m=n+1,nl     !tau'
            k=n-m+1+nl    !tau-tau' (wrapped due to PBC)
            indkp  = jkp  + (m-1)*Nc ! K', tau'
            indmkp = jmkp + (m-1)*Nc ! -K', tau' 
            indqkp = jqkp + (m-1)*Nc ! K'+Q, tau'
            
            chi(k,ic)=chi(k,ic) + r1*r2*( ! Cluster d-wave or s-wave susc.
c                    up K+Q,tau; K'+Q,tau'   dn   -K,tau;-K',tau'
     &                temat2(indqk,indqkp)*temat3(indmk,indmkp) +
c                    dn K+Q,tau; K'+Q,tau'   up   -K,tau;-K',tau'              
     &                temat3(indqk,indqkp)*temat2(indmk,indmkp) )
          end do
          end do

          end do
          end do
          end do


          else ! optimized cluster pair-field susceptibility

!         --------------------------------         
!         rearrange data in temat2, temat3
!         --------------------------------         
          allocate( T2(nl,nl,Nc,Nc), T3(nl,nl,Nc,Nc) )
          do jkp=1,Nc
          do jk=1,Nc
          do m=1,nl
          do n=1,nl

            indk = jk + (n-1)*Nc 
            indkp = jkp + (m-1)*Nc 

!           ------------------------------------------
!           T2(n,m,jk,jkp) = temat2( [jk,n], [jkp,m] )
!           T3(n,m,jk,jkp) = temat3( [jk,n], [jkp,m] )
!           ------------------------------------------

            T2(n,m,jk,jkp) = temat2( indk, indkp )
            T3(n,m,jk,jkp) = temat3( indk, indkp )
          enddo
          enddo
          enddo
          enddo

c         make the d or s wave form factor
          r1jk(1:Nc) = gdwave(1:Nc)*halfr*(1+isu)+halfr*(1-isu)
                       
c         Now calculate the susceptibility          
          chi(:,:) = zeroc
          do ic=1,Nc
          do jk=1,Nc
          do jkp=1,Nc
            r1 = r1jk(jk)
            r2 = r1jk(jkp)
            jmk  = ickdiff(ick0,jk)
            jmkp = ickdiff(ick0,jkp)
            jqk  = ickplus(ic,jk)
            jqkp = ickplus(ic,jkp)

            chi_k(:) = zeroc

          do m=1,nl
          do n=1,nl
            k = n-m

!      ----------------------------------------
!  original code performed
!      temat2( [s(ic,jk),n],[s(ic,jk'),m] ) * 
!      temat3( [jk,n],      [jk',m] ) 
!      +
!      temat3( [s(ic,jk),n],[s(ic,jk'),m] ) * 
!      temat2( [jk,n],      [jk',m] ) 
!      ----------------------------------------
            
            chi_k(k)=chi_k(k) + ( ! Cluster d-wave or s-wave susc.
c                     up     K+Q K'+Q  dn     -K    -K'
     &                T2(n,m,jqk,jqkp)*T3(n,m,jmk,jmkp) +
c                     dn     K+Q K'+Q  up     -K    -K'
     &                T3(n,m,jqk,jqkp)*T2(n,m,jmk,jmkp) )
           enddo
           enddo
      


               do k=0,(nl-1)
                  kk = k+1 
                  chi(kk,ic) = chi(kk,ic) + (r1*r2)*chi_k(k)
               enddo
               do k=-(nl-1),-1
                  kk = k+nl+1 
                  chi(kk,ic) = chi(kk,ic) + (r1*r2)*chi_k(k)
               enddo


          end do
          end do
          end do

          deallocate( T2, T3  )

          endif

c         FT back to cluster coordinates.  It should be real.
          data=zeror
          do n=1,nl
          do ic=1,Nc
          do ik=1,Nc
            data(n,ic)=data(n,ic)+FTCoefs_K_to_R(ik,ic)*chi(n,ik)
          end do
          end do
          end do
          scmeas(1:nl,1:Nc)= scmeas(1:nl,1:Nc) + 
     &                       sgn*data(1:nl,1:Nc)/(Nc*Nc)

          if (iauto.eq.2) then
c         Now integrate to calculate the susceptibility.  If we just
c         sum data over all the space time indices, then we should
c         obtain a result which is too large by 32 Nc^2 nl/dtau.  To obtain
c         a properly normalized result, we divide by this factor.
            r1=zeror
            do ic=1,Nc
              r2=twor/threer
              do l=1,nl   ! the integral in tau           
                r1=r1+r2*data(l,ic)  
                r2=twor-r2  
              end do 
            end do  
            autodat(nmeas+(nrun-1)*meas)=r1*sgn*dt/real(32*nt*Nc,kind)  
            autosgn(nmeas+(nrun-1)*meas)=sgn
          end if ! iauto
        end if 
          
        if(iter.eq.niter.and.isg.ge.1.or.
     &     iter.eq.niter.and.isb.ge.1) then

c       CONDITIONALLY MAKE MEASUREMENTS NEEDED ONLY ON THE LAST ITERATION.

c         Measure the cluster circulating current (Q=(pi,pi)) in ccmeas and 
c         Raman B1g susceptibility in b1gmeas and the B2g susceptibility in 
c         b2gmeas
c
c              up
c       Q+K  -->--- Q+K'
c  g(K')                g(K) = cos(Kx)-cos(Ky)
c  tau                  tau'
c         K  --<--- K'
c              up
c
c
c         g(K) Gu(K'+Q,tau';K+Q,tau) Gu(K,tau;K',tau') g(K')'
c
          chi(:,:) = zeroc
          chi2(:,:) = zeroc
          chi3(:,:) = zeroc
          do ic=1,Nc    !Q
cdir$ concurrent
          do jk =1,Nc   !K
            r1=gdwave(jk)
            r3=gB1g(jk,ic)
            r5=gB2g(jk,ic)
          do jkp=1,Nc   !K'
            r2=gdwave(jkp)
            r4=gB1g(jkp,ic)
            r6=gB2g(jkp,ic)
              jqk  = ickplus(ic,jk)
              jqkp = ickplus(ic,jkp)

          if (use_original) then

          do n=1,nl     !tau
              indk   = jk   + (n-1)*Nc
              indqk  = jqk  + (n-1)*Nc
          do m=1,n     !tau'
              k=n-m+1
              indkp  = jkp  + (m-1)*Nc
              indqkp = jqkp + (m-1)*Nc
              
              chi(k,ic)=chi(k,ic) - r1*r2*(     ! - due to closed Fermi loop
     &                  temat2(indk,indkp)*temat2(indqkp,indqk) +
     &                  temat3(indk,indkp)*temat3(indqkp,indqk) )
              chi2(k,ic)=chi2(k,ic) - r3*r4*(     ! - due to closed Fermi loop
     &                  temat2(indk,indkp)*temat2(indqkp,indqk) +
     &                  temat3(indk,indkp)*temat3(indqkp,indqk) )
              chi3(k,ic)=chi3(k,ic) - r5*r6*(     ! - due to closed Fermi loop
     &                  temat2(indk,indkp)*temat2(indqkp,indqk) +
     &                  temat3(indk,indkp)*temat3(indqkp,indqk) )
          end do
          do m=n+1,nl     !tau'
              k=n-m+1+nl
              indkp  = jkp  + (m-1)*Nc
              indqkp = jqkp + (m-1)*Nc
              
              chi(k,ic)=chi(k,ic) - r1*r2*(     ! - due to closed Fermi loop
     &                  temat2(indk,indkp)*temat2(indqkp,indqk) +
     &                  temat3(indk,indkp)*temat3(indqkp,indqk) )
              chi2(k,ic)=chi2(k,ic) - r3*r4*(     ! - due to closed Fermi loop
     &                  temat2(indk,indkp)*temat2(indqkp,indqk) +
     &                  temat3(indk,indkp)*temat3(indqkp,indqk) )
              chi3(k,ic)=chi3(k,ic) - r5*r6*(     ! - due to closed Fermi loop
     &                  temat2(indk,indkp)*temat2(indqkp,indqk) +
     &                  temat3(indk,indkp)*temat3(indqkp,indqk) )
          end do
          end do

          else ! Use the optimzed version

          chi_k(:) = zeror
          do n=1,nl
            indk   = jk   + (n-1)*Nc
            indqk  = jqk  + (n-1)*Nc

!dir$      concurrent
            do m=1,nl

            k = n-m
            indkp  = jkp  + (m-1)*Nc
            indqkp = jqkp + (m-1)*Nc
            
            chi_k(k) = chi_k(k) - (     ! - due to closed Fermi loop
     &                  temat2(indk,indkp)*temat2(indqkp,indqk) +
     &                  temat3(indk,indkp)*temat3(indqkp,indqk) )
            enddo
           enddo

           do k=0,(nl-1)
              kk = k+1
              chi(kk,ic) = chi(kk,ic) + (r1*r2)*chi_k(k)
              chi2(kk,ic) = chi2(kk,ic) + (r3*r4)*chi_k(k)
              chi3(kk,ic) = chi3(kk,ic) + (r5*r6)*chi_k(k)
           enddo
           do k=-(nl-1),-1
              kk = k+1+nl
              chi(kk,ic) = chi(kk,ic) + (r1*r2)*chi_k(k)
              chi2(kk,ic) = chi2(kk,ic) + (r3*r4)*chi_k(k)
              chi3(kk,ic) = chi3(kk,ic) + (r5*r6)*chi_k(k)
           enddo
          endif


          end do
          end do
          end do

          
c         We need to fix the tau=tau' K=K' limit of the cluster susc.
          do ic=1,Nc    !Q
          do jk =1,Nc   !K
            r1=gdwave(jk)
            r3=gB1g(jk,ic)
            r5=gB2g(jk,ic)
          do n=1,nl     !tau
            indk   = jk   + (n-1)*Nc

            chi(1,ic)=chi(1,ic) + r1*r1*( temat2(indk,indk) +
     &                                    temat3(indk,indk)    )
            chi2(1,ic)=chi2(1,ic) + r3*r3*( temat2(indk,indk) +
     &                                    temat3(indk,indk)    )
            chi3(1,ic)=chi3(1,ic) + r5*r5*( temat2(indk,indk) +
     &                                    temat3(indk,indk)    )
          end do
          end do
          end do
          
c         Calculate the disconnected parts of the cluster cc susc.
c
c             
c       Q+K  ->-    ->- Q+K'                   
c  g(K')        \  /      g(K) = cos(Kx)-cos(Ky)
c  tau          /  \      tau'
c         K  -<-    -<-   K'
c             
c
c         g(K) Gu(K,tau;Q+K,tau) Gu(Q+K',tau';K',tau') g(K')'
c
c         When Q=0, this should be
c
c         g(K) [Gu(K,tau;K,tau)-1] [Gu(K',tau';K',tau')-1] g(K')'
c
c         however, as SUM_K g(K) =0 we may neglect the -1
c
          do ic=1,Nc    !Q
cdir$ concurrent
          do jk =1,Nc   !K
            r1=gdwave(jk)
            r3=gB1g(jk,ic)
            r5=gB2g(jk,ic)
          do jkp=1,Nc   !K'
            r2=gdwave(jkp)
            r4=gB1g(jkp,ic)
            r6=gB2g(jkp,ic)
            jqk  = ickplus(ic,jk)
            jqkp = ickplus(ic,jkp)
          do n=1,nl     !tau
            indk   = jk   + (n-1)*Nc
            indqk  = jqk  + (n-1)*Nc
cdir$ concurrent
          do m=1,n     !tau'
            k=n-m+1
            indkp  = jkp  + (m-1)*Nc
            indqkp = jqkp + (m-1)*Nc
            
            chi(k,ic)=chi(k,ic) + r1*r2*(         
     &                temat2(indk,indqk)*temat2(indqkp,indkp) +
     &                temat3(indk,indqk)*temat3(indqkp,indkp) +
     &                temat2(indk,indqk)*temat3(indqkp,indkp) +
     &                temat3(indk,indqk)*temat2(indqkp,indkp) )
            chi2(k,ic)=chi2(k,ic) + r3*r4*(         
     &                temat2(indk,indqk)*temat2(indqkp,indkp) +
     &                temat3(indk,indqk)*temat3(indqkp,indkp) +
     &                temat2(indk,indqk)*temat3(indqkp,indkp) +
     &                temat3(indk,indqk)*temat2(indqkp,indkp) )
            chi3(k,ic)=chi3(k,ic) + r5*r6*(         
     &                temat2(indk,indqk)*temat2(indqkp,indkp) +
     &                temat3(indk,indqk)*temat3(indqkp,indkp) +
     &                temat2(indk,indqk)*temat3(indqkp,indkp) +
     &                temat3(indk,indqk)*temat2(indqkp,indkp) )
          end do
          do m=n+1,nl     !tau'
            k=n-m+nl+1
            indkp  = jkp  + (m-1)*Nc
            indqkp = jqkp + (m-1)*Nc
            
            chi(k,ic)=chi(k,ic) + r1*r2*(         
     &                temat2(indk,indqk)*temat2(indqkp,indkp) +
     &                temat3(indk,indqk)*temat3(indqkp,indkp) +
     &                temat2(indk,indqk)*temat3(indqkp,indkp) +
     &                temat3(indk,indqk)*temat2(indqkp,indkp) )
            chi2(k,ic)=chi2(k,ic) + r3*r4*(         
     &                temat2(indk,indqk)*temat2(indqkp,indkp) +
     &                temat3(indk,indqk)*temat3(indqkp,indkp) +
     &                temat2(indk,indqk)*temat3(indqkp,indkp) +
     &                temat3(indk,indqk)*temat2(indqkp,indkp) )
            chi3(k,ic)=chi3(k,ic) + r5*r6*(         
     &                temat2(indk,indqk)*temat2(indqkp,indkp) +
     &                temat3(indk,indqk)*temat3(indqkp,indkp) +
     &                temat2(indk,indqk)*temat3(indqkp,indkp) +
     &                temat3(indk,indqk)*temat2(indqkp,indkp) )
          end do
          end do
          end do
          end do
          end do
          
c         FT cc back to cluster coordinates
          data=zeror
          do n=1,nl
          do ic=1,Nc
          do ik=1,Nc
            data(n,ic)=data(n,ic)+FTCoefs_K_to_R(ik,ic)*chi(n,ik)
          end do
          end do
          end do
          ccmeas(1:nl,1:Nc) = ccmeas(1:nl,1:Nc) + sgn*data(1:nl,1:Nc)
c         FT b1g back to cluster coordinates
          data=zeror
          do n=1,nl
          do ic=1,Nc
          do ik=1,Nc
            data(n,ic)=data(n,ic)+FTCoefs_K_to_R(ik,ic)*chi2(n,ik)
          end do
          end do
          end do
          b1gmeas(1:nl,1:Nc) = b1gmeas(1:nl,1:Nc) + sgn*data(1:nl,1:Nc)
c         FT b2g back to cluster coordinates
          data=zeror
          do n=1,nl
          do ic=1,Nc
          do ik=1,Nc
            data(n,ic)=data(n,ic)+FTCoefs_K_to_R(ik,ic)*chi3(n,ik)
          end do
          end do
          end do
          b2gmeas(1:nl,1:Nc) = b2gmeas(1:nl,1:Nc) + sgn*data(1:nl,1:Nc)
        end if

c       Conditionally measure the two-particle susceptibility bubbles.
        if(iter.eq.niter.and.isg.ge.1) then

          alp=(1.0_kindgp,0.0_kindgp)
          bet=(0.0_kindgp,0.0_kindgp)

c         Measure the dynamic two-particle greens functions.  These will
c         be formed from products of the QMC single-particle green
c         function G(k,k',iwn,iwn') corresponding to a particular
c         configuration of the HHS fields {s} and disorder fields {V}.
c
c         first form the temporary vectors which store G_u-1/2I
          temat3=gu
          do i=1,nt
            temat3(i,i)=temat3(i,i)-halfr
          end do
c                    +
c         temat2=fmat * temat3
cint!     temat2=matmul(conjg(transpose(fmat)),temat3)

          alpha_c = alp
          beta_c = bet
          if (use_original) then
          call cgemm(Cstring,Nstring,nt,nt,nt,alpha_c,
     &               fmat,nt,temat3,nt,beta_c,temat2,nt)
          else

          nn = nt
          nna = nl
          nnb = Nc

          ld1 = size(Afmat,1)
          ld2 = size(Bfmat,1)
          ld3 = size(temat3,1)
          ld4 = size(temat2,1)

          call kron_matmulL(Cstring,Nstring,nn,nna,nnb,
     &       alpha_c, Afmat,ld1, Bfmat,ld2, temat3,ld3, 
     &       beta_c, temat2,ld4)


          endif
c                                     +
c         temat3= temat2 * fmat = fmat  * G_up * fmat
cint!     temat3=matmul(temat2,fmat)

          alpha_c = alp
          beta_c = bet
          if (use_original) then

          call cgemm(Nstring,Nstring,nt,nt,nt,alpha_c,temat2,nt,
     &               fmat,nt,beta_c,temat3,nt)

          else

          nn = nt
          nna = nl
          nnb = Nc

          ld1 = size(temat2,1)
          ld2 = size(Afmat,1)
          ld3 = size(Bfmat,1)
          ld4 = size(temat3,1)

          call kron_matmulR( Nstring, Nstring, nn,nna,nnb,
     &        alpha_c, temat2,ld1, Afmat,ld2, Bfmat,ld3,   
     &        beta_c, temat3,ld4 )


          endif


c         then form the temporary vectors which store G_d-1/2I
          temat1=gd
          do i=1,nt
            temat1(i,i)=temat1(i,i)-halfr
          end do
c                    +
c         temat2=fmat * temat1
cint!     temat2=matmul(conjg(transpose(fmat)),temat1)


          alpha_c = alp
          beta_c = bet
          if (use_original) then

          call cgemm(Cstring,Nstring,nt,nt,nt,alpha_c,
     &               fmat,nt,temat1,nt,beta_c,temat2,nt)

          else

          nn = nt
          nna = nl
          nnb = Nc

          ld1 = size(Afmat,1)
          ld2 = size(Bfmat,1)
          ld3 = size(temat1,1)
          ld4 = size(temat2,1)

          call kron_matmulL(Cstring,Nstring,nn,nna,nnb,
     &       alpha_c, Afmat,ld1, Bfmat,ld2, temat1,ld3, 
     &       beta_c, temat2,ld4)

          endif

c                                     +
c         temat1= temat2 * fmat = fmat  * G_d * fmat
cint!     temat1=matmul(temat2,fmat)

          alpha_c = alp
          beta_c = bet
          if (use_original) then

          call cgemm(Nstring,Nstring,nt,nt,nt,alpha_c,
     &               temat2,nt,fmat,nt,beta_c,temat1,nt)

          else

          nn = nt
          nna = nl
          nnb = Nc

          ld1 = size(temat2,1)
          ld2 = size(Afmat,1)
          ld3 = size(Bfmat,1)
          ld4 = size(temat1,1)


          call kron_matmulR( Nstring, Nstring, nn,nna,nnb,
     &        alpha_c, temat2,ld1, Afmat,ld2, Bfmat,ld3,   
     &        beta_c, temat1,ld4 )

          endif


c                +
c         Thus,(F G    F)_q is in temat1
c                  down
c         and
c                +
c              (F G  F)_q=0 is in temat3
c                  up
c
c         The connected parts of the susceptibilities have the general form:
c
c         k+Q       k'+Q     k+Q       k'+Q
c         -<----------<-     -<----------<-
c        particle-particle   particle hole
c         -<----------<-     ->---------->-
c         -k         -k'      k         k'
c
c
c         So, eg. for the particle-hole part, we need to form the direct 
c         product of temat3 and the transpose of temat1.
          do ic=1,Nc
cdir$ concurrent
          do jc=1,Nc
cdir$ concurrent
          do il=1,nl
cdir$ concurrent
          do jl=1,nl
            i = ic + (il-1)*Nc
            j = jc + (jl-1)*Nc
            icw = ick2map(ic,ic)       ! confined to the wedge
            jcw = ick2map(ic,jc)       ! not confined to the wedge
            iw = icw + (il-1)*Ncw      ! confined to the wedge
            jw = jcw + (jl-1)*Nc
            
            icQ = ickplus(ic,icKpi)    ! indices for k+Q 
            jcQ = ickplus(jc,icKpi)    ! where Q=(pi,pi)
            iQ = icQ + (il-1)*Nc
            jQ = jcQ + (jl-1)*Nc
c            gamatcenter(iw,jw)=gamatcenter(iw,jw)-         !ie Chi +-
c     &                       temat3(i,j)*temat1(j,i)*sgn
c            gamatcorner(iw,jw)=gamatcorner(iw,jw)-
c     &                       temat3(i,j)*temat1(jQ,iQ)*sgn
            gamatcenter(iw,jw)=gamatcenter(iw,jw)-
     &                         (temat1(i,j)*temat1(j,i)+
     &                          temat3(i,j)*temat3(j,i))*sgn*halfr
c           k+Q       k'+Q
c           -<----------<-    |
c           particle hole     |
c           ->---------->-   \ /
c            k         k'     *
            gamatcorner(iw,jw)=gamatcorner(iw,jw)-
     &                       (temat1(i,j)*temat1(jQ,iQ)+
     &                        temat3(i,j)*temat3(jQ,iQ))*sgn*halfr     
            gamatscenter(iw,jw)=gamatscenter(iw,jw)+real(
     &               halfr*(temat3(i,j)*conjg(temat1(i,j))+
     &               temat1(i,j)*conjg(temat3(i,j)))*sgn,kindg)
c            k+Q       k'+Q            k          k'
c       -s   -<----------<-       s   -<----------<-    |
c           particle-particle  +                        |
c       +s   -<----------<-      -s   -<----------<-   \ /
c            -k         -k'           -k-Q     -k'-Q    *
            gamatscorner(iw,jw)=gamatscorner(iw,jw)+ 
     &               halfr*(temat3(i,j)*conjg(temat1(iQ,jQ))+
     &               temat1(iQ,jQ)*conjg(temat3(i,j)))*sgn
            gamatccenter(iw,jw)=gamatccenter(iw,jw)-
     &                          (temat1(i,j)*temat1(j,i)+
     &                           temat3(i,j)*temat3(j,i))*sgn
c           k+Q       k'+Q
c           -<----------<-    |
c           particle hole     |
c           ->---------->-   \ /
c            k         k'     *
            gamatccorner(iw,jw)=gamatccorner(iw,jw)-
     &                        (temat1(i,j)*temat1(jQ,iQ)+
     &                         temat3(i,j)*temat3(jQ,iQ))*sgn
          end do
          end do
          end do
          end do

c         Measure the disconnected parts of the charge and 
c         spin diagrams.  They have the general form:
c
c         k+Q       k'+Q     k+Q       k'+Q
c         -<---    ---<-     -<---    ---<-
c         pp  |    |         ph  |    |
c         -<---    ---<-     ->---    --->-
c         -k         -k'      k         k'
c
          do ic=1,Nc
cdir$ concurrent
          do jc=1,Nc
cdir$ concurrent
          do il=1,nl
cdir$ concurrent
          do jl=1,nl
            i = ic + (il-1)*Nc
            j = jc + (jl-1)*Nc
            icw = ick2map(ic,ic)        ! confined to the wedge
            jcw = ick2map(ic,jc)        ! not confined to the wedge
            iw = icw + (il-1)*Ncw       ! confined to the wedge
            jw = jcw + (jl-1)*Nc
            
            icQ = ickplus(ic,icKpi)
            jcQ = ickplus(jc,icKpi)
            iQ = icQ + (il-1)*Nc
            jQ = jcQ + (jl-1)*Nc
            
            gamatccenter(iw,jw)=gamatccenter(iw,jw)
     &                         +(temat3(i,i)+temat1(i,i))*
     &                          (temat3(j,j)+temat1(j,j))*sgn
            gamatccorner(iw,jw)=gamatccorner(iw,jw)
     &                         +(temat3(i,iQ)+temat1(i,iQ))*
     &                          (temat3(jQ,j)+temat1(jQ,j))*sgn
            gamatcenter(iw,jw)=gamatcenter(iw,jw)
     &                        +(temat3(i,i)-temat1(i,i))*
     &                         (temat3(j,j)-temat1(j,j))*sgn*halfr
            gamatcorner(iw,jw)=gamatcorner(iw,jw)
     &                       +(temat3(i,iQ)-temat1(i,iQ))*
     &                        (temat3(jQ,j)-temat1(jQ,j))*sgn*halfr
         end do
         end do
         end do
         end do
c
c        Finally calculate the vacuum part (only finite for 
c        charge sus).  They have the general form:
c
c        k+Q    
c        -<---  
c        ph  |  
c        ->---  
c         k     
c

         do ic=1,Nc
cdir$ concurrent
         do il=1,nl
           i = ic + (il-1)*Nc
           icw = ick2map(ic,ic)        ! confined to the wedge
           iw = icw + (il-1)*Ncw       ! confined to the wedge
           icQ = ickplus(ic,icKpi)
           iQ = icQ + (il-1)*Nc
           gamatc0center(iw)=gamatc0center(iw)+
     &                       (temat3(i,i)+temat1(i,i))*sgn
           gamatc0corner(iw)=gamatc0corner(iw)+
     &                       (temat3(iQ,i)+temat1(iQ,i))*sgn
         end do
         end do

        end if  ! iter.eq.niter.and.isg.ge.1

        time2=real(MPI_Wtime(),kind)
        time_meas=time_meas+time2-time1

        if(myrank.eq.0.and.iprint.ge.9) then
            write(lud,*) 'done measure';flush(lud)
        endif
        return
        end

        subroutine FT_RRp_toKKp(green,Mat1,Mat2)
c**********************************************************************
c       Fourier Transform green(X,l;X',l') to Mat2(k,l;K',l').   This
c       subroutine scales like Nc^3 nl^2, and green is the input, Mat1 is 
c       a work array and Mat2 is the output.
c       note: precission difference kind and kindgp (i.e. Mat1 vs Green)
c**********************************************************************
        use Global
        implicit none
c**********************************************************************
        integer         :: l,lp,ik,ikp,ix,ixp,indx,indxp,indk,indkp    
        real(kind)      :: green(nt,nt)
        complex(kindgp) :: Mat1(nt,nt), Mat2(nt,nt)
!

        logical, parameter :: use_original = .false.

        complex(kindgp)    :: M1(Nc,Nc),M2(Nc,Nc),Gtmp(Nc,Nc)
        complex(kindgp)    :: F(Nc,Nc), FH(Nc,Nc)

        integer :: ik_lo,ik_hi, ikp_lo,ikp_hi
        integer :: ix_lo,ix_hi, ixp_lo,ixp_hi
        integer :: ixl,ixplp, ikl,ikplp, ioffset
        
c**********************************************************************
        
        if (use_original) then
c       G(R,t; R',t') --> G(K,t; R',t')

        Mat1=zeroc
        do l=1,nl             ! t
        do lp=1,nl            ! t'
        do ik=1,Nc            ! K
        do ixp=1,Nc           ! R'
          indk=ik+(l-1)*Nc    ! K,t
          indxp=ixp+(lp-1)*Nc ! R',t'
          do ix=1,Nc          ! sum over R
            indx=ix+(l-1)*Nc  ! R,t
            Mat1(indk,indxp)=Mat1(indk,indxp)+
     &                       FTCoefs_R_to_K(ik,ix)*green(indx,indxp)
          end do
        end do
        end do 
        end do
        end do
        
c       G(K,t; R',t') --> G(K,t; K',t')

        Mat2=zeroc
        do l=1,nl               ! t
        do lp=1,nl              ! t'
        do ik=1,Nc              ! K
        do ikp=1,Nc             ! K'
          indk=ik+(l-1)*Nc      ! K,t
          indkp=ikp+(lp-1)*Nc   ! K',t'
          do ixp=1,Nc           ! R'
            indxp=ixp+(lp-1)*Nc ! R',t'
            Mat2(indk,indkp)=Mat2(indk,indkp)+
     &                       conjg(FTCoefs_R_to_K(ikp,ixp))*
     &                       Mat1(indk,indxp)
          end do
        end do
        end do 
        end do
        end do



        else

!       ------------------------------
!       use matrix multiply to compute
!       Mat1 and Mat2
!       ------------------------------
!

        ik_lo = 1
        ik_hi = Nc
        ikp_lo = 1
        ikp_hi = Nc
        ix_lo = 1
        ix_hi = Nc
        ixp_lo = 1
        ixp_hi = Nc

        F =  FTCoefs_R_to_K
        do ik=ik_lo,ik_hi
        do ix=ix_lo,ix_hi
          FH(ix,ik) = conjg(F(ik,ix))
        enddo
        enddo

        do lp=1,nl
        do l=1,nl
!          --------------------------
!          copy Gtmp(ix,ixp) <- green( [ix,l], [ixp,lp] )
!          --------------------------
           do ixp=ixp_lo,ixp_hi
             ixplp = ixp + (lp-1)*Nc
             ioffset = (l-1)*Nc

             do ix=ix_lo,ix_hi
               ixl = ioffset + ix 
               Gtmp(ix,ixp) = green( ixl, ixplp )
             enddo
           enddo

!          --------------------------------------------
!          compute part of Mat1([ik,l],[ixp,lp]) = 
!                       matmul(F(ik,ix),Gtmp( [ix,ixp] ) )
!          --------------------------------------------
             M1(ik_lo:ik_hi,ixp_lo:ixp_hi) = 
     &             matmul( F(ik_lo:ik_hi,ix_lo:ix_hi), 
     &                  Gtmp(ix_lo:ix_hi,ixp_lo:ixp_hi))

!          ------------------------------------------
!          copy M1(ik,ixp) into Mat1([ik,l],[ixp,lp])
!          ------------------------------------------
           do ixp=ixp_lo,ixp_hi
            ixplp = ixp + (lp-1)*Nc
            ioffset = (l-1)*Nc

            do ik=ik_lo,ik_hi
              ikl = ioffset + ik 
              Mat1( ikl, ixplp ) = M1(ik,ixp)
            enddo

           enddo
!          ------------------------------------------------------------
!          M2( ik,ikp ) = matmul( M1(ik,ixp), conj(trans(F(ikp,ixp)) )
!          ------------------------------------------------------------


             M2(ik_lo:ik_hi,ikp_lo:ikp_hi) = 
     &         matmul( M1(ik_lo:ik_hi,ixp_lo:ixp_hi),
     &                     FH(ixp_lo:ixp_hi,ikp_lo:ikp_hi) )
!          ------------------------------------------
!          copy M2(ik,ikp) into Mat2([ik,l],[ikp,lp])
!          ------------------------------------------
           do ikp=ikp_lo,ikp_hi
            ikplp = ikp + (lp-1)*Nc
            ioffset = (l-1)*Nc

            do ik=ik_lo,ik_hi
              ikl = ioffset + ik 
              Mat2( ikl,ikplp ) = M2(ik,ikp)
            enddo

           enddo
        

         enddo ! do l
         enddo ! do lp

         
       
        endif
        
        return
        end
!======================================================================
! This subroutine define the phase form factor in S. Sachdev's paper
! for a square lattice
!----------------------------------------------------------------------
        subroutine create_theta(theta,indx)
        use Global
        implicit none
        include 'mpif.h'
        !--------------------------------------------------------------
        ! Purpose
        ! =======
        ! initialize the arrays for calculating bond susceptibility
        !
        ! Arguements
        ! ==========
        !
        integer,intent(in):: indx
        complex(kind),intent(inout)::theta(4,Nc,0:2)
        integer :: ic,ix,iy
        complex(kind):: phase(4)=(/onec,ii,-onec,-ii/)
        !phase:  |
        !        2
        !     -3-|-1-
        !        4
        !        |

        theta = zeroc

        ! the first phase parameter for RVB
        ! (1),(2),..., is the index of bond
        ! -i   -i   -i   -i
        !  * +1 * -1 * +1 * -1
        !  i    i    i    i
        !  * +1 * -1 * +1 * -1
        ! -i   -i   -i   -i
        !  * +1 * -1 * +1 * -1
        !  i    i    i    i
        !  * +1 * -1 * +1 * -1
        !

           do ic = 1,Nc
              ix= Rc(1,ic)
              iy= Rc(2,ic)
              if(mod(ix+iy,2).eq.0) then
                 theta(:,ic,1)=phase*(-1)**(ix-1)
              endif
           enddo

        ! the second phase parameter for RVB
        ! (1),(2),..., is the index of bond
        ! -i   +i   -i   +i
        !  * -1 * +1 * -1 * +1
        !  i   -i    i   -i
        !  * +1 * -1 * +1 * -1
        ! -i    i   -i    i
        !  * -1 * +1 * -1 * +1
        !  i   -i    i   -i
        !  * +1 * -1 * +1 * -1
        !
           do ic = 1,Nc
              ix= Rc(1,ic)
              iy= Rc(2,ic)
              if(mod(ix+iy,2).eq.0) then
                 theta(:,ic,2)=phase
              endif
           enddo

        ! the homogeneous phase parameter
        ! (1),(2),..., is the index of bond
        !  1    1    1    1
        !  * 1  *  1 *  1 *  1
        !  1    1    1    1
        !  * 1  *  1 *  1 *  1
        !  1    1    1    1
        !  * 1  *  1 *  1 *  1
        !  1    1    1    1
        !  *  1 *  1 *  1 *  1
        !
           do ic = 1,Nc
              ix= Rc(1,ic)
              iy= Rc(2,ic)
              if(mod(ix+iy,2).eq.0) then
                 theta(:,ic,0)=oner
              endif
           enddo

        return
        end subroutine create_theta
!======================================================================
! Measure bond order wave susceptibility 
!----------------------------------------------------------------------
        subroutine bow_suscept(i,j,it1,k,l,it2,value)
        use Global
        implicit none
        include 'mpif.h'
        !
        ! Purpose
        ! =======
        ! measuring bond order wave operator susceptibility
        ! of bond (i,j, it1) and (k,l,it2)
        ! Note: i =x= j and k =x= l
        !
        ! Arguements
        ! ==========
        integer,intent(in):: i,j,it1
        integer,intent(in):: k,l,it2
        real(kind),intent(out)::value
        real(kind) :: tmp_suscept

        !write(*,*) 'Start to measure BOW!'


        value = zeror
        call Wick_contract_bow(i,j,it1,k,l,it2,tmp_suscept)
        value = value + tmp_suscept
        call Wick_contract_bow(i,j,it1,l,k,it2,tmp_suscept)
        value = value + tmp_suscept
        call Wick_contract_bow(j,i,it1,k,l,it2,tmp_suscept)
        value = value + tmp_suscept
        call Wick_contract_bow(j,i,it1,l,k,it2,tmp_suscept)
        value = value + tmp_suscept

        return
        end subroutine bow_suscept
!----------------------------------------------------------------------
! measuring Wick contraction of BOW
!----------------------------------------------------------------------
        subroutine Wick_contract_bow(i,j,it1,k,l,it2,value)
        use Global
        implicit none
        include 'mpif.h'

        integer,intent(in)::i,j,it1
        integer,intent(in)::k,l,it2
        !integer,intent(in)::Nc
        real(kind),intent(out)::value

        integer :: ind_i,ind_j,ind_k,ind_l
        real(kind) :: tmp1,tmp2

        value = zeror
        ind_i = i + (it1-1)*Nc
        ind_j = j + (it1-1)*Nc
        ind_k = k + (it2-1)*Nc
        ind_l = l + (it2-1)*Nc

        ! The connected graph:   ind /------\imd
        !                            \------/
        ! Unequal time Green's function product
        !
        tmp1 = zeror
        tmp2 = zeror

        tmp1 = -gu(ind_l,ind_i)*gu(ind_j,ind_k)
        tmp2 = -gd(ind_l,ind_i)*gd(ind_j,ind_k)

        value = value + tmp1 + tmp2


        ! The disconnected graph:  ind / \  / \ imd
        !                              \ /  \ /
        ! equal time Green's function product
        tmp1 = zeror
        tmp2 = zeror

        tmp1 = gu(ind_j, ind_i) + gd(ind_j, ind_i)
        tmp2 = gu(ind_l, ind_k) + gd(ind_l, ind_k)

        value = value + tmp1*tmp2

        !  equal-time equal-space contribution
        if(i.eq.l .and. it1.eq.it2)then         
           value = value + (gu(ind_j,ind_k) + gd(ind_j,ind_k))
        end if

        return
        end subroutine Wick_contract_bow
!----------------------------------------------------------------------
! measure bond order wave operator
!----------------------------------------------------------------------
        subroutine meas_bow(i,j,it,value)
        use Global
        implicit none
        include 'mpif.h'

        integer,intent(in)::i,j,it
        real(kind),intent(out)::value

        integer::ind_i,ind_j

        ind_i = i + (it-1)*Nc
        ind_j = j + (it-1)*Nc

        value = - gu(ind_j,ind_i) - gu(ind_i,ind_j)
     &          - gd(ind_j,ind_i) - gd(ind_i,ind_j)

        return
        end subroutine meas_bow
!======================================================================
! Measuring spin-spin susceptibility at ZZ direction
!----------------------------------------------------------------------
        subroutine spinzz_suscept(il1,il2,it1,il3,il4,it2,value)
        use Global
        implicit none
        include 'mpif.h'
        !
        ! Purpose
        ! =======
        ! measuring spin Z-Z bond operator susceptibility
        ! of bond (i,j, it1) and (k,l,it2)
        ! Note: i =x= j and k =x= l
        ! See Shuxiang Yang's  note of RVB measurements
        !
        ! Arguements
        ! ==========
        integer,intent(in):: il1,il2,it1
        integer,intent(in):: il3,il4,it2
        real(kind),intent(out)::value
        real(kind) :: tmp_suscept
        integer :: ind_il1,ind_il2,ind_il3,ind_il4

        !write(*,*) 'Start to measure spin Z-Z!'
        value = zeror

        call Wick_contract_spinzz(il1,il2,it1,it1,
     &il3,il4,it2,it2,tmp_suscept)
        value = value + tmp_suscept

        call Wick_contract_spinzz(il1,il3,it1,it2,
     &il2,il4,it1,it2,tmp_suscept)
        value = value + tmp_suscept

        call Wick_contract_spinzz(il1,il4,it1,it2,
     &il2,il3,it1,it2,tmp_suscept)
        value = value + tmp_suscept

        ind_il1 = il1 + (it1-1)*Nc
        ind_il2 = il2 + (it1-1)*Nc
        ind_il3 = il3 + (it2-1)*Nc
        ind_il4 = il4 + (it2-1)*Nc
        tmp_suscept = -2.d0*(-gu(ind_il1,ind_il1)+gd(ind_il1,ind_il1))
     &                      *(-gu(ind_il2,ind_il2)+gd(ind_il2,ind_il2))
     &                      *(-gu(ind_il3,ind_il3)+gd(ind_il3,ind_il3))
     &                      *(-gu(ind_il4,ind_il4)+gd(ind_il4,ind_il4)) 
        value = value + tmp_suscept

        return
        end subroutine spinzz_suscept 
!----------------------------------------------------------------------
! measuring Wick contraction for spin Z-Z suscept
!----------------------------------------------------------------------
        subroutine Wick_contract_spinzz(il1,il2,it1,it2,
     &il3,il4,it3,it4,value)
        use Global
        implicit none
        include 'mpif.h'
        !
        ! Arguements
        ! ==========
        integer,intent(in):: il1,il2,it1,it2
        integer,intent(in):: il3,il4,it3,it4
        real(kind),intent(out)::value
        real(kind) :: CM12,CM34,CN1,CN2,CN3,CN4
        integer :: ind_il1,ind_il2,ind_il3,ind_il4
        !
        ! Execute
        ! =======
        
        CM12 = zeror
        ind_il1= il1+(it1-1)*Nc
        ind_il2= il2+(it2-1)*Nc
        if(il1.eq.il2.and.it1.eq.it2) then
           CM12 = CM12 -gu(ind_il1,ind_il2)*(oner-gu(ind_il2,ind_il1))
     &                 -gd(ind_il1,ind_il2)*(oner-gd(ind_il2,ind_il1))
        else
           CM12 = CM12 +gu(ind_il1,ind_il2)*gu(ind_il2,ind_il1)
     &                 +gd(ind_il1,ind_il2)*gd(ind_il2,ind_il1)
        end if
        CN1 = -gu(ind_il1,ind_il1)+gd(ind_il1,ind_il1)
        CN2 = -gu(ind_il2,ind_il2)+gd(ind_il2,ind_il2)
        CM12 = CM12 - CN1*CN2
         

        CM34 = zeror
        ind_il3= il3+(it3-1)*Nc
        ind_il4= il4+(it4-1)*Nc
        if(il3.eq.il4.and.it3.eq.it4) then
           CM34 = CM34 -gu(ind_il3,ind_il4)*(oner-gu(ind_il4,ind_il3))
     &                 -gd(ind_il3,ind_il4)*(oner-gd(ind_il4,ind_il3))
        else
           CM34 = CM34 +gu(ind_il3,ind_il4)*gu(ind_il4,ind_il3)
     &                 +gd(ind_il3,ind_il4)*gd(ind_il4,ind_il3)
        end if
        CN3 = -gu(ind_il3,ind_il3)+gd(ind_il3,ind_il3)
        CN4 = -gu(ind_il4,ind_il4)+gd(ind_il4,ind_il4)
        CM34 = CM34 - CN3*CN4

        value = CM12*CM34

        return
        end subroutine Wick_contract_spinzz
!----------------------------------------------------------------------
! measure spin-spin Z-Z bond operator
!----------------------------------------------------------------------
        subroutine meas_spinzz(i,j,it,value)
        use Global
        implicit none
        include 'mpif.h'
        !
        ! Arguements
        ! ==========
        integer,intent(in)::i,j,it
        real(kind),intent(out)::value
        real(kind) :: CM12,CN1,CN2
        integer::ind_i,ind_j
        !
        ! Execute
        ! =======
        
        CM12 = zeror
        ind_i= i+(it-1)*Nc
        ind_j= j+(it-1)*Nc
        CM12 = CM12 +gu(ind_i,ind_j)*gu(ind_j,ind_i)
     &              +gd(ind_i,ind_j)*gd(ind_j,ind_i)
        CN1 = -gu(ind_i,ind_i)+gd(ind_i,ind_i)
        CN2 = -gu(ind_j,ind_j)+gd(ind_j,ind_j)

        value = CM12 - CN1*CN2
        return
        end subroutine meas_spinzz
!======================================================================
! Measuring spin-spin susceptibility <BondSS_ij(T_1)BondSS_kl(T_2)>
! where BondSS=<Sx.Sx+Sy.Sy+Sz.Sz>
! Details can be found in note by Zhaoxin Xu
!----------------------------------------------------------------------
        subroutine spinSS_suscept(ind_i,ind_j,it1,ind_k,ind_l,it2,flag,
     &value)
        use Global
        implicit none
        include 'mpif.h'
        ! Purpose
        ! =======
        ! This subroutine measure the spin-spin bond susceptibility
        ! <BondSS_ij(T_1)BondSS_kl(T_2)>
        ! where BondSS=<Sx.Sx+Sy.Sy+Sz.Sz>
        !
        ! Arguements
        ! ==========
        integer,intent(in):: ind_i,ind_j,it1
        integer,intent(in):: ind_k,ind_l,it2
        ! flag=1 : singlet bond
        ! flag=-1: triplet bond
        integer,intent(in):: flag
        real(kind),intent(out)::value
        real(kind) :: wick_cnt_up, wick_cnt_dn
        real(kind) :: tmp_sus, rflag
        ! Execute
        ! =======
        value = zeror
        rflag = DBLE(flag)
        !-----------------------------------------------------
        !(1)
        tmp_sus = wick_cnt_up(ind_i,ind_i,it1,ind_k,ind_k,it2)
     &           *wick_cnt_dn(ind_j,ind_j,it1,ind_l,ind_l,it2)
        value = value + tmp_sus
        !(2)
        tmp_sus = wick_cnt_up(ind_i,ind_i,it1,ind_k,ind_l,it2)
     &           *wick_cnt_dn(ind_j,ind_j,it1,ind_l,ind_k,it2)
     *           *(-rflag)
        value = value + tmp_sus
        !(3)
        tmp_sus = wick_cnt_up(ind_i,ind_i,it1,ind_l,ind_k,it2)
     &           *wick_cnt_dn(ind_j,ind_j,it1,ind_k,ind_l,it2)
     *           *(-rflag)
        value = value + tmp_sus
        !(4)
        tmp_sus = wick_cnt_up(ind_i,ind_i,it1,ind_l,ind_l,it2)
     &           *wick_cnt_dn(ind_j,ind_j,it1,ind_k,ind_k,it2)
        value = value + tmp_sus
        !-----------------------------------------------------
        !(5)
        tmp_sus = wick_cnt_up(ind_i,ind_j,it1,ind_k,ind_k,it2)
     &           *wick_cnt_dn(ind_j,ind_i,it1,ind_l,ind_l,it2)
     &           *(-rflag)
        value = value + tmp_sus
        !(6)
        tmp_sus = wick_cnt_up(ind_i,ind_j,it1,ind_k,ind_l,it2)
     &           *wick_cnt_dn(ind_j,ind_i,it1,ind_l,ind_k,it2)
        value = value + tmp_sus
        !(7)
        tmp_sus = wick_cnt_up(ind_i,ind_j,it1,ind_l,ind_k,it2)
     &           *wick_cnt_dn(ind_j,ind_i,it1,ind_k,ind_l,it2)
        value = value + tmp_sus
        !(8)
        tmp_sus = wick_cnt_up(ind_i,ind_j,it1,ind_l,ind_l,it2)
     &           *wick_cnt_dn(ind_j,ind_i,it1,ind_k,ind_k,it2)
     &           *(-rflag)
        value = value + tmp_sus
        !-----------------------------------------------------
        !(9)
        tmp_sus = wick_cnt_up(ind_j,ind_i,it1,ind_k,ind_k,it2)
     &           *wick_cnt_dn(ind_i,ind_j,it1,ind_l,ind_l,it2)
     &           *(-rflag)
        value = value + tmp_sus
        !(10)
        tmp_sus = wick_cnt_up(ind_j,ind_i,it1,ind_k,ind_l,it2)
     &           *wick_cnt_dn(ind_i,ind_j,it1,ind_l,ind_k,it2)
        value = value + tmp_sus
        !(11)
        tmp_sus = wick_cnt_up(ind_j,ind_i,it1,ind_l,ind_k,it2)
     &           *wick_cnt_dn(ind_i,ind_j,it1,ind_k,ind_l,it2)
        value = value + tmp_sus
        !(12)
        tmp_sus = wick_cnt_up(ind_j,ind_i,it1,ind_l,ind_l,it2)
     &           *wick_cnt_dn(ind_i,ind_j,it1,ind_k,ind_k,it2)
     &           *(-rflag)
        value = value + tmp_sus
        !-----------------------------------------------------
        !(13)
        tmp_sus = wick_cnt_up(ind_j,ind_j,it1,ind_k,ind_k,it2)
     &           *wick_cnt_dn(ind_i,ind_i,it1,ind_l,ind_l,it2)
        value = value + tmp_sus
        !(14)
        tmp_sus = wick_cnt_up(ind_j,ind_j,it1,ind_k,ind_l,it2)
     &           *wick_cnt_dn(ind_i,ind_i,it1,ind_l,ind_k,it2)
     &           *(-rflag)
        value = value + tmp_sus
        !(15)
        tmp_sus = wick_cnt_up(ind_j,ind_j,it1,ind_l,ind_k,it2)
     &           *wick_cnt_dn(ind_i,ind_i,it1,ind_k,ind_l,it2)
     &           *(-rflag)
        value = value + tmp_sus
        !(16)
        tmp_sus = wick_cnt_up(ind_j,ind_j,it1,ind_l,ind_l,it2)
     &           *wick_cnt_dn(ind_i,ind_i,it1,ind_k,ind_k,it2)
        value = value + tmp_sus

        return
        end subroutine spinSS_suscept
!----------------------------------------------------------------------
! Function for measuring spin_up Wick contraction
!
        function wick_cnt_up(ind_i,ind_j,it1,ind_k,ind_l,it2)
        use Global
        implicit none
        include 'mpif.h'
        ! Purpose
        ! =======
        ! Measuring Wick contraction for spin up
        ! 
        ! Arguements
        ! ==========
        !
        real(kind)::wick_cnt_up
        integer,intent(in) :: ind_i,ind_j,it1
        integer,intent(in) :: ind_k,ind_l,it2
        integer    :: il1, il2, il3, il4
        real(kind) :: tmp1, tmp2, tmp3, tmp4
        !
        ! Execute
        ! =======
        wick_cnt_up=zeror

        il1 = ind_i + (it1-1)*Nc
        il2 = ind_j + (it1-1)*Nc
        il3 = ind_k + (it2-1)*Nc
        il4 = ind_l + (it2-1)*Nc

        if(ind_i.eq.ind_j) then
           tmp1 = oner - gu(il2,il1)
        else
           tmp1 = -gu(il2,il1)
        endif

        if(ind_k.eq.ind_l) then
           tmp2 = oner - gu(il4,il3)
        else
           tmp2 = -gu(il4,il3)
        endif

        if(it1.eq.it2.and.ind_i.eq.ind_l) then
           tmp3 = oner - gu(il4,il1)
        else
           tmp3 = gu(il4,il1)
        endif

        tmp4 = -gu(il2,il3)
        
        wick_cnt_up = tmp1*tmp2 - tmp3*tmp4

        return
        end function wick_cnt_up
!----------------------------------------------------------------------
! Function for measuring spin_up Wick contraction
!
        function wick_cnt_dn(ind_i,ind_j,it1,ind_k,ind_l,it2)
        use Global
        implicit none
        include 'mpif.h'
        ! Purpose
        ! =======
        ! Measuring Wick contraction for spin up
        ! 
        ! Arguements
        ! ==========
        !
        real(kind)::wick_cnt_dn
        integer,intent(in) :: ind_i,ind_j,it1
        integer,intent(in) :: ind_k,ind_l,it2
        integer    :: il1, il2, il3, il4
        real(kind) :: tmp1, tmp2, tmp3, tmp4
        !
        ! Execute
        ! =======
        wick_cnt_dn=zeror

        il1 = ind_i + (it1-1)*Nc
        il2 = ind_j + (it1-1)*Nc
        il3 = ind_k + (it2-1)*Nc
        il4 = ind_l + (it2-1)*Nc

        if(ind_i.eq.ind_j) then
           tmp1 = oner - gd(il2,il1)
        else
           tmp1 = -gd(il2,il1)
        endif

        if(ind_k.eq.ind_l) then
           tmp2 = oner - gd(il4,il3)
        else
           tmp2 = -gd(il4,il3)
        endif

        if(it1.eq.it2.and.ind_i.eq.ind_l) then
           tmp3 = oner - gd(il4,il1)
        else
           tmp3 = gd(il4,il1)
        endif

        tmp4 = -gd(il2,il3)
        
        wick_cnt_dn = tmp1*tmp2 - tmp3*tmp4

        return
        end function wick_cnt_dn
!----------------------------------------------------------------------
! measure spin-spin S-S bond operator
!----------------------------------------------------------------------
        subroutine meas_spinSS(i,j,it,flag,value)
        use Global
        implicit none
        include 'mpif.h'
        !
        ! Arguements
        ! ==========
        integer,intent(in)::i,j,it
        ! flag=1 : singlet bond
        ! flag=-1: triplet bond
        integer,intent(in):: flag
        real(kind),intent(out)::value
        real(kind) :: rflag
        integer::ind_i,ind_j
        !
        ! Execute
        ! =======

        value  = zeror
        rflag = DBLE(flag)
        ind_i= i+(it-1)*Nc
        ind_j= j+(it-1)*Nc

        value = value + (oner-gu(ind_i,ind_i))*(oner-gd(ind_j,ind_j))
     &                - rflag*gu(ind_j,ind_i) *      gd(ind_i,ind_j)
     &                - rflag*gu(ind_i,ind_j) *      gd(ind_j,ind_i)
     &                + (oner-gu(ind_j,ind_j))*(oner-gd(ind_i,ind_i))


        return
        end subroutine meas_spinSS

!======================================================================
! Measuring bond-order wave susceptibilities for 3D cubic lattice
!----------------------------------------------------------------------
        subroutine BOW_SUSMEAS(x_kkt,x_kdt,x_ddt,x_eet,k_avg,d_avg)
        use Global
        implicit none
        include 'mpif.h'
        ! Purpose
        ! =======
        ! This subroutine calculate the susceptibility of the hopping and 
        ! double occupancy of the 3D lattice for the linear response
        ! to the dynamic modulation for optical lattice
        !
        ! Arguements
        ! ==========
        real(kind), intent(inout) :: x_kkt(nl),x_kdt(nl)
        real(kind), intent(inout) :: x_ddt(nl),x_eet(nl)
        integer :: ll,it,jt,ib,jb,i_neibor,j_neibor
        integer :: ind_i,ind_j,ind_k,ind_l,i1,il,jl
        integer :: NADJ
        integer, allocatable:: neibor_lnk(:,:)
        real(kind) :: k_avg,d_avg
        real(kind) :: tmp_dd1, tmp_dd2
        ! Execute
        ! =======

        ! NADJ: number of bond
        ! 1) n.n. neighbor of 2D squre, NADJ=4
        ! 2) n.n. neighbor of 3D cubin, NADJ=6
        !
        ! neibor_link(NADJ, Nc): link table for bond
        !
        if(ndim.eq.2) NADJ=4
        if(ndim.eq.3) NADJ=6

        allocate(neibor_lnk(NADJ,Nc))
   
        ! neighboring table for cubic lattice
        do i1 = 1, Nc
           neibor_lnk(1,i1)=nneighbor(1,i1)     ! nn in the +x direction!     2  5
           neibor_lnk(2,i1)=nneighbor(2,i1)     ! nn in the +y direction!     | / 
           neibor_lnk(3,i1)=nneighbor(1+ndim,i1)! nn in the -x direction! 3 - i1 - 1
           neibor_lnk(4,i1)=nneighbor(2+ndim,i1)! nn in the -y direction!   / |
           if(ndim.eq.3) then                   !                          6  4
              neibor_lnk(5,i1)=nneighbor(3,i1)  ! nn in the +z direction! 
              neibor_lnk(6,i1)=nneighbor(3+ndim,i1)! nn in the -z direction
           endif
        enddo

        ! clear arrays
        x_kkt=zeror
        x_kdt=zeror
        x_ddt=zeror
        x_eet=zeror

        do ll = 1, nl                          ! dt = T-T'

        !-------------------------------------------------------------
        ! measure <K(T)K(T')>
           do il = 1, Nc                       ! index of i
           do ib = 1, NADJ                     ! index of i neighbors
           i_neibor = neibor_lnk(ib,il)
           do jl = 1, Nc 
           do jb = 1, NADJ
           j_neibor = neibor_lnk(jb,jl)

        !-------------------------------------------------------------
           do it = 1, nl                          ! index of T
              jt = 1 + mod(it+ll-2,nl)            ! index of T', jt=it+dt

              ind_i = il       + (it-1)*Nc
              ind_j = i_neibor + (it-1)*Nc
              ind_k = jl       + (jt-1)*Nc
              ind_l = j_neibor + (jt-1)*Nc
        ! connected graph
              if(it.eq.jt.and.ind_l.eq.ind_i) then
                 x_kkt(ll) = x_kkt(ll) 
     & + (oner-gu(ind_l,ind_i))*gu(ind_j,ind_k)
     & + (oner-gd(ind_l,ind_i))*gd(ind_j,ind_k)
              else
                 x_kkt(ll) = x_kkt(ll) 
     & - gu(ind_l,ind_i)*gu(ind_j,ind_k)
     & - gd(ind_l,ind_i)*gd(ind_j,ind_k)
              endif
        ! disconnected graph
              x_kkt(ll) = x_kkt(ll) 
     & + (gu(ind_j,ind_i)+gd(ind_j,ind_i))
     & * (gu(ind_l,ind_k)+gd(ind_l,ind_k))
           enddo
           enddo
           enddo           
           enddo
           enddo
        !-------------------------------------------------------------
        ! measure <K(T)D(T')>
           do il = 1, Nc
           do ib = 1, NADJ
           i_neibor = neibor_lnk(ib,il)
           do jl = 1, Nc
        !-------------------------------------------------------------
           do it = 1, nl                          ! index of T
              jt = 1 + mod(it+ll-2,nl)            ! index of T', jt=it+dt
           
              ind_i = il       + (it-1)*Nc
              ind_j = i_neibor + (it-1)*Nc
              ind_k = jl       + (jt-1)*Nc
        ! connected graph
              if(it.eq.jt.and.ind_k.eq.ind_i) then
                 x_kdt(ll) = x_kdt(ll) 
     & + (oner-gu(ind_k,ind_i))*gu(ind_j,ind_k)*(oner-gd(ind_k,ind_k))
     & + (oner-gd(ind_k,ind_i))*gd(ind_j,ind_k)*(oner-gu(ind_k,ind_k))
              else
                 x_kdt(ll) = x_kdt(ll)
     & - gu(ind_k,ind_i)*gu(ind_j,ind_k)*(oner-gd(ind_k,ind_k))
     & - gd(ind_k,ind_i)*gd(ind_j,ind_k)*(oner-gu(ind_k,ind_k))
              endif
        ! disconnected graph
              x_kdt(ll) = x_kdt(ll)
     & - gu(ind_j,ind_i)*(oner-gu(ind_k,ind_k))*(oner-gd(ind_k,ind_k))
     & - gd(ind_j,ind_i)*(oner-gd(ind_k,ind_k))*(oner-gu(ind_k,ind_k))

           enddo !end do it
           enddo
           enddo
           enddo
        !-------------------------------------------------------------
        ! measure <D(T)K(T')>
           do il = 1, Nc
           do jl = 1, Nc
           do jb = 1, NADJ
           j_neibor = neibor_lnk(jb,jl)
        !-------------------------------------------------------------
           do it = 1, nl                          ! index of T
              jt = 1 + mod(it+ll-2,nl)            ! index of T', jt=it+dt

              ind_i = il       + (it-1)*Nc
              ind_k = jl       + (jt-1)*Nc
              ind_l = j_neibor + (jt-1)*Nc
        ! connected graph
              if(it.eq.jt.and.ind_l.eq.ind_i) then
                 x_kdt(ll) = x_kdt(ll) 
     & + (oner-gu(ind_l,ind_i))*gu(ind_i,ind_k)*(oner-gd(ind_i,ind_i))
     & + (oner-gd(ind_l,ind_i))*gd(ind_i,ind_k)*(oner-gu(ind_i,ind_i))
              else
                 x_kdt(ll) = x_kdt(ll)
     & - gu(ind_l,ind_i)*gu(ind_i,ind_k)*(oner-gd(ind_i,ind_i))
     & - gd(ind_l,ind_i)*gd(ind_i,ind_k)*(oner-gu(ind_i,ind_i))
              endif
        ! disconnected graph
              x_kdt(ll) = x_kdt(ll)
     & - (oner-gu(ind_i,ind_i))*gu(ind_l,ind_k)*(oner-gd(ind_i,ind_i))
     & - (oner-gd(ind_i,ind_i))*gd(ind_l,ind_k)*(oner-gu(ind_i,ind_i))

           enddo !end do it
           enddo
           enddo
           enddo
        !-------------------------------------------------------------
        ! measure <D(T)D(T')>
           do il = 1, Nc
           do jl = 1, Nc
        !-------------------------------------------------------------
           do it = 1, nl                          ! index of T
              jt = 1 + mod(it+ll-2,nl)            ! index of T', jt=it+dt

              ind_i = il + (it-1)*Nc
              ind_k = jl + (jt-1)*Nc

              if(it.eq.jt.and.ind_k.eq.ind_i) then
                 tmp_dd1 = (oner-gu(ind_i,ind_i))*(oner-gu(ind_k,ind_k))
     & + (oner-gu(ind_k,ind_i))*gu(ind_i,ind_k)
                 tmp_dd2 = (oner-gd(ind_i,ind_i))*(oner-gd(ind_k,ind_k))
     & + (oner-gd(ind_k,ind_i))*gd(ind_i,ind_k)
              else
                 tmp_dd1 =(oner-gu(ind_i,ind_i))*(oner-gu(ind_k,ind_k))
     & - gu(ind_k,ind_i)*gu(ind_i,ind_k)
                 tmp_dd2 =(oner-gd(ind_i,ind_i))*(oner-gd(ind_k,ind_k))
     & - gd(ind_k,ind_i)*gd(ind_i,ind_k)
              endif
              x_ddt(ll) = x_ddt(ll) + tmp_dd1*tmp_dd2

           enddo !end do it
           enddo
           enddo

        enddo ! end do ll

        x_kkt = x_kkt*sgn / real(Nc*Nc*nl,kind)
        x_kdt = x_kdt*sgn / real(Nc*Nc*nl,kind)
        x_ddt = x_ddt*sgn / real(Nc*Nc*nl,kind)
        x_eet(:) = x_kkt(:) - alphax*x_kdt(:) + alphax**2*x_ddt(:)

        k_avg=zeror
        d_avg=zeror
        !-------------------------------------------------------------
        ! measure <K> and <D>
        do ll = 1, nl
        do il = 1, Nc
           ind_i = il + (ll-1)*Nc
           do ib = 1, NADJ
              i_neibor = neibor_lnk(ib,il)
              ind_j = i_neibor + (ll-1)*Nc                          
              k_avg = k_avg - gu(ind_j,ind_i) - gd(ind_j,ind_i)
           enddo
           d_avg = d_avg + (oner-gu(ind_i,ind_i))*(oner-gd(ind_i,ind_i))
        enddo
        enddo
        k_avg = k_avg*sgn / real(nl*Nc,kind)
        d_avg = d_avg*sgn / real(nl*Nc,kind)

        deallocate(neibor_lnk)
  
        return
        end subroutine BOW_SUSMEAS
