c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
c***************************************************************************
        module Global
c****************************************************************************
c       This module should appear in all the main blocks of the code.  
c       It replaces the common block and declares and passes all the 
c       global variables and parameters (those globally accessible).
c       Note that many of the vaiables are allocatable.  They are
c       allocated/deallocated in the calls below for 1 and 2 particle
c       variables.
c****************************************************************************
        implicit none
        save
c**************************************************************************** 
c       set some parameters.
c**************************************************************************** 

c**************************************************************************** 
c       character parameters.
c**************************************************************************** 
        character(1), parameter :: Nstring='N',Cstring='C'

c**************************************************************************** 
c       integer parameters.
c**************************************************************************** 
        integer, parameter :: kind=8    ! on most system 8=>DP and 4=>SP (8!)
        integer, parameter :: kindg=4    ! prec. of gamat (4!)
        integer, parameter :: kindgp=8   ! prec. of gamat calc. (4 or 8)
        integer, parameter :: iprint=0   ! if large, print lots see below
        integer, parameter :: ifilter=0  ! 0 dont use, 1 use low pass filter
        integer, parameter :: brufor=1   ! 1D CoarsGrain integration brute
        integer, parameter :: lub=9      ! logical unit <--> 'bins.dat'
        integer, parameter :: lus=40     ! logical unit <--> 'sigma.dat'
        integer, parameter :: lug=45     ! logical unit <--> 'gamat.dat'
        integer, parameter :: lui=15     ! logical unit <--> 'indat'
        integer, parameter :: lud=66     ! logical unit <--> 'data'
        integer, parameter :: skipmax=60 ! maximum value of skip
        integer, parameter :: irvb=0     ! >0 for RVB meas, 1 for Cluster RVB, 2 for both cluster and local
        integer, parameter :: ibow=1     ! >0 for chiRVB and chiBOW meas
        
c**************************************************************************** 
c       real parameters.
c**************************************************************************** 
        real(kind), parameter ::    eps=0.000100000000000_kind
        real(kind), parameter ::  zeror=0.000000000000000_kind
        real(kind), parameter ::  halfr=0.500000000000000_kind
        real(kind), parameter ::   oner=1.000000000000000_kind
        real(kind), parameter ::   twor=2.000000000000000_kind
        real(kind), parameter :: threer=3.000000000000000_kind
        real(kind), parameter ::  fourr=4.000000000000000_kind
        real(kind), parameter ::     pi=3.141592653589793_kind
        real(kind), parameter :: autoln=1.000000000000000_kind
        real(kind), parameter :: alphax=10.00000000000000_kind
c**************************************************************************** 
c       complex parameters.
c**************************************************************************** 
        complex(kind), parameter :: zeroc=(0.0_kind,0.0_kind)
        complex(kind), parameter :: onec=(1.0_kind,0.0_kind)
        complex(kind), parameter :: ii=(0.0_kind,1.0_kind)

c****************************************************************************
c       Define some variables.
c****************************************************************************
c       To reduce storage, 32 bin precision (kind 4) is used on some machines
c       for the calculation (kindgp) and storage (kindg) of the two-
c       particle matrices.  In principle, fmat and temat# should also be 
c       sized kindg; however, this can cause complications with the used of 
c       cgemm on the t3e since a 32 bit BLAS is unavailable there.
c       Thus, they will be declared size kindgp (4 on 32 bit machines
c       8 on the t3e).
c**************************************************************************** 
c       Whenever possible, the following is used to set the
c       nomeclature of subscripts
c
c       Gc              cluster greens function
c       K               momentum
c       s               space
c       f               frequency
c       t               euclidean time
c       sc              script (i.e. site or cluster excluded)
c       pt              perturbation theory result
c**************************************************************************** 

c**************************************************************************** 
c       character variables.
c**************************************************************************** 
        character*80 :: abortstring
        character*3 :: cluster
c**************************************************************************** 
c       integer variables.
c**************************************************************************** 
        integer :: nl,Nc,N_sp,nwn,NKmax,nt,lrun,
     &             iseed,iter,niter,Ncw,ntw,run,warm,skip,ntr,
     &             meas,nmeas,nrun,nslr,nuse,ntot,nover,ndim,ngroup,
     &             nsgn,isu,isi,iso,isb,isd,ised,isg,isf,isv,
     &             ibetts,iosft,npt,icond,abortf,umode,Nblock,
     &             a(3,3),icK0,icKpi,icR0,myrank,ierr,nproc,
     &             nwsc,iskip,iauto,igroupxy
c
c       integer allocatable variables
        integer, allocatable  :: Rc(:,:),icrequ(:,:),ltaudiff(:,:),
     &                           icrdiff(:,:),ickdiff(:,:),
     &                           ickplus(:,:),ickequ(:,:),
     &                           ickdeg(:),nneighbor(:,:),
     &                           ickmap(:),ick2map(:,:),s(:)
  

c**************************************************************************** 
c       real variables        
c**************************************************************************** 
        real(kind) :: beta,temp,ed,tprime,tperp,u,V0,alpha,ftarget,
     &                acrat,dt,dfac,ifac,nfocc,sgn,test,testl,
     &                time_meas,time_put,time_date,time_geod,
     &                time_sumup,g(3,3),sgnm,acount,ucount,
     &                moom,moof,mooft,moofs,sgnf,sgnfs,curtdata(4),
     &                RVB0, RVB0d, RVB0xx,AFM0,BB0,xyswitch,
     &                bow_meas,dbop_meas

c
c       Allocatable real arrays which are size kind
        real(kind), allocatable :: V(:),gf(:,:),gfs(:,:),gft(:,:),
     &                             chf(:,:),chfs(:,:),chft(:,:),
     &                             scf(:,:),scfs(:,:),scft(:,:),
     &                             ccf(:,:),ccfs(:,:),ccft(:,:),
     &                             data(:,:),chaf(:,:),chafs(:,:),
     &                             chaft(:,:),gmeas(:,:),chmeas(:,:),
     &                             scmeas(:,:),cham(:,:),ccmeas(:,:),
     &                             autodat(:),autosgn(:),
     &                             chii(:),chii0(:),vari(:),vari0(:),
     &                             expm1(:),expregm1(:),
     &                             Kc(:,:),wn(:),tau(:),
     &                             Epsbar(:),Gcstsc(:,:),Gcstosft(:,:),
     &                             GcKt(:,:),Gcst_not(:,:,:),
     &                             gu(:,:),gd(:,:), gfine(:),Bspline(:),
     &                             Cspline(:),Dspline(:),FNspline(:),
     &                             XNspline(:),HELPspline(:,:),
     &                             mpi_rbufout(:),mpi_rbufin(:),
     &                             mpi_rbufgf(:,:),Bmat(:,:),Dmat(:,:),
     &                            kt(:,:),gdwave(:),gB1g(:,:),gB2g(:,:),
     &                             b1gmeas(:,:),b2gmeas(:,:),
     &                             RVB1(:),RVB1d(:),AFM1(:), RVB1xx(:),
     &                             BB1(:),BBp1(:),BBp2(:),BBp3(:),
     &                             RVBchixxm(:,:),RVBchixym(:,:),
     &                             RVBchixxt(:,:),RVBchixyt(:,:),
     &                             ChiHHSm(:,:),ChiHHSt(:,:),data1(:,:),
     &chiRVB_tau(:,:),chiBOW_tau(:,:),chizz_tau(:,:),chiss_tau(:,:),
     &sus_kkt(:),sus_kdt(:),sus_ddt(:),sus_eet(:)

c       Allocatable real arrays of size kindg (32 bit)
        real(kindg),    allocatable :: gamatscenter(:,:),
     &                                 tematrh(:,:)

c**************************************************************************** 
c       Complex variables
c**************************************************************************** 
c       Allocatable complex variables of size kind
        complex(kind), allocatable  :: GcKf(:,:),GcKfsc(:,:),
     &                                 GcKfpt(:,:),Gcsfosft(:,:),
     &                                 Gcsf(:,:),chi(:,:),chi2(:,:),
     &                                 chi3(:,:),
     &                                 sigma(:,:),sigmapt(:,:),
     &                                 acond(:),bcond(:),
     &                                 mpi_cbufgcsf(:,:),
     &                                 mpi_cbufsig(:,:),
     &                                 FTCoefs_K_to_R(:,:),
     &                                 FTCoefs_R_to_K(:,:)
     
c       Allocatable complex arrays of size kindgp
        complex(kindgp), allocatable :: fmat(:,:),temat1(:,:),
     &                                  temat2(:,:),temat3(:,:),
     &                                  Afmat(:,:), Bfmat(:,:)

c       Allocatable complex arrays of size kindg (32 bit)
        complex(kindg), allocatable :: gamatscorner(:,:),
     &                                 mpi_cbufgmatc0(:),
     &                                 gamatcenter(:,:),
     &                                 gamatccenter(:,:),
     &                                 gamatc0center(:),temath(:,:),
     &                                 gamatcorner(:,:),
     &                                 gamatccorner(:,:),
     &                                 gamatc0corner(:)
c
        complex(kind) :: chiRVB0(0:2), chiBOW0(0:2),
     &chizz0(0:2),chiss0(0:2)


c**************************************************************************** 
c       Table of Variables
c**************************************************************************** 
c
c       ndim    Number of spatial dimensions
c       ngroup  Number of group operations which leave the lattice ivariant
c       nwn     number of matsubara frequencies used
c       Nc      Size of the cluster
c       nl      number of time slices
c       nt      total number of points on the space-time lattice
c       iseed   seed integer the random number generator
c       iter    present iteration number
c       niter   total number of iterations to be done
c       run     number of runs to be executed per iteration
c       warm    number of warmup sweeps between iterations
c       skip    number of sweeps between measurments
c       meas    number of measurements made per run
c       ntr     number of sweeps before G is recalculated from {s(l)}
c       nslr    number of sweeps sinse the last such recalculation
c       s(p)    HS field at "site" p= ic +(l-1)*Nc
c       nuse    number of frequencies where Sigma is calculated from
c               the QMC data (should now be nl/2)
c       isu     switch, if U<0 (>0), then isu=-1 (+1)
c       isi     switch, if isi>=1, increase the # of runs by ifac each iter.
c       iso     switch, if iso=1 (0) then initiate with old sigma (pt result)
c       isb     switch, if isb=>1, calculate bins of data
c       isd     switch, if isd=1 use damping (dfac) in calc. of sigma
c       isg     switch, if isg=1 then calc. the two-particle properties
c       isf     switch, if isf=1 initalize with an externally provided s(i) 
c       ised    switch, if ised=1, adjust ed to obtain 2*nfocc=ftarget
c       isv     switch, if isv=1, diagonal disorder is to be included.
c       iprint  switch, if >0 (up to 99) debugging information is printed
c                       if >=1 done messages from main code sections
c                       if >=9 done from lesser code sections (update..)
c                       if >=19 all tables output is printed
c       ibetts  switch, if ibetts=1, use the Betts tables
c       iskip   switch  if 0 (1) (do)nt adjust skip
c       iauto   switch  if 1 spin- , 2 pair- , 3 charge-,
c                          4- Green's f G(k,w)-, 5 -G(r,t)-, autotime  
c       umode   switch  
c                       if 0 use Hirsch-Fye update
c                       if 1 use Ed's delayed update
c                       if 2 use Block update
c       irvb    switch  if 1 do RVB measurement
c       Nblock  size of update block (adjacent in time)
c       beta    inverse temperature
c       temp    temperature
c       pi      3.14159.... acos(-1)
c       ed      bare orbital energy (measured relative to the chem. pot.)
c       u       Hubbard U
c       V0      strength of binary (continuous) disorder when V>0 (V<0).
c       V(ic)   strength of disorder potential at cluster site ic
c       tprime  ratio of next-near to near-neighbor hopping
c       tperp   ratio of interplanar to near-neighbor hopping
c       tau     array of euclidean times (0,beta,dtau)
c       ftarget target filling
c       test    convergence criteria
c       testl   convergence criteria from last iteration
c       acrat   acceptance ratio
c       dt      delta-tau
c       dfac    damping factor (for sigma, see also isd)
c       ifac    see isi
c       nfocc   orbital occupancy for a single spin species
c       sgn     present sign of determinant
c       expm1   lookup table for updating g
c       expregm1   lookup table for recalculating G from s
c       Kc      lookup table for K (axis,point index)
c       wn      lookup table of matsubara frequencies
c       indsx   lookup table for pbc in space (x)
c       indsy   lookup table for pbc in space (y)
c       gmeas   single run accumulator for the single-particle green function
c       chmeas  single run accumulator for the cluster spin susc.
c       cham    single run accumulator for the cluster charge susc.
c       sgnm    single run accumulator average sign.
c       moom    single run accumulator unscreened moment.
c       gf      all runs accumulator for the single-particle green function
c               sign divided out after each run
c       gfs     all runs accumulator for the square single-particle green
c               function sign divided out after each run
c       gft     all runs accumulator for the single-particle green function
c               multiplied by the sign.
c       chf,chfs,chft
c               same as gf,gfs,gft, but for spin
c       chaf,chafs,chaft
c               same as gf,gfs,gft, but for charge
c       moof    all runs accumulator for the unscreened moment*sgn
c               sign divided out after each run
c       mooft   all runs accumulator for the unscreened moment*sgn
c       moofs   all runs accumulator for the square unscreened moment*sgn
c               sign divided out after each run
c       sgnf    all runs accumulator for the sign
c       sgnfs   all runs accumulator for the square sign
c       gu      QMC up-spin green function
c       gd      QMC down-spin green fuction
c       Gcstsc  Bare green function => QMC green function for all s=0,V=0
c       Gcst_not 
c               Bare green function => QMC green function for all s=0
c               but with finite random local fields V(i)
c       Gcstosft Cluster-excluded G_sc(r,t) used for the OSFT
c       GcKf    G(k,w) on the cluster
c       GcKfsc  cluster-excluded G_sc(k,w)
c       GcKfpt  PT G(k,w)
c       Gcsfosft Cluster-excluded G_sc(r,w) used for the OSFT
c       Gcsf    Cluster G(r,w)
c       chi     temporary array for calculating ph bubble
c       chi2    temporary array for calculating ph bubble
c       chi3    temporary array for calculating ph bubble
c       ii      sqrt(-1)
c       sigma   self energy
c       sigmapt PT self energy
c       fmat    Fourier transform matrix
c       gamatcenter
c               spin two-particle sus at Q=(0,0)
c       gamatscenter
c               singlet two-particle sus at Q=(0,0)
c       gamatccenter
c               charge two-particle sus at Q=(0,0)
c       simlar quantites exist for the zone corner [Q=(pi,pi)].
c       temat#  temporary complex nt x nt arrays
c       data    temporary array.
c**************************************************************************** 

        end module Global

        subroutine allocate_2p
c       allocate some arrays used to measure two-particle properties.
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        integer info,infot,icount

        infot=0
        icount=0
        allocate (gamatscenter(ntw,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gamatscorner(ntw,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (fmat(nt,nt),Afmat(nl,nl),Bfmat(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gamatcenter(ntw,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gamatccenter(ntw,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gamatc0center(ntw),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gamatcorner(ntw,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gamatccorner(ntw,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gamatc0corner(ntw),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ick2map(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
       
        if(infot.ne.0) then
          write(6,*) 'allocate_2p',infot,' failures at step',icount
          stop
        end if

        return
        end subroutine allocate_2p



        subroutine allocate_1p
c       allocate some arrays to the desired size.
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        integer info,infot,icount

        infot=0
        icount=0
        
        allocate (gu(nt,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (gd(nt,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gfine(nover*nl+1),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Bspline(0:NKmax-1),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Cspline(0:NKmax-1),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Dspline(0:NKmax-1),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(FNspline(0:NKmax),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(XNspline(0:NKmax),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(HELPspline(-2:NKmax+1,1:4),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(V(Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gf(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gfs(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gft(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chf(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chfs(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chft(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(scf(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(scfs(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(scft(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ccf(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ccfs(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ccft(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(data(-nl:nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(data1(-nl:nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chaf(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chafs(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chaft(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gmeas(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chmeas(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(scmeas(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(cham(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ccmeas(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(b1gmeas(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(b2gmeas(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(expm1(-1:1),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(expregm1(-1:1),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Kc(3,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(wn(-nwn:nwn),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(tau(nl+1),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Epsbar(Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Gcstsc(-nl:nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Gcstosft(-nl:nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(GcKt(-nl:nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Gcst_not(-nl:nl,Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(GcKf(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(GcKfsc(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(GcKfpt(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Gcsfosft(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Gcsf(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chi(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chi2(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chi3(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(sigma(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(sigmapt(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(acond(Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(bcond(Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(mpi_rbufout(8),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(mpi_rbufin(8),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(mpi_rbufgf(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(mpi_cbufgcsf(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(mpi_cbufsig(-nwn:nwn,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(mpi_cbufgmatc0(nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Rc(3,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(icrequ(Nc,ngroup),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(icrdiff(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(nneighbor(6,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ickdiff(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ickplus(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ickequ(Nc,ngroup),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ickdeg(Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ickmap(Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(s(nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(FTCoefs_K_to_R(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(FTCoefs_R_to_K(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gdwave(Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gB1g(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(gB2g(Nc,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Bmat(nt,Nblock),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(Dmat(Nblock,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(kt(3,max(1,(2*(ntot+1))**ndim/Nc)),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (temat1(nt,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (temat2(nt,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (temat3(nt,nt),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (sus_kkt(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (sus_kdt(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (sus_ddt(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (sus_eet(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (chiRVB_tau(nl,0:2),stat=info)
        infot=infot+info
        icount=icount+1   
        allocate (chiBOW_tau(nl,0:2),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (chizz_tau(nl,0:2),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (chiss_tau(nl,0:2),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (RVB1(nl),stat=info)
        infot=infot+info
        icount=icount+1   
        allocate (RVB1d(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (RVB1xx(nl),stat=info)
        infot=infot+info
        icount=icount+1   
        allocate (AFM1(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (BB1(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (BBp1(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (BBp2(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (BBp3(nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (RVBchixxm(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (RVBchixxt(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (RVBchixym(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (RVBchixyt(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(ltaudiff(nl,nl),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (ChiHHSm(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1
        allocate (ChiHHSt(nl,Nc),stat=info)
        infot=infot+info
        icount=icount+1

        if(infot.ne.0.and.myrank.eq.0) then
          write(lud,*) 'allocate_1p',infot,' failures at step',icount
          flush(lud)
          stop
        end if

        return
        end subroutine allocate_1p

        subroutine deallocate_2p

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************

        deallocate (gamatscenter)
        deallocate (gamatscorner)
        deallocate (fmat,Afmat,Bfmat)
        deallocate (gamatcenter)
        deallocate (gamatccenter)
        deallocate (gamatc0center)
        deallocate (gamatcorner)
        deallocate (gamatccorner)
        deallocate (gamatc0corner)
        deallocate (ick2map)         
        return
        end subroutine deallocate_2p
        
        
        subroutine deallocate_1p

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************

        deallocate (gu)
        deallocate (gd)
        deallocate (gfine)
        deallocate (Bspline)
        deallocate (Cspline)
        deallocate (Dspline)
        deallocate (FNspline)
        deallocate (XNspline)
        deallocate (HELPspline)
        deallocate (V)
        deallocate (gf)
        deallocate (gfs)
        deallocate (gft)
        deallocate (chf)
        deallocate (chfs)
        deallocate (chft)
        deallocate (scf)
        deallocate (scfs)
        deallocate (scft)
        deallocate (ccf)
        deallocate (ccfs)
        deallocate (ccft)
        deallocate (data)
        deallocate (data1)
        deallocate (chaf)
        deallocate (chafs)
        deallocate (chaft)
        deallocate (gmeas)
        deallocate (chmeas)
        deallocate (scmeas)
        deallocate (cham)
        deallocate (ccmeas)
        deallocate (b1gmeas)
        deallocate (b2gmeas)
        deallocate (expm1)
        deallocate (expregm1)
        deallocate (Kc)
        deallocate (wn)
        deallocate (tau)
        deallocate (Epsbar)
        deallocate (Gcstsc)
        deallocate (Gcstosft)
        deallocate (GcKt)
        deallocate (Gcst_not)
        deallocate (GcKf)
        deallocate (GcKfsc)
        deallocate (GcKfpt)
        deallocate (Gcsfosft)
        deallocate (Gcsf)
        deallocate (chi)
        deallocate (chi2)
        deallocate (chi3)
        deallocate (sigma)
        deallocate (sigmapt)
        deallocate (acond)
        deallocate (bcond)
        deallocate(mpi_rbufout)
        deallocate(mpi_rbufin)
        deallocate(mpi_rbufgf)
        deallocate(mpi_cbufgcsf)
        deallocate(mpi_cbufsig)
        deallocate(mpi_cbufgmatc0)
        deallocate (Rc)
        deallocate (icrequ)
        deallocate (icrdiff)
        deallocate (nneighbor)
        deallocate (ickdiff)
        deallocate (ickplus)
        deallocate (ickequ)
        deallocate (ickdeg)
        deallocate (ickmap)
        deallocate (s)
        deallocate(FTCoefs_K_to_R)        
        deallocate(FTCoefs_R_to_K)        
        deallocate(gdwave)      
        deallocate(gB1g)          
        deallocate(gB2g)        
        deallocate(Bmat)        
        deallocate(Dmat)        
        deallocate(kt)        
        deallocate (temat1)
        deallocate (temat2)
        deallocate (temat3)
        deallocate (RVB1)
        deallocate (RVB1d)
        deallocate (AFM1)
        deallocate (BB1)
        deallocate (BBp1)
        deallocate (BBp2)
        deallocate (BBp3)
        deallocate (RVBchixxm)
        deallocate (RVBchixxt)
        deallocate (RVBchixym)
        deallocate (RVBchixyt)
        deallocate (ltaudiff)         
        deallocate (ChiHHSm)
        deallocate (ChiHHSt)
        deallocate (sus_kkt,sus_kdt,sus_ddt,sus_eet)
        deallocate (chiRVB_tau,chiBOW_tau,chizz_tau,chiss_tau)

        return
        end subroutine deallocate_1p


      module Dispersion
      implicit none

c        interface Epsk
c
c        function Epsk_1D(k)
c        use Global
c        real(kind)  :: k,Epsk_1D
c        end function
c
c! This is for a 2-chain calculation (which determines whether
c! even (which=0) or odd (which=1))
c        function Epsk_1D_2C(k,which)
c        use Global
c        real(kind)  :: k,Epsk_1D_2C
c        integer     :: which
c        end function
c
c        function Epsk_nD(k)
c        use Global
c        real(kind)  :: k(3),Epsk_nD
c        end function
c
c        end interface

      contains

c      function Epsk_1D(k)
c      use Global
c      real(kind)  :: k,Epsk_1D
c      Epsk_1D = -halfr*cos(k)
c      end function Epsk_1D
c        
c! This is for a 2-chain calculation (which determines whether
c! even (which=0) or odd (which=1))
c      function Epsk_1D_2C(k,which)
c      use Global
c      real(kind)  :: k,Epsk_1D_2C
c      integer     :: which
c      Epsk_1D_2C=-halfr*(cos(k) + tprime*cos(twor*k))-(1+2*which)*tperp
c      end function Epsk_1D_2C
c
      function Epsk_nD(k)
      use Global
      real(kind)  :: k(3),Epsk_nD

      if (ndim.eq.2) then
         Epsk_nD = -halfr*(cos(k(1))+cos(k(2))) 
     &           - tprime*(cos(k(1))*cos(k(2))-oner)
      else if (ndim.eq.3.and.ibetts.eq.0) then
         Epsk_nD = -halfr*(cos(k(1))+cos(k(2))) 
     &           - tprime*(cos(k(1))*cos(k(2))-oner) 
     &           - halfr*tperp*cos(k(3))
      else if (ndim.eq.3.and.ibetts.eq.1) then
         Epsk_nD = -halfr*(cos(k(1))+cos(k(2))+cos(k(3)))
     &           - tprime*(cos(k(1))*cos(k(2))
     &                    +cos(k(1))*cos(k(3))
     &                    +cos(k(2))*cos(k(3)))
      endif

cPK: We cannot error check here: the routine will not inline
c and will not vectorize if it contains a stop or i/o. 
c This kills performance on the X1 while coarsegraining.
c
c         if(myrank.eq.0) write(0,*) "What am I doing here?????"
c         stop
c      end if
      end function Epsk_nD

      end module Dispersion


        subroutine deallocate_auto

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************

        deallocate (autodat)
        deallocate (autosgn)
        deallocate (vari)
        deallocate (vari0)
        deallocate (chii)
        deallocate (chii0)

        return
        end subroutine deallocate_auto
      
        subroutine allocate_auto
c       allocate some arrays to the desired size.
c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        integer info,infot,icount

        infot=0
        icount=0
        
        allocate(autodat(meas*run),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(autosgn(meas*run),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(vari(meas),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(vari0(meas),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chii(meas),stat=info)
        infot=infot+info
        icount=icount+1
        allocate(chii0(meas),stat=info)
        infot=infot+info
        icount=icount+1

        if(infot.ne.0) then
          write(lud,*) 'allocate_auto',infot,' failures at step',icount
          flush(lud)
          stop
        end if

        return
        end subroutine allocate_auto

      
        module module_kron
        implicit none

!       --------------------------------------------
!       set use_matmul = .true. on cray in MSP mode
!       to enable multistreaming
!
!       otherwise, set use_matmul = .false. to use
!       BLAS cgemm/zgemm
!       --------------------------------------------
        logical, parameter :: use_matmul = .false.
        integer, parameter :: wp=8   ! prec. of gamat calc. (4 or 8)                        
        integer, parameter :: sp = selected_real_kind(6,20)
        integer, parameter :: dp = selected_real_kind(14,30)
  
         
        interface gemm
        module procedure gemm_c, gemm_z
        end interface gemm


        public 

        contains


        subroutine gemm_d(transA,transB,m,n,k, alpha, A,lda, B,ldb, 
     &                 beta, C, ldc )
        implicit none
        character         :: transA, transB
        integer           :: m,n,k,lda,ldb,ldc
        real(kind=dp)  :: alpha,beta
        real(kind=dp)  :: A(lda,*), B(ldb,*)
        real(kind=dp)  :: C(ldc,*)

        call dgemm( transA,transB,m,n,k,alpha,A,lda,B,ldb,
     &           beta,C,ldc )

        return
        end subroutine gemm_d

    
        subroutine gemm_s(transA,transB,m,n,k, alpha, A,lda, B,ldb, 
     &                 beta, C, ldc )
        implicit none
        character         :: transA, transB
        integer           :: m,n,k,lda,ldb,ldc
        real(kind=sp)  :: alpha,beta
        real(kind=sp)  :: A(lda,*), B(ldb,*)
        real(kind=sp)  :: C(ldc,*)

        call sgemm( transA,transB,m,n,k,alpha,A,lda,B,ldb,
     &           beta,C,ldc )

        return
        end subroutine gemm_s

        subroutine gemm_c(transA,transB,m,n,k, alpha, A,lda, B,ldb, 
     &                 beta, C, ldc )
        implicit none
        character         :: transA, transB
        integer           :: m,n,k,lda,ldb,ldc
        complex(kind=sp)  :: alpha,beta
        complex(kind=sp)  :: A(lda,*), B(ldb,*)
        complex(kind=sp)  :: C(ldc,*)

        call cgemm( transA,transB,m,n,k,alpha,A,lda,B,ldb,
     &           beta,C,ldc )

        return
        end subroutine gemm_c


        subroutine gemm_z(transA,transB,m,n,k, alpha, A,lda, B,ldb, 
     &                 beta, C, ldc )
        implicit none
        character         :: transA, transB
        integer           :: m,n,k,lda,ldb,ldc
        complex(kind=dp)  :: alpha,beta
        complex(kind=dp)  :: A(lda,*), B(ldb,*)
        complex(kind=dp)  :: C(ldc,*)

        call zgemm( transA,transB,m,n,k,alpha,A,lda,B,ldb,
     &           beta,C,ldc )

        return
        end subroutine gemm_z

        subroutine crandom_number(Z)
        implicit none
        complex(kind=wp), intent(inout), dimension(:,:) :: Z
        integer, parameter :: nb = 256
        real(kind=wp), dimension(nb,nb) :: Zr, Zc
        real(kind=wp), parameter :: zero = 0.0d0
        real(kind=wp), parameter :: one = 1.0d0
        complex(kind=wp) :: zi
        integer :: i,iend,isize, j,jend,jsize
        

        zi = cmplx(zero,one)

        do j=1,size(Z,2),nb
        do i=1,size(Z,1),nb
           iend = min(size(Z,1),i+nb-1)
           jend = min(size(Z,2),j+nb-1)

           isize = iend-i+1
           jsize = jend-j+1

           call random_number(Zr(1:isize,1:jsize))
           call random_number(Zc(1:isize,1:jsize))

           Z(i:iend,j:jend) =  Zr(1:isize,1:jsize)+
     &                 zi*Zc(1:isize,1:jsize)

        enddo
        enddo

        return
        end subroutine crandom_number





        subroutine kron(A,B, C )
        implicit none
        complex(kind=wp), intent(in), dimension(:,:) :: A,B
        complex(kind=wp), intent(inout), dimension(:,:) :: C
!
!       ----------------------------------------------------------------
!       explicitly form C([ib,ia],[jb,ja])  = kron( A(ia,ja), B(ib,jb) )
!       ----------------------------------------------------------------
        integer :: ia,ja, ib,jb, ic,jc
        integer :: nra,nca, nrb,ncb, nrc,ncc

        nra = size(A,1)
        nca = size(A,2)
        nrb = size(B,1)
        ncb = size(B,2)
        nrc = nra*nrb
        ncc = nca*ncb

        do ja=1,nca
        do jb=1,ncb
        do ia=1,nra
        do ib=1,nrb

          ic = ib + (ia-1)*nrb
          jc = jb + (ja-1)*ncb

          C(ic,jc) = A(ia,ja)*B(ib,jb)

        enddo
        enddo
        enddo
        enddo
        
        return
        end subroutine kron

        subroutine kron_matmulL( trans1, trans2, n,na,nb,
     &        alpha, A,lda, B,ldb,  X,ldx, beta, Y,ldy )
        implicit none
        character, intent(in) :: trans1, trans2
        integer, intent(in) :: n, na,nb, lda,ldb,ldx,ldy

        complex(kind=wp), intent(in) :: alpha,beta
        complex(kind=wp), intent(in) :: A(lda,na), B(ldb,nb)
        complex(kind=wp), intent(in) :: X(ldx,*)
        complex(kind=wp), intent(inout) :: Y(ldy,*)

!
        complex(kind=wp) :: AA(na,na), BB(nb,nb), AAt(na,na)
        complex(kind=wp) :: Xj(nb,na), Yj(nb,na), BBXj(nb,na)
        complex(kind=wp) :: YYj(na*nb,1)

        logical :: is_transX, is_ctransX
        logical :: is_trans_ab, is_ctrans_ab
        integer :: m,j,  ia,ib,indx
        real(kind=wp), parameter :: one = 1.0d0
        real(kind=wp), parameter :: zero = 0.0d0
        complex(kind=wp), parameter :: czero = 0.0d0
        complex(kind=wp), parameter :: cone = 1.0d0

        integer :: mm,nn,kk,  ld1,ld2,ld3
        complex(kind=wp) :: lalpha, lbeta


!       compute 
!       Y = beta * Y + alpha * op(C) * op(X)
!       C = kron(A,B)
!
        is_trans_ab = ((trans1.eq.'T').or.(trans1.eq.'t'))
        is_ctrans_ab = ((trans1.eq.'C').or.(trans1.eq.'c'))

        if (is_trans_ab) then

               AA(1:na,1:na) = transpose(A(1:na,1:na))
               BB(1:nb,1:nb) = transpose(B(1:nb,1:nb))

        elseif (is_ctrans_ab) then

               AA(1:na,1:na) = transpose(A(1:na,1:na))
               AA(1:na,1:na) = conjg( AA(1:na,1:na) )

               BB(1:nb,1:nb) = transpose(B(1:nb,1:nb))
               BB(1:nb,1:nb) = conjg( BB(1:nb,1:nb) )
        else
               AA(1:na,1:na) = A(1:na,1:na)
               BB(1:nb,1:nb) = B(1:nb,1:nb)
        endif

        AAt(1:na,1:na) = transpose(AA(1:na,1:na))

!       ------------------------
!       Y(:,j) = C(:,:) * X(:,j)
!       Y(:,j) = BB * X(:,j) * transpose(AA)
!       ------------------------
        is_transX = (trans2.eq.'T').or.(trans2.eq.'t')
        is_ctransX = (trans2.eq.'C').or.(trans2.eq.'c')
        m = na*nb

        if (beta.eq.czero) then
             Y(1:m,1:n) = czero
        elseif (beta.ne.cone) then
             Y(1:m,1:n) = beta*Y(1:m,1:n)
        endif


        do j=1,n
          if (is_transX) then
              do ia=1,na
              do ib=1,nb
                indx = ib + (ia-1)*nb
                Xj(ib,ia) = X(j, indx )
              enddo
              enddo

          elseif (is_ctransX) then
              do ia=1,na
              do ib=1,nb
                indx = ib + (ia-1)*nb
                Xj(ib,ia) = conjg(X(j, indx ))
              enddo
              enddo
          else
              do ia=1,na
              do ib=1,nb
                indx = ib + (ia-1)*nb
                Xj(ib,ia) = X(indx,j)
              enddo
              enddo
          endif

          if (use_matmul) then

          BBXj(1:nb,1:na) = matmul( BB(1:nb,1:nb), Xj(1:nb,1:na) )
          Yj(1:nb, 1:na) = matmul( BBXj(1:nb,1:na), AAt(1:na,1:na) )

          else
             mm = nb
             nn = na
             kk = nb
             ld1 = size(BB,1)
             ld2 = size(Xj,1)
             ld3 = size(BBXj,1)
             lalpha = cone
             lbeta = czero
             call gemm( 'N', 'N', mm,nn,kk, 
     &                  lalpha, BB,ld1, Xj,ld2,
     &                  lbeta,  BBXj, ld3 )

             mm = nb
             nn = na
             kk = na
             ld1 = size(BBXj,1)
             ld2 = size(AAt,1)
             ld3 = size(Yj,1)
             lalpha = cone
             lbeta = czero
             call gemm( 'N','N', mm,nn,kk,
     &                   lalpha, BBXj,ld1, AAT,ld2,
     &                   lbeta,  Yj, ld3 )  
         

          endif

          if (alpha.ne.cone) then
             Yj(1:nb,1:na) = alpha * Yj(1:nb,1:na)
          endif

          do ia=1,na
          do ib=1,nb
            indx = ib + (ia-1)*nb
            Y(indx,j) = Y(indx,j) + Yj(ib,ia)
          enddo
          enddo


          enddo

         return
         end subroutine kron_matmulL





        subroutine kron_matmulR( trans1, trans2, m,na,nb,
     &        alpha, X,ldx, A,lda, B,ldb,   beta, Y,ldy )
        implicit none
        character, intent(in) :: trans1, trans2
        integer, intent(in) :: m, na,nb, lda,ldb,ldx,ldy

        complex(kind=wp), intent(in) :: alpha,beta
        complex(kind=wp), intent(in) :: A(lda,na), B(ldb,nb)
        complex(kind=wp), intent(in) :: X(ldx,*)
        complex(kind=wp), intent(inout) :: Y(ldy,*)

!
        complex(kind=wp) :: AA(na,na), BB(nb,nb), BBt(nb,nb)
        complex(kind=wp) :: Xj(nb,na), Yj(nb,na), XjAA(nb,na)

        logical :: is_transX, is_ctransX
        logical :: is_trans_ab, is_ctrans_ab
        integer :: n,j,  ia,ja, ib,jb, indx

        real(kind=wp), parameter :: one = 1.0d0
        real(kind=wp), parameter :: zero = 0.0d0
        complex(kind=wp), parameter :: czero = 0.0d0
        complex(kind=wp), parameter :: cone = 1.0d0

        integer :: mm,nn,kk, ld1,ld2,ld3
        complex(kind=wp) :: lalpha,lbeta


!       compute 
!       Y = beta * Y + alpha *  op(X) * op(C)
!       C = kron(A,B)
!
        is_trans_ab = ((trans2.eq.'T').or.(trans2.eq.'t'))
        is_ctrans_ab = ((trans2.eq.'C').or.(trans2.eq.'c'))


        is_transX = (trans1.eq.'T').or.(trans1.eq.'t')
        is_ctransX = (trans1.eq.'C').or.(trans1.eq.'c')

        if (is_trans_ab) then

               AA(1:na,1:na) = transpose(A(1:na,1:na))
               BB(1:nb,1:nb) = transpose(B(1:nb,1:nb))

        elseif (is_ctrans_ab) then

               AA(1:na,1:na) = transpose(A(1:na,1:na))
               AA(1:na,1:na) = conjg( AA(1:na,1:na) )

               BB(1:nb,1:nb) = transpose(B(1:nb,1:nb))
               BB(1:nb,1:nb) = conjg( BB(1:nb,1:nb) )
        else
               AA(1:na,1:na) = A(1:na,1:na)
               BB(1:nb,1:nb) = B(1:nb,1:nb)
        endif

        BBt(1:nb,1:nb) = transpose(BB(1:nb,1:nb))

!       ------------------------
!       Y(j,:) = X(j,:)*C(:,:) 
!       Y(j,:) = transpose(BB) * X(j,:) * A
!       ------------------------
        n = na*nb


!        -----------------------------------------------------------
!        Y([jb,ja]) = X( [ib,ia] ) * C( [ib,ia],[jb,ja] )
!                   = X( [ib,ia] ) * A(ia,ja) * B(ib,jb)
!                   = transpose(B(ib,jb)) * (X([ib,ia]) * A(ia,ja))
!        -----------------------------------------------------------
        if (beta.eq.czero) then
             Y(1:m,1:n) = czero
        elseif (beta.ne.cone) then
             Y(1:m,1:n) = beta*Y(1:m,1:n)
        endif


        do j=1,m
          if (is_transX) then
              do ia=1,na
              do ib=1,nb
                indx = ib + (ia-1)*nb
                Xj(ib,ia) = X(indx,j)
              enddo
              enddo

          elseif (is_ctransX) then
              do ia=1,na
              do ib=1,nb
                indx = ib + (ia-1)*nb
                Xj(ib,ia) = conjg(X(indx,j ))
              enddo
              enddo
          else
              do ia=1,na
              do ib=1,nb
                indx = ib + (ia-1)*nb
                Xj(ib,ia) = X(j,indx)
              enddo
              enddo
          endif

          if (use_matmul) then

          XjAA(1:nb,1:na) = matmul( Xj(1:nb,1:na), AA(1:na,1:na) )
          Yj(1:nb, 1:na) = matmul( BBt(1:nb,1:nb), XjAA(1:nb,1:na) )

          else

            mm = nb
            nn = na
            kk = na
            ld1 = size(Xj,1)
            ld2 = size(AA,1)
            ld3 = size(XjAA,1)
            lalpha = cone
            lbeta = czero

            call gemm( 'N','N', mm,nn,kk, 
     &                 lalpha, Xj,ld1, AA,ld2,
     &                 lbeta, XjAA, ld3 )

            mm = nb
            nn = na
            kk = nb
            ld1 = size(BBt,1)
            ld2 = size(XjAA,1)
            ld3 = size(Yj,1)
            lalpha = cone
            lbeta = czero

            call gemm('N','N', mm,nn,kk,
     &                 lalpha, BBt,ld1, XjAA,ld2, 
     &                 lbeta, Yj, ld3 )

          endif

          if (alpha.ne.cone) then
             Yj(1:nb,1:na) = alpha * Yj(1:nb,1:na)
          endif

          do ja=1,na
          do jb=1,nb
            indx = jb + (ja-1)*nb
            Y(j,indx) = Y(j,indx) + Yj(jb,ja)
          enddo
          enddo


          enddo

         return
         end subroutine kron_matmulR

        
        subroutine test_kron_matmul(n,na,nb)
        integer, intent(in) :: n,na,nb

        complex(kind=wp) :: A(na,na), B(nb,nb)
        complex(kind=wp) :: zi
        real(kind=wp), parameter :: zero = 0.0d0
        real(kind=wp), parameter :: one = 1.0d0
        complex(kind=wp) ::  X(na*nb,na*nb)
        complex(kind=wp), dimension(na*nb,n) :: Y,Y0,Y1
        complex(kind=wp), dimension(n,na*nb) :: Z,Z0,Z1
        complex(kind=wp) :: alpha,beta
        complex(kind=wp) :: C(na*nb,na*nb)

        complex(kind=wp) :: CC(na*nb,na*nb)
        complex(kind=wp) :: XX(na*nb,na*nb)

        integer :: m, lda,ldb,ldx,ldy,ldz
        integer :: icase, jcase
        character*3 :: ctable
        character :: trans1, trans2, char1,char2


        zi = cmplx(zero,one)
        call crandom_number(A)
        call crandom_number(B)
        call crandom_number(X)
        call crandom_number(Y0)
        call crandom_number(Z0)

        call kron(A,B,C)

        lda = size(A,1)
        ldb = size(B,1)
        ldx = size(X,1)
        ldy = size(Y,1)
        ldz = size(Z,1)

        alpha = A(1,1)
        beta = B(1,1)


        m = na*nb
        ctable='NTC'

        write(*,*) 'n,na,nb ',n,na,nb

        do icase=1,3
        do jcase=1,3

          trans1 = ctable(icase:icase)
          trans2 = ctable(jcase:jcase)

          if (trans1.eq.'T') then
               CC = transpose(C)
          elseif (trans1.eq.'C') then
               CC = conjg(transpose(C))
          else
               CC = C
          endif

          if (trans2.eq.'T') then
              XX = transpose(X)
          elseif (trans2.eq.'C') then
              XX = conjg(transpose(X))
          else 
              XX = X
          endif
          
        
        Y = Y0
        call kron_matmulL(trans1,trans2,n,na,nb,
     &        alpha, A,lda,B,ldb, X, ldx, beta, Y, ldy )

        m = na*nb
        Y1 = beta*Y0(1:m,1:n) + 
     &         alpha*matmul( CC(1:m,1:m), XX(1:m,1:n) )


         write(*,*) 'matmulL: trans1,trans2 ', 
     &        trans1,trans2,maxval(abs(Y-Y1))


         Z = Z0
         char1 = trans2
         char2 = trans1
         call kron_matmulR(char1,char2,n,na,nb,
     &         alpha, X,ldx, A,lda,B,ldb, beta, Z, ldz )


         Z1(1:n,1:m) = beta*Z0(1:n,1:m) + 
     &           alpha*matmul(XX(1:n,1:m),CC(1:m,1:m))

        write(*,*) 'matmulR: char1,char2 ',
     &          char1,char2,maxval(abs(Z-Z1))

        enddo
        enddo

        return
        end subroutine test_kron_matmul
        end module module_kron
