        subroutine output
c*******************************************************************
c       This block is designed to process the binned data contained
c       in gf, gfs, gft... etc. (the binning was done sequentially).
c       It works optimally for both clean and disordered systems.
c       For clean systems, the average over all the processors is
c       taken before the measurements are divided by the sign.  For
c       disordered ssytems (isv=1), the data from each processor
c       (which contains a different disorder configuration) is 
c       divided by the corresponding average sign.  Afterwards,
c       the data is averaged over the processors.
c*******************************************************************
        use Global
        use module_kron
        implicit none
        include 'mpif.h'
c*****************************************************************
        integer :: i,l,j,k,ix,iy,ic,jc,n,m,mp,
     &             ind,imd,lnsgn,ick,ickq,icr,iequiv,
     &             il,jl,icw,jcw,iw,jw,info,icQ,jcQ,iQ,jQ,
     &             mpi_ibuf,nsize,nbins,nbin,itype,itop
        integer :: ld1,ld2,ld3,ld4, nn,nna,nnb, i1,i2,i3,i4

        real(kind) :: sigg,goo,nd,sigmoo,sigch,sigcha,sigsc,susmf3,
     &                chaclu,chiclu,chaafclu,chiafclu,chaafzt,chiafzt,
     &                time1,time2,h,r1,r2,suscf3,sig_nfocc,r3,r4,
     &                ta,dtfine,edold,sgn0,avdat,avsgn,r1d,r2d,r8,r8b,
     &                r1xx,r2xx,
     &                autotime,suspf3,scfclu,r5,r6,r7,r5b,r6b,r7b,
     &                RVB0buf,RVB1buf(nl),RVBMOM,
     &                RVB0dbuf,RVB1dbuf(nl),RVBdMOM,
     &                 RVB0xxbuf,RVB1xxbuf(nl),RVBxxMOM,
     &                AFM0buf,AFM1buf(nl),AFMMOM,
     &                BB0buf,BB1buf(nl),BB1MOM,
     &                BBp1buf(nl),BBp1MOM,
     &                BBp2buf(nl),BBp2MOM,
     &                BBp3buf(nl),BBp3MOM,
     &                Bondxx(nl,Nc), Bondxy(nl,Nc)

        logical, parameter :: use_original = .false.
        complex(kind) :: csum,csum1
        complex(kindgp) :: bet,alp
        complex(kind=kindgp) :: alpha_c, beta_c
        complex(kindg) :: z,z1
c       Declarations for using the akima spline 
        integer :: MACHPD,Ispline,Nspline
        real(kind) :: SPVAL,BETAspline
        ! for column and staggered bond susceptibility
        complex(kind):: chiRVB0_buf(0:2)
        real(kind):: chiRVB_w0(0:2),chiRVB_t0(0:2)
        real(kind):: chiRVB_tau_buf(nl,0:2)
        ! for bond order wave operator and its susceptibility
        complex(kind):: chiBOW0_buf(0:2)
        real(kind):: chiBOW_w0(0:2),chiBOW_t0(0:2)
        real(kind):: chiBOW_tau_buf(nl,0:2)
        ! for spin-spin Z-Z bond operator and its susceptibility
        complex(kind):: chizz0_buf(0:2)
        real(kind):: chizz_w0(0:2),chizz_t0(0:2)
        real(kind):: chizz_tau_buf(nl,0:2)
        ! for spin-spin Z-Z bond operator and its susceptibility
        complex(kind):: chiss0_buf(0:2)
        real(kind):: chiss_w0(0:2),chiss_t0(0:2)
        real(kind):: chiss_tau_buf(nl,0:2)
c*****************************************************************

        time1=real(MPI_Wtime(),kind)
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
        
c
c***********************************************************************
c       DATA REDUCTION.
c
c       Move data to proc 0.  These must be called from all processors
c***********************************************************************

        call MPI_REDUCE(run,mpi_ibuf,1,MPI_INTEGER,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then                                            !0
          lrun=mpi_ibuf                                                 !0
          write(lud,*) 'total runs reduced to proc 0 =',lrun            !0
        endif                                                           !0
        call MPI_REDUCE(nsgn,mpi_ibuf,1,MPI_INTEGER,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then                                            !0
          lnsgn=mpi_ibuf                                                !0
          write(lud,*) 'runs with finite sgn reduced to proc 0 =',lnsgn !0
          if(lnsgn.ne.lrun) write(lud,*) 'SIGN PROBLEM'                 !0
        endif                                                           !0
c
c       get the average sign over all processors
        call MPI_REDUCE(sgnf,sgn0,1,MPI_REAL8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then
           sgn0=sgn0/real(lrun,kind)
        endif

        if(irvb.ne.0.and.Nc.gt.1.and.ndim.eq.2) then
        ! RVB column and staggered bond suscept.
        call MPI_REDUCE(chiRVB0,chiRVB0_buf,3,MPI_COMPLEX16,MPI_SUM,0,
     &MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(chiRVB_tau,chiRVB_tau_buf,nl*3,MPI_REAL8,
     &MPI_SUM,0,MPI_COMM_WORLD,ierr)
        ! BOW and its suscept 
        call MPI_REDUCE(chiBOW0,chiBOW0_buf,3,MPI_COMPLEX16,MPI_SUM,0,
     &MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(chiBOW_tau,chiBOW_tau_buf,nl*3,MPI_REAL8,
     &MPI_SUM,0,MPI_COMM_WORLD,ierr)
        ! spin-spin ZZ bond order and its suscept 
        call MPI_REDUCE(chizz0,chizz0_buf,3,MPI_COMPLEX16,MPI_SUM,0,
     &MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(chizz_tau,chizz_tau_buf,nl*3,MPI_REAL8,MPI_SUM,
     &0,MPI_COMM_WORLD,ierr)
        ! spin-spin SS bond order and its suscept 
        call MPI_REDUCE(chiss0,chiss0_buf,3,MPI_COMPLEX16,MPI_SUM,0,
     &MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(chiss_tau,chiss_tau_buf,nl*3,MPI_REAL8,MPI_SUM,
     &0,MPI_COMM_WORLD,ierr)
!       Calculate the RVB moment of the cluster
        call MPI_REDUCE(RVB0,RVB0buf,1,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(RVB1,RVB1buf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
 
        call MPI_REDUCE(RVB0d,RVB0dbuf,1,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(RVB1d,RVB1dbuf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)

        call MPI_REDUCE(RVB0xx,RVB0xxbuf,1,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(RVB1xx,RVB1xxbuf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)



        call MPI_REDUCE(AFM0,AFM0buf,1,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(AFM1,AFM1buf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)

        call MPI_REDUCE(BB0,BB0buf,1,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(BB1,BB1buf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(BBp1,BBp1buf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(BBp2,BBp2buf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(BBp3,BBp3buf,nl,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)

        call MPI_REDUCE(RVBchixxt,RVBchixxm,nl*Nc,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(RVBchixyt,RVBchixym,nl*Nc,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(ChiHHSt,ChiHHSm,nl*Nc,MPI_REAL8,MPI_SUM,0,
     &                  MPI_COMM_WORLD,ierr)


        if(myrank.eq.0) then   
          !for bond suscept.
          chiRVB0=chiRVB0_buf/(sgn0*meas*lrun*nl*Nc)
          chiRVB_tau=chiRVB_tau_buf/(sgn0*meas*lrun*nl*Nc*Nc)

          chiRVB_w0(0) = sum(chiRVB_tau(:,0))/nl
          chiRVB_w0(1) = sum(chiRVB_tau(:,1))/nl
          chiRVB_w0(2) = sum(chiRVB_tau(:,2))/nl
          chiRVB_w0 = chiRVB_w0 - real(conjg(chiRVB0)*chiRVB0)

          chiRVB_t0 = chiRVB_tau(1,:)
          chiRVB_t0 = chiRVB_t0 - real(conjg(chiRVB0)*chiRVB0)

          chiRVB_tau(:,0) = chiRVB_tau(:,0) - real(conjg(chiRVB0(0))
     &*chiRVB0(0))
          chiRVB_tau(:,1) = chiRVB_tau(:,1) - real(conjg(chiRVB0(1))
     &*chiRVB0(1))
          chiRVB_tau(:,2) = chiRVB_tau(:,2) - real(conjg(chiRVB0(2))
     &*chiRVB0(2))

          !for BOW and its suscept.
          chiBOW0 = chiBOW0_buf/(sgn0*meas*lrun*nl*Nc)
          chiBOW_tau = chiBOW_tau_buf/(sgn0*meas*lrun*nl*Nc*Nc)

          chiBOW_w0(0) = sum(chiBOW_tau(:,0))/nl
          chiBOW_w0(1) = sum(chiBOW_tau(:,1))/nl
          chiBOW_w0(2) = sum(chiBOW_tau(:,2))/nl
          chiBOW_w0 = chiBOW_w0 - real(conjg(chiBOW0)*chiBOW0)

          chiBOW_t0 = chiBOW_tau(1,:)
          chiBOW_t0 = chiBOW_t0 - real(conjg(chiBOW0)*chiBOW0)

          chiBOW_tau(:,0) = chiBOW_tau(:,0) - real(conjg(chiBOW0(0))
     &*chiBOW0(0))
          chiBOW_tau(:,1) = chiBOW_tau(:,1) - real(conjg(chiBOW0(1))
     &*chiBOW0(1))
          chiBOW_tau(:,2) = chiBOW_tau(:,2) - real(conjg(chiBOW0(2))
     &*chiBOW0(2))

          !for spin-spin bond ZZ and its suscept.
          chizz0 = chizz0_buf/(sgn0*meas*lrun*nl*Nc)
          chizz_tau = chizz_tau_buf/(sgn0*meas*lrun*nl*Nc*Nc)

          chizz_w0(0) = sum(chizz_tau(:,0))/nl
          chizz_w0(1) = sum(chizz_tau(:,1))/nl
          chizz_w0(2) = sum(chizz_tau(:,2))/nl
          chizz_w0 = chizz_w0 - real(conjg(chizz0)*chizz0)

          chizz_t0 = chizz_tau(1,:)
          chizz_t0 = chizz_t0 - real(conjg(chizz0)*chizz0)

          chizz_tau(:,0) = chizz_tau(:,0) - real(conjg(chizz0(0))
     &*chizz0(0))
          chizz_tau(:,1) = chizz_tau(:,1) - real(conjg(chizz0(1))
     &*chizz0(1))
          chizz_tau(:,2) = chizz_tau(:,2) - real(conjg(chizz0(2))
     &*chizz0(2))

          !for spin-spin bond SS and its suscept.
          chiss0 = chiss0_buf/(sgn0*meas*lrun*nl*Nc)
          chiss_tau = chiss_tau_buf/(sgn0*meas*lrun*nl*Nc*Nc)

          chiss_w0(0) = sum(chiss_tau(:,0))/nl
          chiss_w0(1) = sum(chiss_tau(:,1))/nl
          chiss_w0(2) = sum(chiss_tau(:,2))/nl
          chiss_w0 = chiss_w0 - real(conjg(chiss0)*chiss0)

          chiss_t0 = chiss_tau(1,:)
          chiss_t0 = chiss_t0 - real(conjg(chiss0)*chiss0)

          chiss_tau(:,0) = chiss_tau(:,0) - real(conjg(chiss0(0))
     &*chiss0(0))
          chiss_tau(:,1) = chiss_tau(:,1) - real(conjg(chiss0(1))
     &*chiss0(1))
          chiss_tau(:,2) = chiss_tau(:,2) - real(conjg(chiss0(2))
     &*chiss0(2))

          !for RVB moment  
          RVB0=RVB0buf/(sgn0*meas*lrun*nl*Nc)
          RVB0d=RVB0dbuf/(sgn0*meas*lrun*nl*Nc)
          RVB0xx=RVB0xxbuf/(sgn0*meas*lrun*nl*Nc)
          AFM0=AFM0buf/(sgn0*meas*lrun*nl*Nc)
          BB0=BB0buf/(sgn0*meas*lrun*nl*Nc)
          RVBchixxm=RVBchixxm/(sgn0*lrun)
          RVBchixym=RVBchixym/(sgn0*lrun)
          ChiHHSm=ChiHHSm/(sgn0*lrun)

          r1 = sum(RVB1buf)/(sgn0*meas*lrun*(Nc*nl)**2)
          r1d = sum(RVB1dbuf)/(sgn0*meas*lrun*(Nc*nl)**2)
          r1xx = sum(RVB1xxbuf)/(sgn0*meas*lrun*(Nc*nl)**2)
          r3= sum(AFM1buf)/(sgn0*meas*lrun*(Nc*nl)**2)
 
          RVBMOM=r1 - RVB0**2
          RVBdMOM=r1d - RVB0d**2
          RVBxxMOM=r1xx - RVB0xx**2
          AFMMOM=r3 - AFM0**2


          r5 = sum(BB1buf)/(sgn0*meas*lrun*Nc*nl**2)
          r6= sum(BBp1buf)/(sgn0*meas*lrun*Nc*nl**2)
          r7= sum(BBp2buf)/(sgn0*meas*lrun*Nc*nl**2)
          r8= sum(BBp3buf)/(sgn0*meas*lrun*Nc*nl**2)

          BB1MOM=r5 - BB0**2
          BBp1MOM=r6 - BB0**2
          BBp2MOM=r7 - BB0**2
          BBp3MOM=r8 - BB0**2
c         RVBchixx(l,i) = <s(i,l)s(i+x,l)s(0,0)s(x,0)>
c                       - <s(i,l)s(i+x,l)> <s(0,0)s(x,0)>
c                       - <s(i,l)s(0,0)>   <s(i+x,l)s(x,0)>
c                       - <s(i,l)s(x,0)>   <s(i+x,l)s(0,0)>
c         or
c         RVBchixx(l,i) = <s(i,l)s(i+x,l)s(0,0)s(x,0)>
c                       - <s(-x,0)s(0,0)>  <s(-x,0)s(0,0)>
c                       - <s(i,l)s(0,0)>   <s(i,l)s(0,0)>
c                       - <s(i-x,l)s(0,0)> <s(i+x,l)s(0,0)>
c         
c            i2=nneighbor(1,i1)     ! nn in the +x direction (2)  !     i3  i4
c            i3=nneighbor(2,i1)     ! nn in the +y direction (3)  !     
c            i4=nneighbor(2,i2)     ! nn in the +x+y direction (4)! i5  i1  i2
c            i5=nneighbor(1+ndim,i1)! nn in the -x direction (5)  ! 
c            i6=nneighbor(2+ndim,i1)! nn in the -y direction (6)  !     i6

          do ic=1,Nc
          do l=1,nl
            RVBchixxt(l,ic) = RVBchixxm(l,ic) 
     & - (ChiHHSm(1,nneighbor(1+ndim,icR0)))**2
     & - (ChiHHSm(l,ic))**2
     & - ChiHHSm(l,nneighbor(1+ndim,ic))*ChiHHSm(l,nneighbor(1,ic))

            Bondxx(l,ic) = RVBchixxm(l,ic) 
     &                     - (ChiHHSm(1,nneighbor(1+ndim,icR0)))**2
          end do
          end do
c         RVBchixxt = RVBchixxm - BB0**2

c         RVBchixy(l,i) = <s(i,l)s(i+x,l)s(0,0)s(y,0)>
c                       - <s(i,l)s(i+x,l)> <s(0,0)s(y,0)>
c                       - <s(i,l)s(0,0)>   <s(i+x,l)s(y,0)>
c                       - <s(i,l)s(y,0)>   <s(i+x,l)s(0,0)>
c         or
c         RVBchixy(l,i) = <s(i,l)s(i+x,l)s(0,0)s(x,0)>
c                       - <s(-x,0)s(0,0)>  <s(-y,0)s(0,0)>
c                       - <s(i,l)s(0,0)>   <s(i+x-y,l)s(0,0)>
c                       - <s(i-y,l)s(0,0)> <s(i+x,l)s(0,0)>
c         

          do ic=1,Nc
          do l=1,nl
            RVBchixyt(l,ic) = RVBchixym(l,ic) 
     & - (ChiHHSm(1,nneighbor(1+ndim,icR0)))
     & * (ChiHHSm(1,nneighbor(2+ndim,icR0)))
     & - (ChiHHSm(l,ic))*(ChiHHSm(l,nneighbor(2+ndim,nneighbor(1,ic))))
     & - ChiHHSm(l,nneighbor(2+ndim,ic))*ChiHHSm(l,nneighbor(1,ic))
            Bondxy(l,ic) = RVBchixym(l,ic)
     & - (ChiHHSm(1,nneighbor(1+ndim,icR0)))
     & * (ChiHHSm(1,nneighbor(2+ndim,icR0)))
          end do
          end do
c         RVBchixyt = RVBchixym - BB0**2

         
        
          r2 = RVB1buf(1)/(sgn0*meas*lrun*nl*Nc**2)
          r2 = r2 - RVB0**2
          r2d = RVB1dbuf(1)/(sgn0*meas*lrun*nl*Nc**2)
          r2d= r2d - RVB0d**2
          r2xx = RVB1xxbuf(1)/(sgn0*meas*lrun*nl*Nc**2)
          r2xx = r2xx - RVB0xx**2
          
          r4 = AFM1buf(1)/(sgn0*meas*lrun*nl*Nc**2)
          r4 = r4 - AFM0**2

          r5b = BB1buf(1)/(sgn0*meas*lrun*nl*Nc)
          r5b=r5b-BB0**2
          r6b= BBp1buf(1)/(sgn0*meas*lrun*nl*Nc)
          r6b=r6b-BB0**2
          r7b= BBp2buf(1)/(sgn0*meas*lrun*nl*Nc)
          r7b=r7b-BB0**2

          r8b= BBp3buf(1)/(sgn0*meas*lrun*nl*Nc)
          r8b=r8b-BB0**2

          if(ibow.ge.1.and.Nc.gt.1)then
!--------------------------------------------------------------------
! column and staggered RVB bond suscept.
          write(lud,*)
     &    '      T   chiRVB0_re     chiRVB0_im     chiRVB0(iw=0)  
     &chiRVB0(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiRVB0(0),
     &chiRVB_w0(0),chiRVB_t0(0)

          write(lud,*)
     &    '      T   chiRVB_cl0_re  chiRVB_cl0_im  chiRVB_cl(iw=0) 
     &chiRVB_cl(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiRVB0(1),
     &chiRVB_w0(1),chiRVB_t0(1)

          write(lud,*)
     &    '      T   chiRVB_st0_re  chiRVB_st0_im  chiRVB_st(iw=0) 
     &chiRVB_st(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiRVB0(2),
     &chiRVB_w0(2),chiRVB_t0(2)

c         Print out chiRVB_cl and chiRVB_st as a function of tau
          write(lud,*)
     &    '    tau   chiRVB(tau)    chiRVB_cl(tau)    chiRVB_st(tau)'
          do l=1,nl
            write(lud,"(1x,f7.4,3(4x,f10.4))")tau(l),chiRVB_tau(l,0),
     &chiRVB_tau(l,1),chiRVB_tau(l,2)
          end do

!--------------------------------------------------------------------
! BOW and its suscept
          write(lud,*)
     &    '      T   chiBOW0_re     chiBOW0_im     chiBOW0(iw=0)  
     &chiBOW0(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiBOW0(0),
     &chiBOW_w0(0),chiBOW_t0(0)

          write(lud,*)
     &    '      T   chiBOW0_cl_re  chiBOW0_cl_im  chiBOW_cl(iw=0) 
     &chiBOW_cl(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiBOW0(1),
     &chiBOW_w0(1),chiBOW_t0(1)

          write(lud,*)
     &    '      T   chiBOW0_st_re  chiBOW0_st_im  chiBOW_st(iw=0)  
     &chiBOW_st(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiBOW0(2),
     &chiBOW_w0(2),chiBOW_t0(2)

! Print out chiBOW_cl and chiBOW_st as a function of tau
          write(lud,*)
     &    '    tau   chiBOW(tau)    chiBOW_cl(tau)    chiBOW_st(tau)'
          do l=1,nl
            write(lud,"(1x,f7.4,3(4x,f10.4))")tau(l),chiBOW_tau(l,0),
     &chiBOW_tau(l,1),chiBOW_tau(l,2)
          end do
!----------------------------------------------------------------------
! spinzz bond order and its suscept
          write(lud,*)
     &    '      T   chizz0_re     chizz0_im     chizz0(iw=0)  
     &chizz0(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chizz0(0),chizz_w0(0),
     &chizz_t0(0)

          write(lud,*)
     &    '      T   chizz_cl0_re  chizz_cl0_im  chizz_cl(iw=0) 
     &chizz_cl(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chizz0(1),chizz_w0(1),
     &chizz_t0(1)

          write(lud,*)
     &    '      T   chizz_st0_re  chizz_st0_im  chizz_st(iw=0)
     &  chizz_st(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chizz0(2),chizz_w0(2),
     &chizz_t0(2)

! Print out chizz_cl and chizz_st as a function of tau
          write(lud,*)
     &    '    tau   chizz(tau)    chizz_cl(tau)    chizz_st(tau)'
          do l=1,nl
            write(lud,"(1x,f7.4,3(4x,f10.4))")tau(l),chizz_tau(l,0),
     &chizz_tau(l,1),chizz_tau(l,2)
          end do
!----------------------------------------------------------------------
! spinss bond order and its suscept
          write(lud,*)
     &    '      T   chiss0_re     chiss0_im      chiss0(iw=0)  
     &  chiss0(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiss0(0),chiss_w0(0),
     &chiss_t0(0)

          write(lud,*)
     &    '      T   chiss_cl0_re  chiss_cl0_im   chiss_cl(iw=0) 
     &  chiss_cl(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiss0(1),chiss_w0(1),
     &chiss_t0(1)

          write(lud,*)
     &    '      T   chiss_st0_re  chiss_st0_im   chiss_st(iw=0)
     &  chiss_st(tau=0)'
          write(lud,"(1x,f7.5,4(2x,f13.6))") temp,chiss0(2),chiss_w0(2),
     &chiss_t0(2)

! Print out chiss_cl and chiss_st as a function of tau
          write(lud,*)
     &    '    tau   chiss(tau)    chiss_cl(tau)    chiss_st(tau)'
          do l=1,nl
            write(lud,"(1x,f7.4,3(4x,f10.4))")tau(l),chiss_tau(l,0),
     &chiss_tau(l,1),chiss_tau(l,2)
          end do

          end if ! if(iBOW.ge.1.and.Nc.gt.1)
          
          !RVB and its suscept
!----------------------------------------------------------------------
          write(lud,*)
c         temperature,  static bond value, bond susceptibility iw=0, zero time bond susceptibility
     &           '    T         RVB0            RVB1         RVBSMOM',
     &           '       RVBUSMOM '! cluster, s-wave symmetry
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, RVB0, r1, RVBMOM,
     &                                       r2
         write(lud,*) 
     &           '    T         RVB0d          RVB1d         RVBdSMOM',
     &           '       RVBdUSMOM '                                    ! cluster, d-wave symmetry
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, RVB0d, r1d, RVBdMOM,
     &                                       r2d
           write(lud,*) 

     &           '    T         RVB0xx        RVB1xx         RVBxxSMOM',
     &           '       RVBxxUSMOM '! cluster, RVBxx
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, RVB0xx, r1xx, 
     &RVBxxMOM,
     &                                       r2xx
         write(lud,*) 


     &           '    T         AFM0          AFM1           AFMSMOM',  ! AFM operator ,.i.e. ABS(S^z(Q)), Q=(pi,pi)
     &           '       AFMUSMOM ' 
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, AFM0, r3, AFMMOM,
     &                                       r4

          write(lud,*) 
     &           '    T         BB0            BB1         BB1SMOM',
     &           '       BB1USMOM '  !local bond susceptibility
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, BB0, r5, BB1MOM,
     &                                       r5b

          write(lud,*) 
     &    '    T         BB0            BBperp         BBperpSMOM',
     &           '       BBperpUSMOM ' !nearest -neighbor perpendicular bond susceptibility
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, BB0, r6, BBp1MOM,
     &                                       r6b

             write(lud,*) 
     &    '    T         BB0            BBparal         BBparalSMOM',
     &           '       BBparalUSMOM ' !nearest -neighbor parallel bond susceptibility
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, BB0, r7, BBp2MOM,
     &                                 r7b

            if(Nc.gt.4) write(lud,*) 
     &    '    T         BB0            BBline        BBlineSMOM',
     &           '       BBlineUSMOM '                                  !nearest -neighbor bond neighbors, same direction
          write(lud,"(1x,f7.5,4(2x,e13.6))") temp, BB0, r8, BBp3MOM,
     &                                 r8b
     
c         Print out RVBchixx and RVBchixy as a function of X and tau
          write(lud,*) '  tau   Rx   Ry  RVBchixx   RVBchixy'
          do ic=1,Nc
          do l=1,nl
            write(lud,"(1x,f6.3,2x,i2,2x,i2,4x,f6.3,4x,f6.3)")
     &        tau(l),Rc(1,ic),Rc(2,ic),RVBchixxt(l,ic),RVBchixyt(l,ic)
          end do
          end do
          write(lud,*) '  '
          write(lud,*) '   T      RVBchixx   RVBchixy  '
          r1=sum(RVBchixxt)/nt; r2=sum(RVBchixyt)/nt
          write(lud,"(1x,f7.4,2x,f9.6,2x,f9.6)") temp,r1,r2
        end if !(myrank.eq.0)
        end if !(irvb.ne.0)
        if (Nc.eq.4.and.myrank.eq.0) then
        write(lud,*) "tau,S1,S2,S3,S4"
        i1=1
        i2=nneighbor(1,i1) ! nn in the +x direction (2)
        i3=nneighbor(2,i1) ! nn in the +y direction (3)
        i4=nneighbor(2,i2) ! nn in the +x+y direction (4)
        do l=1,nl
          write(lud,*) (l-1)*dt,
     &             s(i1+(l-1)*Nc),s(i2+(l-1)*Nc),
     &             s(i3+(l-1)*Nc),s(i4+(l-1)*Nc)
        end do
        end if

        
c       Reduce matrices to proc. 0, normalize and symmetrize.
        call Red_Symm(gf,lnsgn)
        call Red_Symm(gfs,lnsgn)
        call Red_Symm(chaf,lnsgn)
        call Red_Symm(chafs,lnsgn)
        call Red_Symm(chf,lnsgn)
        call Red_Symm(chfs,lnsgn)
        call Red_Symm(scf,lnsgn)
        call Red_Symm(scfs,lnsgn)
        
c       Reduce matrices to proc. 0, normalize, symmetrize, and divide by sgn
        call Red_Symmt(gft,sgn0,lrun)
        call Red_Symmt(chft,sgn0,lrun)
        call Red_Symmt(scft,sgn0,lrun)
        call Red_Symmt(chaft,sgn0,lrun)
       
        if (iter.eq.niter.or.iskip.eq.1) then
       
c       Calculate the autocorrelation time of a measurement determined
c       by the value of iauto contained in the arrays autodat(meas*run)
c       and autosgn(meas*run).  This will be done by calculating the
c       variance of the data, vari(l) as a function of bin size.
c       the number of datum is meas*run, we will allow l, the bin 
c       size, to run from 1 to meas.
c
c                    l vari(n)                 1   N             2
c       autolth(n) = ---------     vari(n) = ---- SUM (x_m - <x>)
c                    2 vari(1)               N-1  m=1
c
c       where x_m is the bin averages.  The autocorrelation length is
c       done this way for each Markov process and then the results are
c       reduced to process 0.
c
        avdat=sum(autodat(1:meas*run))
        avsgn=sum(abs(autosgn(1:meas*run))) ! just a normalization
        vari(:)=zeror; chii(:)=zeror; vari0(:)=zeror; chii0(:)=zeror
        do nsize=1,meas         ! the bin size
          nbins=(meas*run)/nsize  ! the number of bins of this size
          do nbin=1,nbins
c           calculate the bin averages
            r1=zeror;r2=zeror
            do i=(nbin-1)*nsize+1,nbin*nsize
              r1=r1+autodat(i)
              r2=r2+abs(autosgn(i))  ! just a normalization
            end do
            if(abs(r2).gt.eps.and.abs(avsgn).gt.eps) then 
              vari(nsize)=vari(nsize)+(r1/r2-avdat/avsgn)**2
              chii(nsize)=chii(nsize)+r1/r2
            end if
          end do
          if (nbins.gt.1)
     &    vari(nsize)=vari(nsize)/real(nbins-1,kind)
          chii(nsize)=chii(nsize)/real(nbins,kind)
        end do
c       Now put the autocorrelation length in vari
        if (abs(vari(1)).gt.eps) then
        r1=vari(1)
        do nsize=1,min(meas,40)
          vari(nsize)=halfr*nsize*vari(nsize)/r1
        end do
        end if 
        call MPI_REDUCE(vari,vari0,meas,MPI_REAL8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        call MPI_REDUCE(chii,chii0,meas,MPI_REAL8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then
          vari0=vari0/real(nproc,kind)
          chii0=chii0/real(nproc,kind)
          write(lud,*) '***********************************************'
          if (iauto.eq.1) then
          write(lud,*) ' Autocorrelation Length of the cluster chi_spin'
          else if (iauto.eq.2) then
          write(lud,*) 
     &    ' Autocorrelation Length of the cluster chi_pairing'
           else if (iauto.eq.3) then
          write(lud,*) 
     &    ' Autocorrelation Length of the cluster chi_charge' 
         else if (iauto.eq.4) then
          write(lud,*) 
     &    ' Autocorrelation Length of the cluster G(k=0,w(0))'   
       else if (iauto.eq.5) then
          write(lud,*) 
     &    ' Autocorrelation Length of the cluster G(r=0,t=0)'
          end if
          if(iter.eq.niter) then
          if (iauto.eq.1) then
             write(lud,*) 'bin size    chii(binsize)   tau(binsize)'
          else if(iauto.eq.2) then
             write(lud,*) 'bin size    Pdii(binsize)   tau(binsize)'
          else if(iauto.eq.3) then
             write(lud,*) 'bin size    Chaii(binsize)   tau(binsize)'
          else if(iauto.eq.4) then
             write(lud,*) 'bin size    Gkw_ii(binsize)   tau(binsize)'
         else if(iauto.eq.5) then
             write(lud,*) 'bin size    Grt_ii(binsize)   tau(binsize)'

          end if
            do nsize=1,min(40,meas)
              write(lud,*) nsize,chii0(nsize),vari0(nsize)
            end do
          end if
          autotime=zeror
          do nsize=10,min(40,meas)
            autotime=autotime+vari0(nsize)
          end do
          if(min(40,meas).gt.9) 
     &       autotime=autotime/real(min(40,meas)-9,kind)
          write(lud,*) '  '
          write(lud,*) 'autotime= ',autotime
          write(lud,*) '***********************************************'

        end if ! myrank.eq.0

        end if
        
        if(iter.eq.niter.and.isg.eq.1) then
c         Now reduce the two-particle properties.  First, we must 
c         deallocate temat1 and temat2 to make room for temp arrays.
          deallocate (temat1)
          deallocate (temat2)
          allocate (temath(ntw,nt),stat=info)
          allocate (tematrh(ntw,nt),stat=info)
c         we need size-specific MPI datatypes: MPI_<TYPE>n, where
c         <TYPE> is REAL, COMPLEX, etc. and n is the number of byte!!
c         needed for the variable.

c         Reduce matrices to proc. 0, normalize, symmetrize, and divide by sgn
          call Red_Gamat(gamatcenter,sgn0,lrun*meas)
          call Red_Gamat_real(gamatscenter,sgn0,lrun*meas)
          call Red_Gamat(gamatccenter,sgn0,lrun*meas)
          call Red_Gamat(gamatcorner,sgn0,lrun*meas)
          call Red_Gamat(gamatscorner,sgn0,lrun*meas)
          call Red_Gamat(gamatccorner,sgn0,lrun*meas)
c           
          call Red_Gamat0(gamatc0center,sgn0,lrun*meas)     
          call Red_Gamat0(gamatc0corner,sgn0,lrun*meas)    
c                                                                       
c         Now reallocate temat1 and temat2 and deallocate 
c         tematrh and temath
          deallocate (temath)
          deallocate (tematrh)
          allocate (temat1(nt,nt),stat=info)
          allocate (temat2(nt,nt),stat=info)
        end if                                                          
                                                                        
                                                                        
c
c       reduce the total number of updates
c        call MPI_REDUCE(ucount,mpi_ibuf,1,MPI_INTEGER,MPI_SUM,0,
c     &                  MPI_COMM_WORLD,ierr)
c
c       We now reduce, and normalize a few scalars.
        mpi_rbufout(1) = sgnf
        mpi_rbufout(2) = sgnfs
        mpi_rbufout(3) = moof
        mpi_rbufout(4) = moofs
        mpi_rbufout(5) = acrat
        if(isv.eq.1) then
          mpi_rbufout(6) = run*mooft/sgnf
        else
          mpi_rbufout(6) = mooft
        end if
        mpi_rbufout(7) = acount
        mpi_rbufout(8) = ucount
        call MPI_REDUCE(mpi_rbufout,mpi_rbufin,8,
     &                  MPI_REAL8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then                  
          ucount=mpi_rbufin(8)
          sgnf  = mpi_rbufin(1)/real(lrun,kind)
          sgnfs = mpi_rbufin(2)/real(lrun,kind)
          moof  = mpi_rbufin(3)/real(lnsgn,kind)
          moofs = mpi_rbufin(4)/real(lnsgn,kind)
          acrat = mpi_rbufin(5)/ucount
          mooft = mpi_rbufin(6)/(real(lrun,kind))
          if (abs(mpi_rbufin(7)).lt.eps) then  ! prevent floating exception
                  acount=zeror
          else
                  acount = acrat*ucount/
     &                     mpi_rbufin(7)
          end if
          if(isv.eq.0) mooft=mooft/sgn0                                 !0
          if(iskip.eq.1) then
c           adjust skip to keep autotime=autoln
            skip=max(1,int(skip*autotime/autoln))
            skip=min(skipmax,skip)
            write(lud,*) 'skip= ',skip
          end if
        end if
        if(iskip.eq.1) 
     &    call MPI_BCAST(skip,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)      ! skip
 

c***********************************************************************
c
c       DATA REDUCTION COMPLETED.
c
c       At this point all the needed variables should be reduced to
c       proc. 0, symmetrized and normalized.
c***********************************************************************
        
c***********************************************************************
c       Now calculate some error bars, and write some data
c***********************************************************************
        if(myrank.eq.0) then                                            !0
          if(lnsgn.gt.1)                                                !0
     &      sigmoo=sqrt(abs(moofs-moof**2)/real(lnsgn-1,kind))          !0
          if(lrun.gt.1)                                                 !0
     &      sigg=sqrt(abs(sgnf**2-sgnfs)/real(lrun-1,kind))             !0
          write(lud,*) '    mu**2           +/- '                       !0
          write(lud,406) mooft,sigmoo                                   !0
 406      format(3x,f10.7,5x,f10.7)                                     !0
          write(lud,404)                                                !0
     &    sgnf,sigg,acrat,acount					!0
 404      format(' sgn=',f9.6,' +/- ',f9.6,' acrat= ',f9.6,		!0
     &           ' avg. Nblocka=',f9.6) 				!0
                                                                        !0
          write(lud,*) ' '                                              !0
          if(iter.eq.niter) then                                        !0
            write(lud,*) ' n ic  Gtau(n,ic)    +/-   CHItau(n,ic)  +/-',!0
     &                '  CHAtau(n,ic)  +/-   SCtau(n,ic)  +/- '         !0
          else                                                          !0
            write(lud,*) ' n ic  Gtau(n,ic)    +/-   CHAtau(n,ic)  +/-' !0
          end if                                                        !0
c23456789a123456789b123456789c123456789d123456789e123456789f123456789012!0
          do i=1,nl                                                     !0
          do ic=1,Nc                                                    !0
            if(lnsgn.gt.1) then                                         !0
              sigg=sqrt(abs(gf(i,ic)**2-gfs(i,ic))/(lnsgn-1))           !0
              sigch=sqrt(abs(chf(i,ic)**2-chfs(i,ic))/(lnsgn-1))        !0
              sigsc=sqrt(abs(scf(i,ic)**2-scfs(i,ic))/(lnsgn-1))        !0
              sigcha=sqrt(abs(chaf(i,ic)**2-                            !0
     &                    chafs(i,ic))/(lnsgn-1))                       !0
            end if                                                      !0
c           Now subtract the non-interacting part of the local          !0
c           charge susceptibility (it is a very small contribution to   !0
c           the error bar).                                             !0
            chaft(i,ic)=chaft(i,ic)-(2.0*(1.0-gft(1,icR0)))**2          !0
            if(iter.eq.niter) then                                      !0
              write(lud,401) i,ic,gft(i,ic),sigg,chft(i,ic),sigch,      !0
     &                      chaft(i,ic),sigcha,scft(i,ic),sigsc         !0
            else if(i.eq.1) then   ! write the tau=0 result             !0
              write(lud,401) i,ic,gft(i,ic),sigg,chaft(i,ic),sigcha     !0
            end if                                                      !0
c note the next line is a potential problem as it is the only           !0
c place where sig_nfocc is initialized                                  !0
            if(i.eq.1.and.ic.eq.1) sig_nfocc=sigg                       !0
          end do                                                        !0
          end do                                                        !0
 401      format(i3,1x,i2,1x,f9.6,1x,7f10.6)                            !0
 408      format(f7.3,2x,f7.5,2x,f7.5,2x,f7.5,2x,f7.5)                  !0
          nfocc=1.0-gft(1,icR0)                                         !0
          if(iter.eq.niter) then                                        !0
            write(lud,*) '   '                                          !0
            write(lud,*) '  tau   <n(tau)n(0)>  +/-  '                  !0
            do i=1,nl                                                   !0
              if(lnsgn.gt.1) then                                       !0
                sigcha=sqrt(abs(chaf(i,icR0)**2-                        !0
     &                      chafs(i,icR0))/(lnsgn-1))                   !0
              end if                                                    !0
              write(lud,408) (i-1)*dt,chaft(i,icR0),sigcha              !0
            end do                                                      !0
            write(lud,*) '   '                                          !0
          end if                                                        !0
                                                                        !0
c                                                                       !0
c         R=0 Cluster charge, spin and pairing susceptibility           !0
c                                                                       !0
          susmf3=zeror                                                  !0
          suscf3=zeror                                                  !0
          suspf3=zeror                                                  !0
          h=twor/threer                                                 !0
          do l=1,nl                                                     !0
            susmf3=susmf3+2.0*h*chft(l,icR0)                            !0
            suscf3=suscf3+h*chaft(l,icR0)                               !0
            suspf3=suspf3+h*scft(l,icR0)                                !0
            h=2.0-h                                                     !0
          end do                                                        !0
          susmf3=susmf3*beta/real(nl,kind)                              !0
          suscf3=suscf3*beta/real(nl,kind)                              !0
          suspf3=suspf3*beta/real(nl,kind)                              !0
c                                                                       !0
c         Q=0 Cluster charge, spin and pairing susceptibility           !0
c                                                                       !0
          chaclu=zeror                                                  !0
          chiclu=zeror                                                  !0
          scfclu=zeror                                                  !0
          do ic=1,Nc                                                    !0
            h=twor/threer                                               !0
            do l=1,nl            ! the integral in tau                  !0
              chaclu=chaclu+h*chaft(l,ic)                               !0
              chiclu=chiclu+h*chft(l,ic)                                !0
              scfclu=scfclu+h*scft(l,ic)                                !0
              h=twor-h                                                  !0
            end do                                                      !0
          end do                                                        !0
          chaclu=chaclu*beta/real(nl,kind)                              !0
          chiclu=chiclu*beta/real(nl,kind)                              !0
          scfclu=scfclu*beta/real(nl,kind)                              !0
          write(lud,*) '  '                                             !0
c                                                                       !0
c         Q=Pi Cluster charge and spin susceptibility                   !0
c                                                                       !0
          chaafclu=zeror                                                !0
          chiafclu=zeror                                                !0
          do ic=1,Nc                                                    !0
            h=twor/threer                                               !0
            r1=(-1)**sum(Rc(:,ic))                                      !0
            do l=1,nl            ! the integral in tau                  !0
              chaafclu=chaafclu+h*r1*chaft(l,ic)                        !0
              chiafclu=chiafclu+h*r1*chft(l,ic)                         !0
              h=twor-h                                                  !0
            end do                                                      !0
          end do                                                        !0
          chaafclu=chaafclu*beta/real(nl,kind)                          !0
          chiafclu=chiafclu*beta/real(nl,kind)                          !0
c                                                                       !0
c         Q=Pi zero-time Cluster charge and spin susceptibility         !0
c                                                                       !0
          chaafzt=zeror                                                 !0
          chiafzt=zeror                                                 !0
          do ic=1,Nc                                                    !0
            r1=(-1)**sum(Rc(:,ic))                                      !0
            chaafzt=chaafzt+r1*chaft(1,ic)                              !0
            chiafzt=chiafzt+r1*chft(1,ic)                               !0
          end do                                                        !0
c                                                                       !0
c         Write some of the results to lud                              !0
c                                                                       !0
          write(lud,*) '  '                                             !0
          write(lud,*) 'Chaclu=',chaclu,' Cha_ii=',suscf3               !0
          if(iter.eq.niter) then                                        !0
            write(lud,*) 'Chiclu=',chiclu,' Scfclu=',scfclu             !0
            write(lud,*) 'Chiafclu=',chiafclu                           !0
            write(lud,*) 'Chaafclu=',chaafclu                           !0
            write(lud,*) 'Chiafzt=',chiafzt                             !0
            write(lud,*) 'Chaafzt=',chaafzt                             !0
          end if                                                        !0
c                                                                       !0
c                                                                       !0
c***********************************************************************!0
c         Now calculate the cluster susceptibilities chi(R,inu) and     !0
c         chi(K,inu) using Akima oversampling.                          !0
c***********************************************************************!0
          dtfine=beta/real(nl*nover,kind)                               !0
c                                                                       !0
          if(irvb.ne.0.and.Nc.gt.1.and.ndim.eq.2) then                  !0
	    itop = 5                                                    !0
	  else                                                          !0
	    itop = 1                                                    !0
	  end if                                                        !0
          if(iter.eq.niter) then                                        !0
          do itype=-1,itop                                              !0
             if(itype.eq.-1) then                                       !0
              write(lud,*) ' '                                          !0
              write(lud,*) ' Cluster Magnetic Susceptibility '          !0
            else if(itype.eq.0) then                                    !0
              write(lud,*) ' '                                          !0
              write(lud,*) ' Cluster Charge Susceptibility '            !0
            else if(itype.eq.1) then                                    !0
              write(lud,*) ' '                                          !0
              write(lud,*) ' Cluster Pairing Susceptibility '           !0
            else if(itype.eq.2) then                                    !0
              write(lud,*) ' '                                          !0
              write(lud,*) ' Cluster xx Zbond Susceptibility '          !0
            else if(itype.eq.3) then                                    !0
              write(lud,*) ' '                                          !0
              write(lud,*) ' Cluster xy Zbond Susceptibility '          !0
            else if(itype.eq.4) then                                    !0
              write(lud,*) ' '                                          !0
              write(lud,*) ' Cluster xx  four-spin cumulant '          !0
            else if(itype.eq.5) then                                    !0
              write(lud,*) ' '                                          !0
              write(lud,*) ' Cluster xy four-spin cumulant  '          !0
           end if                                                      !0
                                                                        !0
            do ic=1,Nc                                                  !0
c             Form the data to be splined.                              !0
              XNspline(0:nl)=tau(1:nl+1)                                !0
              if(itype.eq.-1) then                                      !0
                FNspline(0:nl-1) = chft(1:nl,ic)                        !0
                FNspline(nl)     = chft(1,ic)                           !0
              else if(itype.eq.0) then                                  !0
                FNspline(0:nl-1) = chaft(1:nl,ic)                       !0
                FNspline(nl)     = chaft(1,ic)                          !0
              else if(itype.eq.1) then                                  !0
                FNspline(0:nl-1) = scft(1:nl,ic)                        !0
                FNspline(nl)     = scft(1,ic)                           !0
              else if(itype.eq.2) then                                  !0
                FNspline(0:nl-1) = Bondxx(1:nl,ic)                      !0
                FNspline(nl)     =  Bondxx(1,ic)                        !0
              else if(itype.eq.3) then                                  !0
                FNspline(0:nl-1) = Bondxy(1:nl,ic)                      !0
                FNspline(nl)     =  Bondxy(1,ic)                        !0
              else if(itype.eq.4) then                                  !0
                FNspline(0:nl-1) = RVBchixxt(1:nl,ic)                   !0
                FNspline(nl)     = RVBchixxt(1,ic)                      !0
              else if(itype.eq.5) then                                  !0
                FNspline(0:nl-1) = RVBchixyt(1:nl,ic)                   !0
                FNspline(nl)     = RVBchixyt(1,ic)                      !0


              end if                                                    !0
              BETAspline=0.5_kind                                       !0
              Nspline=nl                                                !0
              il=0                                                      !0
c             Form the spline coeffients                                !0
              call akima1(Nspline,XNspline,FNspline,NKmax,              !0
     &                  BETAspline,Bspline,Cspline,Dspline,il,          !0
     &                  HELPspline,Ispline)                             !0
              if(Ispline.ne.0)                                          !0
     &                   write(lud,*) 'Akima Failure, Ispline=',Ispline !0
                                                                        !0
                                                                        !0
c             Now form the dense splined data for oversampling          !0
              do l=1,nover*nl                                           !0
                ta=beta*(l-1)/real(nover*nl,kind)                       !0
                gfine(l)=spval(ta,Nspline,XNspline,FNspline,            !0
     &                       Bspline,Cspline,Dspline)                   !0
              end do                                                    !0
c             Now do the FT, i.e. integrate the oversampled chi         !0
c             times a Fourier factor.                                   !0
              do n=0,nl/2                                               !0
                csum=zeroc                                              !0
                r1=2.0/3.0                                              !0
                r2=twor*pi*temp*n                                       !0
                do l=1,nl*nover                                         !0
                  csum=csum +dtfine*r1                                  !0
     &               *exp(ii*r2*(l-1)*dtfine)*gfine(l)                  !0
                  r1=2.0-r1                                             !0
                end do                                                  !0
                Gcsf(n,ic)=-csum                                  !  Note the sign change !!!!!!
                Gcsf(-n,ic)=conjg(Gcsf(n,ic))                           !0
              end do                                                    !0
            end do  ! ic                                                !0
            if(itype.lt.2) then                                         !0
              write(lud,*) '	nu	       Chi(R=0,inu) '           !0
              do n=0,nl/2                                               !0
                r2=twor*pi*temp*n                                       !0
                write(lud,"(3(f9.4,2x))")  r2,Gcsf(n,icR0)              !0
              end do                                                    !0
            else if (itype.eq.2) then
              write(lud,*) '  '
              write(lud,*) '   vn        Rx       Ry       ChiZBondxx(R,ivn)'
              do n=0,min(3,nl/2)
              do ic=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,i3,2x,i3,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Rc(1,ic),Rc(2,ic),Gcsf(n,ic)
              end do
              end do                                              !0
            else if (itype.eq.3) then
   	      write(lud,*) '  '
              write(lud,*) '   vn        Rx       Ry       ChiZBondxy(R,ivn)'
              do n=0,min(3,nl/2)
              do ic=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,i3,2x,i3,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Rc(1,ic),Rc(2,ic),Gcsf(n,ic)
              end do
              end do
            else if (itype.eq.4) then
              write(lud,*) '  '
              write(lud,*) '   vn        Rx       Ry       FourSpinxx(R,ivn)'
              do n=0,min(3,nl/2)
              do ic=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,i3,2x,i3,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Rc(1,ic),Rc(2,ic),Gcsf(n,ic)
              end do
              end do                                              !0
	    else if (itype.eq.5) then
   		write(lud,*) '  '
              write(lud,*) '   vn        Rx       Ry       FourSpinxy(R,ivn)'
              do n=0,min(3,nl/2)
              do ic=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,i3,2x,i3,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Rc(1,ic),Rc(2,ic),Gcsf(n,ic)
              end do
              end do

            endif

                                                                        !0
c           Now fourier transform Gcsf to obtain GcKf                   !0
            call FT_Rf_to_Kf(Gcsf,GcKf)                                 !0

            if(itype.eq.-1) then
              write(lud,*) '  '
              write(lud,*) 'ivn   Chis(pi,ivn)'
              do n=0,nl/2
                write(lud,448) twor*pi*temp*n, GcKf(n,icKpi)
              end do

c             also print 1/Nc sum_kk' (coskx-cosky) chi(k-k') (coskx'-coxky')  
              write(lud,*) '  '
              write(lud,*) "ivn  sum_KK' (cosKx-cosKy)Chis(K-K',ivn)(cosKx'-cosKy')"
              do n=0,nl/2
                csum = zeroc
                do ic=1,Nc
                do jc=1,Nc
                  r1=gdwave(ic) ; r2=gdwave(jc) ; ick=ickdiff(ic,jc)
                  csum = csum + r1*r2*GcKf(n,ick)
                end do
                end do 
                write(lud,448) twor*pi*temp*n, csum/float(Nc)
              end do
              write(lud,*) '  '
              write(lud,*) '   vn        Kx       Ky        Chis(K,ivn)'
              do n=0,min(3,nl/2)
	      do ick=1,ncw  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,f7.4,2x,f7.4,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Kc(1,ick),Kc(2,ick),GcKf(n,ick)
              end do
              end do
            else if (itype.eq.1) then
              write(lud,*) '  '
              write(lud,*) 'ivn   Chic(0,ivn)'
              do n=0,nl/2
                write(lud,448) twor*pi*temp*n, GcKf(n,icK0)
              end do
            else if (itype.eq.2) then
              write(lud,*) '  '
              write(lud,*) '   vn        Kx       Ky       ChiZBondxx(K,ivn)'
              do n=0,min(3,nl/2)
	      do ick=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,f7.4,2x,f7.4,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Kc(1,ick),Kc(2,ick),GcKf(n,ick)
              end do
              end do
            else if (itype.eq.3) then
              write(lud,*) '  '
              write(lud,*) '   vn        Kx       Ky       ChiZBondxy(K,ivn)'
              do n=0,min(3,nl/2)
              do ick=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,f7.4,2x,f7.4,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Kc(1,ick),Kc(2,ick),GcKf(n,ick)
              end do
              end do
            else if (itype.eq.4) then
              write(lud,*) '  '
              write(lud,*) '   vn        Kx       Ky       FourSpinxx(K,ivn)'
              do n=0,min(3,nl/2)
	      do ick=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,f7.4,2x,f7.4,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Kc(1,ick),Kc(2,ick),GcKf(n,ick)
              end do
              end do
            else if (itype.eq.5) then
              write(lud,*) '  '
              write(lud,*) '   vn        Kx       Ky        FourSpinxy(K,ivn)'
              do n=0,min(3,nl/2)
              do ick=1,Nc  ! on K in the IRW
                write(lud,"(1x,f8.4,2x,f7.4,2x,f7.4,2x,f8.3,2x,f8.3)") 
     &            twor*pi*temp*n, Kc(1,ick),Kc(2,ick),GcKf(n,ick)
              end do
              end do


            end if
          end do ! itype                                                
          end if ! iter.eq.niter                                        
                                                                        !0
 448      format(f7.3,2x,f8.3,2x,f8.3)                                  !0
c***********************************************************************!0
c         Now calculate gfwn and sigma from gf                          !0
c                                                                       !0
c         This will be done by splining the the difference              !0
c                                                                       !0
c         gft(tau) - Gcstosft(tau) ,                                    !0
c                                                                       !0
c         and then oversampling with a sampling rate (in imag. time):   !0
c                                                                       !0
c***********************************************************************!0
          dtfine=beta/real(nl*nover,kind)                               !0
c                                                                       !0
          do ic=1,Nc                                                    !0
                                                                        !0
c           AKIMA SPLINE is IMPORTANT, since it does not generate       !0
c           spurious high-frequency features.    Numerical Recipes      !0
c           spline and splint functions are inferior.                   !0
c           Form the data to be splined.                                !0
            XNspline(0:nl)=tau(1:nl+1)                                  !0
            FNspline(0:nl-1)=gft(1:nl,ic)-Gcstosft(0:nl-1,ic)           !0
            FNspline(nl)=Gcstosft(0,ic)-gft(1,ic)                       !0
            BETAspline=0.5_kind                                         !0
            Nspline=nl                                                  !0
            il=0                                                        !0
c           Form the spline coeffients                                  !0
            call akima1(Nspline,XNspline,FNspline,NKmax,                !0
     &                  BETAspline,Bspline,Cspline,Dspline,il,          !0
     &                  HELPspline,Ispline)                             !0
            if(Ispline.ne.0)                                            !0
     &                   write(lud,*) 'Akima Failure, Ispline=',Ispline !0
                                                                        !0
                                                                        !0
c           Now form the dense splined data for oversampling            !0
            do l=1,nover*nl                                             !0
              ta=beta*(l-1)/real(nover*nl,kind)                         !0
              gfine(l)=spval(ta,Nspline,XNspline,FNspline,              !0
     &                       Bspline,Cspline,Dspline)                   !0
c             write(lud,*) ta,gfine(l)                                  !0
            end do                                                      !0
c           Now do the FT, i.e. integrate the oversampled difference    !0
c           gft(tau) - Gcstosft(tau) times a Fourier factor.            !0
            do n=0,nuse                                                 !0
              csum=zeroc                                                !0
              r1=2.0/3.0                                                !0
              do l=1,nl*nover                                           !0
                csum=csum +dtfine*r1                                    !0
     &               *exp(ii*wn(n)*(l-1)*dtfine)*gfine(l)               !0
                r1=2.0-r1                                               !0
              end do                                                    !0
c             The next step is a low-pass filter used to reduce noise   !0
c             This step may be unneccessary when Akima is used.         !0
c             Presently, the cutoff frequency is pi/dtau, the Nyquist   !0
c             frequency                                                 !0
              if(ifilter.eq.1)                                          !0
     &          csum=csum/(exp(halfr*(real(iabs(2*n+1)-nl)))+1.0)       !0
c             Now form the Green fuction by adding back Gcsfosft(iwn)   !0
              Gcsf(n,ic)=-csum + Gcsfosft(n,ic)                         !0
              Gcsf(-n-1,ic)=conjg(Gcsf(n,ic))                           !0
            end do                                                      !0
          end do                                                        !0
                                                                        !0
c         Now fourier transform Gcsf to obtain GcKf                     !0
          call FT_Rf_to_Kf(Gcsf,GcKf)                                   !0
                                                                        !0
c         Symmetrize GcKf                                               !0
          call Symm_PG(GcKf)                                            !0
          if(abs(tprime).lt.eps.and.abs(ed).lt.eps)                     !0
     &      call Symm_PH(GcKf)                                          !0
                                                                        !0
c         Extract the self energy by inverting                          !0
c                                                                       !0
c                            Gscr(iwn,K)                                !0
c         G (iwn,K)= -----------------------------                      !0
c                     1 - Gscr(iwn,K)*sigma(iwn,K)                      !0
c                                                                       !0
          if(isd.eq.1.and.iter.gt.1.or.isd.eq.1.and.iso.ge.1) then      !0
            sigma(-nuse-1:nuse,:)= dfac*(1.0/GcKfsc(-nuse-1:nuse,:)-    !0
     &                                   1.0/GcKf(-nuse-1:nuse,:)    )  !0
     &                           + (1.0-dfac)*sigma(-nuse-1:nuse,:)     !0 
          else                                                          !0
            sigma(-nuse-1:nuse,:)=1.0/GcKfsc(-nuse-1:nuse,:)            !0
     &                           -1.0/GcKf(-nuse-1:nuse,:)              !0
          end if                                                        !0
                                                                        !0
c         Now (optionally) append the perturbation theory self-energy   !0
c         or the HF form onto the QMC non-perturbative self energy      !0
          if(icond.eq.0) then                                           !0
            write(lud,*) 'Sigma includes asymptotic part from P.T.'     !0
            do ic=1,Nc                                                  !0
            do n=nuse+1,nwn-1                                           !0
              sigma(n,ic)=sigmapt(n,ic)                                 !0
              sigma(-n-1,ic)= conjg(sigmapt(n,ic))                      !0
            end do                                                      !0
            end do                                                      !0
          else if(icond.eq.1) then                                      !0
            write(lud,*) 'self energy includes asymptotic part from HFC'!0
c           Parameterize the high-frequency form of the self energy     !0
c           This only makes sense if nuse <= nl/2 (Nyquist cutoff pi/dt)!0
            do ic=1,Nc                                                  !0
              acond(ic)=(wn(nuse)**2)                                   !0
     &                 *real(sigma(nuse,ic)-u*(nfocc-halfr),kind)       !0
              bcond(ic)=wn(nuse)**3*(aimag(sigma(nuse,ic))+             !0
     &                 (u**2*nfocc*(oner-nfocc))/wn(nuse))              !0
            end do                                                      !0
                                                                        !0
            do ic=1,Nc                                                  !0
            do n=nuse+1,nwn-1                                           !0
              sigma(n,ic)=u*(nfocc-halfr)-ii*u**2*nfocc*(oner-nfocc)    !0
     &               /wn(n)+acond(ic)/(wn(n)**2)+ii*bcond(ic)/(wn(n)**3)!0
              Sigma(-n-1,ic)= conjg(sigma(n,ic))                        !0
            end do                                                      !0
            end do                                                      !0
          end if                                                        !0
                                                                        !0
c         test for convergence by inspecting Im[sigma(0,ic) ]           !0
          testl=test                                                    !0
          test=zeror                                                    !0
          do ic=1,Nc                                                    !0
            test = test + abs(aimag(sigma(0,ic)))                       !0
          end do                                                        !0
          if(iter.ge.2) then                                            !0
            r1=abs(testl-test)/real(Nc,kind)                            !0
            write(lud,*) 'iter= ',iter,' test= ',r1                     !0
          end if                                                        !0
                                                                        !0
        end if  ! myrank.eq.0                                           !0
                                                                        
        call Calculate_ed(chaclu,sig_nfocc) !recalc. ed and broadcast

c       Now send the calculated sigma to all processors

        call MPI_BCAST(sigma,(2*nwn+1)*Nc,
     &                 MPI_COMPLEX16,
     &                 0,MPI_COMM_WORLD,ierr)

c       Now that we have the new ed adn sigma, write out sigma, ed,...
c       tio sigma.dat if this is the last iteration. 
        if(iter.eq.niter.and.myrank.eq.0) then                        !0
          rewind(lus)                                                 !0
          call Header_Sigma   ! parameters on top of sigma.dat        !0
                                                                      !0
          write(lus,*) '          mu**2            T*chi_ii(T)'       !0
          write(lus,*) mooft,temp*susmf3                              !0
          write(lus,*) '     tau        ic     gf(tau)     '          !0
                                                                      !0
          do n=1,nl                                                   !0
          do ic=1,Nc                                                  !0
            ta=(n-1)*dt                                               !0
            write(lus,"(1x,e12.5,2x,i4,2x,e12.5,2x,e12.5)")           !0 
     &        ta,ic,gft(n,ic)                                         !0
          end do                                                      !0
          end do                                                      !0
          write(lus,*) '  n    s(n)  the HHS fields    '              !0
          write(lus,*) sgn                                            !0
          do n=1,nt                                                   !0
            write(lus,*) n,s(n)                                       !0
          end do                                                      !0
          write(lus,*) '  n   ic           sigma(n,ic)        ',      !0
     &                '        sigmapt(n,ic)'                         !0
          do ic=1,Ncw         ! Only write for K in the IW            !0
          do n=0,nwn                                                  !0
            z = sigma(n,ic)       !convert to kindg                   !0
            z1= sigmapt(n,ic)     !convert to kindg                   !0
            write(lus,44) n,ic,z,z1                                   !0
          end do                                                      !0
          end do                                                      !0
        end if        ! iter.eq.niter.and.myrank.eq.0                 !0
                                                                      !0
 44     format(1x,i3,2x,i3,2x,'(',e13.6,',',e13.6,')',x,              !0
     &         '(',e13.6,',',e13.6,')' )                              !0
c                                                                     !0
c***********************************************************************
c       On the last iteration, condition and write out the 2-particle
c       properties.
c***********************************************************************
        if(iter.eq.niter.and.isg.ge.1) then                             
c
c         Use all procs to make a new GcKf used for conditioning        
          call CoarseGrain(GcKf,sigma)                                  
          if(myrank.eq.0) then                                          !0
c           Condition the high-frequency part of the two-particle       !0
c           susceptibilities.  This is done by first repeating the      !0
c           measurement of the two-particle parts using the averaged    !0
c           green function gf:                                          !0
            write(lud,*) 'cond. the high-frequency part of gamma      *'!0
                                                                        !0
                                                                        !0
c           First transform GcKf(n,ic) to space-time.  Do the time      !0
c           part first, which is real due to inversion symmetry.        !0
            call FT_Kf_to_Kt(GcKf,GcKt,1)                               !0
                                                                        !0
c           Now the spatial part, which again is real                   !0
            call FT_Kt_to_Rt(GcKt,Gcstsc)                               !0
                                                                        !0
                                                                        !0
c           Q=(0,0)                                                     !0
c           Zone Center susceptibilities:                               !0
c                                                                       !0
c                                                                       !0
c           first form the temporary vectors which store G-1/2I         !0
            do m=1,nl                                                   !0
            do n=1,nl                                                   !0
            do ic=1,Nc                                                  !0
            do jc=1,Nc                                                  !0
              k=n-m                                                     !0
              i=icrdiff(ic,jc)                                          !0
              ind=ic + (n-1)*Nc                                         !0
              imd=jc + (m-1)*Nc                                         !0
              gd(ind,imd)=Gcstsc(k,i)                                   !0
            end do                                                      !0
            end do                                                      !0
            end do                                                      !0
            end do                                                      !0
                                                                        !0
            do n=1,nt                                                   !0
            do m=1,nt                                                   !0
              temat3(n,m)=gd(n,m)                                       !0
            end do                                                      !0
            end do                                                      !0
            do i=1,nt                                                   !0
              temat3(i,i)=gd(i,i)-halfr                                 !0
            end do                                                      !0
            alp=(1.0_kindgp,0.0_kindgp)                                 !0
            bet=(0.0_kindgp,0.0_kindgp)                                 !0
                                                                        !0
c                      +                                                !0
c           temat2=fmat * temat3                                        !0
cint!       temat2=matmul(conjg(transpose(fmat)),temat3)                !0

          alpha_c = alp
          beta_c = bet
          if(use_original) then

            call cgemm(Cstring,Nstring,nt,nt,nt,alp,                    !0
     &                 fmat,nt,temat3,nt,bet,temat2,nt)                 !0

          else

          nn = nt
          nna = nl
          nnb = Nc

          ld1 = size(Afmat,1)
          ld2 = size(Bfmat,1)
          ld3 = size(temat3,1)
          ld4 = size(temat2,1)

          call kron_matmulL(Cstring,Nstring,nn,nna,nnb,
     &       alpha_c, Afmat,ld1, Bfmat,ld2, temat3,ld3, 
     &       beta_c, temat2,ld4)


          endif


c                                       +                               !0
c           temat3= temat2 * fmat = fmat  * G_up * fmat                 !0
cint!       temat3=matmul(temat2,fmat)                                  !0

          alpha_c = alp
          beta_c = bet
          if(use_original) then

            call cgemm(Nstring,Nstring,nt,nt,nt,alp,temat2,             !0
     &                 nt,fmat,nt,bet,temat3,nt)                        !0
                                                                        !0
          else

          nn = nt
          nna = nl
          nnb = Nc

          ld1 = size(temat2,1)
          ld2 = size(Afmat,1)
          ld3 = size(Bfmat,1)
          ld4 = size(temat3,1)

          call kron_matmulR( Nstring, Nstring, nn,nna,nnb,
     &        alpha_c, temat2,ld1, Afmat,ld2, Bfmat,ld3,
     &        beta_c, temat3,ld4 )


          endif
                              
c           then form the temporary vectors which store G_d-1/2I        
            do n=1,nt                                                   
            do m=1,nt                                                   
              temat1(n,m)=gd(n,m)                                       
            end do                                                      
            end do                                                      
            do 206 i=1,nt                                               
              temat1(i,i)=gd(i,i)-halfr                                 
 206        continue                                                    
c                      +                                                
c           temat2=fmat * temat1                                        
cint!       temat2=matmul(conjg(transpose(fmat)),temat1)                

          alpha_c = alp
          beta_c = bet
          if(use_original) then

            call cgemm(Cstring,Nstring,nt,nt,nt,alp,fmat,               
     &                 nt,temat1,nt,bet,temat2,nt)                      

          else

          nn = nt
          nna = nl
          nnb = Nc

          ld1 = size(Afmat,1)
          ld2 = size(Bfmat,1)
          ld3 = size(temat1,1)
          ld4 = size(temat2,1)

          call kron_matmulL(Cstring,Nstring,nn,nna,nnb,
     &       alpha_c, Afmat,ld1, Bfmat,ld2, temat1,ld3, 
     &       beta_c, temat2,ld4)

          endif


c                                       +                               !0
c           temat1= temat2 * fmat = fmat  * G_d * fmat                  !0
cint!       temat1=matmul(temat2,fmat)                                  !0

          alpha_c = alp
          beta_c = bet
          if(use_original) then

            call cgemm(Nstring,Nstring,nt,nt,nt,alp,temat2,             !0
     &                 nt,fmat,nt,bet,temat1,nt)                        !0
                                                                        !0
          else

          nn = nt
          nna = nl
          nnb = Nc

            ld1 = size(temat2,1)
            ld2 = size(Afmat,1)
            ld3 = size(Bfmat,1)
            ld4 = size(temat1,1)


          call kron_matmulR( Nstring, Nstring, nn,nna,nnb,
     &        alpha_c, temat2,ld1, Afmat,ld2, Bfmat,ld3,   
     &        beta_c, temat1,ld4 )

          endif


c                  +                                                    !0
c           Thus,(F G    F)_q is in temat1                              !0
c                    down                                               !0
c           and                                                         !0
c                  +                                                    !0
c           Thus,(F G  F)_q is in temat3                                !0
c                    up                                                 !0
c                                                                       !0
c           Now form the connected parts and subtract them              !0
c           from the full measurement.  These have the same form        !0
c           as their counterparts in meas.f                             !0
                                                                        !0
            do il=1,nl                                                  !0
            do ic=1,Nc                                                  !0
            do jl=1,nl                                                  !0
            do jc=1,Nc                                                  !0
              i = ic + (il-1)*Nc                                        !0
              j = jc + (jl-1)*Nc                                        !0
              icw = ick2map(ic,ic)    ! confined to the wedge           !0
              jcw = ick2map(ic,jc)    ! not confined to the wedge       !0
              iw = icw + (il-1)*Ncw   ! confined to the wedge           !0
              jw = jcw + (jl-1)*Nc                                      !0
                                                                        !0
              icQ = ickplus(ic,icKpi)                                   !0
              jcQ = ickplus(jc,icKpi)                                   !0
              iQ = icQ + (il-1)*Nc                                      !0
              jQ = jcQ + (jl-1)*Nc                                      !0
              gamatcenter(iw,jw)=gamatcenter(iw,jw)+                    !0
     &                           temat3(i,j)*temat1(j,i)                !0
              gamatcorner(iw,jw)=gamatcorner(iw,jw)+                    !0
     &                           temat3(i,j)*temat1(jQ,iQ)              !0
              gamatscenter(iw,jw)=gamatscenter(iw,jw)-real(             !0
     &                          halfr*(temat3(i,j)*conjg(temat1(i,j))+  !0
     &                          temat1(i,j)*conjg(temat3(i,j))),kindg)  !0
              gamatscorner(iw,jw)=gamatscorner(iw,jw)-                  !0
     &                        halfr*(temat3(i,j)*conjg(temat1(iQ,jQ))+  !0
     &                        temat1(iQ,jQ)*conjg(temat3(i,j)))         !0
              gamatccenter(iw,jw)=gamatccenter(iw,jw)+                  !0
     &                            (temat1(i,j)*temat1(j,i)+             !0
     &                             temat3(i,j)*temat3(j,i))             !0
              gamatccorner(iw,jw)=gamatccorner(iw,jw)+                  !0
     &                            (temat1(i,j)*temat1(jQ,iQ)+           !0
     &                             temat3(i,j)*temat3(jQ,iQ))           !0
            end do                                                      !0
            end do                                                      !0
            end do                                                      !0
            end do                                                      !0
                                                                        !0
                                                                        !0
c           Now add the bare bubbles back (they are diagonal in         !0
c           frequency and K)                                            !0
            do ick=1,Nc                                                 !0
            ickQ=ickplus(ick,icKpi)                                     !0
            do n=-nl/2,nl/2-1                                           !0
              ind=ick2map(ick,ick) + (n+nl/2)*Ncw                       !0
              imd=ick2map(ick,ick) + (n+nl/2)*Nc                        !0
              gamatcenter(ind,imd)=gamatcenter(ind,imd)-                !0
     &                     ((Nc*beta)**2)*GcKf(n,ick)**2                !0
              gamatcorner(ind,imd)=gamatcorner(ind,imd)-                !0
     &                     ((Nc*beta)**2)*GcKf(n,ick)*GcKf(n,ickQ)      !0
              gamatccenter(ind,imd)=gamatccenter(ind,imd)-              !0
     &                     ((Nc*beta)**2)*2.0*GcKf(n,ick)**2            !0
              gamatccorner(ind,imd)=gamatccorner(ind,imd)-              !0
     &                     ((Nc*beta)**2)*2.0*GcKf(n,ick)*              !0
     &                           GcKf(n,ickQ)                           !0
              gamatscenter(ind,imd)=gamatscenter(ind,imd)+real(         !0
     &                     ((Nc*beta)**2)*GcKf(-n-1,ickdiff(icK0,ick))  !0
     &                                   *GcKf(n,ick),kindg)            !0
              gamatscorner(ind,imd)=gamatscorner(ind,imd)+              !0
     &                     ((Nc*beta)**2)*GcKf(-n-1,ickdiff(icK0,ick))  !0
     &                                   *GcKf(n,ickQ)                  !0
                                                                        !0
            end do                                                      !0
            end do                                                      !0
c           done conditioning of two-particle properties.  Now divide   !0
c           by the degeneracy and print.                                !0
            do ic=1,Ncw                ! confined to the wedge          !0
            do il=1,nl                                                  !0
              iw = ic + (il-1)*Ncw     ! confined to the wedge          !0
              do jc=1,Nc                                                !0
              do jl=1,nl                                                !0
                jw = jc + (jl-1)*Nc                                     !0
                gamatcenter(iw,jw) = gamatcenter(iw,jw)/                !0
     &                                           real(ickdeg(ic),kindg) !0
                gamatscenter(iw,jw)= gamatscenter(iw,jw)/               !0
     &                                           real(ickdeg(ic),kindg) !0
                gamatccenter(iw,jw)= gamatccenter(iw,jw)/               !0
     &                                           real(ickdeg(ic),kindg) !0
                gamatcorner(iw,jw) = gamatcorner(iw,jw)/                !0
     &                                           real(ickdeg(ic),kindg) !0
                gamatscorner(iw,jw)= gamatscorner(iw,jw)/               !0
     &                                           real(ickdeg(ic),kindg) !0
                gamatccorner(iw,jw)= gamatccorner(iw,jw)/               !0
     &                                           real(ickdeg(ic),kindg) !0
              end do                                                    !0
              end do                                                    !0
              gamatc0center(iw) = gamatc0center(iw)/                    !0
     &                                           real(ickdeg(ic),kindg) !0
              gamatc0corner(iw) = gamatc0corner(iw)/                    !0
     &                                           real(ickdeg(ic),kindg) !0
            end do                                                      !0
            end do                                                      !0

#ifndef PUT_GAMAT
            write(lus,*) 'gamatcenter    gamatscenter     gamatccenter' !0
#endif
            do n=1,ntw                                                  !0
            do l=1,nl                                                   !0
            do jc=1,Nc                                                  !0
              m= jc + (l-1)*Nc                                          !0
              mp = ick2map(jc,jc) + (l-1)*Ncw                           !0
              gamatccenter(n,m)=gamatccenter(n,m)                       !0
     &                 -gamatc0center(n)*gamatc0center(mp)              !0
            end do                                                      !0
            end do                                                      !0
            end do                                                      !0
#ifdef PUT_GAMAT
            write(lug) ntw , nl, Nc
            write(lug) gamatcenter(:,:),
     &                    gamatscenter(:,:), 
     &                    gamatccenter(:,:)
#else
            do n=1,ntw                                                  !0
            do l=1,nt                                                   !0
            write(lus,*) gamatcenter(n,l),
     &              cmplx(gamatscenter(n,l),zeror,kindg), 
     &                    gamatccenter(n,l)
            end do                                                      !0
            end do                                                      !0
#endif

#ifndef PUT_GAMAT      
      write(lus,*)'gamatcorner    gamatscorner     gamatccorner'        !0
#endif
            do n=1,ntw                                                  !0
            do l=1,nl                                                   !0
            do jc=1,Nc                                                  !0
              m= jc + (l-1)*Nc                                          !0
              mp = ick2map(jc,jc) + (l-1)*Ncw                           !0
              gamatccorner(n,m)=gamatccorner(n,m)                       !0
     &                 -gamatc0corner(n)*gamatc0corner(mp)              !0
            end do                                                      !0
            end do                                                      !0
            end do                                                      !0
#ifdef PUT_GAMAT
            write(lug) gamatcorner(:,:),                            !0
     &                    gamatscorner(:,:),                            !0
     &                    gamatccorner(:,:)                             !0
#else
            do n=1,ntw                                                  !0
            do l=1,nt                                                   !0
            write(lus,*) gamatcorner(n,l),                            !0
     &                    gamatscorner(n,l),                            !0
     &                    gamatccorner(n,l)                             !0
            end do
            end do
#endif

        end if ! myrank.eq.0                         

        endif !  iter.eq.niter.and.isg.ge.1                         
c***********************************************************************
c       Done the 2-particle properties
c***********************************************************************
                     

        time2=real(MPI_Wtime(),kind)
        if(myrank.eq.0) write(lud,*) 'time in put= ',time2-time1        !0
        time_put=time_put+time2-time1

        if(myrank.eq.0.and.iprint.ge.9) write(lud,*) 'done output'      !0
        return
        end
      
      
        subroutine Red_Symm(mat,num)
c       reduce to proc 0, normalize and symmetrize matrices.

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        real(kind) :: mat(nl,Nc)
        integer :: i,ic,iequiv,num
        
        call MPI_REDUCE(mat,mpi_rbufgf,nl*Nc,
     &                  MPI_REAL8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then
          mat=zeror                                                     !0
          do i=1,nl                                                     !0
          do ic=1,Nc                                                    !0
            do iequiv=1,ngroup  ! impose some point-group symmetries    !0
              mat(i,ic) = mat(i,ic) + mpi_rbufgf(i,icrequ(ic,iequiv))   !0
            end do                                                      !0
          enddo                                                         !0
          enddo                                                         !0
          mat=mat/real(num*ngroup,kind)
        endif                                                   
        
        return
        end subroutine Red_Symm

        
        subroutine Red_Symmt(mat,sgn0,num)
c       reduce to zero and symmetrize matrices.

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        real(kind) :: mat(nl,Nc),sgn0
        integer :: num,i,ic,iequiv
        
        if(isv.eq.1) mat=run*mat/sgnf  ! then there is disorder
        call MPI_REDUCE(mat,mpi_rbufgf,nl*Nc,
     &                  MPI_REAL8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then
          mat=zeror                                                     !0
          do i=1,nl                                                     !0
          do ic=1,Nc                                                    !0
            do iequiv=1,ngroup  ! impose some point-group symmetries    !0
              mat(i,ic) = mat(i,ic) + mpi_rbufgf(i,icrequ(ic,iequiv))   !0
            end do                                                      !0
          enddo                                                         !0
          enddo                                                         !0
          if(isv.eq.0) mat=mat/sgn0                                     !0
          mat=mat/real(num*ngroup,kind)                                 !0
        endif                                                        
        
        return
        end subroutine Red_Symmt


        subroutine Red_Gamat(mat,sgn0,num)
c       reduce to zero and symmetrize matrices.

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        complex(kindg) :: mat(ntw,nt)
        real(kind) :: sgn0
        integer :: num,ix,iy

        if(isv.eq.1) mat=run*mat/sgnf
        call MPI_REDUCE(mat,temath,nt*ntw,
     &                  MPI_COMPLEX8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then                                          !0
          do ix=1,ntw                                                 !0
          do iy=1,nt                                                  !0
            mat(ix,iy) = temath(ix,iy)                                !0
          enddo                                                       !0
          enddo                                                       !0
          if(isv.eq.0) mat=mat/sgn0                                   !0
          mat=mat/real(num,kindg)                                     !0
        endif                                                         
        
        return
        end subroutine Red_Gamat

      
        subroutine Red_Gamat_real(mat,sgn0,num)
c       reduce to zero and symmetrize matrices.

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        real(kindg) :: mat(ntw,nt)
        real(kind) :: sgn0
        integer :: num,ix,iy

        if(isv.eq.1) mat=run*mat/sgnf                                 
        call MPI_REDUCE(mat,tematrh,nt*ntw,                   
     &                  MPI_REAL4,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then                                          !0
          do ix=1,ntw                                                 !0
          do iy=1,nt                                                  !0
            mat(ix,iy) = tematrh(ix,iy)                               !0
          enddo                                                       !0
          enddo                                                       !0
          if(isv.eq.0) mat=mat/sgn0                                   !0
          mat=mat/real(num,kindg)                                     !0
        endif                                                        
        
        return
        end subroutine Red_Gamat_real


        subroutine Red_Gamat0(mat,sgn0,num)
c       reduce to zero and symmetrize matrices.

c************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        complex(kindg) :: mat(ntw)
        real(kind) :: sgn0
        integer :: num,ix

c                                                                       
        if(isv.eq.1) mat=run*mat/sgnf
        call MPI_REDUCE(mat,mpi_cbufgmatc0,ntw,       
     &                  MPI_COMPLEX8,
     &                  MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then                                          !0
          do ix=1,ntw                                                 !0
            mat(ix) = mpi_cbufgmatc0(ix)                              !0
          enddo                                                       !0
          if(isv.eq.0) mat=mat/sgn0                                   !0
          mat=mat/real(num,kindg)                                     !0
        endif                                                         !0

        return
        end subroutine Red_Gamat0



![KA{P 13}{Akima and Renner Subsplines}                                 
![        {Akima and Renner Subsplines}*)                               
      SUBROUTINE AKIMA1(N,XN,FN,NK,BETA,B,C,D,ISWTCH,HELP,IERR)     
!                                                                       
!*****************************************************************      
!                                                                *      
!  The program AKIMA1 computes the coefficients  B(I), C(I) and  *      
!  D(I) for I=0, ... , N-1 of an interpolating cubic AKIMA       *      
!  spline, which can be either periodic or nonperiodic.          *      
!  The subspline has the representation                          *      
!                                                                *      
!  S(X)=FN(I)+B(I)(X-XN(I))+C(I)(X-XN(I))**2+D(I)(X-XN(I))**3    *      
!                                                                *      
!  for any point X in the subinterval [XN(I), XN(I+1)] for       *      
!  I=0,..., N-1.                                                 *      
!                                                                *      
!                                                                *      
!  ASSUMPTIONS:                                                  *      
!  ============                                                  *      
!                1. N >= 4 or NK >= 4                            *      
!                2. If 0.0 < BETA < 1.0, we must have            *      
!                   NK >= N + INT((N+1)/2), otherwise NK = N     *      
!                3. The nodes XN(I), I=0, ..., N, must be        *      
!                   ordered monotonically, i.e., XN(I) < XN(I+1) *      
!                   for I=0, ... , N-1                           *      
!                                                                *      
!                                                                *      
!  INPUT PARAMETERS:                                             *      
!  =================                                             *      
!  N       : Index of the last node                              *      
!  XN      : DOUBLE PRECISION (NK+1)-vector XN(0:NK), containing *      
!            the nodes XN(I) for I=0, ... , N                    *      
!  FN      : DOUBLE PRECISION (NK+1)-vector FN(0:NK), containing *      
!            the functional values FN(I) at XN(I) for I=0,...,N  *      
!  NK      : NK = N+INT((N+1)/2) maximal number of nodes allowed *      
!            when using rounded corners. If corners are not      *      
!            rounded, we use: NK = N                             *      
!  BETA    : If 0.0 < BETA < 1.0 the corners are rounded, other- *      
!            wise corners are kept                               *      
!            In the periodic case, we do not round a corner that *      
!            may exist at XN(0) = XN(N), even if 0.0 < BETA < 1.0*      
!  ISWTCH  : = 0, nonperiodic spline                             *      
!            = 1, periodic spline                                *      
!            In the periodic case, the interval [XN(0), XN(N)]   *      
!            must be an interval of periodicity with FN(0) =     *      
!            FN(N).                                              *      
!                                                                *      
!                                                                *      
!  AUXILIARY VARIABLES:                                          *      
!  ====================                                          *      
!  HELP    : DOUBLE PRECISION array HELP(-2:NK+1,1:4)            *      
!                                                                *      
!                                                                *      
!  OUTPUT PARAMETERS:                                            *      
!  =================                                             *      
!  N       : Number of the last node. If corners are rounded,    *      
!            i.e., if 0.0 < BETA < 1.0 , then the output value   *      
!            of N can differ from its input. When corners are    *      
!            rounded, the set of nodes can maximally be enlarged *      
!            by INT((N+1)/2) new nodes.                          *      
!  XN      : DOUBLE PRECISION vector XN(0:NK) containing the     *      
!            nodes XN(I), I=0,...,N. If 0.0 < BETA < 1.0, then   *      
!            the output nodes can differ from the input nodes.   *      
!  FN      : DOUBLE PRECISION vector FN(0:NK), containing the    *      
!            functional values FN(I) at XN(I) for I = 0, ..., N. *      
!            If 0.0 < BETA < 1.0, the node and functional values *      
!            XN(I) and FN(I) can differ from their input values. *      
!  B       : DOUBLE PRECISION vector B(0:NK-1) ] B, C and D con- *      
!  C       : DOUBLE PRECISION vector C(0:NK-1) ] tain the coeffi-*      
!  D       : DOUBLE PRECISION vector D(0:NK-1) ] cients of the   *      
!                                     subspline for I=0 to NK-1. *      
!                                                                *      
!  IERR    : error parameter                                     *      
!            = 0,  all is ok                                     *      
!            =-1, N < 4 or NK < 4                                *      
!            =-2, the XN(I) are not monotonically ordered        *      
!            =-3, NK < N+INT((N+1)/2) while 0.0 < BETA < 1.0     *      
!                                                                *      
!----------------------------------------------------------------*      
!                                                                *      
!  Required subroutine: MACHPD                                   *      
!                                                                *      
!*****************************************************************      
!                                                                *      
!  Author   : Gisela Engeln-M�llges                              *      
!  Date     : 04.09.1993                                         *      
!  Source   : FORTRAN 77                                         *      
!                                                                *      
!*****************************************************************      
!                                                                       
!     Declarations                                                      
!                                                                       
      IMPLICIT real(8) (A - H, O - Z) 
      real(8) XN (0:NK), FN (0:NK), B (0:NK - 1), C (0:NK - 1), 
     &        D (0:NK - 1), HELP ( - 2:NK + 1, 1:4)                             
!                                                                       
!     Check input parameters                                            
!                                                                       
      IERR = 0 
      IF (N.LT.4.OR.NK.LT.4) THEN 
         IERR = - 1 
         RETURN 
      ENDIF 
      IF (BETA.GT.0.0_8.AND.BETA.LT.1.0_8.AND.NK.LT.N + INT (0.5_8 *
     &    (N + 1) ) ) THEN                                                  
         IERR = - 3 
         RETURN 
      ENDIF 
      IF (ISWTCH.NE.0) THEN 
         ISWTCH = 1 
      ENDIF 
!                                                                       
!     Compute the machine constant                                      
!                                                                       
      FMACHP = 1.0_8 
    5 FMACHP = 0.5_8 * FMACHP 
      IF (MACHPD (1.0_8 + FMACHP) .EQ.1) GOTO 5 
      FMACHP = 2.0_8 * FMACHP 
      EPS = 4.0_8 * FMACHP 
!                                                                       
!     Compute the lengths of the subintervals and store in the first    
!     column of the auxiliary array HELP                                
!                                                                       
      DO 10, I = 0, N - 1 
         HELP (I, 1) = XN (I + 1) - XN (I) 
         IF (HELP (I, 1) .LE.EPS) THEN 
            IERR = - 2 
            RETURN 
         ENDIF 
   10 END DO 
!                                                                       
!     Compute the secant slopes and store in the second column          
!     of the auxiliary array HELP                                       
!                                                                       
      DO 20, I = 0, N - 1 
         HELP (I, 2) = (FN (I + 1) - FN (I) ) / HELP (I, 1) 
   20 END DO 
!                                                                       
!     Compute the magnitude of the slope differences and store          
!     in the third column of HELP                                       
!                                                                       
      DO 30, I = 0, N - 2 
         HELP (I, 3) = ABS (HELP (I + 1, 2) - HELP (I, 2) ) 
   30 END DO 
                                                                        
      IF (ABS (FN (0) - FN (N) ) .GT.EPS) THEN 
         ISWTCH = 0 
      ENDIF 
                                                                        
      IF (BETA.GT.0.0_8.AND.BETA.LT.1.0_8) THEN 
         IF (ISWTCH.EQ.1) THEN 
!                                                                       
!     Prepare loop                                                      
!                                                                       
            I = 1 
            IMAX = N - 1 
            HELP ( - 1, 3) = ABS (HELP (0, 2) - HELP (N - 1, 2) ) 
            HELP (N - 1, 3) = HELP ( - 1, 3) 
         ELSE 
            I = 2 
            IMAX = N - 2 
         ENDIF 
                                                                        
   35    XL = HELP (I - 2, 3) + HELP (I, 3) 
         XR = HELP (I - 1, 3) 
!                                                                       
!     Eliminate existing corners                                        
!                                                                       
         IF (XL.LE.EPS.AND.XR.GT.EPS) THEN 
!                                                                       
!     Relabel points I to N                                             
!                                                                       
            DO 40, J = N, I, - 1 
               XN (J + 1) = XN (J) 
               FN (J + 1) = FN (J) 
   40       END DO 
!                                                                       
!     Reassign interval lengths and slopes from I to N-1                
!                                                                       
            DO 50, J = N - 1, I, - 1 
               HELP (J + 1, 1) = HELP (J, 1) 
               HELP (J + 1, 2) = HELP (J, 2) 
   50       END DO 
!                                                                       
!     Reassign slope differences from I to IMAX                         
!                                                                       
            DO 60, J = IMAX, I, - 1 
               HELP (J + 1, 3) = HELP (J, 3) 
   60       END DO 
!                                                                       
!     Generate new points labelled I and I+1                            
!                                                                       
            XL = HELP (I - 1, 1) 
            XR = HELP (I + 1, 1) 
            XB = BETA * MIN (XL, XR) 
            XLAMDA = XB / XL 
            XMUE = XB / XR 
            XN (I) = XN (I) - XLAMDA * HELP (I - 1, 1) 
            FN (I) = FN (I) - XLAMDA * (FN (I) - FN (I - 1) ) 
            XN (I + 1) = XN (I + 1) + XMUE * HELP (I + 1, 1) 
            FN (I + 1) = FN (I + 1) + XMUE * (FN (I + 2) - FN (I + 1) ) 
!                                                                       
!     Compute new interval lengths                                      
!                                                                       
            DO 70, J = I - 1, I + 1 
               HELP (J, 1) = XN (J + 1) - XN (J) 
   70       END DO 
!                                                                       
!     Compute new slopes and slope differences                          
!                                                                       
            HELP (I, 2) = (FN (I + 1) - FN (I) ) / HELP (I, 1) 
            DO 80, J = I - 1, I 
               HELP (J, 3) = ABS (HELP (J + 1, 2) - HELP (J, 2) ) 
   80       END DO 
!                                                                       
!     Increase number of nodes                                          
!                                                                       
            N = N + 1 
            IMAX = IMAX + 1 
         ENDIF 
!                                                                       
!     Set index for next point                                          
!                                                                       
         I = I + 1 
         IF (I.LE.IMAX) GOTO 35 
      ENDIF 
                                                                        
      IF (ISWTCH.EQ.1) THEN 
!                                                                       
!     For the periodic case, form additional slope data                 
!                                                                       
         HELP ( - 2, 2) = HELP (N - 2, 2) 
         HELP ( - 1, 2) = HELP (N - 1, 2) 
         HELP (N, 2) = HELP (0, 2) 
         HELP (N + 1, 2) = HELP (1, 2) 
      ELSE 
!                                                                       
!     For the nonperiodic case, provide additional slopes as well       
!                                                                       
         HELP ( - 2, 2) = 3.0_8 * HELP (0, 2) - 2.0_8 * HELP (1, 2) 
         HELP ( - 1, 2) = 2.0_8 * HELP (0, 2) - HELP (1, 2) 
         HELP (N, 2) = 2.0_8 * HELP (N - 1, 2) - HELP (N - 2, 2) 
         HELP (N + 1, 2) = 3.0_8 * HELP (N - 1, 2) - 2.0_8 * 
     &                     HELP (N - 2, 2)                                                          
      ENDIF 
!                                                                       
!     Compute additional slope differences                              
!                                                                       
      HELP ( - 2, 3) = ABS (HELP ( - 1, 2) - HELP ( - 2, 2) ) 
      HELP ( - 1, 3) = ABS (HELP (0, 2) - HELP ( - 1, 2) ) 
      HELP (N - 1, 3) = ABS (HELP (N, 2) - HELP (N - 1, 2) ) 
      HELP (N, 3) = ABS (HELP (N + 1, 2) - HELP (N, 2) ) 
!                                                                       
!     Compute the left and right handed slopes at the points 0 and N-1, 
!     and store in column 4 of HELP and in B                            
!                                                                       
      DO 90, I = 0, N - 1 
         XL = HELP (I - 2, 3) 
         XR = HELP (I, 3) 
         IF (XL + XR.GT.EPS) THEN 
            ALPHA = XL / (XL + XR) 
            HELP (I, 4) = HELP (I - 1, 2) + ALPHA * (HELP (I, 2)        
     &                    - HELP (I - 1, 2) )  
            B (I) = HELP (I, 4) 
         ELSE 
            HELP (I, 4) = HELP (I - 1, 2) 
            B (I) = HELP (I, 2) 
         ENDIF 
   90 END DO 
!                                                                       
!     Compute the left handed slope at the point N                      
!                                                                       
      XL = HELP (N - 2, 3) 
      XR = HELP (N, 3) 
      IF (XL + XR.GT.EPS) THEN 
         ALPHA = XL / (XL + XR) 
         HELP (N, 4) = HELP (N - 1, 2) + ALPHA * (HELP (N, 2) - 
     &                 HELP (N - 1, 2) )                                                      
      ELSE 
         HELP (N, 4) = HELP (N - 1, 2) 
      ENDIF 
!                                                                       
!     Compute the coefficients C(I) and D(I)                            
!                                                                       
      DO 100, I = 0, N - 1 
         H = 1.0_8 / HELP (I, 1) 
         C (I) = (3.0_8 * HELP (I, 2) - 2.0_8 * B (I) - 
     &            HELP (I + 1, 4) ) * H  
         D (I) = (B (I) + HELP (I + 1, 4) - 2.0_8 * HELP (I, 2) )       
     &            * H * H 
  100 END DO 
                                                                        
      RETURN 
      END SUBROUTINE AKIMA1                         




      real(8) FUNCTION SPVAL (X, N, XN, A, B, C, D) 
!                                                                       
!*****************************************************************      
!                                                                *      
!  The FUNCTION SPVAL determines the functional value of a cubic *      
!  spline S of the form:                                         *      
!                                                                *      
!  S(X) = P(I)(X) = A(I) + B(I)*(X-XN(I)) + C(I)*(X-XN(I))**2 +  *      
!                                         + D(I)*(X-XN(I))**3    *      
!                                                                *      
!  for X in the interval [XN(I),XN(I+1)], where I=0,1,...,N-1.   *      
!                                                                *      
!  If X < XN(0) we evaluate the first polynom P(0), if X > XN(N) *      
!  we evaluate the polynom P(N-1).                               *      
!  We do not check the input X in any way.                       *      
!                                                                *      
!                                                                *      
!  INPUT PARAMETERS:                                             *      
!  =================                                             *      
!  X  :  value for which we want to find the value of the spline *      
!  N  :  index of the final node                                 *      
!  XN :  vector of nodes XN(0:N)                                 *      
!  A  :  ] N+1-vectors ..(0:N);                                  *      
!  B  :  ] the spline coefficients A(I), B(I), C(I), D(I) for    *      
!  C  :  ] I = 0, ..., N-1; used for storage for I = N.          *      
!  D  :  ]                                                       *      
!                                                                *      
!----------------------------------------------------------------*      
!                                                                *      
!  Subroutines required: none                                    *      
!                                                                *      
!*****************************************************************      
!                                                                *      
!  Author   : G�nter Palm                                        *      
!  Date     : 01.06.1991                                         *      
!  Source   : FORTRAN 77                                         *      
!                                                                *      
!*****************************************************************      
!                                                                       
      IMPLICIT real(8) (A - H, O - Z) 
!                                                                       
!  Declarations                                                         
!                                                                       
      real(8) XN (0:N), A (0:N), B (0:N), C (0:N), D (0:N) 
      SAVE I 
!                                                                       
!  Initializing                                                         
!                                                                       
      DATA I / 0 / 
      IF (I.GE.N) I = 0 
!                                                                       
!  In a repeat call we determine the interval [XN(I), XN(I+1)]          
!  that contains X only if X does not lie in the same interval          
!  as last used.                                                        
!                                                                       
      IF (X.LT.XN (I) .OR.X.GE.XN (I + 1) ) THEN 
         I = 0 
         K = N 
   10    M = (I + K) / 2 
         IF (X.LT.XN (M) ) THEN 
            K = M 
         ELSE 
            I = M 
         ENDIF 
         IF (K.GT.I + 1) GOTO 10 
      ENDIF 
!                                                                       
!  Compute X-XN(I) in order to evaluate the polynomial                  
!                                                                       
      XL = X - XN (I) 
!                                                                       
!  Compute the value of the spline via a Horner scheme                  
!                                                                       
      SPVAL = ( (D (I) * XL + C (I) ) * XL + B (I) ) * XL + A (I) 
      RETURN 
      END FUNCTION SPVAL                            


      INTEGER FUNCTION MACHPD (X) 
      implicit none
      real(8) X 
      MACHPD = 0 
      IF (1.0_8.LT.X) MACHPD = 1 
      RETURN 
      END FUNCTION MACHPD                           
      
