extern "C" void launch_reg_delayed(Kernel_Params*);

__global__ void reg_delayed_kernel(int ind, TYPE* Gu, TYPE* Gd , int* pS, TYPE* au, TYPE* bu , TYPE* ad, TYPE* bd
          ,TYPE* d_expRu_Global, TYPE* d_expRd_Global, TYPE* d_diagu , TYPE* d_diagd , TYPE alpha , int Nc , int L, int nDelays) {

  int spin = pS[ind]; //Get the value of the spin from the Ising field.
  int imd ;
  TYPE Ru,Rd;
  TYPE* d_expu  = d_expRu_Global + 1;
  TYPE* d_expd  = d_expRd_Global + 1;

  // Is this the thread that will compute diagonal[ind]
  //imd  = threadIdx.x + blockIdx.x*blockDim.x;
  int threadCount = blockDim.x;  
  for (imd = threadIdx.x; imd < Nc*L; imd += threadCount) {
    d_diagu[imd] = GetG(Gu, imd, imd, Nc*L);
    d_diagd[imd] = GetG(Gd, imd, imd, Nc*L); 
  }
  __syncthreads();

  TYPE avalued = 0;
  TYPE avalueu = 0;
  TYPE bvalued = 0;
  TYPE bvalueu = 0;

  for(int n_block_count = 0 ; n_block_count < nDelays ; n_block_count++) {
    Ru  =  (1-(d_diagu[ind] - 1) * d_expu[spin]);
    Rd  =  (1-(d_diagd[ind] - 1) * d_expd[spin]);
   for (imd = threadIdx.x; imd < Nc*L; imd += threadCount) {
    // Lots of optimizations needed here. We are reading from global memory.
    avalueu = GetG(Gu, imd, ind, Nc*L);//Columns
    avalued = GetG(Gd, imd, ind, Nc*L);
    
    bvalueu = GetG(Gu, ind, imd, Nc*L);//Rows
    bvalued = GetG(Gd, ind, imd, Nc*L);

    if (n_block_count > 0) {
       for (int delay = 0; delay < n_block_count; delay++) {
	 const int idx_1 = FlattenIdx(imd,  delay, nDelays, Nc*L);
	 const int idx_2 = FlattenIdx(delay,ind  ,Nc*L, nDelays);
	 const int idx_3 = FlattenIdx(ind,  delay, nDelays, Nc*L);
         const int idx_4 = FlattenIdx(delay,imd, Nc*L, nDelays);

         avalueu += au[idx_1]*bu[idx_2];
         avalued += ad[idx_1]*bd[idx_2];
         bvalueu += au[idx_3]*bu[idx_4];
         bvalued += ad[idx_3]*bd[idx_4];
       }
     }
     avalueu = (avalueu - delta(imd, ind)) * d_expu[spin] / Ru;
     avalued = (avalued - delta(imd, ind)) * d_expd[spin] / Rd;
     {
     int tmp = FlattenIdx(n_block_count, imd , Nc*L, nDelays);
     bu[tmp] = bvalueu;
     bd[tmp] = bvalued;
     // *****************************************************************
     tmp = FlattenIdx(imd, n_block_count, nDelays, Nc*L);
     au[tmp]  = avalueu;
     ad[tmp]  = avalued;
    }
    // *****************************************************************
    //Updating the diagonal that we read first from the green function atthe very begining.
    //we do this because thediaonals we read will be invalidated by the delayed update, since we don't
    //update the green function we need to update our diagonals so that we cab go through the delays
    d_diagu[imd] = d_diagu[imd] + avalueu * bvalueu;
    d_diagd[imd] = d_diagd[imd] + avalued * bvalued;
   }
    ind++;
    spin = pS[ind]; //Get the value of the spin from the Ising field.     
    __syncthreads();
  }
};

void launch_reg_delayed(Kernel_Params* params) {

  int  ind = 0;
  dim3 ThreadsPerBlock_1_Dim;
 
  if ((params->Nc)*(params->L) > 512) {
    ThreadsPerBlock_1_Dim = 512; 
  } else {
    ThreadsPerBlock_1_Dim = (params->Nc)*(params->L);
  }

  cudaGetLastError();  // Clear the Global error state before calling the kernel,look at the Programming Guide page 50.
  CheckCublasError(cublasCreate(&handle), "Creating handle for cublas");
  cudaThreadSynchronize();
  printf("Iteration \t\tKernel_Events \tKernel_Time  \t Sgemm_Events \t sgemm_Time\n");

  Timer KernelTimer;
  Timer SgemmTimer;
   for(int k = 0; k < (params->Nc)*(params->L); k+= params->nDelays) {
    KernelTimer.Start();
    reg_delayed_kernel<<< 1, ThreadsPerBlock_1_Dim >>>(ind,
        				     params->d_Gu, params->d_Gd ,
        				     params->d_pS,
        				     params->d_au, params->d_bu,
        				     params->d_ad, params->d_bd,
        				     params->d_expRu_global,
        				     params->d_expRd_global,
        				     params->d_diagu,
        				     params->d_diagd,
        				     params->alpha, params->Nc, params->L, params->nDelays);
    cudaDeviceSynchronize();
    cutilCheckMsg("reg_delayed_kernel execution failed.\n");
    KernelTimer.Stop();
    printf(" \t\t %i \t  %2.10f \t  %2.10f ", k/(params->nDelays), KernelTimer.elapsedTime_GPU, KernelTimer.elapsedTime_OS);
    ind += params->nDelays;
    SgemmTimer.Start();
    Multiply_CUBLAS(params->d_au, params->d_bu, params->d_Gu , params);
    Multiply_CUBLAS(params->d_ad, params->d_bd, params->d_Gd , params);
    SgemmTimer.Stop();
    printf("\t %2.10f ",   SgemmTimer.elapsedTime_GPU);
    printf("\t %2.10f \n", SgemmTimer.elapsedTime_OS);
  }

  CheckCublasError(cublasDestroy(handle) , "Destroyig the handle for cublas");  
  printf("\nREG Delayed kernel\n"
        "Delays\t\t"
        "Kernel OS:\t\t"
        "Kernel GPU:\t\t"
        "Kernel Overhead:\t"\
        "Sgemm GPU:\t\t"\
        "Sgemm OS:\t\t"\
        "Sgemm Overhead:\t\t"\
        "All:\t\t\n");

  printf("%i\t\t"
        "%2.10f\t\t%2.10f\t\t%2.10f\t\t"
        "%2.10f\t\t%2.10f\t\t%2.10f\t\t"
        "%2.10f"
        "\n-------------------------------------------------------------\n",
        params->nDelays,
        KernelTimer.Total_OS,
        KernelTimer.Total_GPU,
        KernelTimer.Total_OS - KernelTimer.Total_GPU,
        SgemmTimer.Total_GPU,
        SgemmTimer.Total_OS,
        SgemmTimer.Total_OS - SgemmTimer.Total_GPU,
        KernelTimer.Total_OS + SgemmTimer.Total_OS);
};
