extern "C" void launch_reg_delayed_cuda_5(Kernel_Params* params);

void reg_delayed_cuda_5(Kernel_Params* params) {
  dim3 ThreadsPerBlock_1_Dim(params->ISIZE , 1 , 1);
  int ind=0;
  //for (int ind=0; ind < params->ISIZE; ind += params->nDelays) {
  Timer KernelTimer;
  KernelTimer.Start();
  reg_delayed_kernel_cuda_5<<<1, ThreadsPerBlock_1_Dim >>>(ind, params->d_Gu, params->d_Gd,
        				     params->d_pS,
        				     params->d_au, params->d_bu,
        				     params->d_ad, params->d_bd,
        				     params->d_expRu_global,
        				     params->d_expRd_global,
        				     params->d_diagu,
        				     params->d_diagd,
        				     params->alpha, params->Nc, params->L, params->nDelays);
  cudaDeviceSynchronize(); // Need to synchronize before we can check the status for kernel failures.Programming guide 3.2.8.
  KernelTimer.Stop();
  printf("reg_delayed_cuda_5 took: %f seconds.\n", KernelTimer.elapsedTime_GPU);
  cutilCheckMsg("cuda_5_reg_delayed_kernel execution failed.\n");
 //}
};


__global__ void launch_reg_delayed_cuda_5(int ind_start, TYPE* Gu, TYPE* Gd , int* pS, TYPE* au, TYPE* bu , TYPE* ad, TYPE* bd
          , TYPE* d_expRu_Global, TYPE* d_expRd_Global, TYPE* d_diagu , TYPE* d_diagd , TYPE alpha , int Nc , int L, int nDelays) {

  int spin; 
  int imd ;
  TYPE avalueu, bvalueu;
  TYPE avalued, bvalued;
  TYPE Ru,Rd;
  TYPE* d_expu  = d_expRu_Global + 1;
  TYPE* d_expd  = d_expRd_Global + 1;

  cublasHandle_t handle;
  cublasStatus_t status ;//= cublasCreate(&handle);

  TYPE colu_sum = 0; 
  TYPE cold_sum = 0;
  TYPE rowu_sum = 0;
  TYPE rowd_sum = 0;

  imd  = threadIdx.x ;
  //int ind = ind_start;
  // A clean update loops over all the Green function only once
  status = cublasCreate(&handle);
  if (status != 0) { 
    if (imd ==0)
      printf("handle creation failed\n");
  }

 for (int ind = 0; ind < Nc*L;) {

  d_diagu[imd] = GetG(Gu, imd, imd, Nc*L);
  d_diagd[imd] = GetG(Gd, imd, imd, Nc*L);
  __syncthreads();

  for(int n_block_count = 0; n_block_count < nDelays; n_block_count++) {
    spin = pS[ind];
    Ru  =  (1-(d_diagu[ind] - 1) * d_expu[spin]);
    Rd  =  (1-(d_diagd[ind] - 1) * d_expd[spin]);
    // Lots of optimizations needed here. We are reading from global memory.
    avalueu = GetG(Gu, imd, ind, Nc*L); // Columns
    avalued = GetG(Gd, imd, ind, Nc*L);
    bvalueu = GetG(Gu, ind, imd, Nc*L); // Rows
    bvalued = GetG(Gd, ind, imd, Nc*L);

    if (n_block_count > 0) {
      colu_sum = 0;
      cold_sum = 0;
      rowu_sum = 0;
      rowd_sum = 0;
       for (int delay = 0; delay < n_block_count; delay++) {
         colu_sum = au[FlattenIdx(imd, delay, nDelays, Nc*L)]*bu[ FlattenIdx(delay,ind ,Nc*L, nDelays)] + colu_sum;
         cold_sum = ad[FlattenIdx(imd, delay, nDelays, Nc*L)]*bd[ FlattenIdx(delay,ind ,Nc*L, nDelays)] + cold_sum;
         rowu_sum = au[FlattenIdx(ind, delay, nDelays, Nc*L)]*bu[ FlattenIdx(delay,imd, Nc*L, nDelays)] + rowu_sum;
         rowd_sum = ad[FlattenIdx(ind, delay, nDelays, Nc*L)]*bd[ FlattenIdx(delay,imd, Nc*L, nDelays)] + rowd_sum;
       }
      avalueu += colu_sum;
      avalued += cold_sum;
      bvalueu += rowu_sum;
      bvalued += rowd_sum;
    }
    
    avalueu = (avalueu - delta(imd, ind)) * d_expu[spin]/Ru;
    avalued = (avalued - delta(imd, ind)) * d_expd[spin]/Rd;
    int tmp = FlattenIdx(n_block_count, imd , Nc*L, nDelays);
    bu[tmp] = bvalueu;
    tmp = FlattenIdx(n_block_count, imd, Nc*L, nDelays);
    bd[tmp] = bvalued;
    tmp = FlattenIdx(imd, n_block_count, nDelays, Nc*L);
    au[tmp]  = avalueu;
    tmp = FlattenIdx(imd, n_block_count, nDelays, Nc*L) ;
    ad[tmp]  = avalued;
    //*****************************************************************
    // Updating the diagonal that we read first from the green function at the very begining.
    // We do this because thediaonals we read will be invalidated by the delayed update, since we don't
    // Update the green function we need to update our diagonals so that we cab go through the delays
    d_diagu[imd] = d_diagu[imd] + avalueu * bvalueu;
    d_diagd[imd] = d_diagd[imd] + avalued * bvalued;
    ind++;
    __syncthreads();
  } // for(int n_block_count = 0; n_block_count < nDelays
  // Now multiply:
  if (0 == threadIdx.x) {
    CublasFromDevice(au, bu, Gu, Nc*L, nDelays, handle);
    CublasFromDevice(ad, bd, Gd, Nc*L, nDelays, handle);
  }
  cudaDeviceSynchronize();
 } // for(int ind=0 ;ind < Nc*L  ....
  cublasDestroy(handle);
};

