        subroutine sumup
c**********************************************************************
c       This subroutine is called at the end of a run.  It collects
c       the mesurements and forms the bin averages.  Some of the data
c       is collected from all processors and then printed into the file
c       bins.dat.  This data may be used later to calculate dynamics
c       using MEM.
c**********************************************************************
        use Global
        implicit none
        include 'mpif.h'
c************************************************************
        real(kind) :: time1,time2
c************************************************************

        if(myrank.eq.0.and.iprint.ge.9) then
           write(lud,*) 'start sumup';flush(lud)
        endif
        time1=real(MPI_Wtime(),kind)

!c       normalize the data from a run and load into the 
!c       accumulators gf and gfs.
        sgnm  = sgnm/real(meas,kind)
        sgnf  = sgnf+sgnm
        sgnfs = sgnfs+sgnm**2

        gft   = gft   + gmeas/real(meas*nt,kind)
        chft  = chft  + chmeas/real(2*meas*nt,kind)
        scft  = scft  + scmeas/real(32*meas*nt,kind)
        ccft  = ccft  + ccmeas/real(32*meas*nt,kind)
        chaft = chaft + cham/real(meas*nt,kind)
        mooft = mooft + twor*moom/real(meas*nt,kind)    
        RVBchixxt = RVBchixxt + RVBchixxm/(real(meas*nt,kind)
     &*(oner+xyswitch))
        RVBchixyt = RVBchixyt + RVBchixym/(real(meas*nt,kind)
     &*(oner+xyswitch))
        ChiHHSt = ChiHHSt + ChiHHSm/real(meas*nt,kind)
        

        gmeas   = gmeas/real(meas*nt,kind)
        chmeas  = chmeas/real(2*meas*nt,kind)
        scmeas  = scmeas/real(32*meas*nt,kind)
        ccmeas  = ccmeas/real(32*meas*nt,kind)
        b1gmeas = b1gmeas/real(32*meas*nt,kind)
        b2gmeas = b2gmeas/real(32*meas*nt,kind)
        cham    = cham/real(meas*nt,kind)
      
        !--------------------------------------------------------------
        ! normalize susceptibilities for dynamic modulation of optical
        ! lattice and write data bins to specific file

        if(iter.eq.niter.and.isb.eq.8) then
           bow_meas  = bow_meas /real(meas,kind)
           dbop_meas = dbop_meas/real(meas,kind)

!           sus_kkt = sus_kkt/real(meas,kind) - bow_meas**2
!           sus_kdt = sus_kdt/real(meas,kind) - 2.d0*bow_meas*dbop_meas
!           sus_ddt = sus_ddt/real(meas,kind) - dbop_meas**2
!           sus_eet = sus_eet/real(meas,kind) - dbop_meas**2
!     & + 2.d0*bow_meas*dbop_meas*alphax - dbop_meas**2*alphax**2

           sus_kkt = sus_kkt/real(meas,kind)
           sus_kdt = sus_kdt/real(meas,kind)
           sus_ddt = sus_ddt/real(meas,kind)
           sus_eet = sus_eet/real(meas,kind)
           call writebins_dynmod(sus_kkt,sus_kdt,sus_ddt,sus_eet,
     &bow_meas,dbop_meas,sgnm)

           sus_kkt=zeror
           sus_kdt=zeror
           sus_ddt=zeror
           sus_eet=zeror
           bow_meas = zeror
           dbop_meas= zeror
        endif
        !--------------------------------------------------------------


        if(abs(sgnm).gt.eps) then
          nsgn=nsgn+1
          gf     = gf+gmeas/sgnm
          gfs    = gfs+(gmeas/sgnm)**2

          chf    = chf+chmeas/sgnm
          chfs   = chfs+(chmeas/sgnm)**2

          scf    = scf+scmeas/sgnm
          scfs   = scfs+(scmeas/sgnm)**2

          ccf    = ccf+ccmeas/sgnm
          ccfs   = ccfs+(ccmeas/sgnm)**2

          chaf   = chaf+cham/sgnm
          chafs  = chafs+(cham/sgnm)**2

          moom=twor*moom/(sgnm*real(meas*nt,kind))
          moof=moof+moom
          moofs=moofs+moom**2
          
        end if      !    if(abs(sgnm).gt.eps)


!c       Now store the bins of data for analytic continuation in files.
!c       We need to send and recieve even if sgnm=0. 
        if(iter.eq.niter.and.isb.ge.1) call writebins(gmeas ,1)
        if(iter.eq.niter.and.isb.ge.2) call writebins(chmeas,2)
        if(iter.eq.niter.and.isb.ge.3) call writebins(cham  ,3)
        if(iter.eq.niter.and.isb.ge.4) call writebins(ccmeas,4)
        if(iter.eq.niter.and.isb.ge.5) call writebins(scmeas,5)
        if(iter.eq.niter.and.isb.ge.6) call writebins(b1gmeas,6)
        if(iter.eq.niter.and.isb.ge.7) call writebins(b2gmeas,7)

!c       zero the measurements     
        gmeas  = zeror
        chmeas = zeror
        scmeas = zeror
        ccmeas = zeror
        b1gmeas = zeror
        b2gmeas = zeror
        cham   = zeror
        moom   = zeror
        sgnm   = zeror
        RVBchixxm = zeror
        RVBchixym = zeror
        ChiHHSm = zeror
        

        time2=real(MPI_Wtime(),kind)
        time_sumup=time_sumup+time2-time1
        
        if(myrank.eq.0.and.iprint.ge.9) then
           write(lud,*) 'done sumup';flush(lud)
        endif

        return
        end


        subroutine writebins(measurement,meastype)
c**********************************************************************
c       Write real measurements out into bins of data.  Both sgnm and
c       measurement are buffered to ensure that they are unchanged by
c       writebins.
c**********************************************************************
        use Global
        implicit none
        include 'mpif.h'
c*********************************************************************
        integer       :: i,j,ic,iproc,i1,istat(MPI_STATUS_SIZE),req(2),
     &                   meastype
        real(kind)    :: measurement(nl,Nc),measbuf(nl,Nc),sgnbuf
c*********************************************************************
        i1=1
        sgnbuf=sgnm  ! we need to buffer sgnm and measurement
        measbuf=measurement
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
        if(myrank.eq.0) then
          do iproc=0,nproc-1  ! blocking recieve calls
            if(iproc.ne.0) call MPI_RECV(measbuf,nt,MPI_REAL8,
     &                            iproc,i1,MPI_COMM_WORLD,istat,ierr)
            if(iproc.ne.0) call MPI_RECV(sgnbuf,1,MPI_REAL8,
     &                            iproc,i1,MPI_COMM_WORLD,istat,ierr)
c           The bins of Gf  are determined printed onto the data file.
            write(lub,*) ' '

              write(lub,*) ' '
              write(lub,*) '   '
              if(meastype.eq.1) 
     &          write(lub,*) ' bin of  Gtau(n) from process  ',iproc
              if(meastype.eq.2) 
     &          write(lub,*) ' bin of  Chitau(n) from process  ',iproc
              if(meastype.eq.3) 
     &          write(lub,*) ' bin of  Chatau(n) from process  ',iproc
              if(meastype.eq.4) 
     &          write(lub,*) ' bin of  CCtau(n) from process  ',iproc
              if(meastype.eq.5) 
     &          write(lub,*) ' bin of  Sctau(n) from process  ',iproc
              if(meastype.eq.6) 
     &          write(lub,*) ' bin of  B1gtau(n) from process  ',iproc
              if(meastype.eq.7) 
     &          write(lub,*) ' bin of  B2gtau(n) from process  ',iproc
              write(lub,401) ((measbuf(j,ic), j=1,nl),ic=1,Nc),sgnbuf

          end do
        else             ! myrank not= 0  nonblocking sends
          call MPI_ISEND(measbuf,nt,MPI_REAL8,0,
     &                                  i1,MPI_COMM_WORLD,req(1),ierr)
          call MPI_ISEND(sgnbuf,1,MPI_REAL8,0,
     &                                  i1,MPI_COMM_WORLD,req(2),ierr)

          call MPI_WAITALL(2,req,istat,ierr)
c          call MPI_WAIT(req,istat,ierr)
        end if

 401    format(1x,f12.8,1x,f12.8,1x,f12.8,1x,f12.8,1x,f12.8,1x,f12.8)
        return
        end

!======================================================================
        subroutine writebins_dynmod(xkk,xkd,xdd,xee,bow_avg,dbop_avg,
     &sgn_avg)
        use Global
        implicit none
        include 'mpif.h'
        !--------------------------------------------------------------
        ! Purpose
        ! =======
        ! Write the averaged data bins to separate files
        !
        ! Arguements
        ! ==========
        real(kind),intent(in):: xkk(nl),xkd(nl),xdd(nl),xee(nl)
        real(kind),intent(in):: bow_avg,dbop_avg,sgn_avg
        real(kind) :: xkk_buf(nl),xkd_buf(nl),xdd_buf(nl),xee_buf(nl)
        real(kind) :: bow_buf,dbop_buf,sgn_buf
        integer :: il,iproc,i1,unit_t
        integer :: istat(MPI_STATUS_SIZE),req(7)
        !
        ! Execute
        ! =======
        i1=1

        xkk_buf=xkk  ! we need to buffer data 
        xkd_buf=xkd
        xdd_buf=xdd
        xee_buf=xee
        bow_buf=bow_avg
        dbop_buf=dbop_avg
        sgn_buf=sgn_avg

        call MPI_BARRIER(MPI_COMM_WORLD,ierr)

        if(myrank.eq.0) then

        ! Write header to data file
           if(nrun.eq.1) then 
              write(10000,*) 'Results of <K(T)K(0)>'
              write(10000,*) nl,nl,beta,0.d0
              write(20000,*) 'Results of <K(T)D(0)>'
              write(20000,*) nl,nl,beta,0.d0
              write(30000,*) 'Results of <D(T)K(0)>'
              write(30000,*) nl,nl,beta,0.d0
              write(40000,*) 'Results of <D(T)D(0)>'
              write(40000,*) nl,nl,beta,0.d0
           endif

           do iproc = 0 , nproc-1
        ! Blocking receive calls  
              if(iproc.ne.0) call MPI_RECV(xkk_buf,nl,MPI_REAL8,
     &iproc,i1,MPI_COMM_WORLD,istat,ierr)
              if(iproc.ne.0) call MPI_RECV(xkd_buf,nl,MPI_REAL8,
     &iproc,i1,MPI_COMM_WORLD,istat,ierr)
              if(iproc.ne.0) call MPI_RECV(xdd_buf,nl,MPI_REAL8,
     &iproc,i1,MPI_COMM_WORLD,istat,ierr)
              if(iproc.ne.0) call MPI_RECV(xee_buf,nl,MPI_REAL8,
     &iproc,i1,MPI_COMM_WORLD,istat,ierr)
              if(iproc.ne.0) call MPI_RECV(bow_buf,1,MPI_REAL8,
     &iproc,i1,MPI_COMM_WORLD,istat,ierr)
              if(iproc.ne.0) call MPI_RECV(dbop_buf,1,MPI_REAL8,
     &iproc,i1,MPI_COMM_WORLD,istat,ierr)
              if(iproc.ne.0) call MPI_RECV(sgn_buf,1,MPI_REAL8,
     &iproc,i1,MPI_COMM_WORLD,istat,ierr)

        ! The bins of data are determined printed onto the data file
              unit_t = 10000
              write(unit_t,"('#No. bin=',i5,' from processor',i5)") 
     & nrun,iproc
              do il = 1, nl
                 write(unit_t,"(1X,i4,1x,E16.8)") il, xkk_buf(il)
              enddo
              write(unit_t,"('#sgn=',E16.8)") sgn_buf
              write(unit_t,"('#<K>=',E16.8)") bow_buf
              write(unit_t,"('#<D>=',E16.8)") dbop_buf

              unit_t = 20000
              write(unit_t,"('#No. bin=',i5,' from processor',i5)") 
     & nrun,iproc
              do il = 1, nl
                 write(unit_t,"(1X,i4,1x,E16.8)") il, xkd_buf(il)
              enddo
              write(unit_t,"('#sgn=',E16.8)") sgn_buf
              write(unit_t,"('#<K>=',E16.8)") bow_buf
              write(unit_t,"('#<D>=',E16.8)") dbop_buf

              unit_t = 30000
              write(unit_t,"('#No. bin=',i5,' from processor',i5)") 
     & nrun,iproc
              do il = 1, nl
                 write(unit_t,"(1X,i4,1x,E16.8)") il, xdd_buf(il)
              enddo
              write(unit_t,"('#sgn=',E16.8)") sgn_buf
              write(unit_t,"('#<K>=',E16.8)") bow_buf
              write(unit_t,"('#<D>=',E16.8)") dbop_buf

              unit_t = 40000
              write(unit_t,"('#No. bin=',i5,' from processor',i5)") 
     & nrun,iproc
              do il = 1, nl
                 write(unit_t,"(1X,i4,1x,E16.8)") il, xee_buf(il)
              enddo
              write(unit_t,"('#sgn=',E16.8)") sgn_buf
              write(unit_t,"('#<K>=',E16.8)") bow_buf
              write(unit_t,"('#<D>=',E16.8)") dbop_buf

           enddo !end do iproc

        else ! myrank =x= 0  nonblocking sends
           call MPI_ISEND(xkk_buf,nl,MPI_REAL8,0,i1,MPI_COMM_WORLD,
     & req(1),ierr)
           call MPI_ISEND(xkd_buf,nl,MPI_REAL8,0,i1,MPI_COMM_WORLD,
     & req(2),ierr)
           call MPI_ISEND(xdd_buf,nl,MPI_REAL8,0,i1,MPI_COMM_WORLD,
     & req(3),ierr)
           call MPI_ISEND(xee_buf,nl,MPI_REAL8,0,i1,MPI_COMM_WORLD,
     & req(4),ierr)
           call MPI_ISEND(bow_buf,1,MPI_REAL8,0,i1,MPI_COMM_WORLD,
     & req(5),ierr)
           call MPI_ISEND(dbop_buf,1,MPI_REAL8,0,i1,MPI_COMM_WORLD,
     & req(6),ierr)
           call MPI_ISEND(sgn_buf,1,MPI_REAL8,0,i1,MPI_COMM_WORLD,
     & req(7),ierr)

           call MPI_WAITALL(7,req,istat,ierr)
c          call MPI_WAIT(req,istat,ierr)
        end if
        
        return
        end subroutine writebins_dynmod
