      program tables_2D_Betts      ! TWO DIMENSIONAL BETTS TABLES
c***************************************************************************
c     CHANGES REQUIRED FOR USE IN QMC CODES:
c     1. remove comment(s) in from of
c          use Global
c          define the  fmat matrix
c          call allocate_2p
c     2. change program and end statements to subroutine and return
c     3. delete the "Stuff that belongs in module_Global"...
c     4. remove "write(6,*) 'enter cluster'"...
c     5. set  use_brute_force = .false. and use_lin_alg = .true.
c     7. remove these comments
c     CHANGES REQUIRED FOR USE IN Analysis CODES:
c     1. do the steps above
c     2. remove the delcarations of ict, ig, jg and kg
c***************************************************************************
c     This subroutine makes the lookup tables used throughout the
c     full TWO DIMENSIONAL code.  It employs the definitions and 
c     cluster geometries developed by D.D. Betts, H.Q. Lin, and 
c     J.S. Flynn Canadian Journal of Physics, 77, 353-369.
c***************************************************************************
c     use Global
      implicit none
      interface transform2D
      function gtransform2D(a,igroup,ixy,isx,isy)
      integer :: a(3,3),gtransform2D(3,3),igroup,ixy(8),isx(8),isy(8)
      end
      
      function Rtransform2D(Rc,igroup,ixy,isx,isy)
      integer :: Rc(3),Rtransform2D(3),igroup,ixy(8),isx(8),isy(8)
      end
      
      function Ktransform2D(Kc,igroup,ixy,isx,isy)
      integer :: igroup, ixy(8),isx(8),isy(8)
      real(8) :: Kc(3),Ktransform2D(3)
      end
      end interface 

c***************************************************************************
c     Stuff that belongs in module_Global
c***************************************************************************
      character*3 :: cluster
      integer, parameter :: kind=8, lud=6, iprint=10, ntot=400,isg=1,
     &                      myrank=0, ndim=2
      integer :: Nc,Ncw,nl,ntw,Rc(3,32),a(3,3),
     &           ngroup,ios,icrequ(32,8),icrdiff(32,32),icK0,icKpi,
     &           icR0,icR1,N_sp,ickdiff(32,32),ickplus(32,32),ickequ(32,8),
     &           ickdeg(32),ickmap(32),ick2map(32,32),nwsc
      real(kind), parameter :: zeror=0.0_kind,halfr=0.5_kind,
     &                         oner=1.0_kind,twor=2.0_kind,
     &                         eps=0.0001_kind,pi=3.14159269
      real(kind) :: g(3,3),Kc(3,32), Epsbar(32),Dwavebar(32),tprime,
     &              Dwic(32,32),gdwave(32),kt(3,(2*ntot+1)**2/Nc)
      complex(kind) :: FTCoefs_K_to_R(32,32),FTCoefs_R_to_K(32,32),
     &                 Dwave(32,32)
      complex(kind), parameter :: ii=(0.0_kind,1.0_kind)
c***************************************************************************
c     Locally defined
c***************************************************************************
      logical    :: allowed,denied,denied_lin_alg
      integer :: i,i1,i2,ix,iy,it,igroup,ind,ic,ic1,ic2,ick,idim,ik,
     &           itest,j,jmd,jm,j1,j2,jc,jk,jkc,jlow,jhigh,
     &           k,k2,kroniker,m,m1,m2,n,n1,n2,l,p,nshell,
     &           ixy(8),isx(8),isy(8),iv1(3),iv2(3),iv3(3),irwgroup(32),
     &           a1(3,3),group(8),
     &           ict(-32:32,-32:32),ig(32),jg(32),kg(8)
      real(kind) :: r1,r2,r3,r4,r5,Dkt,c1(3,3),c2(3,3),
     &              rv1(3),rv2(3),rv3(3),Ai(3,3)
c For lin algebra test
      real(kind) :: ainv(3,3),a1ainv(3,3)
      complex(kind) :: cr1
c Select use of "brute force" vs linear algebra method for 
c checking lattice coincidence. Both methods should agree; lin alg
c method is much faster.
      logical, parameter :: use_brute_force = .false.
      logical, parameter :: use_lin_alg = .true.
c***************************************************************************
c     Local arrays
c     ixy(igroup)      D=2 square point group table(see below)
c     isx(igroup)      D=2 square point group table(see below)
c     isy(igroup)      D=2 square point group table(see below)
c     group(i)         Indices of allowed group operations
c     irwgroup(ic)     the group operation maps K(ic) into IRW
c     ig(ic)           i in the g1 and j in the g2 direction
c     jg(jc)           corresponding to K point ic
c     ict(i,j)         maps i*g1+j*g2 to ic
c     Ai(i,j)          The inverse transpose of a (see below)
c     the rest are just temporary arrays.
c*************************************************************************
      
c     CLUSTER MOMENTUM and DUAL SPACE for the 2D square lattice 
c                                                           _   _ 
c     Different Real-space Cluster Geometries.  The vectors a1, a2
c     define the parallelpipeds which define the periodically replicated 
c     clusters.  Here, we list only the clusters from table 1A with Nc
c     even.
c              _          _         
c  cluster     a1         a2   bipartide  IF  IB  used
c    1A     ( 1, 0)    ( 0, 1)    no               no
c
c    2A     ( 1, 1)    (-1, 1)    yes              no
c
c    4A     ( 2, 0)    ( 0, 2)    yes              no
c
c    8A     ( 2, 2)    (-2, 2)    yes      0   0   yes
c    8C     ( 2, 3)    (-2, 1)    no       0   -   no
c
c   10A     ( 2, 3)    (-2, 2)    no       0   -   no
c   10B     ( 1, 3)    (-3, 1)    yes      1   0   no
c
c   12A     ( 2, 2)    (-4, 2)    yes      2   0   no
c   12B     ( 2, 3)    (-2, 3)    no       0   -   no
c   12D     ( 1, 3)    ( 4, 0)    yes      2   0   no
c
c   14A     ( 3, 2)    (-1, 4)    no       0   -   no
c   14B     ( 4, 2)    (-1, 3)    yes      2   0   no
c
c   16A     ( 4, 2)    ( 0, 4)    yes      1   0   yes
c   16B     ( 4, 0)    ( 0, 4)    yesI     3   1   no
c   16C     ( 3, 2)    (-2, 4)    no       0   -   no
c   16D     ( 4, 1)    ( 0, 4)    no       1   -   no
c   16E     ( 1, 5)    (-3, 1)    yesI     3   1   no
c
c   18A     ( 3, 3)    (-3, 3)    yes      0   0   yes
c   18B     ( 3, 3)    (-2, 4)    yes      0   0   yes
c   18C     ( 4, 2)    (-1, 4)    no       0   -   no
c   18D     ( 5, 3)    (-1, 3)    yesI     4   4   no
c
c   20A     ( 4, 2)    (-2, 4)    yes      1   0   yes
c   20B     ( 3, 4)    (-2, 4)    no       0   -   no
c   20C     ( 1, 4)    (-5, 0)    no       1   -   no
c   20E     ( 5, 1)    ( 0, 4)    yesI     3   1   no
c
c   22A     ( 2, 4)    (-5, 1)    yes      2   0   yes
c   22B     ( 4, 2)    (-3, 4)    no       0   -   no
c   22D     ( 1, 4)    (-5, 2)    no       1   -   no
c
c   24A     ( 4, 4)    (-4, 2)    yes      3   0   yes
c   24B     ( 4, 0)    ( 2, 6)    yesI     5   1   no
c   24C     ( 4, 0)    ( 0, 6)    yesI     7   2   no
c   24D     ( 1, 5)    ( 5, 1)    yes      3   0   yes
c   24E     ( 3, 3)    (-5, 3)    yes      3   0   yes
c   24F     ( 4, 4)    (-3, 3)    yes      3   0   yes
c   24G     ( 4, 3)    (-4, 3)    no       0   -   no
c                                   
c   26A     ( 4, 2)    (-3, 5)    yes      3   0   yes
c   26B     ( 1, 5)    (-5, 1)    yesI     5   1   yes
c   26C     ( 5, 2)    (-3, 4)    no       0   -   no
c
c   28A     ( 2, 4)    (-6, 2)    yesI     4   1   no
c   28B     ( 5, 3)    (-1, 5)    yes      1   0   yes
c   28C     ( 2, 5)    (-4, 4)    no       0   -   no
c   28D     ( 4, 3)    (-4, 4)    no       0   -   no
c
c   30A     ( 5, 3)    (-5, 3)    yes      1   0   yes
c   30B     ( 4, 3)    (-2, 6)    no       0   -   no
c   30C     ( 2, 5)    (-6, 0)    no       1   -   no
c   30D     ( 1, 5)    (-6, 0)    yesI     5   2   no
c   30G     ( 5, 5)    (-2, 4)    yesI     5   2   no
c
c   32A     ( 4, 4)    (-4, 4)    yes      0   0   yes
c   32B     ( 2, 6)    (-4, 4)    yes      0   0   yes
c   32C     ( 3, 5)    (-4, 4)    yes      0   0   no
c   32D     ( 1, 6)    (-5, 2)    no       1   -   no
c   32G     ( 3, 4)    (-5, 4)    no       0   -   no
c 
c  used determined form Table 3, bipartide from Ib in table 1A, yesI indicates
c  that the cluster is imperfect
c
c  We need to automate the process of finding which of the 8 point group 
c  operations are retained by the different clusters.  To do this, we must 
c  perform all a cubic group operations on the {a} and see if they form 
c  an equivalent parallelpiped.  
c
c  Group Operations for a square lattice:
c
c  x --> +/- x                  There are 2^D D! operations in D dimensions
c    \/                         or 8 operations in 2D.
c    /\
c  y --> +/- y   
c        .
c  a'=transform2D(a,igroup,ixy,isx,isy)
c  ixy=0(1) dont (do) exchange x and y
c  isx      sign of x 
c  isy      sign of y
c
c  x <-- isx*[(1-ixy)*x + ixy*y] 
c  y <-- isy*[(1-ixy)*y + ixy*x] 
c

      ixy( 1)=0; isx( 1)=+1; isy( 1)=+1 !identity  xyz e
      ixy( 2)=0; isx( 2)=-1; isy( 2)=+1 !
      ixy( 3)=0; isx( 3)=+1; isy( 3)=-1 !
      ixy( 4)=0; isx( 4)=-1; isy( 4)=-1 !
      ixy( 5)=1; isx( 5)=+1; isy( 5)=+1 !
      ixy( 6)=1; isx( 6)=-1; isy( 6)=+1 !
      ixy( 7)=1; isx( 7)=+1; isy( 7)=-1 !
      ixy( 8)=1; isx( 8)=-1; isy( 8)=-1 !
      
      write(6,*) 'enter cluster'
      read(5,*) cluster
      read(cluster,"(i2)",iostat=ios) Nc           ! should ignore characters
      if (ios.gt.0) then
        read(cluster,"(i1)",iostat=ios) Nc
        if (ios.gt.0) then
           write(lud,*) "Cannot read Nc from string ",cluster
           stop
        endif
      endif

c***********************************************************************        
c       determine the principle translation vectors a1, and a2
c***********************************************************************
           If(cluster.eq.'1A') then; a(1,1:2)=(/1,0/); a(2,1:2)=(/ 0,1/)
      else if(cluster.eq.'2A') then; a(1,1:2)=(/1,1/); a(2,1:2)=(/-1,1/)
      else if(cluster.eq.'4A') then; a(1,1:2)=(/2,0/); a(2,1:2)=(/ 0,2/)
      else if(cluster.eq.'8A') then; a(1,1:2)=(/2,2/); a(2,1:2)=(/-2,2/)
      else if(cluster.eq.'8C') then; a(1,1:2)=(/2,3/); a(2,1:2)=(/-2,1/)
      else if(cluster.eq.'10A')then; a(1,1:2)=(/2,3/); a(2,1:2)=(/-2,2/)
      else if(cluster.eq.'10B')then; a(1,1:2)=(/1,3/); a(2,1:2)=(/-3,1/)
      else if(cluster.eq.'12A')then; a(1,1:2)=(/2,2/); a(2,1:2)=(/-4,2/)
      else if(cluster.eq.'12B')then; a(1,1:2)=(/2,3/); a(2,1:2)=(/-2,3/)
      else if(cluster.eq.'12D')then; a(1,1:2)=(/1,3/); a(2,1:2)=(/ 4,0/)
      else if(cluster.eq.'14A')then; a(1,1:2)=(/3,2/); a(2,1:2)=(/-1,4/)
      else if(cluster.eq.'14B')then; a(1,1:2)=(/4,2/); a(2,1:2)=(/-1,3/)
      else if(cluster.eq.'16A')then; a(1,1:2)=(/4,2/); a(2,1:2)=(/ 0,4/)
      else if(cluster.eq.'16B')then; a(1,1:2)=(/4,0/); a(2,1:2)=(/ 0,4/)
      else if(cluster.eq.'16C')then; a(1,1:2)=(/3,2/); a(2,1:2)=(/-2,4/)
      else if(cluster.eq.'16D')then; a(1,1:2)=(/4,1/); a(2,1:2)=(/ 0,4/)
      else if(cluster.eq.'16E')then; a(1,1:2)=(/1,5/); a(2,1:2)=(/-3,1/)
      else if(cluster.eq.'18A')then; a(1,1:2)=(/3,3/); a(2,1:2)=(/-3,3/)
      else if(cluster.eq.'18B')then; a(1,1:2)=(/3,3/); a(2,1:2)=(/-2,4/)
      else if(cluster.eq.'18C')then; a(1,1:2)=(/4,2/); a(2,1:2)=(/-1,4/)
      else if(cluster.eq.'18D')then; a(1,1:2)=(/5,3/); a(2,1:2)=(/-1,3/)
      else if(cluster.eq.'20A')then; a(1,1:2)=(/4,2/); a(2,1:2)=(/-2,4/)
      else if(cluster.eq.'20B')then; a(1,1:2)=(/3,4/); a(2,1:2)=(/-2,4/)
      else if(cluster.eq.'20C')then; a(1,1:2)=(/1,4/); a(2,1:2)=(/-5,0/)
      else if(cluster.eq.'20E')then; a(1,1:2)=(/5,1/); a(2,1:2)=(/ 0,4/)
      else if(cluster.eq.'22A')then; a(1,1:2)=(/2,4/); a(2,1:2)=(/-5,1/)
      else if(cluster.eq.'22B')then; a(1,1:2)=(/4,2/); a(2,1:2)=(/-3,4/)
      else if(cluster.eq.'22D')then; a(1,1:2)=(/1,4/); a(2,1:2)=(/-5,2/)
      else if(cluster.eq.'24A')then; a(1,1:2)=(/4,4/); a(2,1:2)=(/-4,2/)
      else if(cluster.eq.'24B')then; a(1,1:2)=(/4,0/); a(2,1:2)=(/ 2,6/)
      else if(cluster.eq.'24C')then; a(1,1:2)=(/4,0/); a(2,1:2)=(/ 0,6/)
      else if(cluster.eq.'24D')then; a(1,1:2)=(/1,5/); a(2,1:2)=(/ 5,1/)
      else if(cluster.eq.'24E')then; a(1,1:2)=(/3,3/); a(2,1:2)=(/-5,3/)
      else if(cluster.eq.'24F')then; a(1,1:2)=(/4,4/); a(2,1:2)=(/-3,3/)
      else if(cluster.eq.'24G')then; a(1,1:2)=(/4,3/); a(2,1:2)=(/-4,3/)
      else if(cluster.eq.'26A')then; a(1,1:2)=(/4,2/); a(2,1:2)=(/-3,5/)
      else if(cluster.eq.'26B')then; a(1,1:2)=(/1,5/); a(2,1:2)=(/-5,1/)
      else if(cluster.eq.'26C')then; a(1,1:2)=(/5,2/); a(2,1:2)=(/-3,4/)
      else if(cluster.eq.'28A')then; a(1,1:2)=(/2,4/); a(2,1:2)=(/-6,2/)
      else if(cluster.eq.'28B')then; a(1,1:2)=(/5,3/); a(2,1:2)=(/-1,5/)
      else if(cluster.eq.'28C')then; a(1,1:2)=(/2,5/); a(2,1:2)=(/-4,4/)
      else if(cluster.eq.'28D')then; a(1,1:2)=(/4,3/); a(2,1:2)=(/-4,4/)
      else if(cluster.eq.'30A')then; a(1,1:2)=(/5,3/); a(2,1:2)=(/-5,3/)
      else if(cluster.eq.'30B')then; a(1,1:2)=(/4,3/); a(2,1:2)=(/-2,6/)
      else if(cluster.eq.'30C')then; a(1,1:2)=(/2,5/); a(2,1:2)=(/-6,0/)
      else if(cluster.eq.'30D')then; a(1,1:2)=(/1,5/); a(2,1:2)=(/-6,0/)
      else if(cluster.eq.'30G')then; a(1,1:2)=(/5,5/); a(2,1:2)=(/-2,4/)
      else if(cluster.eq.'32A')then; a(1,1:2)=(/4,4/); a(2,1:2)=(/-4,4/)
      else if(cluster.eq.'32B')then; a(1,1:2)=(/2,6/); a(2,1:2)=(/-4,4/)
      else if(cluster.eq.'32C')then; a(1,1:2)=(/3,5/); a(2,1:2)=(/-4,4/)
      else if(cluster.eq.'32D')then; a(1,1:2)=(/1,6/); a(2,1:2)=(/-5,2/)
      else if(cluster.eq.'32G')then; a(1,1:2)=(/3,4/); a(2,1:2)=(/-5,4/)
      else
         if(myrank.eq.0) write(lud,*) 'Bad cluster specification' !0
         stop
      end if  
      if(iprint.ge.5.and.myrank.eq.0) 
     &   write(lud,"('a=',2('(',i2,1x,i2,')'))") a(1,1:2),a(2,1:2)  

c*********************************************************************************
c     Determine which group operations are allowed    
c*********************************************************************************

      ngroup=0
      if(myrank.eq.0.and.iprint.ge.5) then
        write(lud,*) '----------------------------------'
        write(lud,*) '  trans(a)   igroup L  ixy isx isy'
        write(lud,*) '----------------------------------'
      end if
      do igroup=1,8
         a1=transform2D(a,igroup,ixy,isx,isy)
          
          
         if (use_brute_force) then

c        look at the lattice points generated by a1 and see if they
c        correspond to those produced by a.
c        First, generate the a1 lattice points
         do i=-Nc,Nc
         do j=-Nc,Nc
           denied=.true.
           iv1=(/i,j,0/)
           do l=1,2         ! x,y components of this lattice point
             rv1(l)=sum(iv1(1:2)*a1(1:2,l))
           end do
c          Now generate the a lattice points.  If a and a1 are the 
c          same lattice one of these points MUST correspond to ij
           do n=-6*Nc,6*Nc
           do m=-6*Nc,6*Nc
             iv2=(/n,m,0/)
             do l=1,2         ! x,y components of this lattice point
               rv2(l)=sum(iv2(1:2)*a(1:2,l))
             end do
             if(sum(abs(rv1-rv2)).lt.eps) denied=.false.
           end do
           end do
           if(denied) exit !no corrs. point was found
         end do
         if(denied) exit
         end do
         end if

         if (use_lin_alg) then
c PK If lattice points generated by "a1" vectors can be found in 
c lattice points generated by "a" vectors then (a1)x(a^-1) can only 
c contain integers.
            
            r1 = a(1,1)*a(2,2) - a(2,1)*a(1,2) ! the determinent of a
            ainv=0.
            ainv(1,1) = +a(2,2)/r1
            ainv(1,2) = -a(1,2)/r1       
            ainv(2,1) = -a(2,1)/r1
            ainv(2,2) = +a(1,1)/r1
            a1ainv=matmul(a1,ainv)
            if (all(abs(a1ainv-nint(a1ainv)).le.eps)) then 
               denied_lin_alg=.false. ! matrix only contains integers
            else
               denied_lin_alg=.true.
            end if
         end if

c If both algorithms are enabled, check they agree
         if (use_brute_force.and.use_lin_alg) then
            if (denied.neqv.denied_lin_alg) then
               write (*,*) '*** BUG : Denied algorithm disagree'
               write (*,*) 'a11',a1(:,1)
               write (*,*) 'a12',a1(:,2)
               write (*,*) 'a13',a1(:,3)
               
               write (*,*) 'a1',a(:,1)
               write (*,*) 'a2',a(:,2)
               write (*,*) 'a3',a(:,3)
               write (*,*)
               write (*,*) 'ainv',ainv(:,1)
               write (*,*) 'ainv',ainv(:,2)
               write (*,*) 'ainv',ainv(:,3)
               write (*,*)
               
               write (*,*) 'a1ainv',a1ainv(:,1)
               write (*,*) 'a1ainv',a1ainv(:,2)
               write (*,*) 'a1ainv',a1ainv(:,3)
               write (*,*) 'Denied_lin_alg=',denied_lin_alg
               write (*,*) 'Denied_brute_force=',denied
               write (*,*)
               stop
            end if
         end if
         if (use_lin_alg) denied=denied_lin_alg


         if(denied.eqv..false.) then ! a corresponding point was found
            ngroup=ngroup+1
            group(ngroup)=igroup
         end if    
         if(myrank.eq.0.and.iprint.ge.5) then
           allowed=.true.;if (denied) allowed=.false.
           write(lud,"(2('(',i2,1x,i2,')'),2x,i2,2x,l,3(2x,i2))")
     &     a1(1,1:2),a1(2,1:2),igroup,allowed,ixy(igroup),
     &     isx(igroup),isy(igroup)
         end if
      end do      
     
      if(myrank.eq.0.and.iprint.ge.5) then
        write(lud,*) '----------------------------------'
        write(lud,"('cluster=',a3,' ngroup=',i2,'  groups=',8(i2,1x))") 
     &  cluster,ngroup,group(1:ngroup)
      end if
     




c******************************************************************************
c       calculate the table of cluster point locations Rc(i,ic), i=x,y
c******************************************************************************
c                                    a1 and a2 need not be orthogonal.    
c  ^ y                               Therefore, if v=A n, where n is a
c  |                                 vector containing the projection of
cc-|-c---c-*-c---c---c---c           v onto a, and                          
c| | |   |   |   |   |   |                                                  
cc-|-c---c---c---c---c---c           A= [ a1x  a1y ]  a= [ a1x  a2x ]       
ca2^ * * * * * * * * |   |              [ a2x  a2y ]     [ a1y  a2y ]       
cc-|-c---c---c---c-*-c---c                                                  
c| | |   P   |   | * |   |           which is the transpose of a, then the  
cc-|-c--/c---c---c-*-c---c           projection of v onto a is              
c| | | /v|   |   | * |   |                  -1                              
cc-|-c/--c---c---c-*-c---c             n = A  v,  where                     
c| | /   |   |   | * |   |                                                  
cc-|/0---c---c---c-*-c---c             -1  [ a2y -a2x ] /                   
c| O--------------->---------> x      A  = [-a1y  a1x ]/ (a1x a2y-a2x a1y)  
cc---c---c---c---ca1-c---c                                                  
c|   |   |   |   |   |   |           v is within the cluster, if the        
cc---c---c---c---c---c---c           elements of n, -eps < n_i < 1-eps      
c
c                                                                          

      Ai=zeror
      r1 = a(1,1)*a(2,2) - a(2,1)*a(1,2) ! the determinent of a
      Ai(1,1) = +a(2,2)/r1
      Ai(1,2) = -a(2,1)/r1       
      Ai(2,1) = -a(1,2)/r1
      Ai(2,2) = +a(1,1)/r1

      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '  '                   !0
        write(lud,*) 'ic,Rcx(ic),Rcy(ic)'   !0
      end if
      ic=0
      do j=-Nc,Nc  ! Rc(:,1) must be the origin!  Since a2x may be <0
      do i=-Nc,Nc  ! we need to index i (x) faster than j (y).
        rv1=(/i,j,0/)
        rv2=matmul(Ai,rv1)
        if(rv2(1).lt.oner-eps.and.rv2(1).gt.zeror-eps.and.
     &     rv2(2).lt.oner-eps.and.rv2(2).gt.zeror-eps) then
           ic=ic+1
           Rc(:,ic)=(/i,j,0/) ! Rc, like i.e. a(1,1) is integer
            if(iprint.ge.5.and.myrank.eq.0) 
     &           write(lud,"(i2,4x,i3,4x,i3)") ic,Rc(1:2,ic)   !0
           if(sum(abs(Rc(1:2,ic))).eq.0) icR0=ic
           if(sum(Rc(1:2,ic)**2).eq.1) icR1=ic
         end if
      end do
      end do
      if(ic.ne.Nc) then
         if(myrank.eq.0) write(lud,*) 'Bug in Rc(ic) loop'   !0
         stop
      end if
      if(iprint.ge.5.and.myrank.eq.0) write(lud,*) 'icR0=',icR0

c***********************************************************************        
c       create a table mapping points outside the cluster back into it.
c***********************************************************************        
        

      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '  '                                !0
        write(lud,*) ' Equivalency table r'              !0
        write(lud,*) ' ic  igroup   icrequ(ic,ieqiv)'    !0
      end if
      do ic=1,Nc
        do igroup=1,ngroup
c         Generate the equivalent points using the ngroup operations
          iv1=transform2D(Rc(:,ic),group(igroup),ixy,isx,isy) ! cartesian coordinates
          do n=1,3            
            rv2=matmul(Ai,iv1) ! projection of iv1 onto a1 and a2
            do l=1,2           ! move iv1 and rv2 into the cluster
              if(rv2(l).lt.zeror-eps) then
                 iv1(:)=iv1(:)+a(l,:)
              else if(rv2(l).gt.oner-eps) then
                 iv1(:)=iv1(:)-a(l,:)
              end if
            end do
          end do
c         Figure out which cluster point iv1 is
          itest=0
          do jc=1,Nc
             if(sum(abs(iv1(1:2)-Rc(1:2,jc))).eq.0) then ! iv1=Rc(jc)
                icrequ(ic,igroup)=jc
                itest=1
             end if
          end do
          if(itest.eq.0) then
             write(lud,*) 'No equivalent found in icrequ',ic,igroup
             stop
          end if 
          if(iprint.ge.5.and.myrank.eq.0) 
     &      write(lud,"(2x,i2,5x,i1,8x,i2)") 
     &      ic,igroup,icrequ(ic,igroup) !0
        end do
      end do                               


c***********************************************************************        
c     Now create the tables for the differences of R
c***********************************************************************        
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '  '                      !0
        write(lud,*) ' differences in r'       !0
        write(lud,*) ' ic  jc  icrdiff(ic,jc)' !0
      end if
      do i=1,Nc
      do j=1,Nc
        iv1(:)=Rc(:,i)-Rc(:,j) 
        itest=0          ! We dont know where iv1 is.
        do n=-1,1        ! Look in the surrounding 
        do m=-1,1        ! 9 clusters for iv1.
          iv2=(/n,m,0/)
          do l=1,2
            iv3(l)=iv1(l)+sum(a(1:2,l)*iv2(1:2))
          end do
c         Is iv3 in the main cluster?
          rv2=matmul(Ai,iv3)
          if(rv2(1).lt.oner-eps.and.rv2(1).gt.zeror-eps.and.
     &       rv2(2).lt.oner-eps.and.rv2(2).gt.zeror-eps) then
c            The point is located within the cluster.  
c            Now we must find which point it is.
             do ic=1,Nc
               if(sum(abs(iv3(1:2)-Rc(1:2,ic))).eq.0) then
                 icrdiff(i,j)=ic
                 itest=1
               end if
             end do
             if(iprint.ge.5.and.myrank.eq.0) 
     &         write(lud,"(1x,i2,2x,i2,7x,i2)") i,j,icrdiff(i,j) !0
          end if
        end do
        end do
        if(itest.ne.1) then
          write(lud,*) 'no icrdiff(i,j) found',i,j
          stop
        end if
      end do
      end do  
        

c***********************************************************************        
c       Now set up the K-space tables
c***********************************************************************        

c    The reciprocal lattice vectors g1=(g1_x,g1_y) and g2=(g2_x,g2_y) for 
c    a general set of real space lattice vectors a1=(a1_x,a1_y) and 
c    a2=(a2_x,a2_y) is:
c    
c    g1_x =  a2_y * twopi / area
c    g1_y = -a2_x * twopi / area
c    
c    g2_x = -a1_y * twopi / area
c    g2_y =  a1_x * twopi / area
c    
c    with
c    
c    area = |ax*by-ay*bx|

      r1=a(1,1)*a(2,2) - a(2,1)*a(1,2)
      g(1,1) =+2*pi*a(2,2)/r1
      g(1,2) =-2*pi*a(2,1)/r1
      g(2,1) =-2*pi*a(1,2)/r1
      g(2,2) =+2*pi*a(1,1)/r1
      
      if(iprint.ge.5.and.myrank.eq.0) 
     &  write(lud,"('g=',2(' (',f6.3,1x,f6.3,')  '))") g(1,1:2),g(2,1:2)



c***********************************************************************        
c       Create the cluster momentum tables Kc(i,ic), i=x,y
c***********************************************************************        

c     First find the points in the irreducible wedge.  They will be
c     indexed from 1 to Ncw
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) ' Wedge K points '    !0
        write(lud,*) 'ic  i  j    Kcx(ic)    Kcy(ic)' !0
      end if
      ic=0; Ncw=0; icK0=0; icKpi=0
      do i=-Nc,Nc
      do j=-Nc,Nc
        iv1=(/i,j,0/)    ! i in the g1 and j in the g2 direction
        do l=1,2         ! x,y components 
          rv1(l)=sum(iv1(1:2)*g(1:2,l))
        end do
        if(rv1(1).gt.-pi+eps.and.rv1(1).lt.pi+eps.and.
     &     rv1(2).gt.-pi+eps.and.rv1(2).lt.pi+eps) then
c          The point is in the first SC Brillouin zone.  Now see if the 
c          point falls within the IRW.  First see if it is equivalent to 
c          any known point in the IRW (if so, it cannot be in IRW).
           do igroup=2,ngroup ! exclude the identity
              rv2=transform2D(rv1,group(igroup),ixy,isx,isy)
              itest=0
              do jc=1,Ncw
                 if(sum(abs(Kc(1:2,jc)-rv2(1:2))).lt.eps) then ! rv2 and Kc(ic)
                    itest=1                                    ! are the same 
                    exit                                       ! leave the jc loop
                 end if
              end do
              if(itest.eq.1) exit !leave the igroup loop
           end do
           if(itest.eq.0) then ! rv1 is in 1BZ and not eqivalent to any point in 
              Ncw=Ncw+1        ! the known IRW.  It must be in the IRW
              ic=ic+1
              Kc(:,ic)=rv1(:)
              ig(ic)=i
              jg(ic)=j
              if(sum(abs(rv1(1:2)-pi)).lt.eps) icKpi=ic
              if(sum(abs(rv1(1:2)-zeror)).lt.eps) icK0=ic
              if(iprint.ge.5.and.myrank.eq.0) 
     &          write(lud,5) ic,i,j,Kc(1:2,ic)     !0
           end if
        end if 
      end do 
      end do 
      
      if(icK0.eq.0) then  ! we failed to find (0,0)
        if(myrank.eq.0) 
     &    write(lud,*) 'failed to find (0,0) in the IRW'    !0
        stop
      else if(icKpi.eq.0) then  ! we failed to find (pi,pi)
        icKpi=icK0
        if(myrank.eq.0) then
          write(lud,*) '*****************************************'    !0
          write(lud,*) '* failed to find (pi,pi) in the IRW     *'    !0
          write(lud,*) '* apparently the lattice isnt bipartide *'    !0
          write(lud,*) '* icKpi=icK0 so that the code will run  *'    !0
          write(lud,*) '*****************************************'    !0
        end if
      end if



c     Now find the points in the 1BZ but outside the irreducible wedge.  
c     They will be indexed from Ncw+1 to Nc
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) ' Non Wedge K points '    !0
      end if
      do i=-Nc,Nc              
      do j=-Nc,Nc
         iv1=(/i,j,0/)    ! i in the g1 and j in the g2 direction
         do l=1,2         ! x,y components 
            rv1(l)=sum(iv1(1:2)*g(1:2,l))
         end do
         if(rv1(1).gt.-pi+eps.and.rv1(1).lt.pi+eps.and.
     &      rv1(2).gt.-pi+eps.and.rv1(2).lt.pi+eps) then
c           The point is in the first SC Brillouin zone.  Now see if the 
c           point falls within the IRW.  
            itest=0
            do jc=1,Ncw
               if(sum(abs(Kc(1:2,jc)-rv1(1:2))).lt.eps) then ! rv2 and Kc(ic) 
                  itest=1                                    ! are the same 
                  exit                                       ! leave the jc loop
               end if
            end do
            if(itest.eq.0) then ! rv1 is in 1BZ and not the IRW
               ic=ic+1
               Kc(:,ic)=rv1(:)
               ig(ic)=i
               jg(ic)=j
               if(iprint.ge.5.and.myrank.eq.0) 
     &              write(lud,5) ic,i,j,Kc(1:2,ic)     !0
            end if
         end if 
      end do 
      end do 
      
      
      if(iprint.ge.5.and.myrank.eq.0) 
     &  write(lud,"('icK0=',i2,' icKpi=',i2,' Ncw=',i2,' Nc=',i2)") 
     &  icK0,icKpi,Ncw,ic

 5    format(1x,i2,1x,i2,1x,i2,2x,f9.6,2x,f9.6)
      ntw=nl*Ncw      


c***********************************************************************
c     Create the indirect addressing table ict(i,j,k) which maps a 
c     point indexed by the principle translations in k-space to ic.
c***********************************************************************        
      N_sp=sqrt(float(Nc))
      do i=-N_sp,N_sp
      do j=-N_sp,N_sp
         iv1=(/i,j,0/)
         do l=1,2         !x,y components 
            rv1(l)=sum(iv1(1:2)*g(1:2,l)) ! K point in xyz
         end do
         do it=1,3
           do l=1,2
              if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
              if(rv1(l).gt.pi+eps)  rv1(l)=rv1(l)-twor*pi
           end do
         end do
         itest=0
         do jc=1,Nc
            if(sum(abs(rv1(1:2)-Kc(1:2,jc))).lt.eps) then        
               ict(i,j)=jc
               itest=1
            end if
         end do
         if(itest.eq.0) then
            write(lud,5) 'ict failed for i,j=',i,j
            stop
         end if
           
      end do
      end do

c***********************************************************************        
c     Create the table ickdiff(ic1,ic2), which takes the difference
c     between two K-points
c***********************************************************************        
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                           !0
        write(lud,*) 'ic1   ic2  ickdiff(ic1,ic2)  ' !0
      end if
      do ic1=1,Nc
      do ic2=1,Nc
         rv1=Kc(:,ic1)-Kc(:,ic2)
         do it=1,3           ! move the difference into the 1BZ
           do l=1,2
              if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
              if(rv1(l).gt.pi+eps)  rv1(l)=rv1(l)-twor*pi
           end do
         end do
c        Now we need to figure out where this point is!
         itest=0
         do ic=1,Nc
            if(sqrt(sum((rv1(1:2)-Kc(1:2,ic))**2)).lt.eps) then    
               ickdiff(ic1,ic2)=ic
               itest=1
            end if
         end do
         if(itest.eq.0) then
            write(lud,5) 'ickdiff failed for ic1,ic2=',ic1,ic2
            stop
         end if
         if(iprint.ge.5.and.myrank.eq.0) 
     &        write(lud,"(1x,i2,4x,i2,6x,i2)") ic1,ic2,ickdiff(ic1,ic2) !0
      end do
      end do
          

c***********************************************************************        
c     Create the table ickplus(ic1,ic2), which takes the sum
c     between to K-points
c***********************************************************************        
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                           !0
        write(lud,*) 'ic1   ic2  ickplus(ic1,ic2)  ' !0
      end if
      do ic1=1,Nc
      do ic2=1,Nc
         rv1=Kc(:,ic1)+Kc(:,ic2)
         do it=1,3           ! move the difference into the 1BZ
           do l=1,2
              if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
              if(rv1(l).gt.pi+eps)  rv1(l)=rv1(l)-twor*pi
           end do
         end do
c        Now we need to figure out where this point is!
         itest=0
         do ic=1,nc
            if(sqrt(sum((rv1(1:2)-Kc(1:2,ic))**2)).lt.eps) then    
               ickplus(ic1,ic2)=ic
               itest=1
            end if
         end do
         if(itest.eq.0) then
            write(lud,5) 'ickplus failed for ic1,ic2=',ic1,ic2
            stop
         end if
         if(iprint.ge.5.and.myrank.eq.0) 
     &     write(lud,"(1x,i2,4x,i2,6x,i2)") ic1,ic2,ickplus(ic1,ic2) !0
      end do
      end do

c***********************************************************************
c     Now find the points equivalent to ic, ickequ(ic,igroup) where 
c     igroup is the group operation.
c***********************************************************************        
c
c
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                            !0
        write(lud,*) ' equivalency in k'              !0
        write(lud,*) ' ic  igroup  ickequ(ic,igroup)' !0
      end if
      do ic=1,Nc
         i=ig(ic)
         j=jg(ic)
         iv1=(/i,j,0/)
         do l=1,2
            rv2(l)=sum(iv1(1:2)*g(1:2,l))
         end do
c     Kc(ic) falls within the zone, by construction, now look 
c     for equivalent points
c     
         do igroup=1,ngroup
            rv1=transform2D(rv2,group(igroup),ixy,isx,isy) 
c           These are the equivalent points.  Now map to 1st BZ
            do it=1,3           ! move the difference into the 1BZ
              do l=1,2
                 if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
                 if(rv1(l).gt.+pi+eps) rv1(l)=rv1(l)-twor*pi
              end do
            end do
c           Now we need to figure out where this point is!
            itest=0
            do jc=1,Nc
               if(sqrt(sum((rv1(1:2)-Kc(1:2,jc))**2)).lt.eps) then
                  ickequ(ic,igroup)=jc
                  itest=1
               end if
            end do
            if(iprint.ge.5.and.myrank.eq.0) 
     &        write(lud,"(2x,i2,4x,i2,6x,i2)") 
     &        ic,igroup,ickequ(ic,igroup) !0
            if(itest.eq.0) then
               if(myrank.eq.0) write(lud,*)'no equivalent K-point found' !0
               if(myrank.eq.0) write(lud,*)'j1,n2,m2',j1,n2,m2 !0
               if(myrank.eq.0) write(lud,*)'K=',rv2(1:2) !0
               stop
            end if
         end do
      end do
        
c***********************************************************************        
c     Now generate a table which maps any K to an equivalent point
c     in the irreducible wedge (IW), and a table for the degeneracy 
c     of each point in the IW. 
c 
c     The use of the wedge can greatly reduce the storage required.
c     for example for Nc=8, only 4 of the 8 points are in the irreducible 
c     wedge.  Thus, all storage arrays may be reduced in size by a 
c     factor of 2.  As Nc-->oo, the saving factor approaches 8 for
c     square and diamond clusters.  
c
c***********************************************************************        
c       
      
      ickdeg=0; irwgroup=0
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                            !0
        write(lud,*) ' degeneracy in k'              !0
        write(lud,*) ' ic  ickmap(ic)  ickdeg(ic) irwgroup(ic)' !0
      end if
c     scan over all of the K points in the 1BZ
      do ic=1,Nc
         i=ig(ic)
         j=jg(ic)
         iv1=(/i,j,0/)
         do l=1,2
            rv2(l)=sum(iv1(1:2)*g(1:2,l))
         end do
c        Scan over all allowed point group operations, to find the
c        one that maps ic into the IRW
         itest=0
         groupops: do igroup=1,ngroup
           rv1=transform2D(rv2,group(igroup),ixy,isx,isy) 
c          These are the equivalent points.  Now map to 1st BZ
           do it=1,3              ! move the difference into the 1BZ
             do l=1,2
                if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
                if(rv1(l).gt.+pi+eps) rv1(l)=rv1(l)-twor*pi
             end do
           end do
c          See if this point is in the IRW!
           do jc=1,Ncw
              if(sqrt(sum((rv1(1:2)-Kc(1:2,jc))**2)).le.eps) then
                 ickmap(ic)=jc
                 ickdeg(jc)=ickdeg(jc)+1
                 irwgroup(ic)=igroup
                 itest=1
              end if
           end do
           if(itest.eq.1) exit    ! at first group op. that takes ic into IRW
         end do groupops
         if(itest.eq.0) then
            if(myrank.eq.0) write(lud,*) 'mapping failed for' !0
            if(myrank.eq.0) write(lud,*) 'ic',ic !0
            if(myrank.eq.0) write(lud,*) 'K=',rv2(1:2) !0
            stop
         end if
      end do !ic
      
      if(iprint.ge.5.and.myrank.eq.0) then 
        do ic=1,Nc    
          write(lud,"(2x,i2,6x,i2,9x,i2,9x,i2)") 
     &    ic,ickmap(ic),ickdeg(ic),irwgroup(ic)
        end do
      end if
         


c     Test the degeneracy table.  Since ickdeg holds the degeneracy
c     of points in the IRW, its sum must equal Nc.
      i=0
      do ic=1,Ncw
         i=i+ickdeg(ic)
      end do
      if(i.ne.Nc) then
         if(myrank.eq.0) write(lud,*) 'bug in ickdeg',i
         stop
      end if
        
c     Test the group table.  irwgroup(ic within wedge)=1, the identity
      do ic=1,Ncw
      if(irwgroup(ic).ne.1) then
        if(myrank.eq.0) write(lud,*) 'irwgroup failed for wedge pt.' !0
        if(myrank.eq.0) write(lud,*) 'ic',ic,'irwgroup(ic)',irwgroup(ic) !0
        if(myrank.eq.0) write(lud,"('K=(',f6.3,1x,f6.3,')')") Kc(1:2,ic) !0
      end if
      end do


c***********************************************************************
c     generate the kt-points in the Brillouin-zone of the 
c     super-lattice, i.e. in the Wigner-Seitz cell of the  
c     reciprocal space defined by the cluster K-points
c***********************************************************************
     
c      Dkt=halfr*sqrt(oner*Nc)/real(ntot,kind); rv2=zeror; rv1=zeror
      Dkt=oner/real(ntot,kind); rv2=zeror; rv1=zeror
      nshell=3 ! must be adjusted so that enough G are tested.
      nwsc=0   ! number of k-points in Wigner-Seitz cell 
      do i=-ntot+1,ntot
      do j=-ntot+1,ntot
          iv1=(/i,j,0/)
c          rv1(1)=Dkt*((i-eps)*g(1,1)+(j-eps)*g(2,1))  !ktx
c          rv1(2)=Dkt*((i-eps)*g(1,2)+(j-eps)*g(2,2))  !kty
          rv1(1:2)=Dkt*(iv1(1:2)-halfr)*pi
c         Check if this kt-point is outside the Wigner-Seitz cell if
c
c         (kt.G)/(G.G) > 1/2  for any 
c
c         G = n g1 + m g2 with n,m not both zero
c
          itest=0
          do n=-nshell,nshell
          do m=-nshell,nshell
            if(m.eq.0.and.n.eq.0) then
c             do nothing
            else
              iv2=(/n,m,0/)
              do l=1,ndim
                 rv2(l)=sum(iv2(1:ndim)*g(1:ndim,l))  ! G
              end do
              if(dot_product(rv1,rv2)/dot_product(rv2,rv2).gt.halfr)then
c               rv1=kt is outside the WS cell defined by G=rv2
                itest=1
                exit
              end if
            end if
          end do !m
            if(itest.eq.1) exit
          end do !n
          if(itest.eq.0) then 
             nwsc=nwsc+1
             kt(:,nwsc) = rv1(:)
          endif
       enddo ! j
       enddo ! i
c      
c      test the WS cell
       if(abs(Nc*nwsc-(twor*ntot)**2)/(twor*ntot)**2.gt.eps) then         
          write(lud,*) 'Nc*Number of points in WS, nwsc', Nc*nwsc
          write(lud,*) 'Total number of points in BZ', (2*ntot)**2 
c          stop
       end if
       
c***********************************************************************
c     Form Epsbar(K)
c***********************************************************************        

      if(myrank.eq.0.and.iprint.ge.5) write(lud,*) '    '
      if(myrank.eq.0.and.iprint.ge.5) write(lud,*) 
     &               'ic   Epsbar(ic)   Dwavebar(ic)'


      Epsbar=zeror
      Dwavebar=zeror
      do ic=1,Nc
         do i=1,nwsc
            rv1(1) = Kc(1,ic)+kt(1,i)
            rv1(2) = Kc(2,ic)+kt(2,i)
c           Epsbar(ic)=Epsbar(ic) - halfr*sum(cos(rv1(1:2)))
            Epsbar(ic)= Epsbar(ic) -halfr*(cos(rv1(1))+cos(rv1(2))) 
     &                - tprime*(cos(rv1(1))*cos(rv1(2))-oner)
            Dwavebar(ic)=Dwavebar(ic) -halfr*(cos(rv1(1))-cos(rv1(2))) 
         end do
         Epsbar(ic)=Epsbar(ic)/real(nwsc,kind)
         Dwavebar(ic)=Dwavebar(ic)/real(nwsc,kind)
      if(myrank.eq.0.and.iprint.ge.5) write(lud,"(i2,6x,f9.6,2x,f9.6)") 
     &           ic,Epsbar(ic),Dwavebar(ic)
      end do
c     Check 
      r1=sum(Epsbar); r2=sum(Epsbar*ickdeg)
      if(abs(tprime).lt.eps.and.abs(r1).gt.eps
     &   .or.
     &  abs(tprime).lt.eps.and.abs(r2).gt.eps) then
        if(myrank.eq.0) then
          write(lud,*) 'sum Epsbar wrong.',r1,r2
        end if
        stop
      end if

     
c***********************************************************************        
c     Form the Fourier transform tables.  
c     R-->K,  FTCoefs_K_to_R
c     K-->R,  FTCoefs_K_to_R
c
c                  __          the factor of 1/N is required
c               1  \           so that both G(R)~1/w and G(K)~1/w
c     G(R=0) = --- /  G(K)     for large w
c              N   --
c                  K
c***********************************************************************        
      do ic=1,Nc
        do ik=1,Nc
          r1=zeror
          do idim=1,ndim
             r1=r1+Kc(idim,ik)*Rc(idim,ic)
          end do
          FTCoefs_K_to_R(ik,ic)=exp(-ii*r1)/real(Nc,kind)
          FTCoefs_R_to_K(ik,ic)=exp(+ii*r1)
        end do
      end do
      
c     Now test for consistency of Kc and Rc
      do ic=1,Nc
        r1= real(sum(FTCoefs_K_to_R(1:Nc,ic)))-kroniker(ic,icR0)
        r2=aimag(sum(FTCoefs_K_to_R(1:Nc,ic)))
        r3= real(sum(FTCoefs_K_to_R(ic,1:Nc)))-kroniker(ic,icK0)
        r4=aimag(sum(FTCoefs_K_to_R(ic,1:Nc)))
        
        if(abs(r1).gt.eps.or.abs(r2).gt.eps) then
          write(6,*) 'Kc and Rc are inconsistent'
          write(6,*) 'icr,r1,r2=',ic,r1,r2
          stop
        end if
        if(abs(r3).gt.eps.or.abs(r4).gt.eps) then
          write(6,*) 'Kc and Rc are inconsistent, K=0'
          write(6,*) 'ick,r3,r4=',ic,r3,r4
          stop
        end if
      end do

c***********************************************************************        
c       Create the d-wave form factor
c***********************************************************************        
        gdwave=zeror
        do ik=1,Nc
          do idim=1,2    ! d_x2-y2 form factor
            gdwave(ik)=gdwave(ik)+cos(Kc(idim,ik))*(-1)**(idim-1)
          end do
        end do



c print out the d-wave order parameter in real space
      if(myrank.eq.0.and.iprint.ge.5) write(lud,*) '     '
      if(myrank.eq.0.and.iprint.ge.5) 
     &        write(lud,*) 'Dwave order parameter'
      if(myrank.eq.0.and.iprint.ge.5) 
     &        write(lud,*) 'ic, jc, Dwave(ic,jc)'
      Dwave(1:32,1:32) = cmplx(zeror,zeror)
      do ic =1,Nc   ! for all R1
       do jc=1,Nc   ! for all R2
        do ik=1,Nc  ! sum over K
          Dwave(ic,jc)=Dwave(ic,jc)+FTCoefs_K_to_R(ik,ic)
     &               *conjg(FTCoefs_K_to_R(ik,jc))*Dwavebar(ik)
        end do
        if (abs(Dwave(ic,jc)).gt.eps) then
         if(myrank.eq.0.and.iprint.ge.5) write(lud,"(i2,3x,i2,3x,f9.6)")
     &      ic,jc,real(Dwave(ic,jc)*Nc)
        end if
      end do
      end do
c**********************************************************************
c       TABLES FOR THE TWO-PARTICLE CALCULATIONS
c**********************************************************************
      if(isg.eq.1) then
         
         if(myrank.eq.0) write(lud,*) 'Allocating 2-particle arrays' !0
c        call allocate_2p       ! Needs Ncw from above
        
c**********************************************************************        
c        For two-particle propagators Gamma(K,K'), the group operations
c        may only be applied similtaneously to all K.  Thus, if U_K(K) is 
c        the operation that maps K to the IW, then we need to define the
c        table U_K(K') for all K.  This is done by the table
c        ick2map(a,b) : takes indices a, b to k-points K(a), K'(b) and 
c        returns the index c of a k-point Q(c), which index c is obtained
c        by performing on b whatever transformation would be necessary to
c        perform on a to send K(a) into the irreducible wedge.
c**********************************************************************        

c        First consider the K already in the irreducible wedge, where
c        U_K(K')=K'
         do ic=1,Ncw
c           In IRW, only the identity group operation leaves K in the IRW
            do jc=1,Nc
               ick2map(ic,jc)=jc 
            end do
         end do
c        Now consider the K that fall outside of the IW
         do ic=Ncw+1,Nc
            igroup=irwgroup(ic) ! this group operation maps K(ic) into IRW
            do jc=1,Nc
               i2=ig(jc)
               j2=jg(jc)
               iv1(:)=(/i2,j2,0/)
c              calculate the corresponding K       
               do l=1,2
                  rv1(l)=sum(iv1(1:2)*g(1:2,l))
               end do
c              now apply the sampe map to jc that takes K(ic) to the IRW.    
               rv2=transform2D(rv1,group(igroup),ixy,isx,isy) 
c              Now we have to figure out what the point rv2 is. First,
c              map to 1st BZ
               do it=1,2        ! 
               do l=1,2
                  if(rv2(l).lt.-pi+eps) rv2(l)=rv2(l)+twor*pi
                  if(rv2(l).gt.pi+eps)  rv2(l)=rv2(l)-twor*pi
               end do
               end do
c              Find the jk that corresponds to rv1=K     
               itest=0    
               do jk=1,Nc
                  if(sum(abs(rv2(1:2)-Kc(1:2,jk))).lt.eps) then
                     ick2map(ic,jc) = jk
                     itest=1
                  end if
               end do
               if(itest.eq.0) then
                  if(myrank.eq.0) write(lud,*) 
     &                            'ick2map mapping failed for' !0
                  if(myrank.eq.0) write(lud,*) 'ic',ic,'jc',jc !0
                  stop
               end if
            end do
         end do

c        Perform a test on both ick2map and ickmap
         do ic=1,Nc
            if(ick2map(ic,ic).ne.ickmap(ic).and.myrank.eq.0) then
               write(lud,*) 'problem with ick2map or ickmap' !0
               write(lud,*) 'ic=',ic,' ickmap(ic)=',ickmap(ic) !0
               write(lud,*) 'ick2map(ic,ic)=',ick2map(ic,ic) !0
            end if
         end do
          
         if(iprint.ge.5.and.myrank.eq.0) then
            write(lud,*) ' ic  jc  ick2map(ic,jc)' !0
            do ic=1,Nc          !0
               do jc=1,Nc       !0
                  write(lud,"(1x,i2,2x,i2,5x,i2)") ic,jc,ick2map(ic,jc) !0
               end do           !0
            end do                                                      !0
         end if
        
c***********************************************************************        
c         define the  fmat matrix
c***********************************************************************        
c         do ix=1,Nc             !space label
c            do jk=1,Nc          !k-label
c               do n=1,nl        !time label
c                  do m=1,nl     !frequency label
c                     
c                     ind=ix + (n-1)*Nc ! space-time index
c                     jmd=jk + (m-1)*Nc ! momentum-frequency index
c                     jm=m-1-nl/2
c                     fmat(ind,jmd)=dt*
c     &                exp(-ii*pi*((twor*jm+oner)*(n-oner)/real(nl,kind))
c     &                -ii*Kc(1,jk)*Rc(1,ix)
c     &                -ii*Kc(2,jk)*Rc(2,ix)  )
c                  end do
c               end do
c            end do
c         end do
         
      end if                    ! isg.eq.1


      stop
      end  
      
      
c***********************************************************************        
c***********************************************************************        
c         2D subroutines and functions
c***********************************************************************        
c***********************************************************************        

        
      function gtransform2D(a,igroup,ixy,isx,isy)
c
c  x <-- isx*[(1-ixy)*x + ixy*y] 
c  y <-- isy*[(1-ixy)*y + ixy*x] 
c
      implicit none
      integer :: i,a(3,3),g(3,3),gtransform2D(3,3),igroup,ixy(8),
     &           isx(8),isy(8)
      
      do i=1,3  
         gtransform2D(i,1)=isx(igroup)*((1-ixy(igroup))*a(i,1) 
     &                  + ixy(igroup)*a(i,2))
         gtransform2D(i,2)=isy(igroup)*((1-ixy(igroup))*a(i,2) 
     &                  + ixy(igroup)*a(i,1))
      end do
      
      return
      end
        
      function Rtransform2D(Rc,igroup,ixy,isx,isy)
c
c  x <-- isx*[(1-ixy)*x + ixy*y] 
c  y <-- isy*[(1-ixy)*y + ixy*x] 
c
      implicit none
      integer :: i,Rc(3),g(3),Rtransform2D(3),igroup,ixy(8),
     &           isx(8),isy(8)
      
      Rtransform2D(1)=isx(igroup)*((1-ixy(igroup))*Rc(1)
     &             + ixy(igroup)*Rc(2))
      Rtransform2D(2)=isy(igroup)*((1-ixy(igroup))*Rc(2)
     &             + ixy(igroup)*Rc(1))
                         
      return
      end
      
      function Ktransform2D(Kc,igroup,ixy,isx,isy)
c     
c  x <-- isx*[(1-ixy)*x + ixy*y] 
c  y <-- isy*[(1-ixy)*y + ixy*x] 
c
      implicit none
      integer :: i,igroup, ixy(8),isx(8),isy(8)
      real(8) :: Kc(3),g(3),Ktransform2D(3)
      
      Ktransform2D(1)=isx(igroup)*((1-ixy(igroup))*Kc(1) 
     &             + ixy(igroup)*Kc(2))
      Ktransform2D(2)=isy(igroup)*((1-ixy(igroup))*Kc(2) 
     &             + ixy(igroup)*Kc(1))
      
      return
      end
      
      function kroniker(i,j)
      integer i,j,kroniker
      
      if(i.eq.j) then
        kroniker=1
      else
        kroniker=0   
      end if
      
      return
      end 
