      program tables_3D_Betts      ! THREE DIMENSIONAL BETTS TABLES
c***************************************************************************
c     CHANGES REQUIRED FOR USE IN QMC CODES:
c     1. remove comment(s) in front of
c          use Global
c          define the  fmat matrix
c          call allocate_2p
c     2. change program and end statements to subroutine and return
c     3. delete the "Stuff that belongs in module_Global"
c     4. remove "write(6,*) 'enter cluster'"...
c     5. remove any duplicate defintions of kroniker
c     6. remove these comments
c     CHANGES REQUIRED FOR USE IN Analysis CODES:
c     1. do the steps above
c     2. remove the delcarations of ict, ig, jg and kg
c***************************************************************************
c     This subroutine makes the lookup tables used throughout the
c     full THREE DIMENSIONAL code.  It employs the definitions and 
c     cluster geometries developed by D.D. Betts and G.E. Stewart,
c     Canadian Journal of Physics, 75, 47-66.
c***************************************************************************
c      use Global
      implicit none
c***************************************************************************
      interface transform3D
      function gtransform3D(a,igroup,ixy,ixz,iyz,isx,isy,isz)
      integer :: a(3,3),gtransform3D(3,3),igroup,ixy(48),ixz(48),
     &           iyz(48),isx(48),isy(48),isz(48)
      end
      
      function Rtransform3D(Rc,igroup,ixy,ixz,iyz,isx,isy,isz)
      integer :: Rc(3),Rtransform3D(3),igroup,ixy(48),ixz(48),iyz(48),
     &     isx(48),isy(48),isz(48)
      end
      
      function Ktransform3D(Kc,igroup,ixy,ixz,iyz,isx,isy,isz)
      integer    :: igroup, ixy(48),ixz(48),iyz(48),
     &     isx(48),isy(48),isz(48)
      real(8) :: Kc(3),Ktransform3D(3)
      end
      end interface 
      
      interface Inverse
      function Inverse_3x3(a)
      integer :: a(3,3)
      real(8) :: Inverse_3x3(3,3)
      end function
      end interface
      
      interface determinant
      function Ideterminant(a)
      integer :: a(3,3),Ideterminant
      end function
      function Rdeterminant(a)
      real(8) :: a(3,3),Rdeterminant
      end function
      end interface

c***************************************************************************
c     Stuff that belongs in module_Global
c***************************************************************************
      character*3 :: cluster
      integer, parameter :: kind=8, lud=6, iprint=10, ntot=100, isg=1,
     &                      myrank=0, ndim=3, nl=10
      integer :: Nc,Ncw,N_sp,ntw,icK0,icKpi,Rc(3,32),a(3,3),
     &           icR0,ngroup,icrequ(32,48),nwsc,
     &           icrdiff(32,32),ickdiff(32,32),ickplus(32,32),
     &           ickequ(32,48),ickdeg(32),ickmap(32),ick2map(32,32)
      real(kind), parameter :: zeror=0.0_kind,halfr=0.5_kind,
     &                         oner=1.0_kind,twor=2.0_kind,
     &                         eps=0.0001_kind,pi=3.1415927,
     &                         tprime=0.0_kind
      real(kind) :: g(3,3),Kc(3,32), Epsbar(32),kt(3,(2*ntot+1)**3)
      complex(kind) :: FTCoefs_K_to_R(32,32),FTCoefs_R_to_K(32,32)
      complex(kind), parameter :: ii=(0.0_kind,1.0_kind)
c***************************************************************************
c     Locally defined
c***************************************************************************
      logical    :: allowed,denied,denied_lin_alg
      integer :: i,i1,i2,ix,iy,it,igroup,ic,ic1,ic2,ick,idim,ind,
     &           ik,itest,j,jmd,jm,j1,j2,jc,jk,jkc,k,k2,kroniker,
     &           l,m,m1,m2,n,n1,n2,p,Lc,iv1(3),iv2(3),iv3(3),
     &           ig(32),jg(32),kg(32),ict(-32:32,-32:32,-32:32),
     &           a1(3,3),a2(3,3),nshell,
     &           ixy(48),ixz(48),iyz(48),isx(48),isy(48),isz(48),
     &           irwgroup(32),group(48),neighbor(6,3),nneigh(32,6)
      real(kind) :: r1,r2,r3,r4,r5,r6,r7,r8,r9,Dkt,c1(3,3),c2(3,3),
     &              rv1(3),rv2(3),rv3(3)
c For lin algebra test
      real(kind) :: ainv(3,3),a1ainv(3,3)
c Select use of "brute force" vs linear algebra method for 
c checking lattice coincidence. Both methods should agree; lin alg
c method is much faster.
      logical, parameter :: use_brute_force = .false.
      logical, parameter :: use_lin_alg = .true.

c***************************************************************************
c       Local arrays
c       ixy(igroup)      D=3 cubic point group table(see below)
c       ixz(igroup)      D=3 cubic point group table(see below)
c       iyz(igroup)      D=3 cubic point group table(see below)
c       isx(igroup)      D=3 cubic point group table(see below)
c       isy(igroup)      D=3 cubic point group table(see below)
c       isz(igroup)      D=3 cubic point group table(see below)
c       group(i)         Indices of allowed group operations
c       irwgroup(ic)     the group operation maps K(ic) into IRW
c       ig(ic)           i in the g1 and j in the g2 direction
c       jg(jc)           k in the g3 direction
c       kg(ic)           corresponding to K point ic
c       ict(i,j,k)       maps i*g1+j*g2+k*g3 to ic
c***************************************************************************
        
c       CLUSTER MOMENTUM and DUAL SPACE for the 3D cubic lattice
c                                                             _   _   _
c       Different Real-space Cluster Geometries.  The vectors a1, a2, a3
c       define the parallelpipeds which define the periodically replicated 
c       clusters.  Here, we only use the clusters which Betts considers "good".
c               _             _            _
c  cluster      a1            a2           a3       group  Ng bipartide 
c    1A     ( 1, 0, 0)    ( 0, 1, 0)   ( 0, 0, 1)     Oh   48   no
c
c    8A     ( 2, 0, 0)    ( 0, 2, 0)   ( 0, 0, 2)     Oh   48   yes
c    8B     ( 2, 1, 0)    ( 0, 1, 2)   ( 1,-1, 1)     C2h  4    no  (=C2 x Ci)
c
c   10A     ( 1, 0, 2)    ( 1, 2,-1)   ( 1,-2, 0)     Ci   2    no
c
c   12A     ( 2, 1, 1)    ( 1,-2, 0)   ( 1, 0,-2)     C2h  4    no
c   12B     ( 0, 2, 0)    ( 0, 0, 3)   ( 2, 0, 1)     C2h  4    no
c
c   14A     ( 2, 2, 2)    ( 2,-1, 1)   ( 1, 2,-1)     C3i  6   yes
c   14B     ( 2, 2, 1)    ( 0,-1, 2)   ( 2,-1, 0)     Ci   2    no
c   14C     ( 2, 2, 0)    (-1, 2, 1)   ( 1, 0, 2)     Ci   2    no
c
c   16A     ( 2, 2, 0)    ( 2, 0, 2)   ( 0, 2, 2)     Oh  48   yes*
c   16B     ( 2, 2, 0)    ( 1,-1,-2)   (-1, 1,-2)     D2h  8   yes* (=D2 x Ci)
c   16C     ( 4, 2, 0)    ( 1, 2,-1)   (-1, 2, 1)     D2h  8   yes*
c   16D     ( 2,-2, 0)    ( 0, 2,-2)   ( 1, 2, 1)     C2h  4   yes
c   16E     ( 2, 2, 0)    ( 2,-2, 0)   ( 1, 0,-2)     C2h  4    no
c   16F     ( 1, 2, 0)    (-3, 1,-1)   ( 0, 2, 2)     C2h  4    no
c
c   18A     ( 2, 2, 2)    ( 2,-1,-1)   ( 1, 1,-2)     D3h 12   yes*
c   18B     ( 3, 2, 1)    ( 2,-1,-1)   ( 1, 1,-2)     C3i  6   yes*
c   18C     ( 2, 2, 2)    ( 2, 1,-1)   (-1, 2,-1)      Ci  2   yes
c   18D     ( 0, 0, 3)    ( 2,-2,-1)   ( 2, 1,-1)     C2h  4    no
c   18E     ( 3, 3, 0)    ( 2, 0, 2)   ( 2, 0,-1)     C2h  4    no
c   18F     ( 3, 2, 1)    ( 0, 2,-2)   ( 0, 2, 1)     C2h  4    no
c
c   20A     ( 3, 2, 1)    ( 2,-1,-1)   (-1, 1,-2)     Ci   2    yes
c   20B     ( 4, 1, 1)    ( 0, 2,-2)   ( 1, 2, 1)     Ci   2    yes
c   20C     ( 2, 2, 2)    ( 2, 0,-2)   (-1, 1,-2)     Ci   2    yes
c   20D     ( 3, 1, 0)    ( 0, 2, 2)   ( 1,-2, 1)     Ci   2    yes
c   20E     ( 4, 1, 1)    ( 2,-2, 0)   ( 1,-1,-2)     C2h  4    yes
c   20F     ( 2, 2, 1)    ( 2,-2, 1)   ( 1, 0,-2)     C2h  4    yes
c   20G     ( 2, 2, 1)    ( 2,-2, 0)   ( 1, 1,-2)     C2h  4    no
c
c   22A     ( 4, 2, 0)    ( 0, 3,-1)   (-1, 2, 1)     Ci   2    yes
c   22B     ( 2, 2, 2)    ( 3, 0,-1)   (-1, 1,-2)     Ci   2    yes
c   22C     ( 3, 2, 0)    ( 0, 2, 2)   ( 1,-1, 2)     Ci   2    no
c   22D     ( 1, 2, 0)    ( 2, 1, 2)   ( 2,-2, 1)     Ci   2    no
c   22E     ( 4, 1, 1)    ( 2, 2,-3)   ( 0, 2,-1)     Ci   2    no
c
c   24A     ( 3, 1, 0)    ( 0, 2,-2)   ( 0, 2, 2)     C2h  4    yes
c   24B     ( 2, 2, 2)    ( 0, 2,-2)   ( 2, 0,-2)     D3h 12    yes
c   24C     ( 2, 2, 2)    ( 3, 0,-1)   (-1, 2,-1)     Ci   2    yes
c   24D     ( 0, 0, 3)    ( 1, 3, 0)   (-2, 2, 0)     D2h  8    no
c   24E     ( 3, 2, 0)    ( 3,-2, 0)   ( 1, 0,-2)     C2h  4    no
c                                                           
c   26A     ( 3, 2, 1)    ( 2,-2,-2)   ( 1, 1,-2)     Ci   2    yes
c   26B     ( 1, 3, 2)    ( 3,-2, 1)   ( 3, 1, 0)     C3i  6    yes
c   26C     ( 2, 1, 3)    ( 2,-1,-2)   ( 2, 2,-1)     Ci   2    no
c
c  The bipartide lattices had a eps_0^HA in table 8 (and common sense), 
c  A * means equivlent to others with same Nc.
c  I.e., lattices 16A=16B=16C, 18A=18B,  20B=20C=20D.
c
c  Only 16A has all 48 operations and has the point group Oh.  The remaining clusters 
c  have lower symmetry than the cubic lattice, with symmetries designated by the
c  Schoenflies symbols (From Ibach and Luth):
c  Cj (j=2,3,4, 6) j-fold rotation axis 
c  Sj j-fold rotation-inversion axis 
c  Dj j 2-fold rotation axes perpendicular to a j-fold principle rotation axis 
c  T  4 three-and 3 two-fold rotation axes, as in a tetrahedron 
c  O  4 three-and 3 four-fold rotation axes, as in a octahedron 
c  Ci a center of inversion 
c  Cs a mirror plane
c  In addition, their are sufixes for mirror planes 
c     h: horizontal=perpendicular to the c  rotation axis, 
c     v: vertical=parallel to the main rotation axis in the plane, 
c     d: diagonal=parallel to the main rotation axis in the plane  
c        bisecting the two-fold rotation axes.
c 
c  The Clusters identified by Betts, have several different symmetries
c
c  C2h   A two fold axis, plus a single mirror plane that is perpendicular to the axis
c  Ci    Contains only the identity and inversion
c  C3i   Triagonal with a 3-fold axis and inversion (GUESS)??
c          * rotations by 2pi/3 about the axix x=y=z
c          * inversion wrt the origin.
c  Oh    The full symmetry of the cubic lattice 
c  D2h   2 2-fold axes perp. to the 2-fold rotation axis, plus a horizonal mirror plane
c  D3h   3 2-fold axes perp. to the 3-fold rotation axis, plus a horizonal mirror plane
c
c  We need to automate the process of finding which of the 48 point group operations are
c  retained by the different clusters.  To do this, we must perform all 48 cubic group
c  operations on the {a} and see if they form an equivalent parallelpiped.  
c
c  Group Operations for a cubic lattice:
c
c  x --> +/- x                  There are 2^D D! operations in D dimensions
c    \/                         or 48 operations in 3D.
c    /\
c  y --> +/- y   
c    \/
c    /\
c  z --> +/- z   
c        .
c  a'=transform3D(a,igroup,ixy,ixz,iyz,isx,isy,isz)
c  ixy=0(1) dont (do) exchange x and y
c  ixz=0(1) dont (do) exchange x and z
c  iyz=0(1) dont (do) exchange y and z
c  isx      sign of x 
c  isy      sign of y
c  isz      sign of z
c
c  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
c  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
c  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
c

      ixy( 1)=0;ixz( 1)=0; iyz( 1)=0; isx( 1)=+1; isy( 1)=+1; isz( 1)=+1 !identity  xyz e
      ixy( 2)=1;ixz( 2)=0; iyz( 2)=0; isx( 2)=+1; isy( 2)=+1; isz( 2)=+1 !ref. x=y  yxz o
      ixy( 3)=0;ixz( 3)=1; iyz( 3)=0; isx( 3)=+1; isy( 3)=+1; isz( 3)=+1 !ref. x=z  zyx o
      ixy( 4)=0;ixz( 4)=0; iyz( 4)=1; isx( 4)=+1; isy( 4)=+1; isz( 4)=+1 !ref. y=z  xzy o
      ixy( 5)=1;ixz( 5)=1; iyz( 5)=0; isx( 5)=+1; isy( 5)=+1; isz( 5)=+1 !          yzx e
      ixy( 6)=1;ixz( 6)=0; iyz( 6)=1; isx( 6)=+1; isy( 6)=+1; isz( 6)=+1 !          zxy e
      
      ixy( 7)=0;ixz( 7)=0; iyz( 7)=0; isx( 7)=-1; isy( 7)=+1; isz( 7)=+1 !
      ixy( 8)=1;ixz( 8)=0; iyz( 8)=0; isx( 8)=-1; isy( 8)=+1; isz( 8)=+1 !
      ixy( 9)=0;ixz( 9)=1; iyz( 9)=0; isx( 9)=-1; isy( 9)=+1; isz( 9)=+1 !
      ixy(10)=0;ixz(10)=0; iyz(10)=1; isx(10)=-1; isy(10)=+1; isz(10)=+1 !
      ixy(11)=1;ixz(11)=1; iyz(11)=0; isx(11)=-1; isy(11)=+1; isz(11)=+1 !
      ixy(12)=1;ixz(12)=0; iyz(12)=1; isx(12)=-1; isy(12)=+1; isz(12)=+1 !
     
      ixy(13)=0;ixz(13)=0; iyz(13)=0; isx(13)=+1; isy(13)=-1; isz(13)=+1 !
      ixy(14)=1;ixz(14)=0; iyz(14)=0; isx(14)=+1; isy(14)=-1; isz(14)=+1 !
      ixy(15)=0;ixz(15)=1; iyz(15)=0; isx(15)=+1; isy(15)=-1; isz(15)=+1 !
      ixy(16)=0;ixz(16)=0; iyz(16)=1; isx(16)=+1; isy(16)=-1; isz(16)=+1 !
      ixy(17)=1;ixz(17)=1; iyz(17)=0; isx(17)=+1; isy(17)=-1; isz(17)=+1 !
      ixy(18)=1;ixz(18)=0; iyz(18)=1; isx(18)=+1; isy(18)=-1; isz(18)=+1 !
      
      ixy(19)=0;ixz(19)=0; iyz(19)=0; isx(19)=-1; isy(19)=-1; isz(19)=+1 !
      ixy(20)=1;ixz(20)=0; iyz(20)=0; isx(20)=-1; isy(20)=-1; isz(20)=+1 !
      ixy(21)=0;ixz(21)=1; iyz(21)=0; isx(21)=-1; isy(21)=-1; isz(21)=+1 !
      ixy(22)=0;ixz(22)=0; iyz(22)=1; isx(22)=-1; isy(22)=-1; isz(22)=+1 !
      ixy(23)=1;ixz(23)=1; iyz(23)=0; isx(23)=-1; isy(23)=-1; isz(23)=+1 !
      ixy(24)=1;ixz(24)=0; iyz(24)=1; isx(24)=-1; isy(24)=-1; isz(24)=+1 !
      
      ixy(25)=0;ixz(25)=0; iyz(25)=0; isx(25)=+1; isy(25)=+1; isz(25)=-1 !
      ixy(26)=1;ixz(26)=0; iyz(26)=0; isx(26)=+1; isy(26)=+1; isz(26)=-1 !
      ixy(27)=0;ixz(27)=1; iyz(27)=0; isx(27)=+1; isy(27)=+1; isz(27)=-1 !
      ixy(28)=0;ixz(28)=0; iyz(28)=1; isx(28)=+1; isy(28)=+1; isz(28)=-1 !
      ixy(29)=1;ixz(29)=1; iyz(29)=0; isx(29)=+1; isy(29)=+1; isz(29)=-1 !
      ixy(30)=1;ixz(30)=0; iyz(30)=1; isx(30)=+1; isy(30)=+1; isz(30)=-1 !
      
      ixy(31)=0;ixz(31)=0; iyz(31)=0; isx(31)=-1; isy(31)=+1; isz(31)=-1 !
      ixy(32)=1;ixz(32)=0; iyz(32)=0; isx(32)=-1; isy(32)=+1; isz(32)=-1 !
      ixy(33)=0;ixz(33)=1; iyz(33)=0; isx(33)=-1; isy(33)=+1; isz(33)=-1 !
      ixy(34)=0;ixz(34)=0; iyz(34)=1; isx(34)=-1; isy(34)=+1; isz(34)=-1 !
      ixy(35)=1;ixz(35)=1; iyz(35)=0; isx(35)=-1; isy(35)=+1; isz(35)=-1 !
      ixy(36)=1;ixz(36)=0; iyz(36)=1; isx(36)=-1; isy(36)=+1; isz(36)=-1 !
      
      ixy(37)=0;ixz(37)=0; iyz(37)=0; isx(37)=+1; isy(37)=-1; isz(37)=-1 !
      ixy(38)=1;ixz(38)=0; iyz(38)=0; isx(38)=+1; isy(38)=-1; isz(38)=-1 !
      ixy(39)=0;ixz(39)=1; iyz(39)=0; isx(39)=+1; isy(39)=-1; isz(39)=-1 !
      ixy(40)=0;ixz(40)=0; iyz(40)=1; isx(40)=+1; isy(40)=-1; isz(40)=-1 !
      ixy(41)=1;ixz(41)=1; iyz(41)=0; isx(41)=+1; isy(41)=-1; isz(41)=-1 !
      ixy(42)=1;ixz(42)=0; iyz(42)=1; isx(42)=+1; isy(42)=-1; isz(42)=-1 !
      
      ixy(43)=0;ixz(43)=0; iyz(43)=0; isx(43)=-1; isy(43)=-1; isz(43)=-1 !
      ixy(44)=1;ixz(44)=0; iyz(44)=0; isx(44)=-1; isy(44)=-1; isz(44)=-1 !
      ixy(45)=0;ixz(45)=1; iyz(45)=0; isx(45)=-1; isy(45)=-1; isz(45)=-1 !
      ixy(46)=0;ixz(46)=0; iyz(46)=1; isx(46)=-1; isy(46)=-1; isz(46)=-1 !
      ixy(47)=1;ixz(47)=1; iyz(47)=0; isx(47)=-1; isy(47)=-1; isz(47)=-1 !
      ixy(48)=1;ixz(48)=0; iyz(48)=1; isx(48)=-1; isy(48)=-1; isz(48)=-1 !

      
      write(6,*) 'enter cluster'
      read(5,*) cluster
      read(cluster,"(i2)",iostat=ic) Nc           ! should ignore characters
      if (ic.gt.0) then
        read(cluster,"(i1)",iostat=ic) Nc
        if (ic.gt.0) then
           write(lud,*) "Cannot read Nc from string ",cluster
           stop
        endif
      endif

c***********************************************************************        
c       determine the principle translation vectors a1,a2, and a3 
c***********************************************************************
      if(iprint.ge.5.and.myrank.eq.0) 
     &     write(lud,*) ' cluster= ',cluster !0

      If(cluster.eq.'1A') then                              
         a(1,:)=(/1, 0, 0/); a(2,:)=(/0, 1, 0/); a(3,:)=(/0, 0, 1/)
         
      else if(cluster.eq.'8A') then                           
         a(1,:)=(/2, 0, 0/); a(2,:)=(/0, 2, 0/); a(3,:)=(/0, 0, 2/)
      else if(cluster.eq.'8B') then                           
         a(1,:)=(/2, 1, 0/); a(2,:)=(/0, 1, 2/); a(3,:)=(/1,-1, 1/)
         
      else if(cluster.eq.'10A') then                           
         a(1,:)=(/1, 0, 2/); a(2,:)=(/1, 2,-1/); a(3,:)=(/1,-2, 0/)
         
      else if(cluster.eq.'12A') then                           
         a(1,:)=(/2, 1, 1/); a(2,:)=(/1,-2, 0/); a(3,:)=(/1, 0,-2/)
      else if(cluster.eq.'12B') then                           
         a(1,:)=(/0, 2, 0/); a(2,:)=(/0, 0, 3/); a(3,:)=(/2, 0, 1/)
         
      else if(cluster.eq.'14A') then                           
         a(1,:)=(/2, 2, 2/); a(2,:)=(/2,-1, 1/); a(3,:)=(/1, 2,-1/)
      else if(cluster.eq.'14B') then                           
         a(1,:)=(/2, 2, 1/); a(2,:)=(/0,-1, 2/); a(3,:)=(/2,-1, 0/)
      else if(cluster.eq.'14C') then                           
         a(1,:)=(/2, 2, 0/); a(2,:)=(/-1, 2, 1/); a(3,:)=(/1, 0, 2/) 
         
      else if(cluster.eq.'16A') then                           
         a(1,:)=(/2, 2, 0/); a(2,:)=(/2, 0, 2/); a(3,:)=(/0, 2, 2/) 
      else if(cluster.eq.'16B') then                           
         a(1,:)=(/2, 2, 0/); a(2,:)=(/1,-1,-2/); a(3,:)=(/-1, 1,-2/) 
      else if(cluster.eq.'16C') then                           
         a(1,:)=(/4, 2, 0/); a(2,:)=(/1, 2,-1/); a(3,:)=(/-1, 2, 1/)
      else if(cluster.eq.'16D') then                           
         a(1,:)=(/2,-2, 0/); a(2,:)=(/0, 2,-2/); a(3,:)=(/ 1, 2, 1/) 
      else if(cluster.eq.'16E') then                           
         a(1,:)=(/2, 2, 0/); a(2,:)=(/2,-2, 0/); a(3,:)=(/ 1, 0,-2/) 
      else if(cluster.eq.'16F') then                           
         a(1,:)=(/1, 2, 0/); a(2,:)=(/-3,1,-1/); a(3,:)=(/ 0, 2, 2/) 
         
      else if(cluster.eq.'18A') then                           
         a(1,:)=(/ 2, 2, 2/); a(2,:)=(/ 2,-1,-1/); a(3,:)=(/ 1, 1,-2/)
      else if(cluster.eq.'18B') then                           
         a(1,:)=(/ 3, 2, 1/); a(2,:)=(/ 2,-1,-1/); a(3,:)=(/ 1, 1,-2/)
      else if(cluster.eq.'18C') then                           
         a(1,:)=(/ 2, 2, 2/); a(2,:)=(/ 2, 1,-1/); a(3,:)=(/-1, 2,-1/)
      else if(cluster.eq.'18D') then                           
         a(1,:)=(/ 0, 0, 3/); a(2,:)=(/ 2,-2,-1/); a(3,:)=(/ 2, 1,-1/)
      else if(cluster.eq.'18E') then                           
         a(1,:)=(/ 3, 3, 0/); a(2,:)=(/ 2, 0, 2/); a(3,:)=(/ 2, 0,-1/)
      else if(cluster.eq.'18F') then                           
         a(1,:)=(/ 3, 2, 1/); a(2,:)=(/ 0, 2,-2/); a(3,:)=(/ 0, 2, 1/)
         
      else if(cluster.eq.'20A') then                           
         a(1,:)=(/ 3, 2, 1/); a(2,:)=(/ 2,-1,-1/); a(3,:)=(/-1, 1,-2/)
      else if(cluster.eq.'20B') then                           
         a(1,:)=(/ 4, 1, 1/); a(2,:)=(/ 0, 2,-2/); a(3,:)=(/ 1, 2, 1/)
      else if(cluster.eq.'20C') then                           
         a(1,:)=(/ 2, 2, 2/); a(2,:)=(/ 2, 0,-2/); a(3,:)=(/-1, 1,-2/)
      else if(cluster.eq.'20D') then                           
         a(1,:)=(/ 3, 1, 0/); a(2,:)=(/ 0, 2, 2/); a(3,:)=(/ 1,-2, 1/)
      else if(cluster.eq.'20E') then                           
         a(1,:)=(/ 4, 1, 1/); a(2,:)=(/ 2,-2, 0/); a(3,:)=(/ 1,-1,-2/)
      else if(cluster.eq.'20F') then                           
         a(1,:)=(/ 2, 2, 1/); a(2,:)=(/ 2,-2, 1/); a(3,:)=(/ 1, 0,-2/)
      else if(cluster.eq.'20G') then                           
         a(1,:)=(/ 2, 2, 1/); a(2,:)=(/ 2,-2, 0/); a(3,:)=(/ 1, 1,-2/)
        
      else if(cluster.eq.'22A') then                           
         a(1,:)=(/ 4, 2, 0/); a(2,:)=(/ 0, 3,-1/); a(3,:)=(/-1, 2, 1/)
      else if(cluster.eq.'22B') then                           
         a(1,:)=(/ 2, 2, 2/); a(2,:)=(/ 3, 0,-1/); a(3,:)=(/-1, 1,-2/)
      else if(cluster.eq.'22C') then                           
         a(1,:)=(/ 3, 2, 0/); a(2,:)=(/ 0, 2, 2/); a(3,:)=(/ 1,-1, 2/)
      else if(cluster.eq.'22D') then                           
         a(1,:)=(/ 1, 2, 0/); a(2,:)=(/ 2, 1, 2/); a(3,:)=(/ 2,-2, 1/)
      else if(cluster.eq.'22E') then                           
         a(1,:)=(/ 4, 1, 1/); a(2,:)=(/ 2, 2,-3/); a(3,:)=(/ 0, 2,-1/)
         
      else if(cluster.eq.'24A') then                           
         a(1,:)=(/ 3, 1, 0/); a(2,:)=(/ 0, 2,-2/); a(3,:)=(/ 0, 2, 2/)
      else if(cluster.eq.'24B') then                           
         a(1,:)=(/ 2, 2, 2/); a(2,:)=(/ 0, 2,-2/); a(3,:)=(/ 2, 0,-2/)
      else if(cluster.eq.'24C') then                           
         a(1,:)=(/ 2, 2, 2/); a(2,:)=(/ 3, 0,-1/); a(3,:)=(/-1, 2,-1/)
      else if(cluster.eq.'24D') then                           
         a(1,:)=(/ 0, 0, 3/); a(2,:)=(/ 1, 3, 0/); a(3,:)=(/-2, 2, 0/)
      else if(cluster.eq.'24E') then                           
         a(1,:)=(/ 3, 2, 0/); a(2,:)=(/ 3,-2, 0/); a(3,:)=(/ 1, 0,-2/)
         
      else if(cluster.eq.'26A') then                           
         a(1,:)=(/ 3, 2, 1/); a(2,:)=(/ 2,-2,-2/); a(3,:)=(/ 1, 1,-2/)
      else if(cluster.eq.'26B') then                           
         a(1,:)=(/ 1, 3, 2/); a(2,:)=(/ 3,-2, 1/); a(3,:)=(/ 3, 1, 0/)
      else if(cluster.eq.'26C') then                           
         a(1,:)=(/ 2, 1, 3/); a(2,:)=(/ 2,-1,-2/); a(3,:)=(/ 2, 2,-1/)
      else
         if(myrank.eq.0) write(lud,*) 'Bad cluster, bad, bad' !0
         stop
      end if  
      if(iprint.ge.5.and.myrank.eq.0) 
     &   write(lud,"('a=',3('(',i2,1x,i2,1x,i2,') '))") 
     &   a(1,:),a(2,:),a(3,:)  

c     Test the a vectors      
      if(abs(determinant(a)).ne.Nc) then
        if(myrank.eq.0) write(lud,*) 'This cell volume isnt Nc',
     &                                abs(determinant(a))
        stop
      end if
      
c*********************************************************************************
c     Determine which group operations are allowed    
c*********************************************************************************
      
      ngroup=0; Lc=2*(Nc**.3333+1)
      if(myrank.eq.0.and.iprint.ge.5) then
        write(lud,*) '--------------------------------------------',
     &               '------------------' 
        write(lud,*) '             trans(a)        igroup L  ixy ix',
     &               'z iyz isx isy isz'
        write(lud,*) '--------------------------------------------',
     &               '------------------'
      end if
      do igroup=1,48
         a1=transform3D(a,igroup,ixy,ixz,iyz,isx,isy,isz)

         if (use_brute_force) then

c        look at the lattice points generated by a1 and see if they
c        correspond to those produced by a.
c        First, generate the a1 lattice points
         do i=-Lc,Lc
         do j=-Lc,Lc
         do k=-Lc,Lc
           denied=.true.
           iv1=(/i,j,k/)
           do l=1,3         ! x,ymz components of this lattice point
             rv1(l)=sum(iv1(:)*a1(:,l))
           end do
c          Now generate the a lattice points.  If a and a1 are the 
c          same lattice one of these points MUST correspond to ij
           do n=-4*Lc,4*Lc
           do m=-4*Lc,4*Lc
           do p=-4*Lc,4*Lc
             iv2=(/n,m,p/)
             do l=1,3         ! x,y components of this lattice point
               rv2(l)=sum(iv2(:)*a(:,l))
             end do
             if(sum(abs(rv1-rv2)).lt.eps) denied=.false.
           end do
           end do
           end do
           if(denied) exit !no corrs. point was found
         end do
         if(denied) exit
         end do
         if(denied) exit
         end do
         end if
         if (use_lin_alg) then
c PK If lattice points generated by "a1" vectors can be found in 
c lattice points generated by "a" vectors then (a1)x(a^-1) can only 
c contain integers.
            ainv=Inverse_3x3(a)
            a1ainv=matmul(a1,ainv)
            if (all(abs(a1ainv-nint(a1ainv)).le.eps)) then 
               denied_lin_alg=.false. ! matrix only contains integers
            else
               denied_lin_alg=.true.
            end if
         end if

c If both algorithms are enabled, check they agree
         if (use_brute_force.and.use_lin_alg) then
            if (denied.neqv.denied_lin_alg) then
               write (*,*) '*** BUG : Denied algorithm disagree'
               write (*,*) 'a11',a1(:,1)
               write (*,*) 'a12',a1(:,2)
               write (*,*) 'a13',a1(:,3)
               
               write (*,*) 'a1',a(:,1)
               write (*,*) 'a2',a(:,2)
               write (*,*) 'a3',a(:,3)
               write (*,*)
               write (*,*) 'ainv',ainv(:,1)
               write (*,*) 'ainv',ainv(:,2)
               write (*,*) 'ainv',ainv(:,3)
               write (*,*)
               
               write (*,*) 'a1ainv',a1ainv(:,1)
               write (*,*) 'a1ainv',a1ainv(:,2)
               write (*,*) 'a1ainv',a1ainv(:,3)
               write (*,*) 'Denied_lin_alg=',denied_lin_alg
               write (*,*) 'Denied_brute_force=',denied
               write (*,*)
               stop
            end if
         end if
         if (use_lin_alg) denied=denied_lin_alg

         if(denied.eqv..false.) then ! a corresponding point was found
            ngroup=ngroup+1
            group(ngroup)=igroup
         end if    

         if(myrank.eq.0.and.iprint.ge.5) then
           allowed=.true.;if (denied) allowed=.false.
           write(lud,"(3('(',i2,1x,i2,1x,i2,')'),2x,i2,2x,l,6(2x,i2))")
     &     a1(1,:),a1(2,:),a1(3,:),igroup,allowed,ixy(igroup),
     &     ixz(igroup),iyz(igroup),isx(igroup),isy(igroup),isz(igroup)
         end if
      end do      

     
      if(myrank.eq.0.and.iprint.ge.5) 
     &  write(lud,"('cluster=',a3,' ngroup=',i2,' groups=',48(i2,1x))") 
     &  cluster,ngroup,group(1:ngroup)

      
c*********************************************************************************
c       calculate the table of cluster point locations Rc(i,ic), i=1,2,3 for x,y,z
c*********************************************************************************       
c 
c  ^ y
c  |
cc-|-c---c-*-c---c---c---c               A point P is within the cluster 
c| | |   |   |   |   |   |               if a vector v from the box origin, 
cc-|-c---c---c---c---c---c               O, to the point satisfies
ca2^ * * * * * * * * |   |               
cc-|-c---c---c---c-*-c---c                        
c| | |   P   |   | * |   |                            _   _
cc-|-c--/c---c---c-*-c---c                   -eps  <  v__ai  < 1-eps
c| | | /v|   |   | * |   |                        
cc-|-c/--c---c---c-*-c---c 
c| | /   |   |   | * |   |               for all i=1,3 where v__ai is the
cc-|/0---c---c---c-*-c---c               projection of the vector v along
c| O--------------->---------> x         ai
cc---c---c---c---ca1-c---c                              -1
c|   |   |   |   |   |   |               v = A n;  n = A  v ; where A=transpose(a)
cc---c---c---c---c---c---c
c                                        v = i*a1 + j*a2 + k*a3
c
c                                            [ a1 ]
c                                        a = [ a2 ]
c                                            [ a3 ]
c                                        
      do i=1,3
        a1(:,i)=a(i,:)
      end do
      c1=Inverse(a1)
      if(iprint.ge.5.and.myrank.eq.0) write(lud,*) '  ' !0
      if(iprint.ge.5.and.myrank.eq.0) 
     &     write(lud,*) 'ic,Rcx(ic),Rcy(ic),Rcz(ic)' !0
      ic=0
      do i=-Nc,Nc
      do j=-Nc,Nc
      do k=-Nc,Nc
         rv1=(/i,j,k/)          ! The coordinates of the point.
         rv2=matmul(c1,rv1)
         if(rv2(1).lt.oner-eps.and.rv2(1).gt.zeror-eps.and.
     &      rv2(2).lt.oner-eps.and.rv2(2).gt.zeror-eps.and.
     &      rv2(3).lt.oner-eps.and.rv2(3).gt.zeror-eps ) then
            ic=ic+1
            Rc(:,ic)=(/i,j,k/) ! Rc, like i.e. a(1,1) is integer
            if(iprint.ge.5.and.myrank.eq.0)                             !0
     &           write(lud,"(1x,i2,3x,3(i2,6x))") ic,Rc(1,ic),Rc(2,ic), !0
     &                                            Rc(3,ic)              !0
            if(sum(abs(Rc(:,ic))).eq.0) icR0=ic
         end if
      end do
      end do
      end do
      if(ic.ne.Nc) then
         if(myrank.eq.0) write(lud,*) 'Bug in Rc(ic) loop' !0
         stop
      end if
      if(iprint.ge.5.and.myrank.eq.0) write(lud,*) 'icR0=',icR0
 
c***********************************************************************        
c     create a table mapping points outside the cluster back into it.
c***********************************************************************        
c     Note we use c1 from above!!        

      if(iprint.ge.5.and.myrank.eq.0) then                  !0
         write(lud,*) '  '                                  !0
         write(lud,*) 'Equivalency table r'                 !0
         write(lud,*) 'ic   igroup   icrequ(ic,ieqiv)'      !0
      end if                                                !0
      do ic=1,Nc
        do igroup=1,ngroup
c       Generate the equivalent points using the ngroup operations
          rv3=transform3D(Rc(:,ic),group(igroup),ixy,ixz,iyz,
     &                    isx,isy,isz)
          do it=1,4             !map rv3 back into the cluster.
            rv2=matmul(c1,rv3)  !project rv3 onto a1,a2,a3
            do l=1,3
               if(rv2(l).lt.zeror-eps) then
                  rv3(:)=rv3(:)+a(l,:)
               else if(rv2(l).gt.oner-eps) then
                  rv3(:)=rv3(:)-a(l,:)
               end if
            end do
          end do
c         Figure out which point rv3 is
          itest=0
          do jc=1,Nc
            if(sum(abs(rv3(:)-Rc(:,jc))).lt.eps) then ! rv3=Rc(:,ic)
               icrequ(ic,igroup)=jc
               itest=1
            end if
          end do
          if(itest.ne.1) then
            write(lud,*) 'No equivalent found in icrequ',ic,igroup
            stop
          end if 
          if(iprint.ge.5.and.myrank.eq.0) 
     &      write(lud,"(2x,i2,4x,i2,8x,i2)") ic,igroup,icrequ(ic,igroup) !0
        end do
      end do                               

c***********************************************************************        
c     Now create the tables for the differences of R
c***********************************************************************        
c     Note we use c1 from above!!   
     
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '  ' !0
        write(lud,*) 'differences in r' !0
        write(lud,*) '  ic   jc  icrdiff(ic,jc)' !0
      end if
      do i=1,Nc
      do j=1,Nc
         iv1(:)=Rc(:,i)-Rc(:,j)
         itest=0
         do n=-1,1        ! Look in the surrounding 
         do m=-1,1        ! 27 clusters for R by adding and
         do p=-1,1        ! subtracting a1, a2 and a3
            iv2=iv1 + n*a(1,:)+m*a(2,:)+p*a(3,:)
            rv2=matmul(c1,iv2)
            if(rv2(1).lt.oner-eps.and.rv2(1).gt.zeror-eps.and.
     &         rv2(2).lt.oner-eps.and.rv2(2).gt.zeror-eps.and.
     &         rv2(3).lt.oner-eps.and.rv2(3).gt.zeror-eps ) then
c              This point is within the cluster
c              Now we must find which point it is.
               do ic=1,Nc
                  if(sum(abs(iv2(:)-Rc(:,ic))).eq.0) then
                     icrdiff(i,j)=ic
                     itest=itest+1   ! complain if multiple solutions are found
                  end if
               end do
               if(iprint.ge.5.and.myrank.eq.0) 
     &              write(lud,"(2x,i2,4x,i2,6x,i2)") i,j,icrdiff(i,j) !0
            end if
         end do
         end do
         end do
         if(itest.ne.1) then
            write(lud,*) 'no icrdiff(i,j) found',i,j
            stop
         end if
      end do
      end do  
        
c***********************************************************************        
c     Now create the tables for the neighbors of cluster points
c***********************************************************************        
c     Note we use c1 from above!!   
     
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '  ' !0
        write(lud,*) 'neighbors of ic' !0
        write(lud,*) '  ic   neigh  nneigh(ic,jc)' !0
      end if
      neighbor(1,:)=(/ 1, 0, 0/)
      neighbor(2,:)=(/-1, 0, 0/)
      neighbor(3,:)=(/ 0, 1, 0/)
      neighbor(4,:)=(/ 0,-1, 0/)
      neighbor(5,:)=(/ 0, 0, 1/)
      neighbor(6,:)=(/ 0, 0,-1/)
      do i=1,Nc
      do j=1,6
         iv1(:)=Rc(:,i)-neighbor(j,:)
         itest=0
         do n=-1,1        ! Look in the surrounding 
         do m=-1,1        ! 27 clusters for R by adding and
         do p=-1,1        ! subtracting a1, a2 and a3
            iv2=iv1 + n*a(1,:)+m*a(2,:)+p*a(3,:)
            rv2=matmul(c1,iv2)
            if(rv2(1).lt.oner-eps.and.rv2(1).gt.zeror-eps.and.
     &         rv2(2).lt.oner-eps.and.rv2(2).gt.zeror-eps.and.
     &         rv2(3).lt.oner-eps.and.rv2(3).gt.zeror-eps ) then
c              This point is within the cluster
c              Now we must find which point it is.
               do ic=1,Nc
                  if(sum(abs(iv2(:)-Rc(:,ic))).eq.0) then
                     nneigh(i,j)=ic
                     itest=itest+1   ! complain if multiple solutions are found
                  end if
               end do
               if(iprint.ge.5.and.myrank.eq.0) 
     &              write(lud,"(2x,i2,4x,i2,6x,i2)") i,j,nneigh(i,j) !0
            end if
         end do
         end do
         end do
         if(itest.ne.1) then
            write(lud,*) 'no icrdiff(i,j) found',i,j
            stop
         end if
      end do
      end do  
      
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) 'nearest neighbors of icR0'
        do j=1,6
          write(lud,"(i2)") nneigh(icR0,j)
        end do
      end if

c***********************************************************************        
c       Now set up the K-space tables
c***********************************************************************        

c       Now calculate the principle translations in K-space
c   
c  
c                              _    _
c    _     _    _        2 pi aj x ak          
c    a --> g    g_i =  ------------------      plus cyclic permutations
c                      | ai . (aj x ak)|       
c                                           
c       Principle translation vectors for lattice tiling       
c                                        
c
      r1=determinant(a)
      
      g(1,1)=2*pi*( a(2,2)*a(3,3)-a(3,2)*a(2,3) )/r1  !a2y*a3z-a3y*a2z
      g(1,2)=2*pi*( a(2,3)*a(3,1)-a(3,3)*a(2,1) )/r1  !a2z*a3x-a3z*a2x
      g(1,3)=2*pi*( a(2,1)*a(3,2)-a(3,1)*a(2,2) )/r1  !a2x*a3y-a3x*a2y
      
      g(2,1)=2*pi*( a(3,2)*a(1,3)-a(1,2)*a(3,3) )/r1  !a3y*a1z-a1y*a3z
      g(2,2)=2*pi*( a(3,3)*a(1,1)-a(1,3)*a(3,1) )/r1  !a3z*a1x-a1z*a3x
      g(2,3)=2*pi*( a(3,1)*a(1,2)-a(1,1)*a(3,2) )/r1  !a3x*a1y-a1x*a3y
        
      g(3,1)=2*pi*( a(1,2)*a(2,3)-a(2,2)*a(1,3) )/r1  !a1y*a2z-a2y*a1z
      g(3,2)=2*pi*( a(1,3)*a(2,1)-a(2,3)*a(1,1) )/r1  !a1z*a2x-a2z*a1x
      g(3,3)=2*pi*( a(1,1)*a(2,2)-a(2,1)*a(1,2) )/r1  !a1x*a2y-a2x*a1y
      
      r1=abs(determinant(g))
      if(abs((2*pi)**3/r1-Nc).gt.eps) then
        if(myrank.eq.0) write(lud,*) 'This cell volume aint Nc',r1
        stop
      end if
      do i=1,3
      do j=1,3
        r1=sum(a(i,:)*g(j,:))/(2*pi)
        if(i.eq.j.and.abs(r1-oner).gt.eps) then
	  if(myrank.eq.0) write(lud,*) 'bug in g',i,j,r1
          stop
	end if
        if(i.ne.j.and.abs(r1).gt.eps) then
	  if(myrank.eq.0) write(lud,*) 'bug in g',i,j,r1
          stop
	end if
      end do
      end do
      
      if(iprint.ge.5.and.myrank.eq.0) 
     &  write(lud,"(' g=',3(' (',f5.2,1x,f5.2,1x,f5.2,')'))") 
     &  (g(l,:),l=1,3)
     
     
c***********************************************************************        
c     Create the cluster momentum tables Kc(i,ic), i=x,y,z
c***********************************************************************        

c     First find the points in the irreducible wedge.  They will be
c     indexed from 1 to Ncw
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '  Wedge K points  '    !0
        write(lud,*) 'ic  i  j  k  Kcx(ic)  Kcy(ic)  Kcz(ic)  ' !0
      end if
      ic=0; Ncw=0; icK0=0; icKpi=0
      do i=-Nc,Nc    !times a1
      do j=-Nc,Nc    !times a2
      do k=-Nc,Nc    !times a3
        iv1=(/i,j,k/)
        do l=1,3         !x,y,z components 
          rv1(l)=sum(iv1(:)*g(:,l))
        end do
        if(rv1(1).gt.-pi+eps.and.rv1(1).lt.pi+eps.and.
     &     rv1(2).gt.-pi+eps.and.rv1(2).lt.pi+eps.and.
     &     rv1(3).gt.-pi+eps.and.rv1(3).lt.pi+eps) then
c          The point is in the first SC Brillouin zone.  Now see if
c          the point falls within the IRW.  First see if it is 
c          equivalent to any known point in the IRW (if so it cannot be in IRW)
          do igroup=2,ngroup ! exclude the identity
            rv2=transform3D(rv1,group(igroup),ixy,ixz,iyz,isx,isy,isz)
            itest=0
            do jc=1,Ncw
              if(sum(abs(Kc(:,jc)-rv2(:))).lt.eps) then ! rv2 = Kc(jc)
                 itest=1
                 exit  ! leave the jc loop
              end if
            end do
            if(itest.eq.1) exit !leave the igroup loop
          end do
          if(itest.eq.0) then ! rv1 is in 1BZ and not equivalent to any point  
             Ncw=Ncw+1        ! in the known IRW.  It must be in the IRW
             ic=ic+1
             Kc(:,ic)=rv1(:)
             ig(ic)=i
             jg(ic)=j
             kg(ic)=k
             if(sum(abs(rv1-pi)).lt.eps) icKpi=ic
             if(sum(abs(rv1-zeror)).lt.eps) icK0=ic
             if(iprint.ge.5.and.myrank.eq.0) 
     &            write(lud,5) ic,i,j,k,Kc(1,ic),Kc(2,ic),Kc(3,ic) !0
          end if
        end if 
      end do 
      end do 
      end do 
        
      if(icK0.eq.0) then  ! we failed to find (0,0)
        if(myrank.eq.0) 
     &    write(lud,*) 'failed to find (0,0) in the IRW'    !0
        stop
      else if(icKpi.eq.0) then  ! we failed to find (pi,pi)
        icKpi=icK0
        if(myrank.eq.0) then
          write(lud,*) '*****************************************'    !0
          write(lud,*) '* failed to find (pi,pi,pi) in the IRW  *'    !0
          write(lud,*) '* apparently the lattice isnt bipartide *'    !0
          write(lud,*) '* icKpi=icK0 so that the code will run  *'    !0
          write(lud,*) '*****************************************'    !0
        end if
      end if

c     Now find the points in the 1BZ but outside the irreducible wedge.  
c     They will be indexed from Ncw+1 to Nc
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) ' Non Wedge K points '    !0
      end if
      do i=-Nc,Nc              
      do j=-Nc,Nc
      do k=-Nc,Nc
         iv1=(/i,j,k/)
         do l=1,3         !x,y,z components 
            rv1(l)=sum(iv1(:)*g(:,l)) ! K point in xyz
         end do
         if(rv1(1).gt.-pi+eps.and.rv1(1).lt.pi+eps.and.
     &      rv1(2).gt.-pi+eps.and.rv1(2).lt.pi+eps.and.
     &      rv1(3).gt.-pi+eps.and.rv1(3).lt.pi+eps) then
c           The point is in the first SC Brillouin zone.  Now see if
c           the point falls within the IRW.  
            itest=0
            do jc=1,Ncw
               if(sqrt(sum((Kc(:,jc)-rv1(:))**2)).lt.eps) then
c                 rv1 and Kc(ic) are the same point
                  itest=1
                  exit
               end if
            end do
            if(itest.eq.0) then !rv1 is in 1BZ and not in the IRW.
               ic=ic+1
               Kc(:,ic)=rv1(:)
               ig(ic)=i
               jg(ic)=j
               kg(ic)=k
               if(iprint.ge.5.and.myrank.eq.0) 
     &              write(lud,5) ic,i,j,k,Kc(1,ic),Kc(2,ic),Kc(3,ic) !0
            end if
         end if 
      end do 
      end do 
      end do 
        
 5    format(1x,i2,1x,i2,1x,i2,1x,i2,2x,f6.3,3x,f6.3,3x,f6.3)
      
      ntw=nl*Ncw      
      
      if(ic.ne.Nc) then
         if(myrank.eq.0) write(lud,*) 'Bug in Kc(ic) loop' !0
         stop
      end if
      if(iprint.ge.5.and.myrank.eq.0) 
     &  write(lud,"('icK0=',i2,' icKpi=',i2,' Ncw=',i2,' Nc=',i2)") 
     &  icK0,icKpi,Ncw,ic


c***********************************************************************
c     Create the indirect addressing table ict(i,j,k) which maps a 
c     point indexed by the principle translations in k-space to ic.
c***********************************************************************        
      N_sp=sqrt(float(Nc))
      do i=-N_sp,N_sp
      do j=-N_sp,N_sp
      do k=-N_sp,N_sp
         iv1=(/i,j,k/)
         do l=1,3         !x,y,z components 
            rv1(l)=sum(iv1(:)*g(:,l)) ! K point in xyz
         end do
         do it=1,5
           do l=1,3
              if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
              if(rv1(l).gt.pi+eps)  rv1(l)=rv1(l)-twor*pi
           end do
         end do
         itest=0
         do jc=1,Nc
            if(sum(abs(rv1(:)-Kc(:,jc))).lt.eps) then    
               ict(i,j,k)=jc
               itest=1
            end if
         end do
         if(itest.eq.0) then
            write(lud,*) 'ict failed for i,j,k=',i,j,k
            stop
         end if
      end do
      end do
      end do

c***********************************************************************        
c     Create the table ickdiff(ic1,ic2), which takes the difference
c     between to K-points
c***********************************************************************        
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                           !0
        write(lud,*) 'ic1   ic2  ickdiff(ic1,ic2)  ' !0
      end if
      do ic1=1,Nc
      do ic2=1,Nc
         rv1=Kc(:,ic1)-Kc(:,ic2)
         do it=1,3           ! move the difference into the 1BZ
           do l=1,3
              if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
              if(rv1(l).gt.pi+eps)  rv1(l)=rv1(l)-twor*pi
           end do
         end do
c        Now we need to figure out where this point is!
         itest=0
         do ic=1,Nc
            if(sum(abs(rv1(:)-Kc(:,ic))).lt.eps) then    
               ickdiff(ic1,ic2)=ic
               itest=1
            end if
         end do
         if(itest.eq.0) then
            write(lud,5) 'ickdiff failed for ic1,ic2=',ic1,ic2
            stop
         end if
         if(iprint.ge.5.and.myrank.eq.0) 
     &        write(lud,"(1x,i2,4x,i2,6x,i2)") ic1,ic2,ickdiff(ic1,ic2) !0
      end do
      end do
          
c***********************************************************************        
c     Create the table ickplus(ic1,ic2), which takes the sum
c     between to K-points
c***********************************************************************        
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                           !0
        write(lud,*) 'ic1   ic2  ickplus(ic1,ic2)  ' !0
      end if
      do ic1=1,Nc
      do ic2=1,Nc
         rv1=Kc(:,ic1)+Kc(:,ic2)
         do it=1,3           ! move the difference into the 1BZ
           do l=1,3
              if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
              if(rv1(l).gt.pi+eps)  rv1(l)=rv1(l)-twor*pi
           end do
         end do
c        Now we need to figure out where this point is!
         itest=0
         do ic=1,Nc
            if(sum(abs(rv1(:)-Kc(:,ic))).lt.eps) then    
               ickplus(ic1,ic2)=ic
               itest=1
            end if
         end do
         if(itest.eq.0) then
            write(lud,5) 'ickplus failed for ic1,ic2=',ic1,ic2
            stop
         end if
         if(iprint.ge.5.and.myrank.eq.0) 
     &        write(lud,"(1x,i2,4x,i2,6x,i2)") ic1,ic2,ickplus(ic1,ic2) !0
      end do
      end do

c***********************************************************************
c     Now find the points equivalent to ic, ickequ(ic,igroup) where 
c     igroup is the group operation.
c***********************************************************************        
c
c
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                            !0
        write(lud,*) 'equivalency in k'               !0
        write(lud,*) ' ic  igroup  ickequ(ic,igroup)' !0
      end if
      do ic=1,Nc
         i=ig(ic)
         j=jg(ic)
         k=kg(ic)
         iv1=(/i,j,k/)
         do l=1,3
           rv2(l)=sum(iv1(:)*g(:,l))
         end do
c        Kc(ic) falls within the zone, by construction, now look 
c        for equivalent points
c     
         do igroup=1,ngroup
            rv1=transform3D(rv2,group(igroup),ixy,ixz,iyz,isx,isy,isz) 
c           These are the equivalent points.  Now map to 1st BZ
            do it=1,3           ! move the difference into the 1BZ
              do l=1,3
                 if(rv1(l).lt.-pi+eps) rv1(l)=rv1(l)+twor*pi
                 if(rv1(l).gt.+pi+eps) rv1(l)=rv1(l)-twor*pi
              end do
            end do
c           Now we need to figure out where this point is.
            itest=0
            do jc=1,Nc
               if(sqrt(sum((rv1(:)-Kc(:,jc))**2)).lt.eps) then
                  ickequ(ic,igroup)=jc
                  itest=1
               end if
            end do
            if(iprint.ge.5.and.myrank.eq.0) 
     &        write(lud,"(2x,i2,4x,i2,6x,i2)") 
     &        ic,igroup,ickequ(ic,igroup)                  !0
            if(itest.eq.0) then
               if(myrank.eq.0) then
                 write(lud,*)'no equivalent K-point found' !0
                 write(lud,"('ic,igroup,group(igroup)',3(2x,i2))")
     &             ic,igroup,group(igroup)         !0
                 write(lud,"('K =',3(f6.3,1x))") rv2        !0
                 write(lud,"('K1=',3(f6.3,1x))") rv1       !0
               end if
               stop
            end if
         end do
      end do
        



c***********************************************************************        
c       Now generate a table which maps any K to an equivalent point
c       in the irreducible wedge (IW), and a table for the degeneracy 
c       of each point in the IW. 
c 
c       The use of the wedge can greatly reduce the storage required.
c       for example for Nc=8, only 4 of the 8 points are in the irreducible 
c       wedge.  Thus, all storage arrays may be reduced in size by a 
c       factor of 2.  As Nc-->oo, the saving factor approaches 48 for
c       clusters of high symmetry.  
c
c***********************************************************************        
c       
      
      ickdeg=0; irwgroup=0
      if(iprint.ge.5.and.myrank.eq.0) then
        write(lud,*) '   '                            !0
        write(lud,*) ' degeneracy in k'              !0
        write(lud,*) ' ic  ickmap(ic)  ickdeg(ic) irwgroup(ic)' !0
      end if
c     scan over all of the K points in the 1BZ
      do ic=1,Nc
         i=ig(ic)
         j=jg(ic)
         k=kg(ic)
         iv1=(/i,j,k/)
         do l=1,3
            rv2(l)=sum(iv1(:)*g(:,l))
         end do
c        Scan over all allowed point group operations, to find the
c        one that maps ic into the IRW
         itest=0
         groupops: do igroup=1,ngroup
           rv1=transform3D(rv2,group(igroup),ixy,ixz,iyz,isx,isy,isz) 
c          These are the equivalent points.  Now map to 1st BZ
           do it=1,4              ! move the difference into the 1BZ
             do l=1,3
                if(rv1(l).lt.-pi+eps)  rv1(l)=rv1(l)+twor*pi
                if(rv1(l).gt.+pi+eps)  rv1(l)=rv1(l)-twor*pi
             end do
           end do
c          See if this point is in the IRW!
           do jc=1,Ncw
              if(sum(abs(rv1(:)-Kc(:,jc))).lt.eps) then
                 ickmap(ic)=jc
                 ickdeg(jc)=ickdeg(jc)+1
                 irwgroup(ic)=igroup
                 itest=1
              end if
           end do
           if(itest.eq.1) exit    ! at first group op. that takes ic into IRW
         end do groupops
         if(itest.eq.0) then
            if(myrank.eq.0) write(lud,*) 'mapping failed for' !0
            if(myrank.eq.0) write(lud,*) 'ic',ic !0
            if(myrank.eq.0) write(lud,"('K=',3(f6.3,1x))") rv2 !0
            stop
         end if
         
      end do !ic

      if(iprint.ge.5.and.myrank.eq.0) then 
        do ic=1,Nc    
          write(lud,"(2x,i2,6x,i2,9x,i2,9x,i2)") 
     &    ic,ickmap(ic),ickdeg(ic),irwgroup(ic)
        end do
      end if

c     Test the degeneracy table.  Since ickdeg holds the degeneracy
c     of points in the IRW, its sum must equal Nc.
      if(sum(ickdeg(1:Ncw)).ne.Nc) then
         if(myrank.eq.0) then
           write(lud,*) 'bug in ickdeg',sum(ickdeg(1:Ncw))
           write(lud,*) 'ickdeg=',ickdeg(1:Ncw)
         end if
         stop
      end if
        
c     Test the group table.  irwgroup(ic within wedge)=1, the identity
      do ic=1,Ncw
         if(irwgroup(ic).ne.1) then
            if(myrank.eq.0) then
              write(lud,*) 'irwgroup failed for wedge pt.'     !0
              write(lud,*) 'ic',ic,'irwgroup(ic)',irwgroup(ic) !0
              write(lud,"('K=(',3(f6.3,1x),')')") Kc(:,ic)      !0
            end if
            stop
         end if
      end do

c***********************************************************************
c     generate the kt-points in the Brillouin-zone of the 
c     super-lattice, i.e. in the Wigner-Seitz cell of the  
c     reciprocal space defined by the cluster K-points
c***********************************************************************
     
c      Dkt=halfr*sqrt(oner*Nc)/real(ntot,kind); rv2=zeror; rv1=zeror
      Dkt=oner/real(ntot,kind); rv2=zeror; rv1=zeror
      nshell=1 ! must be adjusted so that enough G are tested.
      nwsc=0   ! number of k-points in Wigner-Seitz cell 
      do i=-ntot+1,ntot
      do j=-ntot+1,ntot
      do k=-ntot+1,ntot
          iv1=(/i,j,k/)
          rv1(:)=Dkt*(iv1(:)-halfr)*pi
c         Check if this kt-point is outside the Wigner-Seitz cell if
c
c         (kt.G)/(G.G) > 1/2  for any 
c
c         G = n g1 + m g2 with n,m not both zero
c
          itest=0
          do n=-nshell,nshell
          do m=-nshell,nshell
          do p=-nshell,nshell
            if(m.eq.0.and.n.eq.0.and.p.eq.0) then
c             do nothing
            else
              iv2=(/n,m,p/)
              do l=1,ndim
                 rv2(l)=sum(iv2(1:ndim)*g(1:ndim,l))  ! G
              end do
              if(dot_product(rv1,rv2)/dot_product(rv2,rv2).gt.halfr)then
c               rv1=kt is outside the WS cell defined by G=rv2
                itest=1
                exit
              end if
            end if
          end do !m
            if(itest.eq.1) exit
          end do !n
            if(itest.eq.1) exit
          end do !p
          if(itest.eq.0) then 
             nwsc=nwsc+1
             kt(:,nwsc) = rv1(:)
          endif
       enddo ! k
       enddo ! j
       enddo ! i
c      
c      test the WS cell
       if(abs(Nc*nwsc-(twor*ntot)**3)/(twor*ntot)**3.gt.eps) then         
          write(lud,*) 'Nc*Number of points in WS, nwsc', Nc*nwsc
          write(lud,*) 'Total number of points in BZ', (2*ntot)**3 
c          stop
       end if
       


c***********************************************************************
c     Form Epsbar(K)
c***********************************************************************        

      if(myrank.eq.0.and.iprint.ge.5) write(lud,*) '    '
      if(myrank.eq.0.and.iprint.ge.5) write(lud,*) 
     &               'ic   Epsbar(ic)'

      Epsbar=zeror
      do ic=1,Nc
         do i=1,nwsc
            rv1(:) = Kc(:,ic)+kt(:,i)
            Epsbar(ic)=Epsbar(ic) - halfr*sum(cos(rv1))
     &                - tprime*(cos(rv1(1))*cos(rv1(2))
     &                         +cos(rv1(1))*cos(rv1(3))
     &                         +cos(rv1(2))*cos(rv1(3)))
         end do
         Epsbar(ic)=Epsbar(ic)/real(nwsc,kind)
      if(myrank.eq.0.and.iprint.ge.5) write(lud,"(i2,6x,f9.6)") 
     &           ic,Epsbar(ic)
      end do
c     Check 
      r1=sum(Epsbar); r2=sum(Epsbar*ickdeg)
      if(abs(tprime).lt.eps.and.abs(r1).gt.eps
     &   .or.
     &	 abs(tprime).lt.eps.and.abs(r2).gt.eps) then
        if(myrank.eq.0) then
          write(lud,*) 'sum Epsbar wrong.',r1,r2
	end if
        stop
      end if

        
c***********************************************************************        
c       Form the Fourier transform tables.  
c       R-->K,  FTCoefs_K_to_R
c       K-->R,  FTCoefs_K_to_R
c
c                    __          the factor of 1/N is required
c                 1  \           so that both G(R)~1/w and G(K)~1/w
c       G(R=0) = --- /  G(K)     for large w
c                N   --
c                    K
c
c***********************************************************************        
      do ic=1,Nc
         do ik=1,Nc
            r1=zeror
            do idim=1,ndim
               r1=r1+Kc(idim,ik)*Rc(idim,ic)
            end do
            FTCoefs_K_to_R(ik,ic)=exp(-ii*r1)/real(Nc,kind)
            FTCoefs_R_to_K(ik,ic)=exp(+ii*r1)
         end do
      end do

c     Now test for consistency of Kc and Rc
      do ic=1,Nc
        r1= real(sum(FTCoefs_K_to_R(1:Nc,ic)))-kroniker(ic,icR0)
        r2=aimag(sum(FTCoefs_K_to_R(1:Nc,ic)))
        r3= real(sum(FTCoefs_K_to_R(ic,1:Nc)))-kroniker(ic,icK0)
        r4=aimag(sum(FTCoefs_K_to_R(ic,1:Nc)))
        
        if(abs(r1).gt.eps.or.abs(r2).gt.eps) then
          write(lud,*) 'Kc and Rc are inconsistent'
          write(lud,*) 'icr,r1,r2=',ic,r1,r2
          stop
        end if
        if(abs(r3).gt.eps.or.abs(r4).gt.eps) then
          write(lud,*) 'Kc and Rc are inconsistent, K=0'
          write(lud,*) 'ick,r3,r4=',ic,r3,r4
          stop
        end if
      end do


c**********************************************************************
c     TABLES FOR THE TWO-PARTICLE CALCULATIONS
c**********************************************************************
      if(isg.eq.1) then
         
         if(myrank.eq.0) write(lud,*) 'Allocating 2-particle arrays' !0
c        call allocate_2p       ! Needs Ncw from above
        
c**********************************************************************        
c        For two-particle propagators Gamma(K,K'), the group operations
c        may only be applied similtaneously to all K.  Thus, if U_K(K) is 
c        the operation that maps K to the IW, then we need to define the
c        table U_K(K') for all K.  This is done by the table
c        ick2map(a,b) : takes indices a, b to k-points K(a), K'(b) and 
c        returns the index c of a k-point Q(c), which index c is obtained
c        by performing on b whatever transformation would be necessary to
c        perform on a to send K(a) into the irreducible wedge.
c**********************************************************************        

c        First consider the K already in the irreducible wedge, where
c        U_K(K')=K'
         do ic=1,Ncw
c           In IRW, only the identity group operation leaves K in the IRW
            do jc=1,Nc
               ick2map(ic,jc)=jc 
            end do
         end do
c        Now consider the K that fall outside of the IW
         do ic=Ncw+1,Nc
           igroup=irwgroup(ic) ! this group operation maps K(ic) into IRW
           do jc=1,Nc
             i2=ig(jc)
             j2=jg(jc)
             k2=kg(jc)
             iv1(:)=(/i2,j2,k2/)
c            calculate the corresponding K       
             do l=1,3
                rv1(l)=sum(iv1(:)*g(:,l))
             end do
c            now apply the sampe map to jc that takes K(ic) to the IRW.    
             rv2=transform3D(rv1,group(igroup),ixy,ixz,iyz,isx,isy,isz) 
c            Now we have to figure out what the point rv2 is. First,
c            map to 1st BZ
             do it=1,3        ! 
               do l=1,3
                  if(rv2(l).lt.-pi+eps) rv2(l)=rv2(l)+twor*pi
                  if(rv2(l).gt.pi+eps)  rv2(l)=rv2(l)-twor*pi
               end do
             end do
c            Find the jk that corresponds to rv1=K     
             itest=0    
             do jk=1,Nc
                if(sum(abs(rv2(:)-Kc(:,jk))).lt.eps) then
                   ick2map(ic,jc) = jk
                   itest=1
                end if
             end do
             if(itest.eq.0) then
                if(myrank.eq.0) then
                  write(lud,*) 'ick2map mapping failed for' !0
                  write(lud,*) 'ic',ic,'jc',jc !0
                end if
                stop
             end if
           end do
         end do

c        Perform a test on both ick2map and ickmap
         do ic=1,Nc
            if(ick2map(ic,ic).ne.ickmap(ic).and.myrank.eq.0) then
               write(lud,*) 'problem with ick2map or ickmap' !0
               write(lud,*) 'ic=',ic,' ickmap(ic)=',ickmap(ic) !0
               write(lud,*) 'ick2map(ic,ic)=',ick2map(ic,ic) !0
            end if
         end do
          
         if(iprint.ge.5.and.myrank.eq.0) then
            write(lud,*) ' ic  jc  ick2map(ic,jc)'                      !0
            do ic=1,Nc                                                  !0
               do jc=1,Nc                                               !0
                  write(lud,"(1x,i2,2x,i2,5x,i2)") ic,jc,ick2map(ic,jc) !0
               end do                                                   !0
            end do                                                      !0
         end if
        
c***********************************************************************        
c         define the  fmat matrix
c***********************************************************************        
c         do ix=1,Nc             !space label
c            do jk=1,Nc          !k-label
c               do n=1,nl        !time label
c                  do m=1,nl     !frequency label
c                     
c                     ind=ix + (n-1)*Nc ! space-time index
c                     jmd=jk + (m-1)*Nc ! momentum-frequency index
c                     jm=m-1-nl/2
c                     fmat(ind,jmd)=dt*
c     &                exp(-ii*pi*((twor*jm+oner)*(n-oner)/real(nl,kind))
c     &                    -ii*Kc(1,jk)*Rc(1,ix)
c     &                    -ii*Kc(2,jk)*Rc(2,ix)
c     &                    -ii*Kc(3,jk)*Rc(3,ix)  )
c                  end do
c               end do
c            end do
c         end do
         
      end if                    ! isg.eq.1


      stop
      end  
      
      
      
c***********************************************************************        
c***********************************************************************        
c         3D subroutines and functions
c***********************************************************************        
c***********************************************************************
      
      
      
        
      function gtransform3D(a,igroup,ixy,ixz,iyz,isx,isy,isz)
c
c  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
c  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
c  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
c
      implicit none
      integer :: i,a(3,3),g(3,3),gtransform3D(3,3),igroup,
     &           ixy(48),ixz(48),iyz(48),isx(48),isy(48),isz(48)
      
      do i=1,3  
         gtransform3D(i,1)=a(i,1)
         gtransform3D(i,2)=(1-iyz(igroup))*a(i,2) + iyz(igroup)*a(i,3)
         gtransform3D(i,3)=(1-iyz(igroup))*a(i,3) + iyz(igroup)*a(i,2)
      end do
               
      do i=1,3  
         g(i,1)= (1-ixz(igroup))*gtransform3D(i,1)
     &         + ixz(igroup)*gtransform3D(i,3)
         g(i,2)=                gtransform3D(i,2)
         g(i,3)= (1-ixz(igroup))*gtransform3D(i,3) 
     &         + ixz(igroup)*gtransform3D(i,1)
      end do

      do i=1,3  
         gtransform3D(i,1)=isx(igroup)*((1-ixy(igroup))*g(i,1) 
     &                  + ixy(igroup)*g(i,2))
         gtransform3D(i,2)=isy(igroup)*((1-ixy(igroup))*g(i,2) 
     &                  + ixy(igroup)*g(i,1))
         gtransform3D(i,3)=isz(igroup)*                 g(i,3)         
      end do
      
      return
      end
        
      function Rtransform3D(Rc,igroup,ixy,ixz,iyz,isx,isy,isz)
c
c  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
c  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
c  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
c
      implicit none
      integer :: i,Rc(3),g(3),Rtransform3D(3),igroup, 
     &           ixy(48),ixz(48),iyz(48),isx(48),isy(48),isz(48)
      
      Rtransform3D(1)=            Rc(1)
      Rtransform3D(2)=(1-iyz(igroup))*Rc(2) + iyz(igroup)*Rc(3)
      Rtransform3D(3)=(1-iyz(igroup))*Rc(3) + iyz(igroup)*Rc(2)
      
      g(1)=(1-ixz(igroup))*Rtransform3D(1) + ixz(igroup)*Rtransform3D(3)
      g(2)=                Rtransform3D(2)
      g(3)=(1-ixz(igroup))*Rtransform3D(3) + ixz(igroup)*Rtransform3D(1)
      
      Rtransform3D(1)=isx(igroup)*((1-ixy(igroup))*g(1) 
     &             + ixy(igroup)*g(2))
      Rtransform3D(2)=isy(igroup)*((1-ixy(igroup))*g(2) 
     &             + ixy(igroup)*g(1))
      Rtransform3D(3)=isz(igroup)*                 g(3) 
      
      return
      end
      
      function Ktransform3D(Kc,igroup,ixy,ixz,iyz,isx,isy,isz)
c     
c  x <-- isx*[(1-ixy)*x + ixy*y]  <-- [(1-ixz)*x + ixz*z]<----------------------
c  y <-- isy*[(1-ixy)*y + ixy*x]  <----------------------<-- [(1-iyz)*y + iyz*z]
c  z <-- isz*<------------------- <-- [(1-ixz)*z + ixz*x]<-- [(1-iyz)*z + iyz*y]
c
      implicit none
      integer :: i,igroup,
     &           ixy(48),ixz(48),iyz(48),isx(48),isy(48),isz(48)
      real(8) :: Kc(3),g(3),Ktransform3D(3)
      
      Ktransform3D(1)=                Kc(1)
      Ktransform3D(2)=(1-iyz(igroup))*Kc(2) + iyz(igroup)*Kc(3)
      Ktransform3D(3)=(1-iyz(igroup))*Kc(3) + iyz(igroup)*Kc(2)
      
      g(1)=(1-ixz(igroup))*Ktransform3D(1) + ixz(igroup)*Ktransform3D(3)
      g(2)=                Ktransform3D(2)
      g(3)=(1-ixz(igroup))*Ktransform3D(3) + ixz(igroup)*Ktransform3D(1)
      
      Ktransform3D(1)=isx(igroup)*((1-ixy(igroup))*g(1) 
     &             + ixy(igroup)*g(2))
      Ktransform3D(2)=isy(igroup)*((1-ixy(igroup))*g(2) 
     &             + ixy(igroup)*g(1))
      Ktransform3D(3)=isz(igroup)*                 g(3) 
      
      return
      end
      
      function Inverse_3x3(a)   
      integer :: a(3,3)
      real(8) :: Inverse_3x3(3,3),r1
c
c       [ a(1,1) a(1,2) a(1,3) ]
c     A=[ a(2,1) a(2,2) a(2,3) ]
c       [ a(3,1) a(3,2) a(3,3) ]
c
c      -1        i+j 
c     A_ji = (-1)   C_ij  /|A|    C_ij is the ij cofactor of A
c     
      
      r1=   a(1,1)*a(2,2)*a(3,3)  ! the determinent of a
     &    + a(1,2)*a(2,3)*a(3,1)
     &    + a(1,3)*a(2,1)*a(3,2)
     &    - a(3,1)*a(2,2)*a(1,3)
     &    - a(3,2)*a(2,3)*a(1,1)
     &    - a(3,3)*a(2,1)*a(1,2) 
      
      Inverse_3x3(1,1) = +( a(2,2)*a(3,3)-a(3,2)*a(2,3) )/r1
      Inverse_3x3(1,2) = -( a(1,2)*a(3,3)-a(3,2)*a(1,3) )/r1
      Inverse_3x3(1,3) = +( a(1,2)*a(2,3)-a(2,2)*a(1,3) )/r1
      
      Inverse_3x3(2,1) = -( a(2,1)*a(3,3)-a(3,1)*a(2,3) )/r1
      Inverse_3x3(2,2) = +( a(1,1)*a(3,3)-a(3,1)*a(1,3) )/r1
      Inverse_3x3(2,3) = -( a(1,1)*a(2,3)-a(2,1)*a(1,3) )/r1
      
      Inverse_3x3(3,1) = +( a(2,1)*a(3,2)-a(3,1)*a(2,2) )/r1
      Inverse_3x3(3,2) = -( a(1,1)*a(3,2)-a(3,1)*a(1,2) )/r1
      Inverse_3x3(3,3) = +( a(1,1)*a(2,2)-a(2,1)*a(1,2) )/r1
      
      
      return
      end
      
      function Ideterminant(a)
      integer::  a(3,3),Ideterminant

      Ideterminant=
     &         a(1,1)*a(2,2)*a(3,3)
     &       + a(1,2)*a(2,3)*a(3,1)
     &       + a(1,3)*a(2,1)*a(3,2)
     &       - a(3,1)*a(2,2)*a(1,3)
     &       - a(3,2)*a(2,3)*a(1,1)
     &       - a(3,3)*a(2,1)*a(1,2)
      return
      end
      
      function Rdeterminant(a)
      real(8)::  a(3,3),Rdeterminant

      Rdeterminant=
     &         a(1,1)*a(2,2)*a(3,3)
     &       + a(1,2)*a(2,3)*a(3,1)
     &       + a(1,3)*a(2,1)*a(3,2)
     &       - a(3,1)*a(2,2)*a(1,3)
     &       - a(3,2)*a(2,3)*a(1,1)
     &       - a(3,3)*a(2,1)*a(1,2)
      return
      end

      function kroniker(i,j)
      integer i,j,kroniker
      
      if(i.eq.j) then
        kroniker=1
      else
        kroniker=0   
      end if
      
      return
      end 
