c
c Interface to update routines and different update methods
c 
c To add a new update method, modify readin checks and satus
c messages, then add to the "switch"es in update,reg,upg below
c

      subroutine update(ntimes)
c
c Update the Hubbard-Statonovich fields ntimes using the method umode
c
      use Global
      implicit none
      integer :: ntimes
c Assume umode is valid: must be verified in readin
c We do no I/O to allow for potential inlining
      if (umode.eq.0) then
         call update_single(ntimes)           ! dateregupg
      else if (umode.eq.1) then
         call update_delayed(ntimes)          ! dateregupg_fast2
      else if (umode.eq.2) then
         call update_blocked_delayed(ntimes)  ! dateregupg_fast
      else if (umode.eq.3) then
         call update_blocked(ntimes)          ! dateregupg_block
      else if (umode.eq.4 .or. umode.eq.5) then
         call update_delayed_cuda(ntimes)           ! for now use regular update
         ! call update_delayed(ntimes)	
      end if
     	 if(myrank.eq.0.and.iprint.ge.9) then
		write(lud,*)'done update';flush(lud)
	endif	
      end subroutine update


      subroutine reg()
c Recalculate the green functions given only the  
c s-field configuration and the initial greens functions Gcstsc  
c (which corresponds to all s=0).
      use Global
      implicit none
      integer m,n,k,jc,ic, ind, imd,gind
      
      if (umode.eq.0) then
         call reg_single()
      else if (umode.eq.1) then
         call reg_delayed()
      else if (umode.eq.2) then
         call reg_blocked_delayed()
      else if (umode.eq.3) then
         call reg_blocked()
      else if (umode.eq.4.or.umode.eq.5) then
         call reg_delayed_cuda()           !use the cuda code
!	  call reg_delayed()
      end if
      
      end subroutine reg

      subroutine upg(lo,acg,gcg)
c
c Updates gd after a change in one of the field variables
c
      use Global
      implicit none
      integer :: lo
      real(kind) :: acg,gcg(nt,nt)
      if (umode.eq.0) then
         call upg_single(lo,acg,gcg)
      else if (umode.eq.2) then
         call upg_blocked_delayed(lo,acg,gcg)
      else if (umode.eq.3) then
         call upg_blocked(lo,acg,gcg)
      end if
      end subroutine upg
      

