extern "C" void launch_update_blocked(Kernel_Params*);


__forceinline__
 __device__ 
void _factorize_no_pivoting(int global_idx, 
                        const int block_size,
                        TYPE Lu[BLOCK][BLOCK], 
                        TYPE Mu[BLOCK][BLOCK], 
                        TYPE tmp[BLOCK], 
                        TYPE& determinant_u) {
//  if (global_idx < 32) {
    determinant_u = 1;
    if (global_idx < block_size) {
      for (int step = 0; step < block_size; step++) {
        // Find the max possible pivot and reorder the rows accordingly (Partial pivoting).
        const int src_row = step;
        // Where the pivot row is (row with the maximum element for pivoting)i
/*        int pivot_row_up   = step;
        for (int search_row = step + 1; search_row < block_size; search_row++) {
          if (Mu[search_row][step] > Mu[pivot_row_up][step]) {
           pivot_row_up = search_row;
          }
        }
        //__syncthreads();
        // Now actually interchange the rows:
        if (pivot_row_up != src_row) {
          // Exchange the M rows
          tmp[global_idx] = Mu[src_row][global_idx];
          Mu[src_row][global_idx]  = Mu[pivot_row_up][global_idx];
          Mu[pivot_row_up][global_idx] = tmp[global_idx];
          // Now exchange the L rows.
          if (global_idx < step - 1) { // Dont flip any thing eyond the diagonal.
            tmp[global_idx] = Lu[src_row][global_idx];
            Lu[src_row][global_idx]  = Lu[pivot_row_up][global_idx];
            Lu[pivot_row_up][global_idx] = tmp[global_idx];
          }
          determinant_u *= -1;
	}
*/
        //  Done reordering rows, print M to trace correctness
        for (int working_row = step + 1; working_row < block_size; working_row++) {
	  // workig row is the row we are 
          const TYPE ratio = Mu[working_row][step] / Mu[step][step];
          // Now update the working row
          Mu[working_row][global_idx] += -1*ratio*Mu[step][global_idx];
          // Now update the lower diagonal Matrix
          // Note that I am avoiding updaing the elements that should stay zeros.
          // this will affect the sign for very small numbers.
          if (step==global_idx) {
            Lu[working_row][step] = 1*ratio*Lu[step][step];
          }
         //__syncthreads();
         }
          determinant_u *= Mu[step][step];        
      }
    }
//  }
}
__device__
__forceinline__
void build_a_MC(TYPE  M_inv[BLOCK][BLOCK], 
         int ind_start,
         TYPE* a,  
         TYPE* G, 
         TYPE A_Tile[5][256 + 1], 
         int block_size,
         int ISIZE, 
         int Nc,
         int global_idx) {


  const int thread_col = threadIdx.x;
  const int thread_row = threadIdx.y;

  if(threadIdx.y ==0) {
    for (int col = threadIdx.x; col < ISIZE ; col += blockDim.x) {
      for (int row = threadIdx.y; row < block_size; row += 1) {
        TYPE val = 0; 
        const int a_idx = FlattenIdx(row, col, ISIZE, block_size);
        //printf ("Now computing (%i,%i)\n", row , col);
        for (int i = 0; i < block_size; i++) {
          val += M_inv[row][i]*GetG(G, (i*Nc + ind_start)%(ISIZE), col, ISIZE);
        }
        a[a_idx] = val;
      }
    }
  }
}

__device__
__forceinline__
void build_a_MC_1_block(TYPE r_inv, 
         int ind_start,
         TYPE* a,  
         TYPE* G, 
         TYPE A_Tile[5][256 + 1], 
         int block_size,
         int ISIZE, 
         int Nc,
         int global_idx) {


  const int thread_col = threadIdx.x;
  const int thread_row = threadIdx.y;

  if(threadIdx.y ==0) {
    for (int col = threadIdx.x; col < ISIZE ; col += blockDim.x) {
        TYPE val = 0; 
        const int a_idx = FlattenIdx(0, col, ISIZE, block_size);
        a[a_idx] =  (r_inv)*GetG(G, ind_start, col, ISIZE);
        //a[a_idx] =  M_inv[row][i]*GetG(G, i*Nc + ind_start, col, ISIZE);
    }
  }
}
__forceinline__
__device__
void print_s(int* ps, int Nc, int L) {
  for (int idx =0; idx < Nc*L; idx++) {
   printf("%i ", ps[idx]); 
  }
  printf("\n");
}

__forceinline__
__device__  
void build_M( int ind_start,
                     int block_size,
                     int Nc,
                     int L,
                     int* pS,
                     TYPE* Gu,
                     TYPE* Gd,
                     TYPE* d_exp,
                     TYPE Mu[BLOCK][BLOCK],
                     TYPE Md[BLOCK][BLOCK],
                     TYPE Lu[BLOCK][BLOCK],
                     TYPE Ld[BLOCK][BLOCK], 
                     TYPE clean_block_u[BLOCK][BLOCK],
                     TYPE clean_block_d[BLOCK][BLOCK]) {

    const int x = ((threadIdx.x)*Nc + ind_start)%(Nc*L);

    for (int col = 0; col < block_size; col ++)
    if (threadIdx.x < block_size && threadIdx.y < block_size) {
      const int col_idx =  ind_start + col*Nc;

      Mu[threadIdx.x][col] = -1*GetG(Gu, x, col_idx, Nc*L)*d_exp[-pS[x]];
      Md[threadIdx.x][col] = -1*GetG(Gd, x, col_idx, Nc*L)*d_exp[pS[x]];

      clean_block_u[threadIdx.x][col] = GetG(Gu, x, col_idx, Nc*L);
      clean_block_d[threadIdx.x][col] = GetG(Gd, x, col_idx, Nc*L);

      // For the diagonal elements (in the green function):
      if (x == col_idx) {
        Mu[threadIdx.x][col] += 1 + d_exp[-pS[x]];
        Md[threadIdx.x][col] += 1 + d_exp[pS[x]];
      }
    }

    for (int col = 0; col < block_size; col++)
    if (threadIdx.x < block_size && threadIdx.y < block_size) {
      if (threadIdx.x == col) {
        // Diagonal elements are ones.
        Lu[col][threadIdx.x] = 1;
        Ld[col][threadIdx.x] = 1;
      } else {
        // Off-diagonal elements
        Lu[col][threadIdx.x] = 0;
        Ld[col][threadIdx.x] = 0;
      }
    }
  
}

__forceinline__ 
__device__
void print_info(int ind_start, int nsweeps, int nTimes, int block_size) {
   //printf("block_size = %i\n", block_size);  
   printf("number of sweeps:%i\nind_start:%i\nnTimes=%i\nblock_size=%i\n", nsweeps, ind_start, nTimes, block_size);
}


// Generate new set of block sizes.
__forceinline__
__device__
void gen_block_size(int* block_sizes, curandState_t* cu_state, int max_block_size) {
    //block_sizes[threadIdx.x]  = (10*getRadomNumber(&cu_state[threadIdx.x])) % max_block_size + 1;
    float z = getRadomNumber(&cu_state[threadIdx.x]);
    block_sizes[threadIdx.x]  = ((int)(10*z)%(max_block_size)) + 1;
}

// get a new block size for the current sweep.
__forceinline__
__device__
TYPE get_block_size(int sweep, int* block_sizes) {
    return block_sizes[sweep];
}

__forceinline__
__device__
bool accept(TYPE det_u, TYPE det_d, curandState_t* cu_state) {
  float z = getRadomNumber(&(cu_state[threadIdx.x]));
  return (abs(det_u*det_d) > z);
}

__forceinline__
__device__
void  update_ind_block_size(int sweep_ind_start, int& ind_start,int& nsweeps, int Nc, int L,int& block_size, int* block_sizes) {

 int new_ind_start = ind_start + block_size*Nc;
 block_size        = get_block_size(nsweeps%(Nc*L), block_sizes);
 
 if (new_ind_start == sweep_ind_start) { // Completed a full wrap 
   // New sweep, get a new bock size.
   nsweeps++;
   block_size = get_block_size(nsweeps%(Nc*L), block_sizes);
   if (threadIdx.x ==0) {
      printf("Switching to block_size = %i at sweep = %i\n", block_size, nsweeps);
    }
  } else if (new_ind_start > Nc*L) {
      new_ind_start -= Nc*L;
  } else if (new_ind_start + block_size*Nc > Nc*L) { // block will over flow
      block_size = 1 + (Nc*L - ind_start)/Nc; // The ind we are at plus whatever we can get.
  } else if (new_ind_start == Nc*L) {
      new_ind_start = 0;  
  } else if ((new_ind_start + block_size*Nc) > sweep_ind_start  && new_ind_start < sweep_ind_start) { 
    // New sweep, get a new bock size.
    nsweeps++;
    block_size = (new_ind_start - sweep_ind_start)/Nc;
  }

  ind_start = new_ind_start;

  if (0 == threadIdx.x) {
    printf ("Updated ind_start to %i\n, block_size to %i\n", ind_start, block_size);
  }

}

__forceinline__
__device__
bool update_spins(int *pS, int Nc, int block_size, int ind_start) { 
 if (threadIdx.x < block_size) {
    pS[ind_start + threadIdx.x*Nc] *= -1;
  }
}


__forceinline__
__device__
void scale_M_exp(TYPE Mu_inv[BLOCK][BLOCK],TYPE  Md_inv[BLOCK][BLOCK], int block_size, int ind_start,int Nc,int L, TYPE* d_exp,int* pS) {
  for (int col = 0 ; col < block_size ; col++) {
    if (threadIdx.x < block_size && threadIdx.y < block_size) {
      Mu_inv[threadIdx.x][col] *= d_exp[-pS[(ind_start + threadIdx.x*Nc)%(Nc*L)]];
      Md_inv[threadIdx.x][col] *= d_exp[ pS[(ind_start + threadIdx.x*Nc)%(Nc*L)]];
    }
  }
} 

__forceinline__
__device__
void build_b(int ind_start,TYPE* bu,TYPE* bd,TYPE* Gu,TYPE* Gd,int Nc,int L, int num_threads, int block_size, int global_idx) {
  for (int col = 0; col < block_size; col++) {
    for (int row = global_idx; row < Nc*L; row += num_threads) {
      int g_col  = ind_start + col*Nc;
      int b_idx  = FlattenIdx (row, col, block_size, Nc*L);
      bu[b_idx]  = GetG (Gu, row, g_col%Nc*L, Nc*L);
      bd[b_idx]  = GetG (Gd, row, g_col%Nc*L, Nc*L);
      // TODO: Check with conrad, why are we doing this !!!
      // This is correct; B = (G - I)[ISIZE, Nblock] so we must 
      // subtract 1 from the diagonal elements !
      if (g_col == row) {
        bu[b_idx] -=1;
        bd[b_idx] -=1;
      }
    }
  }
}


__forceinline__
__device__
void print_rejection() {
  if (threadIdx.x ==0) {
    printf ("Rejection....\n");
  }
}
__global__ void update_blocked(int* d_ind, 
                               TYPE* Gu, 
                               TYPE* Gd, 
                               int* pS, 
                               int* block_sizes, 
                               TYPE* au, 
                               TYPE* bu,
                               TYPE* ad,
                               TYPE* bd,
                               curandState_t* cu_state,
        	               TYPE* d_expR_global_MC,
                               TYPE* d_diagu, 
                               TYPE* d_diagd,
                               TYPE alpha, 
                               int Nc, 
                               int L,
                               int max_block_size) {
  // block size is same as number of delays
  // int ind;
  // const int block_size = 3;
   //printf("Hello from thread %i\n", threadIdx.x);
   const int c_max_block_size = 4;
  __shared__ TYPE Mu[BLOCK][BLOCK];
  __shared__ TYPE Md[BLOCK][BLOCK];
  __shared__ TYPE Mu_inv[BLOCK][BLOCK];
  __shared__ TYPE Md_inv[BLOCK][BLOCK];


  //__shared__ TYPE exp_diagonal[c_max_block_size];

  __shared__ TYPE tmp[BLOCK];
  // The lower triangular matrices.
  __shared__ TYPE Lu[BLOCK][BLOCK];
  __shared__ TYPE Ld[BLOCK][BLOCK];
  __shared__ TYPE    Z[BLOCK][BLOCK];
  __shared__ TYPE determinant_d , determinant_u;


 // Tiles for Multiplication: 
  __shared__ TYPE Au_Tile[c_max_block_size][256+1];
  __shared__ TYPE Ad_Tile[c_max_block_size][256+1];

  __shared__ TYPE clean_block_u[BLOCK][BLOCK];
  __shared__ TYPE clean_block_d[BLOCK][BLOCK];

  
  bool move_accepted = false;
  int acceptances = d_ind[2];
  int rejections  = d_ind[3];

  int ind_start   = d_ind[0]; // Last used Ind_start.
  int nsweeps     = d_ind[1];
  // d_ind[2] = acceptances;
  // d_ind[3] = rejections;
  int sgn             = d_ind[4];
  int nTimes          = d_ind[5];
  int sweep_ind_start = d_ind[6]; // Which ind index marks a begning of a new sweep.
  int block_size      = d_ind[7];
  bool acceptence     = true;

  const int global_idx  = threadIdx.x; //threadIdx.y * blockDim.x + threadIdx.x;
  const int num_threads = blockDim.x * blockDim.y;

  TYPE* d_exp     = d_expR_global_MC + 1;

  if (block_size == -1) { // First Time running, send_mc sets the block size to 1. 
    gen_block_size(block_sizes, cu_state, max_block_size);
    block_size = get_block_size(nsweeps%(Nc*L), block_sizes);
  } else {
    // Not the first time. Coming back from an acceptance, update the ind_start and block size. 
    update_ind_block_size(sweep_ind_start, ind_start, nsweeps, Nc, L,block_size, block_sizes);
  } 


  if (global_idx ==0) {
    print_info(ind_start, nsweeps, nTimes, block_size);
  //printf("block_size = %i\n", block_size);
  //printf("From Update gu = \n");
  //print_g(global_idx, Gu, Nc*L);
  //printf("From Update gd = \n");
  //print_g(global_idx, Gd, Nc*L);
  }

  __syncthreads();
  
  for(;nsweeps < nTimes;) {

    if (block_size != 1) {
      build_M(ind_start, block_size, Nc, L, pS, Gu, Gd, d_exp, Mu, Md, Lu, Ld, clean_block_u, clean_block_d);
      __syncthreads();
      _factorize_no_pivoting(global_idx, block_size, Lu,  Mu, tmp, determinant_u);
      __syncthreads();
      _factorize_no_pivoting(global_idx, block_size, Ld,  Md, tmp, determinant_d);

      if (accept(determinant_d, determinant_u, cu_state)) {

        if (global_idx ==0) 
          printf("Acceptence\n");

        update_spins(pS,Nc, block_size, ind_start);
        acceptances++;
        __syncthreads();
        find_inverse(global_idx, block_size, Lu, Mu, Z, Mu_inv);
        __syncthreads();
        find_inverse(global_idx, block_size, Ld, Md, Z, Md_inv);     
        scale_M_exp(Mu_inv, Md_inv, block_size, ind_start, Nc,L ,d_exp, pS);
        __syncthreads();
        build_b(ind_start, bu, bd, Gu, Gd, Nc, L, num_threads, block_size, global_idx);
        __syncthreads();
        build_a_MC(Mu_inv, ind_start, au, Gu, Au_Tile, block_size, Nc*L, Nc, global_idx);
        build_a_MC(Md_inv, ind_start, ad, Gd, Ad_Tile, block_size, Nc*L, Nc, global_idx);
        // Don't update for acceptance !!
        //update_ind_block_size(sweep_ind_start, ind_start, nsweeps, Nc, L,block_size, block_sizes);
        acceptence = true;
        break;
    } else {
        rejections++;
        print_rejection();
        update_ind_block_size(sweep_ind_start, ind_start, nsweeps, Nc, L,block_size, block_sizes);
    }
   } else { // Delayed updating with delay = 1
      printf ("Yoo, block_size = 1\n");
      // build_M(ind_start, block_size, Nc, L, pS, Gu, Gd, d_exp, Mu, Md, Lu, Ld, clean_block_u, clean_block_d);      
      // TYPE det_u = Mu[0][0];
      // TYPE det_d = Md[0][0];
      TYPE det_u = GetG (Gu, ind_start, ind_start, Nc*L) + 1 + d_exp[-pS[ind_start]];
      TYPE det_d = GetG (Gd, ind_start, ind_start, Nc*L) + 1 + d_exp[ pS[ind_start]];
        
      if (global_idx ==0) {
         printf ("det_u = %f, det_d = %f \n", det_d, det_u);
      } 
     
      if (accept(det_d, det_u, cu_state)) {
       // Build as and Bs the same way we did in the reg_blocked.
        build_b(ind_start, bu, bd, Gu, Gd, Nc, L, num_threads, block_size, global_idx);
        build_a_MC_1_block(1/det_u, ind_start, au, Gu, Au_Tile, block_size, Nc*L, Nc, global_idx);
        build_a_MC_1_block(1/det_d, ind_start, ad, Gd, Ad_Tile, block_size, Nc*L, Nc, global_idx);
        // Don't update for acceptance !!
        //update_ind_block_size(sweep_ind_start, ind_start, nsweeps, Nc, L,block_size, block_sizes);
        acceptence = true;
        break;
      } else {
        rejections++;
        print_rejection();
        update_ind_block_size(sweep_ind_start,ind_start, nsweeps, Nc, L,block_size, block_sizes);
      }
    }
  }

  if (acceptence) {
     acceptence = false;
    if (global_idx == 0) {
      printf ("Mu is: \n");
      print_matrix(global_idx, block_size, Mu);
      printf ("Md is: \n");
      print_matrix(global_idx, block_size, Md);
      printf ("clean_block_u is: \n");
      print_matrix(global_idx, block_size, clean_block_u);
      printf ("clean_block_d is: \n");
      print_matrix(global_idx, block_size, clean_block_d);
    }

    // Now start the LU factorization.
    // Factorization without pivoting is not stable for small numbers we have in the green 
    // function. 
    if ((global_idx == 0) && false) {
      printf("det_u = %f, det_d = %f, det_u*det_d =%f\n", determinant_u,  determinant_d,   determinant_d*determinant_d);
    }

  
  if (global_idx == 0) {
    //printf ("Lower_u is: \n");
    //print_matrix(global_idx, block_size, Lu);
    //printf ("Upper_u is: \n");
    //print_matrix(global_idx, block_size, Mu);
    //printf ("Mu_inv after exp is: \n");
    //print_matrix(global_idx, block_size, Mu);
    //printf ("Md_inv after exp is: \n");
    //print_matrix(global_idx, block_size, Md);
    printf ("Au = \n");
    print_a(au, Nc, L, global_idx, block_size);
    printf ("Ad = \n");
    print_a(ad, Nc, L, global_idx, block_size);
  }

  if(0 == threadIdx.x) {
    d_ind[0] = ind_start;
    d_ind[1] = nsweeps;
    d_ind[2] = acceptances;
    d_ind[3] = rejections;
    d_ind[4] = sgn;
    // 5 reserved for nTimes.
    printf ("Writing back block_size of %i, ind_start=%i\n", block_size, ind_start);
    d_ind[7] = block_size;
    d_ind[8] = sweep_ind_start; // Which ind index marks a begning of a new sweep.
   }

  } else {

    printf ("Leaving without any acceptences !!\n");
    if(0 == threadIdx.x) {
     d_ind[0] = ind_start;
     d_ind[1] = nsweeps;
     d_ind[2] = acceptances;
     d_ind[3] = rejections;
     d_ind[4] = sgn;
     // 5 reserved for nTimes.
     printf ("Writing back block_size of %i, ind_start=%i\n", block_size, ind_start);
     d_ind[7] = -1;
     d_ind[8] = sweep_ind_start; // Which ind index marks a begning of a new sweep.
    }  
  }
  __syncthreads();
}

void launch_update_blocked(Kernel_Params* params) {
  int  ind = 0;
  dim3 thread_block_size;

  thread_block_size.y = 1;//params->nDelays;
 
  if ((params->Nc)*(params->L) > 512) {
    thread_block_size.x = 512/params->nDelays;
    //printf("Launching with :%i threads \n", ThreadsPerBlock_1_Dim);
  } else {
    thread_block_size.x = (params->Nc)*(params->L);
    //printf("Launching with :%i threads \n", ThreadsPerBlock_1_Dim);
  }
  curandState_t*  pCurand_State =  (curandState_t*)(params->d_devStates);
  thread_block_size.x = (params->Nc)*(params->L);
  thread_block_size.y = 1;       //params->nDelays;

  printf("Launching update_blocked with :%i threads \n", thread_block_size.y * thread_block_size.x);

  cudaGetLastError();  
  CheckCublasError(cublasCreate(&handle), "Creating handle for cublas");
  cudaThreadSynchronize();
  printf("Iteration \t\tKernel_Events \tKernel_Time  \t Sgemm_Events \t sgemm_Time\n");

  Timer KernelTimer;
  Timer SgemmTimer;
  int k = 0; 
  KernelTimer.Start();

  int nsweeps = 0;
  int mc_data[20];

  mc_data[5]  = params->nTimes; // Total sweeps to finish (params-nTimes).

  int nDelays = params->nDelays; 
  send_mc_data(params);

  while (nsweeps < params->nTimes) {

    cutilCheckMsg("update_blocked_kernel execution failed.\n");
    KernelTimer.Start();
    update_blocked<<<1, thread_block_size>>>(params->d_ind,
        		   params->d_Gu, params->d_Gd ,
        		   params->d_pS,

                           params->d_block_sizes, 

        		   params->d_au, params->d_bu,
        		   params->d_ad, params->d_bd,
                           pCurand_State,

                           params->d_expR_global_MC,

        		   params->d_diagu,
        		   params->d_diagd,
        		   params->alpha, params->Nc, params->L, 

                           params->nDelays);
      
      cudaDeviceSynchronize();
      cutilCheckMsg("update_blocked_kernel execution failed.\n");
      CopyBackMonteCarloData(params, mc_data);
      nsweeps         = mc_data[1];
      params->nDelays = mc_data[7]; // Effective block size (the block size choosen by the kernel).

      printf("we suggested %i block size, at sweepp %i\n", params->nDelays, nsweeps);
      if ((nsweeps != 0)&&(nsweeps % params->ntr == 0)) {
	  printf("Calling reg from MC\n");
	  launch_reg_blocked(params);
      }
      KernelTimer.Stop();
      printf(" \t\t %i \t  %2.10f \t  %2.10f ", k/(params->nDelays), KernelTimer.elapsedTime_GPU, KernelTimer.elapsedTime_OS);
      ind += params->nDelays;
      SgemmTimer.Start();
      Multiply_CUBLAS(params->d_bu, params->d_au, params->d_Gu , params);
      Multiply_CUBLAS(params->d_bd, params->d_ad, params->d_Gd , params);
      SgemmTimer.Stop();
      printf("\t %2.10f ",  SgemmTimer.elapsedTime_GPU);
      printf("\t %2.10f \n", SgemmTimer.elapsedTime_OS);
  }


  params->nDelays = nDelays;
  CheckCublasError(cublasDestroy(handle) , "Destroyig the handle for cublas");  
  printf("\nupdate_blocked\n"
        "Delays\t\t"
        "Kernel OS:\t\t"
        "Kernel GPU:\t\t"
        "Kernel Overhead:\t"\
        "Sgemm GPU:\t\t"\
        "Sgemm OS:\t\t"\
        "Sgemm Overhead:\t\t"\
        "All:\t\t\n");

  printf("%i\t\t"
        "%2.10f\t\t%2.10f\t\t%2.10f\t\t"
        "%2.10f\t\t%2.10f\t\t%2.10f\t\t"
        "%2.10f"
        "\n-------------------------------------------------------------\n",
        params->nDelays,
        KernelTimer.Total_OS,
        KernelTimer.Total_GPU,
        KernelTimer.Total_OS - KernelTimer.Total_GPU,
        SgemmTimer.Total_GPU,
        SgemmTimer.Total_OS,
        SgemmTimer.Total_OS - SgemmTimer.Total_GPU,
        KernelTimer.Total_OS + SgemmTimer.Total_OS);
  printf("now exitiing ....\n");
  exit(0);
};
