extern "C" void launch_update_delayed(Kernel_Params*);
extern "C" void launch_seed_random(Kernel_Params* params);

__global__ void Seed_Random(curandState_t* States) {
  int id = threadIdx.x ; // Use 1-Dim Index.
  curand_init (1234 , 1234 , 0 , &States[id]) ; // all threads habe the same seesds to generate the same random number
}

__global__ void update_delayed(int* d_ind ,
        	TYPE* Gu, TYPE* Gd,
        	int* pS,
        	TYPE* au, TYPE* bu,
                TYPE* ad, TYPE* bd,
        	curandState_t* curState,
        	TYPE* d_expR_global_MC,
        	TYPE* d_diagu, TYPE* d_diagd,
        	int Nc,
        	int L,
        	int nDelays,
                int nTimes, 
        	float *random_numbers) {
 
  //Find the number of threads.
  //Divided the system size by number of threads.
  int spin;//define the value of the spin from the Ising field.
  int imd ;
  TYPE avalueu, bvalueu;
  TYPE avalued, bvalued;
  TYPE Ru,Rd;
  TYPE* d_exp  = d_expR_global_MC + 1;
  // int ind = d_ind[0]; //set the first value of ind
  // No need to save ind anymore since this is a single run. 

  bool move_accepted = false;
  TYPE RuByRd     = 0;
  int acceptances = 0;
  int rejections  = 0;
  int ind         = d_ind[0];
  int nsweeps     = d_ind[1];
  int sgn = d_ind[4];
  int threadCount = blockDim.x;


  for (imd = threadIdx.x; imd < Nc*L; imd += threadCount) {
    d_diagu[imd] = GetG(Gu, imd, imd, Nc*L);
    d_diagd[imd] = GetG(Gd, imd, imd, Nc*L);
  }
  __syncthreads();

  for (int n_block_count = 0; n_block_count < nDelays ; n_block_count++ /*, nsweeps < nTimes*/) { // Will  change the number of nsweeps comdition later.
    // Shared work among all threads.
    // All the threads are doing the same work to avoid having to communicate
    // The acceptance or the rejection of the proposed move.
    while (!move_accepted) {
      spin = -pS[ind];                //propose a spin flip
      Ru   =   1.0 - ((d_diagu[ind] - 1.0) * d_exp[spin]);
      Rd   =   1.0 - ((d_diagd[ind] - 1.0) * d_exp[-spin]);
      RuByRd = Ru*Rd;
      float z = getRadomNumber(&(curState[threadIdx.x]));
      if(abs(RuByRd) > z) {
        move_accepted = true;
        pS[ind] = spin;
        acceptances++;
	sgn = sgn * (RuByRd < 0 ? -1 : 1);
        break;
      }
      ind++;
      if(ind >= Nc*L) {
        ind = 0 ; //reset spin			
	nsweeps++; // Increase the number of sweeps finished !!
        __syncthreads();
      }
      rejections++;       
     }

    move_accepted = false;
  for (imd = threadIdx.x; imd < Nc*L; imd += threadCount) {
    //Lots of optimisations needed here. We are reading from global memory.
    avalueu = GetG(Gu, imd, ind, Nc*L);//Columns
    avalued = GetG(Gd, imd, ind, Nc*L);
    bvalueu = GetG(Gu, ind, imd, Nc*L);//Rows
    bvalued = GetG(Gd, ind, imd, Nc*L);   
    // Matrix vector product
    if (n_block_count > 0) {
       for (int delay = 0; delay < n_block_count; delay++) {
	 int idx_1 = FlattenIdx(imd, delay, nDelays, Nc*L);	 
         int idx_2 = FlattenIdx(delay,ind ,Nc*L, nDelays);
	 int idx_3 = FlattenIdx(ind, delay, nDelays, Nc*L);
         int idx_4 = FlattenIdx(delay,imd, Nc*L, nDelays);

         avalueu += au[idx_1]*bu[idx_2];
         avalued += ad[idx_1]*bd[idx_2];
         bvalueu += au[idx_3]*bu[idx_4];
         bvalued += ad[idx_3]*bd[idx_4];
       }
    }

    avalueu = (avalueu - delta(imd , ind)) * d_exp[spin]  / Ru;
    avalued = (avalued - delta(imd , ind)) * d_exp[-spin] / Rd;

   {
     int tmp = FlattenIdx(n_block_count, imd , Nc*L, nDelays);
     bu[tmp] = bvalueu;
     bd[tmp] = bvalued;
     // *****************************************************************
     tmp = FlattenIdx(imd, n_block_count, nDelays, Nc*L);
     au[tmp]  = avalueu;
     ad[tmp]  = avalued;
    }
    //Updating the diagonal that we read first from the green function at the very begining.
    //we do this because thediaonals we read will be invalidated by the delayed update, since we don't
    //update the green function we need to update our diagonals so that we cab go through the delays
    d_diagu[imd]   = d_diagu[imd] + avalueu * bvalueu;
    d_diagd[imd]   = d_diagd[imd] + avalued * bvalued;
  }
    ind++;
    if(ind >= Nc*L) {
      ind = 0 ; //reset spin			
      nsweeps++; // Increase the number of sweeps finished !!
    }
    __syncthreads();
  } //  for (int n_block_count 

  // Now multiply with cublas:
  // This will be called on the host though !!!!
  // write out the value of ind before exiting the kernel.
  if(0 == threadIdx.x) {
    d_ind[0] = ind;
    d_ind[1] = nsweeps;
    d_ind[2] = acceptances;
    d_ind[3] = rejections;
    d_ind[4] = sgn;
   }
} // mc_drelayed_kernel


void launch_seed_random(Kernel_Params* params) {
  dim3 ThreadsPerBlock_1_Dim;
  curandState_t*  pCurand_State =  (curandState_t*)(params->d_devStates);
  if ((params->Nc)*(params->L) > 512) {
    ThreadsPerBlock_1_Dim = 512; 
  } else {
    ThreadsPerBlock_1_Dim = (params->Nc)*(params->L);
  }
  Seed_Random<<< 1 , ThreadsPerBlock_1_Dim >>>(pCurand_State);
  cutilCheckMsg("Seed_Random execution failed.\n");
  printf("Back from Seed_Random.\n");  
}


void launch_update_delayed(Kernel_Params* params) {
  dim3 ThreadsPerBlock_1_Dim;

  if ((params->Nc)*(params->L) > 512) {
    ThreadsPerBlock_1_Dim = 512; 
  } else {
    ThreadsPerBlock_1_Dim = (params->Nc)*(params->L);
  }

  cudaGetLastError();   //Clear the Global error state before calling the kernel.
  curandState_t*  pCurand_State =  (curandState_t*)(params->d_devStates);
  CheckCublasError(cublasCreate(&handle), "Creating handle for cublas");
  //Look at the Programming Guide page 50.
  cudaThreadSynchronize();
  printf("Entering launch_update_delayed\n");
  int mc_data[20];
  Timer KernelTimer;
  Timer SgemmTimer;
  int nsweeps=0;
  printf("Iteration \t\tKernel_Events \tKernel_Time  \t Sgemm_Events \t sgemm_Time\n");
  while (nsweeps < params->nTimes) {
    // Send a new parameter to the kernel with the number of trials 
    KernelTimer.Start();
    update_delayed<<< 1, ThreadsPerBlock_1_Dim >>>(params->d_ind,
        				    params->d_Gu, params->d_Gd,
        				    params->d_pS,
        				    params->d_au, params->d_bu,
        				    params->d_ad, params->d_bd,
        				    pCurand_State, 
        				    params->d_expR_global_MC,
        				    params->d_diagu, params->d_diagd,
        				    params->Nc,
        				    params->L,
        				    params->nDelays,
					    params->nTimes,
        				    params->d_random_numbers);
    cutilCheckMsg("update_delayed execution failed.\n");
    // Copy back the number of sweeps:
    CopyBackMonteCarloData(params, mc_data);
    KernelTimer.Stop();    
    nsweeps = mc_data[1];
    printf(" \t\t %i \t  %2.10f \t  %2.10f ", nsweeps, KernelTimer.elapsedTime_GPU, KernelTimer.elapsedTime_OS);
    printf("number of sweeps: %i acceptances: %i rejections: %i sgn: %i \n", nsweeps, mc_data[2], mc_data[3], mc_data[4]);
    if ((nsweeps != 0)&&(nsweeps % params->ntr == 0)) {
	printf("Calling reg from MC\n");
	launch_reg_delayed(params);
    }
    SgemmTimer.Start();
    Multiply_CUBLAS(params->d_au, params->d_bu, params->d_Gu, params);
    Multiply_CUBLAS(params->d_ad, params->d_bd, params->d_Gd, params);
    SgemmTimer.Stop();
    printf("\t %2.10f ",   SgemmTimer.elapsedTime_GPU);
    printf("\t %2.10f \n", SgemmTimer.elapsedTime_OS);
  }
  // Return the sign to the params.
  params->sgn = mc_data[4];
  CheckCublasError(cublasDestroy(handle) , "Destroyig the handle for cublas");
  printf("\nMonte Carlo Kernel\n"
        "Delays\t"
        "Kernel OS:\t"
        "Kernel GPU:\t"
        "Kernel Overhead:\t"\
        "Sgemm GPU:\t"\
        "Sgemm OS:\t"\
        "\tSgemm Overhead:\t"\
        "All:\t\n");

  printf("%i\t"
        "%2.10f\t%2.10f\t\t%2.10f\t"
        "%2.10f\t%2.10f\t\t%2.10f\t"
        "%2.10f"
        "\n-------------------------------------------------------------\n",
        params->nDelays,
        KernelTimer.Total_OS,
        KernelTimer.Total_GPU,
        KernelTimer.Total_OS - KernelTimer.Total_GPU,
        SgemmTimer.Total_GPU,
        SgemmTimer.Total_OS,
        SgemmTimer.Total_OS - SgemmTimer.Total_GPU,
        KernelTimer.Total_OS + SgemmTimer.Total_OS);
  printf("\tLeaving launch_update_delayed\n");
  printf("Leaving launch_update_delayed\n");
};

