extern "C" void launch_update_delayed_cuda_5(Kernel_Params* params);

void dev_mc_delayed_host_cuda_5(Kernel_Params* params) {
  dim3 ThreadsPerBlock_1_Dim(params->ISIZE , 1 , 1);
  cudaGetLastError();   //Clear the Global error state before calling the kernel.
  curandState_t*  pCurand_State =  (curandState_t*)(params->d_devStates); // Move it inside the kernel
  Timer KernelTimer;
  KernelTimer.Start();
  update_delayed_cuda_5<<<1, ThreadsPerBlock_1_Dim >>>(params->d_ind,
        				    params->d_Gu, params->d_Gd,
        				    params->d_pS,
        				    params->d_au, params->d_bu,
        				    params->d_ad, params->d_bd,
        				    pCurand_State, 
        				    params->d_expR_global_MC,
        				    params->d_diagu, params->d_diagd,
        				    params->Nc,
        				    params->L,
        				    params->nDelays,
					    params->nTimes,
        				    params->d_random_numbers);
  cudaDeviceSynchronize(); //Need to synchronize before we can check the status for kernel failures.Programming guide 3.2.8.
  KernelTimer.Stop();
  printf("update_delayed_cuda_5 took: %f seconds.\n", KernelTimer.elapsedTime_GPU);
  cutilCheckMsg("cuda_5_mc_delayed_kernel execution failed.\n");
};
__global__ void update_delayed_cuda_5(int* d_ind ,
        	TYPE* Gu,
        	TYPE* Gd,
        	int* pS,
        	TYPE* au, TYPE* bu , TYPE* ad, TYPE* bd,
        	curandState_t* curState,
        	TYPE* d_expR_global_MC,
        	TYPE* d_diagu , TYPE* d_diagd ,
        	int Nc,
        	int L,
        	int nDelays,
                int nTimes, 
        	float *random_numbers) {
 
  //Find the number of threads.
  //Divided the system size by number of threads.
  int spin;//define the value of the spin from the Ising field.
  int imd ;
  TYPE avalueu, bvalueu;
  TYPE avalued, bvalued;
  TYPE Ru,Rd;
  TYPE* d_exp  = d_expR_global_MC + 1;
  // int ind = d_ind[0]; //set the first value of ind
  // No need to save ind anymore since this is a single run. 
  TYPE colu_sum = 0;
  TYPE cold_sum = 0;
  TYPE rowu_sum = 0;
  TYPE rowd_sum = 0;
  imd = threadIdx.x;
  bool move_accepted = false;
  TYPE RuByRd = 0 ;
  // Define cublas handles:
   cublasHandle_t handle;
   cublasStatus_t status = cublasCreate(&handle);

for (int ind = 0, nsweeps = 0; nsweeps < nTimes;) {
  d_diagu[imd]  = GetG(Gu, imd , imd , Nc*L);
  d_diagd[imd]  = GetG(Gd, imd , imd , Nc*L);
  __syncthreads();

  for (int n_block_count = 0; n_block_count < nDelays ; n_block_count++) {
    // Shared work among all threads.
    // All the threads are doing the same work to avoid having to communicate
    // The acceptance or the rejection of the proposed move.
    while (!move_accepted) {
      spin = -pS[ind];                //propose a spin flip
      Ru   =   1.0 - ((d_diagu[ind] - 1.0) * d_exp[ spin]);
      Rd   =   1.0 - ((d_diagd[ind] - 1.0) * d_exp[ -spin]);
      RuByRd = Ru*Rd;
      float z = getRadomNumber(&(curState[imd]));
      if(RuByRd > z) {
        move_accepted = true;
        pS[ind] = spin;
        break;
      }
      ind++;
      if(ind >= Nc*L) {
        ind = 0 ; //reset spin			
	nsweeps++; // Increase the number of sweeps finished !!
        __syncthreads();
      }
    }
    move_accepted = false;
    //Lots of optimisations needed here. We are reading from global memory.
    avalueu = GetG(Gu, imd , ind , Nc*L);//Columns
    avalued = GetG(Gd, imd , ind , Nc*L);
    bvalueu = GetG(Gu, ind , imd , Nc*L);//Rows
    bvalued = GetG(Gd, ind , imd , Nc*L);
   
    if(n_block_count > 0) {
       colu_sum = 0;
       cold_sum = 0;
       rowu_sum = 0;
       rowd_sum = 0;
       for (int delay = 0; delay < n_block_count; delay++) {
         colu_sum = au[FlattenIdx(imd, delay, nDelays, Nc*L)]*bu[ FlattenIdx(delay,ind ,Nc*L, nDelays)] + colu_sum;
         cold_sum = ad[FlattenIdx(imd, delay, nDelays, Nc*L)]*bd[ FlattenIdx(delay,ind ,Nc*L, nDelays)] + cold_sum;
         rowu_sum = au[FlattenIdx(ind, delay, nDelays, Nc*L)]*bu[ FlattenIdx(delay,imd, Nc*L, nDelays)] + rowu_sum;
         rowd_sum = ad[FlattenIdx(ind, delay, nDelays, Nc*L)]*bd[ FlattenIdx(delay,imd, Nc*L, nDelays)] + rowd_sum;
       }
      avalueu += colu_sum;
      avalued += cold_sum;
      bvalueu += rowu_sum;
      bvalued += rowd_sum;
    }

    avalueu = (avalueu - delta(imd , ind)) * d_exp[spin]  / Ru;
    avalued = (avalued - delta(imd , ind)) * d_exp[-spin] / Rd;

    bu[FlattenIdx(n_block_count, imd , Nc*L, nDelays)] = bvalueu;
    bd[FlattenIdx(n_block_count, imd , Nc*L, nDelays)] = bvalued;

    au[FlattenIdx(imd, n_block_count, nDelays, Nc*L)]  = avalueu;
    ad[FlattenIdx(imd, n_block_count, nDelays, Nc*L)]  = avalued;
    //Updating the diagonal that we read first from the green function at the very begining.
    //we do this because thediaonals we read will be invalidated by the delayed update, since we don't
    //update the green function we need to update our diagonals so that we cab go through the delays
    d_diagu[imd]   = d_diagu[imd] + avalueu * bvalueu;
    d_diagd[imd]   = d_diagd[imd] + avalued * bvalued;
    ind++;
    if(ind >= Nc*L) {
      ind = 0 ; //reset spin			
      nsweeps++; // Increase the number of sweeps finished !!
    }
    __syncthreads();
  } //  for (int n_block_count 

  //Now multiply:
 if (threadIdx.x ==0) { 
    CublasFromDevice(au, bu, Gu, Nc*L, nDelays, handle);
    CublasFromDevice(ad, bd, Gd, Nc*L, nDelays, handle);
   } // if(threadIdx.x ==0 )
  cudaDeviceSynchronize();
 }// for (int ind = 0, nsweeps = 0; nsweeps < nTimes;)

 cublasDestroy(handle);
}; //mc_drelayed_kernel_cuda_5
