extern "C" void CheckError(int err, char* pMsg);
extern "C" void AllocateCudaMemory(Kernel_Params*) ;
extern "C" void CopyDataToGPU(TYPE*, TYPE*, int*, Kernel_Params*);
extern "C" void CopyDataToCPU(TYPE*, TYPE*, int*, Kernel_Params*);
extern "C" void CopyBackMonteCarloData(Kernel_Params* params, int* mc_data);
extern "C" void copyArrayToGPU(Kernel_Params*, float*);
extern "C" void allocateTestArrays(Kernel_Params*);
extern "C" void freeTestArrays();
extern "C" void SetCacheSize();
extern "C" void send_mc_data(Kernel_Params* params);


void CheckError(int err, char* pMsg) {
  if (err != 0) {
    printf(" Error ! %s \n" , pMsg);
    cutilCheckMsg(pMsg);
    exit(-1);
  }
}

void AllocateCudaMemory(Kernel_Params* params) {
  printf("in function AllocateCudaMemory \n");
  printf("(params->Nc)*(params->L): %i\n" , (params->Nc)*(params->L));
  printf("(params->ISIZE): %i\n"  , (params->ISIZE));
  printf("(params->nDelays): %i\n", (params->nDelays));
  printf("(params->alpha): %f\n",   (params->alpha));
  printf("Allocating Mremory from the cuda code \n");
  CheckError(cudaMalloc(&(params->d_au), sizeof(TYPE)*(params->ISIZE)*( params->nDelays)), "Allocating d_au" );
  CheckError(cudaMalloc(&(params->d_bu), sizeof(TYPE)*(params->ISIZE)*( params->nDelays)), "Allocating d_bu" );
  CheckError(cudaMalloc(&(params->d_ad), sizeof(TYPE)*(params->ISIZE)*( params->nDelays)), "Allocating d_ad" );
  CheckError(cudaMalloc(&(params->d_bd), sizeof(TYPE)*(params->ISIZE)*( params->nDelays)), "Allocating d_bd" );
  //Allocate Gu and Gd  
  CheckError(cudaMalloc(&(params->d_Gu), sizeof(TYPE)*(params->GSIZE)), "Allocating Gu");
  CheckError(cudaMalloc(&(params->d_Gd), sizeof(TYPE)*(params->GSIZE)), "Allocating Gd");
  CheckError(cudaMalloc(&(params->d_pS), sizeof(int)*(params->ISIZE)),  "Allocating d_pS");
  //Allocate d_diagd and d_diagu
  CheckError(cudaMalloc(&(params->d_diagd), sizeof(TYPE)*(params->ISIZE)), "Allocating d_diagd");
  CheckError(cudaMalloc(&(params->d_diagu), sizeof(TYPE)*(params->ISIZE)), "Allocating d_diagu");
  


  int size;
  if ((params->Nc)*(params->L) > 512) {
    size = 512; 
  } else {
    size = (params->Nc)*(params->L);
  }

  // Allocate memory for suggested block sizes:
  CheckError(cudaMalloc(&(params->d_block_sizes), sizeof(int)*size), "Allocating block_sizes");
  CheckError(cudaMalloc(&(params->d_devStates), sizeof(curandState_t)*size),"Allocating space for dev_states");
  CheckError(cudaMalloc(&(params->d_ind) , sizeof(int)*10), "Allocating space for d_ind");

  //Used with REG
  CheckError(cudaMalloc(&(params->d_expRu_global), sizeof(TYPE)*4), "Allocating space for d_expRu_global");
  CheckError(cudaMalloc(&(params->d_expRd_global), sizeof(TYPE)*4), "Allocating space for d_expRd_global");
  //Used with MC
  CheckError(cudaMalloc(&(params->d_expR_global_MC), sizeof(TYPE)*4), "Allocating space for d_expR_global_MC");
  //Allopcate Space for random numbers.
}

void CopyDataToGPU(TYPE* Gu, TYPE* Gd, int* S, Kernel_Params* params) {
  printf("Copying data to GPU \n ");
  //Copy Gu and Gd to device Memory.
  CheckError(cudaMemcpy((params->d_Gu) ,Gu, sizeof(TYPE)*(params->GSIZE), cudaMemcpyHostToDevice), "Copying Gu to device");
  printf("After first Copy \n");
  CheckError(cudaMemcpy((params->d_Gd) ,Gd, sizeof(TYPE)*(params->GSIZE), cudaMemcpyHostToDevice), "Copying Gd to device");
  CheckError(cudaMemcpy((params->d_pS) ,S , sizeof(int)*(params->ISIZE), cudaMemcpyHostToDevice),   "Copying pS to device");
  //FillConstMemory();
  //Filling in the  global memory of the lookup table
  TYPE* expforRu = (TYPE*)malloc(sizeof(TYPE)*4);
  TYPE* expforRd = (TYPE*)malloc(sizeof(TYPE)*4);

  //Used with the REG Updating
  expforRu[0] = exp (-(params->alpha)) - 1;
  expforRu[2] = exp ((params->alpha))  - 1;

  expforRd[0] = exp ((params->alpha))  - 1;
  expforRd[2] = exp (-(params->alpha)) - 1;

  CheckError(cudaMemcpy((params->d_expRu_global),expforRu, sizeof(TYPE)*4, cudaMemcpyHostToDevice), "Copying d_expRu to device");
  CheckError(cudaMemcpy((params->d_expRd_global),expforRd, sizeof(TYPE)*4, cudaMemcpyHostToDevice), "Copying d_expRd to device");

  //Used with the MC Updating 
  expforRu[0] = exp (-2*(params->alpha)) - 1;
  expforRu[2] = exp (2*(params->alpha)) - 1;
  CheckError(cudaMemcpy((params->d_expR_global_MC) ,expforRu, sizeof(TYPE)*4, cudaMemcpyHostToDevice), "Copying d_expRu for MC to device");
  free(expforRu);
  free(expforRd);
/*
  This code now moved to send_mc_data 
  int ind_tmp[20];
  ind_tmp[0] = 0; //In/Out: Last used ind_start. 
  ind_tmp[1] = 0; //Out:    Number of sweeps completed.
  ind_tmp[2] = 0; //Out: Number of acceptences.
  ind_tmp[3] = 0; //Out: Number of rejections. 
  ind_tmp[4] = 1; //Out: sgn
  ind_tmp[5] = params->nTimes; //In: Number of sweeps we need to finish.
  ind_tmp[6] = 0;              //In: Initial starting index for ind (0 is "randomly" selected).
  ind_tmp[7] = -1;             //In\Out: Last used block size. 
  CheckError(cudaMemcpy(params->d_ind ,ind_tmp , sizeof(int)*10, cudaMemcpyHostToDevice),   "Copying ind values to device");
*/

}


void send_mc_data(Kernel_Params* params) {
  int ind_tmp[10];
  ind_tmp[0] = 0;
  ind_tmp[1] = 0;
  ind_tmp[2] = 0;
  ind_tmp[3] = 0;
  ind_tmp[4] = 1; // sgn
  ind_tmp[5] = params->nTimes;
  ind_tmp[6] = ((int)rand())%(params->ISIZE); // starting index for the sweep. 
  ind_tmp[0] = ind_tmp[6]; // For calling the first time, start index is same as ind_start.
  ind_tmp[7] = -1;             //In\Out: Last used block size. If it is -1 means that no random blocks sizes were generated.
  //int ind_start  = d_ind[0]; // Last used Ind_start.
  //int nsweeps    = d_ind[1];
  //d_ind[2]       = acceptances;
  //d_ind[3]       = rejections;
  //int sgn             = d_ind[4];
  //int nTimes          = d_ind[5];
  //int block_size      = d_ind[7];
  CheckError(cudaMemcpy(params->d_ind ,ind_tmp , sizeof(int)*10, cudaMemcpyHostToDevice),   "Copying ind values to device");
};


void CopyBackMonteCarloData(Kernel_Params* params, int* mc_data) {
  CheckError(cudaMemcpy(mc_data, params->d_ind, sizeof(int)*10, cudaMemcpyDeviceToHost), "Copying monte carlo values to Host");
}

void CopyDataToCPU(TYPE* Gu, TYPE* Gd , int* S, Kernel_Params* params) {
  //Copy Gu and Gd to device Memory.
  printf("Copy back to the CPU memory \n");
  CheckError(cudaMemcpy(Gu, params->d_Gu, sizeof(TYPE)*(params->GSIZE), cudaMemcpyDeviceToHost), "Copying Gu to Host");
  CheckError(cudaMemcpy(Gd, params->d_Gd, sizeof(TYPE)*(params->GSIZE), cudaMemcpyDeviceToHost), "Copying Gd to Host");
  CheckError(cudaMemcpy(S, params->d_pS ,  sizeof(int)*(params->ISIZE), cudaMemcpyDeviceToHost), "Copying pS to host");
}

void SetCacheSize() {
  cudaFuncSetCacheConfig("reg_delayed_kernel", cudaFuncCachePreferL1);
  cudaFuncSetCacheConfig("mc_delayed_kernel", cudaFuncCachePreferL1);
}

double clock_now() {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (double)now.tv_sec + (double)now.tv_usec/1.0e6;
};

void FillConstMemory() {
  #ifdef __USE_CONST_MEM__
  TYPE* expforRu = (TYPE*)malloc(sizeof(TYPE)*4);
  TYPE* expforRd = (TYPE*)malloc(sizeof(TYPE)*4);
  expforRu[0] = exp (-alpha) - 1;
  expforRu[2] = exp (alpha)  - 1;
  expforRd[0] = exp (alpha)  - 1;
  expforRd[2] = exp (-alpha) - 1;
//CheckError(cudaMemcpyToSymbol("d_expRd", expforRd, sizeof(float)*4 , 0, cudaMemcpyHostToDevice), "Err copying const Memory");
//CheckError(cudaMemcpyToSymbol("d_expRu", expforRu, sizeof(float)*4 , 0, cudaMemcpyHostToDevice), "Err copying to const memory");
  free(expforRu);
  free(expforRd);
  #endif
}

void CheckCublasError(int err, char* pMsg) {
  if(err != 0) {
    printf(" Error ! %s \n" , pMsg);
  }
}

void Multiply_CUBLAS(TYPE* A, TYPE* B, TYPE* C , Kernel_Params* params) {
  TYPE f_Alpha = 1.0f;
  TYPE f_Beta  = 1.0f;
  int M, N ;
  M = (params->Nc)*(params->L);
  N = (params->nDelays);

#ifdef _SINGLE_
  Mult = cublasSgemm(handle,
      CUBLAS_OP_N,
      CUBLAS_OP_N,
      M,  
      M,  
      N,  
      &f_Alpha,
      A,
      M,  
      B,
      N,  
      &f_Beta,
      C,
      M);
#else
  Mult = cublasDgemm(handle,
      CUBLAS_OP_N,
      CUBLAS_OP_N,
      M,  
      M,  
      N,  
      &f_Alpha,
      A,
      M,  
      B,
      N,  
      &f_Beta,
      C,
      M);
#endif
  if(Mult != CUBLAS_STATUS_SUCCESS) {
    printf("Multiplication failed with error %i\n" , Mult);
  }
}

void copyArrayToGPU(Kernel_Params *gpu_params, float *springRs) {
  printf("gpu_params->random_numbers_count = %i\n" , gpu_params->random_numbers_count);  
  allocateTestArrays(gpu_params);
  CheckError(cudaMemcpy(gpu_params->d_random_numbers, springRs,sizeof(TYPE)*(gpu_params->random_numbers_count), cudaMemcpyHostToDevice), "Copying spring random numbers to device");
}

void allocateTestArrays(Kernel_Params *gpu_params) {
  printf("allocateTestArrays\n");
  printf("allocate gpu_params->random_numbers_count = %i\n",gpu_params->random_numbers_count); 
  CheckError(cudaMalloc(&(gpu_params->d_random_numbers), sizeof(TYPE)*(gpu_params->random_numbers_count)), "Allocating space for springRs array");
}

